/*
 * FIbot strategy implementation
 *
 * history:
 *
 *   version 0.93 (ximi)
 *      > code cleanup
 *      > randomized decision in findVictim
 *
 *   version 0.92 (Nikola)
 *	> bug in calculating distance with respect to energy fixed
 *	
 *
 *   version 0.91 (ximi)
 *      > speed improvements and code cleanup
 *   
 *   version 0.9 (the Crew)
 *   	> hybrid (crippled) bfs turned into correct Dijkstra (with hybrid implementation) and
 *   	  imaginary bug removed!
 * 
 *   	> minor bugfixes   
 *    
 *   version 0.8 (Nikola & Milan)
 *      > some bugs in shooting strategy fixed
 *      > resting/recharging at the end of game implemented
 *      > BUG: found a bug in the BFS, could not fix it (sometimes computes strange values, try
 *        test-turnaj2.world and look at the debug info
 *      >
 *   version 0.7 (ximi)
 *	> improved bfs - now collects more information
 *	> several performance issues carried out
 *	> getContent bug fixed?!
 *
 *   version 0.6 (Nikola)
 *	> mostly minor changes, preparation for submitting to the test tournament
 *	> when choosing the closest treasure, thinks about its value & value of its surrounding
 *	  treasures
 *	> now beats OrgBot on all the test-tournament maps
 *
 *   version 0.5 (all three team members [at and before the seminar])
 *	> "find some battery, stand next to it, wait (and fire if posiible) and recharge in the last
 *	  turn", when there is no gold left
 *
 *   version 0.4 (submitted by ximi)
 *	> BFS can be now called for different positions and directions (no new bug created (I hope))
 * 
 *   version 0.3 (submitted by Nikola & Milan)
 *      > a very unstable shooting strategy :-)
 *      > "find the closest gold and go for it" strategy - needs LOTS of tuning :-)
 *
 *   version 0.2 (submitted by ximi)
 *   
 *	> a very unstable decision mechanism (beta) - needs a lot of tuning and many new ideas not
 *	  yet introduced
 *	   
 *	> TODO: improve power vs. treasure priority balancing - perhaps introduce some lorarithmic
 *	  and/or exponential behaviour as well as the relation to the sum of treasures (NeuroBot-
 *	  Strategy.mapTreasure already prepared for this purpose, but some things will probably need
 *	  to be remade)
 *
 *	> TODO: develop some simple shooting algorithm (DONE)
 *
 *	> TODO: much more !?
 *
 *   version 0.1 - initial release (submitted by ximi)
 *
 *	> bfs-search now works correctly (I hope)
 *	> bot still persists at starting position !!!
 * 
 */

package cz.muni.fi.fibot.neurobot;


import java.util.*;
import java.lang.*;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;

public class NeuroBot implements Strategy {

	protected static final int FREE = 0;
	protected static final int WALL = 1;
	protected static final int BOT = 2;
	protected static final int BATT = 3;
	protected static final int GOLD = 4;

	public static String contentToString(int content) {

		switch (content) {

			case FREE: return "free";
			case WALL: return "wall";
			case BOT: return "bot";
			case BATT: return "battery";
			case GOLD: return "gold";
			default: return "unknown";

		}

	}

	protected static final int Left = 0; // Robot.Direction.LEFT.ordinal();
	protected static final int Up = 1; // Robot.Direction.UP.ordinal();
	protected static final int Right = 2; // Robot.Direction.RIGHT.ordinal();
	protected static final int Down = 3; // Robot.Direction.DOWN.ordinal();

	protected static Robot.Direction toDirection(int i) {
/*
		switch (i) {

			case 0: return Robot.Direction.LEFT;
			case 1: return Robot.Direction.UP;
			case 2: return Robot.Direction.RIGHT;
			case 3: return Robot.Direction.DOWN;
			default: return Robot.Direction.LEFT;

		}
*/
		return Robot.Direction.values()[i];

	}

	protected static final int DIRECTIONS[][] = { {-1, 0}, {0, -1}, {1, 0}, {0, 1} };

	private World world;
	private Robot robot;

	private int wWidth;
	private int wHeight;
	private int maxEnergy;
	private int turnsLeft;

	class Element implements Comparable {

		public int x;
		public int y;
		public int hashCode;

		public Element(int x, int y) {

			this.x = x;
			this.y = y;
			this.hashCode = y*wWidth + x;

		}

		public Element(Element element) {

			this.x = element.x;
			this.y = element.y;
			this.hashCode = element.hashCode;

		}
		
		protected boolean testOffsetTo(Element element, int offset[]) {
		
			return
				(offset[0] == element.x - this.x) &&
				(offset[1] == element.y - this.y);
		
		}
		
		public Robot.Direction directionTo(Element element) {
		
			int i;

			for (i = 0; (i < 4) && !this.testOffsetTo(element, DIRECTIONS[i]); ++i);

			if (i == 4) throw new IllegalArgumentException("Elements don't coincide");
			
			return NeuroBot.toDirection(i);
		
		}

		public int hashCode() { return this.hashCode; }

		public boolean equals(Object obj) {

			if (!(obj instanceof Element)) return super.equals(obj);
			
			return (this.hashCode == ((Element)obj).hashCode);

		}

		public int compareTo(Object obj) {
		
			return this.hashCode - ((Element)obj).hashCode;

		}

		public String toString() {

			return "[" + Integer.toString(this.x) + ":" + Integer.toString(this.y) +"]";

		}

	}

	class PathElement extends Element {

		public int distance;
		public int energy;
		public LinkedList<Integer> treasures;
		public int treasure_sum;
		public Robot.Direction direction;
		public PathElement previousElement;

		public PathElement(int x, int y, int distance, int energy, LinkedList<Integer> treasures,
			int treasure_sum, Robot.Direction direction, PathElement previousElement) {

			super(x, y);

			this.distance = distance;
			this.energy = energy;
			this.treasures = treasures;
			this.treasure_sum = treasure_sum;
			this.direction = direction;
			this.previousElement = previousElement;

		}

		public PathElement(Element element, int distance, int energy, LinkedList<Integer> treasures,
			int treasure_sum, Robot.Direction direction, PathElement previousElement) {

			super(element);

			this.distance = distance;
			this.energy = energy;
			this.treasures = treasures;
			this.treasure_sum = treasure_sum;
			this.direction = direction;
			this.previousElement = previousElement;

		}

		public PathElement(PathElement element) {

			super(element);

			this.distance = element.distance;
			this.energy = element.energy;
			this.treasures = element.treasures;
			this.treasure_sum = element.treasure_sum;
			this.direction = element.direction;
			this.previousElement = element.previousElement;

		}

		public int compareTo(Object obj) {

			if (!(obj instanceof PathElement)) return super.compareTo(obj);

			int ref_dist = this.distance - ((PathElement)obj).distance;

			if (ref_dist != 0) return ref_dist;

			int ref_en = ((PathElement)obj).energy - this.energy;

			if (ref_en != 0) return ref_en;

			return ((PathElement)obj).treasure_sum - this.treasure_sum;

		}

		public String toString() {

			StringBuilder retStr = new StringBuilder();

			retStr.append("dist: " + this.distance);
			retStr.append(", dir: " + this.direction);
			retStr.append(", en: " + this.energy);
			retStr.append(", treas: [ ");

			if (this.treasures != null) for (Integer i: this.treasures) retStr.append(i + " ");

			retStr.append("](" + this.treasure_sum + ")");

			if (this.previousElement != null)

				retStr.append(", prev: " + (new Element(this.previousElement)));

			return super.toString() + " " + retStr.toString();

		}

	}

	class PathInfo implements Comparable {

		public double value;
		public int content;
		public PathElement position;
//		public int distance;
//		public int energy;
//		public LinkedList<Integer> treasures;
		public Strategy.BotAction action;

		public PathInfo(double value, int content, PathElement position, /*int distance, int energy, LinkedList<Integer> treasures,*/ Strategy.BotAction action) {

			this.value = value;
			this.content = content;
			this.position = position;
//			this.distance = distance;
//			this.energy = energy;
//			this.treasures = treasures;
			this.action = action;

		}

		public String toString() {

			return this.position +
				", val: " + this.value +
				", con: " + NeuroBot.contentToString(this.content) +
				", act: " + this.action;

		}
		
		public int compareTo(Object obj) {
		
			PathInfo pathInfo = (PathInfo)obj;
			
			return
				(this.value == pathInfo.value)?
				((new Element(this.position)).compareTo(pathInfo.position)):
				((this.value < pathInfo.value)?(-1):(1));
			
		}
		
		public boolean equals(Object obj) {
		
			PathInfo pathInfo = (PathInfo)obj;
			
			return 
				(this.value == pathInfo.value)?
				((new Element(this.position)).equals(pathInfo.position)):
				(false);
		
		}

	}

	private Set<PathInfo> paths;
	private Map<Strategy.BotAction, Double> actions;

	private Set<PathInfo> other_paths[];
	private Map<Strategy.BotAction, Double> other_actions[];

	private Element currentPos;
	private Robot.Direction currentDirection;
	private int currentEnergy;
	private double treasureValueMultiplier;
	private double batteryPowerMultiplier;
	private int mapTreasure,moves;

	private int numRobots;

	private Random generator;

	public NeuroBot() {
	
		this.paths = new TreeSet<PathInfo>();
		this.actions = new HashMap<Strategy.BotAction, Double>();
		this.currentPos = new Element(-1, -1);
		this.generator = new Random();
		
	}

	int intMap[];

	public void setWorld(World w) {

		this.world = w;
		this.wWidth = w.getDimensionX();
		this.wHeight = w.getDimensionY();
		this.maxEnergy = w.getMaxRobotEnergy();
		this.intMap = new int[this.wWidth*this.wHeight];

	}

	public void setRobot(Robot r) {	this.robot = r; }

	public void initialize() {
		numRobots = this.world.getRobots().size();
//		System.out.println (numRobots);
		this.other_paths = (Set<PathInfo>[])(new Set<?>[numRobots - 1]);
		this.other_actions = (Map<Strategy.BotAction, Double>[])(new Map<?, ?>[numRobots - 1]);
		for (int j = 0; j < numRobots - 1; j++) {
			this.other_paths[j] = new TreeSet<PathInfo>();
			this.other_actions[j] = new HashMap<Strategy.BotAction, Double>();
		}
	}

	protected static final int wall = 1;

	protected static final int battery = 2;

	protected static final int robotConst = 4;

	protected static final int treasure = 256;

	protected int getContent(Element element) {

		int sub_content = this.world.getFieldValue(element.x, element.y);

		if ((sub_content & 4) == 4) return BOT;
		if ((sub_content & 1) == 1) return WALL;
		if ((sub_content / 256) > 0) return GOLD;
		if ((sub_content & 2) == 2) return BATT;

		return FREE;

	}

	protected int getGroundContent(Element pos) {

		int sub_content = this.world.getFieldValue(pos.x, pos.y);

		if ((sub_content / 256) > 0) return GOLD;
		if ((sub_content & 2) == 2) return BATT;

		return FREE;

	}

	protected static int getRelativeDistance(Robot.Direction d1, Robot.Direction d2) {

		if (d1 == d2) return 0;

		switch (d1.ordinal()) {

			case Left: if (d2 == Robot.Direction.RIGHT) return 2; break;
			case Up: if (d2 == Robot.Direction.DOWN) return 2; break;
			case Right: if (d2 == Robot.Direction.LEFT) return 2; break;
			case Down: if (d2 == Robot.Direction.UP) return 2; break;

		}

		return 1;

	}

	protected static Strategy.BotAction actionTable[][] = {
	
	 { Strategy.BotAction.STEP, Strategy.BotAction.RIGHT, Strategy.BotAction.LEFT, Strategy.BotAction.LEFT },
	 { Strategy.BotAction.LEFT, Strategy.BotAction.STEP, Strategy.BotAction.RIGHT, Strategy.BotAction.LEFT },
	 { Strategy.BotAction.LEFT, Strategy.BotAction.LEFT, Strategy.BotAction.STEP, Strategy.BotAction.RIGHT },
	 { Strategy.BotAction.RIGHT, Strategy.BotAction.LEFT, Strategy.BotAction.LEFT, Strategy.BotAction.STEP }
	
	};

	protected static Strategy.BotAction getAction(Robot.Direction d1, Robot.Direction d2) {

		return NeuroBot.actionTable[d1.ordinal()][d2.ordinal()];

	}
	
	protected void introducePath(PathElement firstElement, Robot.Direction direction,
		PathElement lastElement, int content, Set<PathInfo> paths,
		Map<Strategy.BotAction, Double> actions) {

		PathElement traverseElement = lastElement;

		while (traverseElement.previousElement != firstElement)

			traverseElement = traverseElement.previousElement;

		double value = 0;

		switch (content) {
		
			case GOLD:
		
/*				int treasure = this.world.getTreasureValue(
					lastElement.x,
					lastElement.y
				); */
				value = this.world.getTreasureValue(lastElement.x, lastElement.y);
				for (int i = lastElement.x-1; i <= lastElement.x+1; i++)

					for (int j = lastElement.y-1; j <= lastElement.y+1; j++) {

						if (this.checkDimensions(i, j) && this.world.containsTreasure(i, j))

							value += this.world.getTreasureValue(i, j);

					}

//					 treasureValueMultiplier*treasure/lastElement.distance;
				
//				this.mapTreasure += treasure; 
				
				break;
				
			case BATT:
			
//				value = batteryPowerMultiplier/lastElement.distance;
				
				break;
		
		}

		Strategy.BotAction action = NeuroBot.getAction(
			direction,
			firstElement.directionTo(traverseElement)
		);
		
		actions.put(action, actions.get(action) + value);

//		System.out.println(lastElement);
		
		paths.add(
			new PathInfo(
				value,
				content,
				lastElement,
/*				lastElement.distance,
				lastElement.energy,
				lastElement.treasures,*/
				action
			)
		);

	}

	protected boolean checkDimensions(Element element) {

		return
			(element.x >= 0) && (element.x < this.wWidth) &&
			(element.y >= 0) && (element.y < this.wHeight);

	}
	
	protected boolean checkDimensions(int x, int y) {

		return (x >= 0) && (x < this.wWidth) && (y >= 0) && (y < this.wHeight);

	}

	protected boolean wantRecharge(int energy) {

		if (energy < 3) return true;
		return (energy < this.world.getMaxRobotEnergy()/2 - this.world.getRestRobotEnergy());

	}

	protected PathElement getPath(PathElement source, Element destination, int content, Robot.Direction direction) {
	
		int rest_energy = source.energy - this.getRelativeDistance(source.direction, direction) - 1;
		int new_distance = source.distance + this.getRelativeDistance(source.direction, direction) + 1;
		LinkedList<Integer> treasures = source.treasures;
		int treasure_sum = source.treasure_sum;

		if (rest_energy < 1) { //tj. ztratil energii behem otaceni, chyba
			rest_energy += this.world.getRestRobotEnergy() - 1;
			new_distance++;
		}

		switch (content) {

			case GOLD:

				treasures = ((treasures == null)?(new LinkedList<Integer>()):(new LinkedList<Integer>(treasures)));
				
                                treasures.add(
                                	this.world.getTreasureValue(
                                                destination.x,
                                                destination.y
                                        )
                                );

				treasure_sum += treasures.getLast();
				//new_distance++;
				//rest_energy--;

				break;

			case BATT:

				if (this.wantRecharge(rest_energy)) {

					rest_energy = this.world.getMaxRobotEnergy();
					new_distance++;

				}

				break;

		}

		if (rest_energy == 1) {

			rest_energy = this.world.getRestRobotEnergy();
			new_distance++;

		}

		return new PathElement(
			destination,
			new_distance,
			rest_energy,
			treasures,
			treasure_sum,
			direction,
			source
		);
		
	}

	boolean debug;

	protected void filteredDebug(PathElement el) {

		if (debug) {

			List<Element> elements = new LinkedList<Element>();

			elements.add(new Element(8, 12));
			elements.add(new Element(9, 13));
			elements.add(new Element(4, 14));
			elements.add(new Element(5, 15));

			for (Element element : elements) {

				if (element.equals(el)) {

					System.out.print(el + "(" + el.distance + ", " + el.energy + ", " + el.treasure_sum + ") ");

					PathElement traverseElement = (PathElement)el;

					while (traverseElement.previousElement != null) {

						System.out.print(traverseElement);

						traverseElement = traverseElement.previousElement;

					}
					
					System.out.println();

				}

			}

		}

	}
	
	protected static void printPathMap(int map[], int width, int height) {
	
		for (int i = 0; i < width; ++i) System.out.print("---");
		
		System.out.println("-");

		for (int j = 0; j < height; ++j) {
		
			for (int i = 0; i < width; ++i) {

				switch (map[j*height + i]) {
				
					case -1: System.out.print("   "); break;
					case 0: System.out.print(" me"); break;
					default:
					
						System.out.print(
							String.format(
								" %1$2d",
								map[j*height + i]
							)
						);
						
						break;
	
				}
				
			}
			
			System.out.println();

		}

	}

	protected void calculatePaths(Robot robot, Set<PathInfo> paths, Map<Strategy.BotAction, Double> actions) {

		for (int i = 0; i < this.wWidth*this.wHeight; ++i) this.intMap[i] = -1;

		paths.clear();

		actions.clear();
		actions.put(Strategy.BotAction.STEP, new Double(0));
		actions.put(Strategy.BotAction.RIGHT, new Double(0));
		actions.put(Strategy.BotAction.LEFT, new Double(0));

		Element source = new Element(robot.getPositionX(), robot.getPositionY());

		PathElement
			start = new PathElement(source, 0, robot.getEnergy(), null, 0, robot.getDirection(), null),
			current;

		Set<Element> elements = new HashSet<Element>();

//		elements.add(source);

		PriorityQueue<PathElement> bfsQueue = new PriorityQueue<PathElement>();

		bfsQueue.add(start);

		while ((current = bfsQueue.poll()) != null) {

			if (!elements.add(current)) continue;

//			this.filteredDebug(current);

			this.intMap[current.hashCode] = current.distance;

			int content = this.getContent(current);

			if (content == GOLD || content == BATT)

				this.introducePath(
					start,
					robot.getDirection(),
					current,
					content,
					paths,
					actions
				);

			for (int i = 0; i < 4; ++i) {

				Element neighbour = new Element(
					current.x + DIRECTIONS[i][0],
					current.y + DIRECTIONS[i][1]
				);

				if (!this.checkDimensions(neighbour)) continue;

				switch (content = this.getContent(neighbour)) {

					case GOLD:
					case BATT:
					case FREE:

						PathElement pathElement = this.getPath(
							current,
							neighbour,
							content,
							NeuroBot.toDirection(i)
						);
/*
						if (content == GOLD || content == BATT)

							this.introducePath(
								start,
								robot.getDirection(),
								pathElement,
								content,
								paths,
								actions
							);
*/
						bfsQueue.offer(pathElement);

						break;

				}

			}

		}

		if (debug) NeuroBot.printPathMap(this.intMap, this.wWidth, this.wHeight);

	}

	private double valuate (int distance, double value) {
		// je treba volit nejakou doboru zavislost
		if (distance > turnsLeft + 1) return 0;
	        return pomValuate(distance, value); // tohle nic moc, zatim
	}
	
	static private double pomValuate(int distance, double value) {
		return value/(distance + 1);
	}
	
	private class PathInfoCompare implements Comparator<PathInfo> {
		public PathInfoCompare() { }
		public int compare(PathInfo p1, PathInfo p2) {
			if (pomValuate(p1.position.distance, p1.value) < pomValuate(p2.position.distance, p2.value)) return 1;
			else if (pomValuate(p1.position.distance, p1.value) > pomValuate(p2.position.distance, p2.value)) return -1;
			else return 0;
		}
	}
	
	private boolean blockedRobot (Robot r) {
		Element tmpEl = new Element(r.getPositionX(), r.getPositionY());
		tmpEl.x += DIRECTIONS[r.getDirection().ordinal()][0];
		tmpEl.y += DIRECTIONS[r.getDirection().ordinal()][1];
		return !checkDimensions(tmpEl) || this.world.containsRobot(tmpEl.x, tmpEl.y) || this.world.containsWall(tmpEl.x, tmpEl.y);
	}

	private Strategy.BotAction findVictim(int dir, int dirShift, boolean checkRotation, int checkEnergy) {

		Element tmpEl = new Element(this.currentPos.x, this.currentPos.y);

		do {

			tmpEl.x += DIRECTIONS[(dir + dirShift)%4][0];
			tmpEl.y += DIRECTIONS[(dir + dirShift)%4][1];

		} while (checkDimensions(tmpEl) && !this.world.containsRobot(tmpEl.x, tmpEl.y) && !this.world.containsWall(tmpEl.x, tmpEl.y));

		if (checkDimensions(tmpEl) && this.world.containsRobot(tmpEl.x, tmpEl.y)) {

//			System.out.println(tmpEl.x+" " + tmpEl.y + this.world.containsRobot(tmpEl.x,tmpEl.y));

			Robot r = this.world.getRobotOn(tmpEl.x, tmpEl.y);

			if (
				(
				 !checkRotation ||  blockedRobot(r) ||
				 r.getDirection() == toDirection((dir + 1)%4) ||
				 r.getDirection() == toDirection((dir + 3)%4)
				) && (r.getEnergy() > checkEnergy)
			   ) 
				switch (dirShift) {

					case 0: return Strategy.BotAction.SHOOT;
					case 1: return Strategy.BotAction.RIGHT;
					case 3: return Strategy.BotAction.LEFT;
					default:
						return (this.generator.nextBoolean())?
							(Strategy.BotAction.RIGHT):
							(Strategy.BotAction.LEFT); // muze si vybrat, ale nas robot je pravicak :-) - now randomized

				}

		}

//		System.out.println("findVictim");
		return Strategy.BotAction.REST;

	}

	private boolean amINextTo(Element pos) {

		int xDist = Math.abs(this.currentPos.x - pos.x);
		int yDist = Math.abs(this.currentPos.y - pos.y);
		return (xDist <= 1 && yDist <= 1 && xDist*yDist == 0);

	}


	public Strategy.BotAction decide() {

//		System.out.println(this.world.getRound());
		this.currentPos.x = this.robot.getPositionX();
		this.currentPos.y = this.robot.getPositionY();
		this.currentDirection = this.robot.getDirection();
		this.currentEnergy = this.robot.getEnergy();
		this.mapTreasure = 0;
		
		this.turnsLeft = this.world.getRoundsCount() - 1 - this.world.getRound();
		
		if (this.currentEnergy == 1) 
			if (this.world.containsBattery(this.currentPos.x, this.currentPos.y) && (turnsLeft > 2 || turnsLeft == 0))
				return Strategy.BotAction.RECHARGE;
			else {/*System.out.println("must");*/return Strategy.BotAction.REST;}

		switch (this.getGroundContent(this.currentPos)) {
		
			case GOLD: return Strategy.BotAction.TAKE;
				
			case BATT:
//				System.out.println (" Stojim na baterce, turnsleft = " + turnsLeft);
				if ((turnsLeft > 2 && wantRecharge(this.currentEnergy)) || this.turnsLeft == 0)
					return Strategy.BotAction.RECHARGE;
				break;
				
		}

//		this.treasureValueMultiplier = (double)this.currentEnergy/this.maxEnergy;
//		this.batteryPowerMultiplier = (double)(this.maxEnergy - this.currentEnergy)/this.maxEnergy;

//		this.debug = true;
		this.calculatePaths(this.robot, this.paths, this.actions);
		this.debug = false;
		for (int i = 0, j = 0; i < numRobots; i++) {
			Robot r = this.world.getRobots().get(i);
			if (r == this.robot) continue;
			this.calculatePaths(r, this.other_paths[j], this.other_actions[j]);
//		        for (PathInfo path: this.other_paths[j]) System.out.println(path);
//			System.out.println("-");	
			j++;
		}
			

//		for (PathInfo path : this.paths) System.out.println(path);

/*		System.out.println("treasure*=" + this.treasureValueMultiplier);
		System.out.println("power*=" + this.batteryPowerMultiplier);

		Strategy.BotAction bestAction = Strategy.BotAction.STEP;

		for (Map.Entry<Strategy.BotAction, Double> entry: this.actions.entrySet()) {
		
			if (entry.getValue() > this.actions.get(bestAction))
			
				bestAction = entry.getKey();

		} */
		PathInfo dummy = new PathInfo(0, WALL, new PathElement(-1,-1, 10000, 0, null, 0, Robot.Direction.UP, null), Strategy.BotAction.REST);

		PathInfo best = dummy;
		PathInfo other_best [] = new PathInfo [numRobots - 1];
		for (int i = 0; i < numRobots - 1; i++) { other_best[i] = dummy; }

		boolean chosen = false; moves = 10000;

		if (!paths.isEmpty()) {

			PriorityQueue<PathInfo> pqPaths = new PriorityQueue<PathInfo> (paths.size(), new PathInfoCompare());

			for (PathInfo path: this.paths) { 
				if (path.content == GOLD && this.world.getTreasureValue(path.position.x, path.position.y) >= 4) {pqPaths.offer(path);
//			System.out.println(path); 
				}
			}

			while (!chosen && !pqPaths.isEmpty()) {

//				int tmpMoves = 10000;
				moves = 10000;
	        	        chosen = true;

				best = pqPaths.poll();
//				System.out.println("BC: " + best);
//				System.out.println("valuate: " + valuate(best.position.distance, best.value));

			for (int i = 0; i < numRobots - 1; i++) {

				for (PathInfo path: this.other_paths[i]) {

					if (best.position.equals(path.position)) {

						moves = (path.position.distance - best.position.distance) < moves ? 
							(path.position.distance - best.position.distance) : moves;

					}

				}

//				System.out.println(i + ": " + other_best[i].position + " (" + best.position + ")");

//				if (best.position.equals(other_best[i].position)) {

//					moves = other_best[i].position.distance - best.position.distance;

//					System.out.println("moves = " + moves);

				if (moves < 0) {
					chosen = false;
//					System.out.println(" Don't want it!");
					break;
				}
			}
			}
		}
		
		boolean battery = false, lowGold = false; 
		PathInfo bestGold = new PathInfo(0, WALL, new PathElement(-1,-1, 10000, 0, null, 0, Robot.Direction.UP, null), Strategy.BotAction.REST);

//		if (best.content == GOLD && this.world.getTreasureValue(best.position.x, best.position.y) < 4) { lowGold = true; bestGold = best; }
		if (best.content != GOLD || lowGold) { battery = true;
			/* dojdi k nejblizsi baterce a cihej tam */
			best = new PathInfo(0, WALL, new PathElement(-1, -1, 1000, 0, null, 0, Robot.Direction.UP, null), Strategy.BotAction.REST);
			for (PathInfo path : this.paths)
				if (path.position.distance < best.position.distance && path.position.distance <= turnsLeft && path.content == BATT)
				{ best = path; }
			if (best.content == BATT) {
//	                        System.out.println("TL" + turnsLeft + " " + "BD:" + best.position.distance);
				
				if (turnsLeft < 2*(best.position.distance) && this.world.getRestRobotEnergy() < this.world.getMaxRobotEnergy() / 3)
					return best.action;
				if (!amINextTo(best.position)) return best.action;
			}
		}
		
                int X = (this.world.getMaxRobotEnergy() - this.currentEnergy) / (this.world.getRestRobotEnergy() - 1);
//		System.out.println(X);
                if (turnsLeft <= X && getGroundContent(this.currentPos) != BATT
			 ) { return Strategy.BotAction.REST; }


		/* strileni */
		int dir = this.currentDirection.ordinal();
		boolean podminka = true;

		Element tmpEl = new Element(this.currentPos);
		Strategy.BotAction tmpAct;
		
/*
		do {
		tmpEl.x += DIRECTIONS[dir][0]; tmpEl.y += DIRECTIONS[dir][1];
		} while (checkDimensions(tmpEl) &&
			!this.world.containsRobot(tmpEl.x,tmpEl.y) && !this.world.containsWall(tmpEl.x,tmpEl.y));

		if (checkDimensions(tmpEl) && this.world.containsRobot(tmpEl.x,tmpEl.y))
		{
			
			Robot r = this.world.getRobotOn(tmpEl.x,tmpEl.y);
			if (r.getEnergy() > 3 && podminka) return Strategy.BotAction.SHOOT;
		}

		tmpEl.x = this.currentPos.x; tmpEl.y = this.currentPos.y;

		do {
			tmpEl.x += DIRECTIONS[(dir+1)%4][0]; tmpEl.y += DIRECTIONS[(dir+1)%4][1];
		} while (checkDimensions(tmpEl) &&
			!this.world.containsRobot(tmpEl.x,tmpEl.y) && !this.world.containsWall(tmpEl.x,tmpEl.y));

		if (checkDimensions(tmpEl) && this.world.containsRobot(tmpEl.x,tmpEl.y))
		{
			Robot r = this.world.getRobotOn(tmpEl.x,tmpEl.y);
			if ((r.getDirection() == toDirection((dir+1)%4) || r.getDirection() == toDirection((dir+3)%4)) &&
			r.getEnergy() > 3 && podminka ) return Strategy.BotAction.RIGHT;
		}
		
		tmpEl.x = this.currentPos.x; tmpEl.y = this.currentPos.y;

		do {
		tmpEl.x += DIRECTIONS[(dir+3)%4][0]; tmpEl.y += DIRECTIONS[(dir+3)%4][1];
		} while (checkDimensions(tmpEl) && 
			!this.world.containsRobot(tmpEl.x,tmpEl.y) && !this.world.containsWall(tmpEl.x,tmpEl.y));

		if (checkDimensions(tmpEl) && this.world.containsRobot(tmpEl.x,tmpEl.y))
		{
			Robot r = this.world.getRobotOn(tmpEl.x,tmpEl.y);
			if ((r.getDirection() == toDirection((dir+1)%4) || r.getDirection() == toDirection((dir+3)%4)) &&
			r.getEnergy() > 3 && podminka ) return Strategy.BotAction.LEFT;
		}
*/
		
		if ( !battery ) { 
			if (moves > 1 || (best.position.distance == 1 && moves == 0)) { tmpAct = findVictim(dir, 0, false, 3); // v primem smeru?
					if (tmpAct != Strategy.BotAction.REST) return tmpAct; }
			if (moves > 3) { tmpAct = findVictim(dir, 1, true, 4); // rozhledni se
					if (tmpAct != Strategy.BotAction.REST) return tmpAct;
					tmpAct = findVictim(dir, 3, true, 4); // rozhledni se
					if (tmpAct != Strategy.BotAction.REST) return tmpAct; }
			if (moves > 5) { tmpAct = findVictim(dir, 2, false, 4);
					if (tmpAct != Strategy.BotAction.REST) return tmpAct; }
		} else {
			int treshold = lowGold ? 0 : 2;
			tmpAct = findVictim(dir, 0, false, treshold);
			if (tmpAct != Strategy.BotAction.REST) return tmpAct;
			if (this.world.getRoundsCount() - 1 - this.world.getRound() > 3) {
				tmpAct = findVictim(dir, 1, false, treshold);
				if (tmpAct != Strategy.BotAction.REST) return tmpAct;
				tmpAct = findVictim(dir, 3, false, treshold);
				if (tmpAct != Strategy.BotAction.REST) return tmpAct;
				if (this.world.getRoundsCount() - 1 - this.world.getRound() > 5) {
					tmpAct = findVictim(dir, 2, false, treshold);
					if (tmpAct != Strategy.BotAction.REST) return tmpAct;
				}
			}	

		/*	do {
			tmpEl.x += DIRECTIONS[(dir+2)%4][0]; tmpEl.y += DIRECTIONS[(dir+2)%4][1];
        	} while (checkDimensions(tmpEl) &&
	            !this.world.containsRobot(tmpEl.x,tmpEl.y) && !this.world.containsWall(tmpEl.x,tmpEl.y));
	
    	    if (checkDimensions(tmpEl) && this.world.containsRobot(tmpEl.x,tmpEl.y))
        	{
	            Robot r = this.world.getRobotOn(tmpEl.x,tmpEl.y);
    	        if ((r.getDirection() == toDirection((dir+1)%4) || r.getDirection() == toDirection((dir+3)%4)) &&
        	    r.getEnergy() > 3 && podminka ) return Strategy.BotAction.LEFT;
	        } */
//			System.out.println("don't know what to do");
			if (this.getGroundContent(this.currentPos) == BATT && turnsLeft < 2) return Strategy.BotAction.SHOOT;
			if (lowGold) return bestGold.action; else return Strategy.BotAction.SHOOT;
		}
//		System.out.println ("Best action: going to " + contentToString(best.content) + " at " + best.position.x + ", " + best.position.y +
//					" (value: " + best.value + ", distance: " + best.position.distance + ")" );
//		System.out.println ("Best action: " + best);

//		System.out.println (world.getTreasureValue(best.position.x, best.position.y));

//		System.out.println(this.getRelativeDistance(Robot.Direction.UP, Robot.Direction.DOWN));
/*		PathElement traverseElement = best.position;

		while (traverseElement.previousElement != null) {

			System.out.println("travese element at " + traverseElement);

			traverseElement = traverseElement.previousElement;

		}
*/
		return best.action;
	}

}
