package cz.muni.fi.fibot.BenderBendingRodriguez;

import java.util.List;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;

public class BenderBendingRodriguez implements Strategy {

	private World world;
	private Robot robot;
	
	private int wX;
	private int wY;
	
	private int rX;
	private int rY;
	
	private int pX;
	private int pY;
	
	private List<Robot> robots;
	
	private double [][] confTreasures = new double[52][52];
	private int [][] confEnemies = new int[52][52];
	private boolean [][] localize = new boolean[52][52];
	
	private boolean [][] fireRanges = new boolean[50][50];
	
	private int numGlobalSteps;

        private int numDeathMatchSteps;

        private int numTreasures;

        private Strategy.BotAction lastAction;
        private int lastRX, lastRY;
        private boolean lastRested;

        private int rounds;

	public void initialize() {
		robots = world.getRobots();
		
		wX = world.getDimensionX();
		wY = world.getDimensionY(); 
		
		for (int i = 1; i <= wX; i++)
		{
			for (int j = 1; j <= wY; j++)
			{
				localize[i][j] = true;
			}
		}
		
		for (int i = 0; i <= wX + 1; i++)
		{
			localize[i][0] = localize[i][wY+1] = false;
		}
		
		for (int i = 0; i <= wY + 1; i++)
		{
			localize[0][i] = localize[wX+1][i] = false;
		}
		
		for (int i = 0; i <= wX + 1; i++)
		{
			for (int j = 0; j <= wY + 1; j++)
			{
				confTreasures[i][j] = 0;
				confEnemies[i][j] = 99999999;
			}
		}
		
		numGlobalSteps = (wX+wY)/4;
		if (numGlobalSteps < 1)
		{
			numGlobalSteps = 1;
		}
                numDeathMatchSteps = 2 * numGlobalSteps;

                lastAction = null;
                lastRX = lastRY = 0;
                lastRested = false;

                rounds = world.getRoundsCount();
	}

	public void setWorld(World w) {
		this.world = w;
	}

	public void setRobot(Robot r) {
		this.robot = r;
	}
	
	private void update()
	{
		int i, j;

                numTreasures = 0;
				
		for (i = 0; i < wX; i++)
		{
			for (j = 0; j < wY; j++)
			{
				fireRanges[i][j] = false;
                                
                                if ((world.containsRobot(i, j)) && (world.getRobotOn(i, j).getId() != robot.getId()) && (world.getRobotOn(i, j).getEnergy() > 0))
                                {
                                        Robot.Direction direction = world.getRobotOn(i, j).getDirection();
                                        confEnemies[i+1][j+1] = 1000;
                                        if (direction.equals(Robot.Direction.LEFT))
                                        {
                                                confEnemies[i+1][j+1] += 100;
                                        }
                                        else if (direction.equals(Robot.Direction.UP))
                                        {
                                                confEnemies[i+1][j+1] += 200;
                                        }
                                        else if (direction.equals(Robot.Direction.RIGHT))
                                        {
                                                confEnemies[i+1][j+1] += 300;
                                        }
                                        else if (direction.equals(Robot.Direction.DOWN))
                                        {
                                                confEnemies[i+1][j+1] += 400;
                                        }
                                }
                                else
                                {
                                        confEnemies[i+1][j+1] = 99999999;
                                }
				
				if (world.containsTreasure(i, j))
				{
					confTreasures[i+1][j+1] = (double)world.getTreasureValue(i, j);
                                        numTreasures++;
				}
				else 
				{
					confTreasures[i+1][j+1] = 0.0; 
				}
			}
		} 
		
		rX = robot.getPositionX();
		rY = robot.getPositionY();
		
		pX = rX + 1;
		pY = rY + 1;
	}
	
	/*private double delta(int i, int j)
	{
		if (world.containsWall(i - 1, j - 1))
		{
			return 0.0;
		}
		
		double neighbors = (double)
			(confTreasures[i-1][j-1] + confTreasures[i][j-1] + confTreasures[i+1][j-1] +
			 confTreasures[i-1][j  ] +                       + confTreasures[i+1][j  ] +
			 confTreasures[i-1][j+1] + confTreasures[i][j+1] + confTreasures[i+1][j+1]);		
			                                                                       
		return neighbors / 24.0 + 4.0 * confTreasures[i][j];
	}*/
	
	private double delta(int i, int j)
	{
		if (world.containsWall(i - 1, j - 1))
		{
			return 0.0;
		}
		
		double neighbors = (double)
			(                          confTreasures[i][j-1] +
			 confTreasures[i-1][j  ] +                       + confTreasures[i+1][j  ] +
			                           confTreasures[i][j+1]                             );	
		
		int divisor = 0;
		if (localize[i-1][j]) divisor++;
		if (localize[i+1][j]) divisor++;
		if (localize[i][j-1]) divisor++;
		if (localize[i][j+1]) divisor++;
			                                                                       
		return neighbors / (3.0 * (double)divisor) + 
			(double)(world.getTreasureValue(i - 1, j - 1) + 1) * confTreasures[i][j];
	}

        private int chumps(int i, int j)
        {
                int steps;
                int minDistance = 9999999;
                int minLineDistance = 0, minAllDistance = 0;
                int minDistanceDirection = 1;
                int allSteps, lineSteps, tmpDirection;
                int dirChange = 0;

                if (world.containsWall(i - 1, j - 1))
                {
                        return 99999999;
                }

                if (world.containsRobot(i - 1, j - 1))
                        return confEnemies[i][j];

                allSteps = confEnemies[i][j-1] / 1000;
                lineSteps = confEnemies[i][j-1] % 100;
                tmpDirection = (confEnemies[i][j-1] / 100) % 10;
                steps = allSteps + lineSteps + 1;
                if (tmpDirection == 2);
                else if (tmpDirection != 4) 
                {
                        steps++;
                        if (steps < minDistance)
                        {
                                minLineDistance = lineSteps;
                                minAllDistance = allSteps;
                                minDistance = steps;
                                minDistanceDirection = 4;
                                dirChange = 1;
                        }
                }
                else if (steps < minDistance)
                {
                        minLineDistance = lineSteps;
                        minAllDistance = allSteps;
                        minDistance = steps;
                        minDistanceDirection = 4;
                        dirChange = 0;
                }
                        
                allSteps = confEnemies[i][j+1] / 1000;
                lineSteps = confEnemies[i][j+1] % 100;
                tmpDirection = (confEnemies[i][j+1] / 100) % 10;
                steps = allSteps + lineSteps + 1;
                if (tmpDirection == 4);
                else if (tmpDirection != 2) 
                {
                        steps++;
                        if (steps < minDistance)
                        {
                                minLineDistance = lineSteps;
                                minAllDistance = allSteps;
                                minDistance = steps;
                                minDistanceDirection = 2;
                                dirChange = 1;
                        }
                }
                else if (steps < minDistance)
                {
                        minLineDistance = lineSteps;
                        minAllDistance = allSteps;
                        minDistance = steps;
                        minDistanceDirection = 2;
                        dirChange = 0;
                }
                        
                allSteps = confEnemies[i-1][j] / 1000;
                lineSteps = confEnemies[i-1][j] % 100;
                tmpDirection = (confEnemies[i-1][j] / 100) % 10;
                steps = allSteps + lineSteps + 1;
                if (tmpDirection == 1);
                else if (tmpDirection != 3) 
                {
                        steps++;
                        if (steps < minDistance)
                        {
                                minLineDistance = lineSteps;
                                minAllDistance = allSteps;
                                minDistance = steps;
                                minDistanceDirection = 3;
                                dirChange = 1;
                        }
                }
                else if (steps < minDistance)
                {
                        minLineDistance = lineSteps;
                        minAllDistance = allSteps;
                        minDistance = steps;
                        minDistanceDirection = 3;
                        dirChange = 0;
                }
                        
                allSteps = confEnemies[i+1][j] / 1000;
                lineSteps = confEnemies[i+1][j] % 100;
                tmpDirection = (confEnemies[i+1][j] / 100) % 10;
                steps = allSteps + lineSteps + 1;
                if (tmpDirection == 3);
                else if (tmpDirection != 1) 
                {
                        steps++;
                        if (steps < minDistance)
                        {
                                minLineDistance = lineSteps;
                                minAllDistance = allSteps;
                                minDistance = steps;
                                minDistanceDirection = 1;
                                dirChange = 1;
                        }
                }
                else if (steps < minDistance)
                {
                        minLineDistance = lineSteps;
                        minAllDistance = allSteps;
                        minDistance = steps;
                        minDistanceDirection = 1;
                        dirChange = 0;
                }

                if (dirChange == 1)
                {
                        return (minDistance * 1000 + minDistanceDirection * 100);
                }
                else 
                {
                        return (minAllDistance * 1000 + minDistanceDirection * 100 + minLineDistance + 1001);
                }

        }
	
	private void calculateTreasurePotentials()
	{
		int i, j, k;
		
		for (k = 0; k < numGlobalSteps; k++)
		{
			for (i = 1; i <= wX; i++)
			{
				for (j = 1; j <= wY; j++)
				{ 
					if (localize[i][j])
					{
						confTreasures[i][j] = delta(i, j);
					}
				}
			}
		}
	}
	
	private void biteMyShinyMetalAss()
	{
		int i, j, k;
		
		for (k = 0; k < numDeathMatchSteps; k++)
		{
			for (i = 1; i <= wX; i++)
			{
				for (j = 1; j <= wY; j++)
				{ 
					if (localize[i][j])
					{
						confEnemies[i][j] = chumps(i, j);
					}
				}
			}
		}
	}
	
	private void calculateFireRanges()
	{
		int i;
		
		for (Robot r : robots)
		{
			if (r.getId() == robot.getId()) continue;
			
			int x = r.getPositionX();
			int y = r.getPositionY();
			
			if (r.getDirection().equals(Robot.Direction.DOWN))
			{				
				for (i = y; i < wY; i++)
				{
					if (!world.containsWall(x, i))
					{
						fireRanges[x][i] = true;
					}
					else
					{
						break;
					}
				}
				
				continue;
			}
			
			if (r.getDirection().equals(Robot.Direction.UP))
			{				
				for (i = y; i >= 0; i--)
				{
					if (!world.containsWall(x, i))
					{
						fireRanges[x][i] = true;
					}
					else
					{
						break;
					}
				}
				
				continue;
			}
			
			if (r.getDirection().equals(Robot.Direction.RIGHT))
			{				
				for (i = x; i < wX; i++)
				{
					if (!world.containsWall(i, y))
					{
						fireRanges[i][y] = true;
					}
					else
					{
						break;
					}
				}
				
				continue;
			}
			
			if (r.getDirection().equals(Robot.Direction.LEFT))
			{				
				for (i = x; i >= 0; i--)
				{
					if (!world.containsWall(i, y))
					{
						fireRanges[i][y] = true;
					}
					else
					{
						break;
					}
				}
				
				continue;
			}
		}
	}
	
	private int shootGain()
	{
		int i;
		
		int x = rX;
		int y = rY;
		
		if (robot.getDirection().equals(Robot.Direction.DOWN))
		{
			for (i = y + 1; i < wY; i++)
			{
				if (world.containsWall(x, i))
				{
					return 0;
				}
				else if (world.containsRobot(x, i))
				{
					if (world.getRobotOn(x, i).getId() == robot.getId())
					{
						return 0;
					}
					
					Robot r = world.getRobotOn(x, i);
					return r.getEnergy() / 2;
				}
			}
		}
		
		if (robot.getDirection().equals(Robot.Direction.UP))
		{			
			for (i = y - 1; i >= 0; i--)
			{
				if (world.containsWall(x, i))
				{
					return 0;
				}
				else if (world.containsRobot(x, i))
				{
					if (world.getRobotOn(x, i).getId() == robot.getId())
					{
						return 0;
					}
					
					Robot r = world.getRobotOn(x, i);
					return r.getEnergy() / 2;
				}
			}
		}
		
		if (robot.getDirection().equals(Robot.Direction.RIGHT))
		{
			for (i = x + 1; i < wX; i++)
			{
				if (world.containsWall(i, y))
				{
					return 0;
				}
				else if (world.containsRobot(i, y))
				{
					if (world.getRobotOn(i, y).getId() == robot.getId())
					{
						return 0;
					}
					
					Robot r = world.getRobotOn(i, y);
					return r.getEnergy() / 2;
				}
			}
		}
		
		if (robot.getDirection().equals(Robot.Direction.LEFT))
		{			
			for (i = x - 1; i >= 0; i--)
			{
				if (world.containsWall(i, y))
				{
					return 0;
				}
				else if (world.containsRobot(i, y))
				{
					if (world.getRobotOn(i, y).getId() == robot.getId())
					{
						return 0;
					}
					
					Robot r = world.getRobotOn(i, y);
					return r.getEnergy() / 2;
				}
			}
		}	
		
		return 0;
	}
	
	private Strategy.BotAction moveTo(Robot.Direction direction)
	{
		Robot.Direction d = robot.getDirection();
		
		if (d.equals(Robot.Direction.UP))
		{
			if (direction.equals(Robot.Direction.UP))
			{
                                if (world.containsRobot(rX, rY - 1))
                                        return Strategy.BotAction.SHOOT;
                                lastAction = Strategy.BotAction.STEP;
				return Strategy.BotAction.STEP;
			}
			if (direction.equals(Robot.Direction.RIGHT))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX + 1, pY).equals(Robot.Direction.UP) && (!world.containsTreasure(rX+1, rY)) && (pY > 1) && (!world.containsWall(rX, rY - 1)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX + 1, pY).equals(Robot.Direction.UP) && (pY > 1) && !world.containsWall(rX, rY - 1))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.RIGHT;
			}
			if (direction.equals(Robot.Direction.LEFT))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX - 1, pY).equals(Robot.Direction.UP) && (!world.containsTreasure(rX-1, rY)) && (pY > 1) && (!world.containsWall(rX, rY - 1)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX - 1, pY).equals(Robot.Direction.UP) && (pY > 1) && !world.containsWall(rX, rY - 1))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.LEFT;
			}
			if (direction.equals(Robot.Direction.DOWN))
			{
				return Strategy.BotAction.LEFT;
			}
		}
		
		if (d.equals(Robot.Direction.RIGHT))
		{
			if (direction.equals(Robot.Direction.UP))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX, pY - 1).equals(Robot.Direction.RIGHT) && (!world.containsTreasure(rX, rY - 1)) && (pX < wX) && (!world.containsWall(rX + 1, rY)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX, pY - 1).equals(Robot.Direction.RIGHT) && (pX < wX) && !world.containsWall(rX + 1, rY))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.LEFT;
			}
			if (direction.equals(Robot.Direction.RIGHT))
			{
                                if (world.containsRobot(rX + 1, rY))
                                        return Strategy.BotAction.SHOOT;
                                lastAction = Strategy.BotAction.STEP;
				return Strategy.BotAction.STEP;
			}
			if (direction.equals(Robot.Direction.LEFT))
			{
				return Strategy.BotAction.LEFT;
			}
			if (direction.equals(Robot.Direction.DOWN))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX, pY + 1).equals(Robot.Direction.RIGHT) && (!world.containsTreasure(rX, rY + 1)) && (pX < wX) && (!world.containsWall(rX + 1, rY)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX, pY + 1).equals(Robot.Direction.RIGHT) && (pX < wX) && !world.containsWall(rX + 1, rY))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.RIGHT;
			}
		}
		
		if (d.equals(Robot.Direction.DOWN))
		{
			if (direction.equals(Robot.Direction.UP))
			{
				return Strategy.BotAction.LEFT;
			}
			if (direction.equals(Robot.Direction.RIGHT))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX + 1, pY).equals(Robot.Direction.DOWN) && (!world.containsTreasure(rX+1, rY)) && (pY < wY) && (!world.containsWall(rX, rY + 1)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX + 1, pY).equals(Robot.Direction.DOWN) && (pY < wY) && !world.containsWall(rX, rY + 1))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.LEFT;
			}
			if (direction.equals(Robot.Direction.LEFT))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX - 1, pY).equals(Robot.Direction.DOWN) && (!world.containsTreasure(rX-1, rY)) && (pY < wY) && (!world.containsWall(rX, rY + 1)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX - 1, pY).equals(Robot.Direction.DOWN) && (pY < wY) && !world.containsWall(rX, rY + 1))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.RIGHT;
			}
			if (direction.equals(Robot.Direction.DOWN))
			{
                                if (world.containsRobot(rX, rY + 1))
                                        return Strategy.BotAction.SHOOT;
                                lastAction = Strategy.BotAction.STEP;
				return Strategy.BotAction.STEP;
			}
		}
		
		if (d.equals(Robot.Direction.LEFT))
		{
			if (direction.equals(Robot.Direction.UP))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX, pY - 1).equals(Robot.Direction.LEFT) && (!world.containsTreasure(rX, rY - 1)) && (pX > 1) && (!world.containsWall(rX - 1, rY)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX, pY - 1).equals(Robot.Direction.LEFT) && (pX > 1) && !world.containsWall(rX - 1, rY))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.RIGHT;
			}
			if (direction.equals(Robot.Direction.RIGHT))
			{
				return Strategy.BotAction.LEFT;
			}
			if (direction.equals(Robot.Direction.LEFT))
			{
                                if (world.containsRobot(rX - 1, rY))
                                        return Strategy.BotAction.SHOOT;
                                lastAction = Strategy.BotAction.STEP;
				return Strategy.BotAction.STEP;
			}
			if (direction.equals(Robot.Direction.DOWN))
			{
                            if (numTreasures != 0)
                                if (bestPotential2(pX, pY + 1).equals(Robot.Direction.LEFT) && (!world.containsTreasure(rX, rY + 1)) && (pX > 1) && (!world.containsWall(rX - 1, rY)))
                                        return Strategy.BotAction.STEP;
                            else
                                if (killAllHumans2(pX, pY + 1).equals(Robot.Direction.LEFT) && (pX > 1) && !world.containsWall(rX - 1, rY))
                                        return Strategy.BotAction.STEP;
				return Strategy.BotAction.LEFT;
			}
		}
		
		return Strategy.BotAction.REST; // should never happen
	}
	
	private Robot.Direction bestPotential2(int x, int y)
	{
		double maxPotential = 0.0;
		Robot.Direction direction = null;
		
		if (confTreasures[x - 1][y] > maxPotential)
		{
			direction = Robot.Direction.LEFT;
			maxPotential = confTreasures[x - 1][y];
		}		
		if (confTreasures[x + 1][y] > maxPotential)
		{
			direction = Robot.Direction.RIGHT;
			maxPotential = confTreasures[x + 1][y];
		}
		if (confTreasures[x][y + 1] > maxPotential)
		{
			direction = Robot.Direction.DOWN;
			maxPotential = confTreasures[x][y + 1];
		}
		if (confTreasures[x][y - 1] > maxPotential)
		{
			direction = Robot.Direction.UP;
			maxPotential = confTreasures[x][y - 1];
		}
		
		return direction;
	}

	private Robot.Direction bestPotential()
	{
		double maxPotential = 0.0;
		Robot.Direction direction = null;
		
		if (confTreasures[pX - 1][pY] > maxPotential)
		{
			direction = Robot.Direction.LEFT;
			maxPotential = confTreasures[pX - 1][pY];
		}		
		if (confTreasures[pX + 1][pY] > maxPotential)
		{
			direction = Robot.Direction.RIGHT;
			maxPotential = confTreasures[pX + 1][pY];
		}
		if (confTreasures[pX][pY + 1] > maxPotential)
		{
			direction = Robot.Direction.DOWN;
			maxPotential = confTreasures[pX][pY + 1];
		}
		if (confTreasures[pX][pY - 1] > maxPotential)
		{
			direction = Robot.Direction.UP;
			maxPotential = confTreasures[pX][pY - 1];
		}
		
		return direction;
	}

        public Robot.Direction killAllHumans2(int X, int Y)
        {
                int i, j;
                double minDistance = 99999999;
		Robot.Direction direction = Robot.Direction.RIGHT;
		
                if (confEnemies[X-1][Y] < minDistance)
                {
                        direction = Robot.Direction.LEFT;
                        minDistance = confEnemies[X - 1][Y];
                }
                if (confEnemies[X+1][Y] < minDistance)
                {
                        direction = Robot.Direction.RIGHT;
                        minDistance = confEnemies[X + 1][Y];
                }
                if (confEnemies[X][Y - 1] < minDistance)
                {
                        direction = Robot.Direction.UP;
                        minDistance = confEnemies[X][Y - 1];
                }
                if (confEnemies[X][Y + 1] < minDistance)
                {
                        direction = Robot.Direction.DOWN;
                        minDistance = confEnemies[X][Y + 1];
                }
                
                return direction;
        }
	
        public Strategy.BotAction killAllHumans(int shoot)
        {
                int i, j;
                double minDistance = 99999999;
		Robot.Direction direction = Robot.Direction.RIGHT;

                if (shoot > 0) return Strategy.BotAction.SHOOT;

                biteMyShinyMetalAss();
		
                if (confEnemies[pX-1][pY] < minDistance)
                {
                        direction = Robot.Direction.LEFT;
                        minDistance = confEnemies[pX - 1][pY];
                }
                if (confEnemies[pX+1][pY] < minDistance)
                {
                        direction = Robot.Direction.RIGHT;
                        minDistance = confEnemies[pX + 1][pY];
                }
                if (confEnemies[pX][pY - 1] < minDistance)
                {
                        direction = Robot.Direction.UP;
                        minDistance = confEnemies[pX][pY - 1];
                }
                if (confEnemies[pX][pY + 1] < minDistance)
                {
                        direction = Robot.Direction.DOWN;
                        minDistance = confEnemies[pX][pY + 1];
                }
                        /*System.out.println(confEnemies[pX-1][pY]+" ");
                        System.out.println(confEnemies[pX+1][pY]+" ");
                        System.out.println(confEnemies[pX][pY-1]+" ");
                        System.out.println(confEnemies[pX][pY+1]+"\n");*/
                
		
                /*for (i = 1;i <= wX;i++)
                {
                        for (j = 1;j<=wY;j++)
                                System.out.println(confEnemies[i][j]);
                        System.out.println();
                }*/
                return moveTo(direction);
        }
	
	public Strategy.BotAction decide() {
				
		// resting!
		if (robot.getEnergy() <= 1)
		{
			return Strategy.BotAction.REST;
		}

                if (world.getRound() == rounds)
                        return Strategy.BotAction.REST;
		
		update();	

                if (lastAction == (Strategy.BotAction.STEP) && (lastRX == rX && lastRY == rY) && !lastRested)
                {
                        lastRested = true;
                        return Strategy.BotAction.REST; 
                }
                lastRX = rX; lastRY = rY;
                lastAction = null;
                lastRested = false;
		
		// shooting!
		int shoot = shootGain();
		if (shoot > 2)  
		{
			calculateFireRanges();
			
			if (fireRanges[rX][rY])
			{
			//	if (shoot - robot.getEnergy() / 2 > 0)
			//	{
					return Strategy.BotAction.SHOOT;
			//	}
			}
			else
			{
				return Strategy.BotAction.SHOOT;
			}
		}
				
		if (world.containsTreasure(rX, rY))
		{
			return Strategy.BotAction.TAKE;
		}
		
		if (world.containsBattery(rX, rY))
		{
			if (robot.getEnergy() < world.getMaxRobotEnergy() - world.getRestRobotEnergy() - 1)
			{
				return Strategy.BotAction.RECHARGE;
			}
		}
		
                if (numTreasures == 0)
                {
                        return killAllHumans(shoot);
                }
                
		calculateTreasurePotentials();		
		Robot.Direction direction = bestPotential();		
		if (direction == null)
		{
			calculateTreasurePotentials();
			direction = bestPotential();
			if (direction == null)
			{
				calculateTreasurePotentials();
				direction = bestPotential();				
				if (direction == null)
				{
					calculateTreasurePotentials();
					direction = bestPotential();					
					if (direction == null)
					{
                                                numTreasures = 0; //ugly hack
                                                return killAllHumans(shoot);
						//return Strategy.BotAction.REST; // we are hopeless here
					}
				}
			}				
		}
			
		Strategy.BotAction action = moveTo(direction);
		return action;
	}
}
