/*
 * CPU.java
 *
 * Created on 15. duben 2006, 17:12
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.chobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;
import java.util.ArrayList;

/**
 *
 * @author Andy1
 */
public class CPU{
    
    public class Score{
        private int val;
        public Score(int i){
            val = i;
        }
        public int getVal(){
            return val;
        }
    }
    private ArrayList<ArrayList<Position>> treeData;
    private ArrayList<Score> treeScore;
    
    
    private boolean gatherOnlyMode;
    private boolean combatOnlyMode;
    private ChoBot strategy;
    
    
    private ArrayList<VirtualTreasure> reachableTreasures;
    private ArrayList<Position> reachableBatteries;
    private ArrayList<VirtualTreasure> treasureList;
    private ArrayList<VirtualOpponent> opponentList;
    private ArrayList<ArrayList<ArrayList<Position>>> pathList;
    
    private int searchDepth;
    
    /** Creates a new instance of CPU */
    public CPU(ChoBot strategy) {
        this.strategy = strategy;
        this.gatherOnlyMode = false;
        this.combatOnlyMode = false;
        this.searchDepth = 3;
        treeData = new ArrayList();
        treeScore = new ArrayList();
        reachableTreasures = new ArrayList();
        reachableBatteries = new ArrayList();
        opponentList = new ArrayList();
        treasureList = new ArrayList();
        pathList = new ArrayList();
    }
    
    public void initialize(){
        
        initOpponentList();
        initTreasureList();
        initBatteryList();
        updateTreasureList();
        checkGatherMode();
        
    }
    
    public BotAction process(){
        BotAction retValue=null;
        Robot me = strategy.getMe();
        World world = strategy.getWorld();
        int time = world.getRoundsCount() - world.getRound();
        
        
        if(time == 1 && world.containsBattery(me.getPositionX(), me.getPositionY())){
            return BotAction.RECHARGE;
        }
        
        updateTreasureList();
        ArrayList<Position>  bestPath;
        
        if(isGatherOnlyMode()){
            bestPath = optimalPathGather();
        } else{
            filterTreasureList();
            bestPath = optimalPathDefault();
        }
        
        //Rest if nothing else to do.
        if(bestPath == null){
            retValue = BotAction.REST;
        } else{
            retValue = this.resolvePath(bestPath);
        }
        
        //Rest when exhausted, but not when doing recharge.
        if(retValue!=BotAction.RECHARGE && world.getRestRobotEnergy() > 0 && me.getEnergy() < 2 && me.getEnergy() < world.getMaxRobotEnergy()) {            
            return BotAction.REST;
        }
        
        return retValue;        
    }
    
    public ArrayList<Position> optimalPathDefault(){
        ArrayList<Position> retValue=null;
        Robot me = strategy.getMe();
        World world = strategy.getWorld();        
        Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(),strategy.getMe().getDirection());
        PathFinder p = strategy.getPathfinder();
        int time = world.getRoundsCount() - world.getRound();
        
        treeData.clear();
        treeData.trimToSize();
        treeScore.clear();
        treeScore.trimToSize();           
        initOpponentList();
        
        int sum=0;
        for(VirtualTreasure t : treasureList){
            sum+= t.getValue();
        }

        //Compute all possible paths
        buildTreeDefault(opponentList,new ArrayList(),searchDepth, treasureList, new ArrayList(), sum, -1);
        
        //Compute direct path to nearest battery to recharge in the end of game.
        ArrayList<Position> rechargePath = findBestRecharge();
        if(rechargePath != null){
            int rechargePathETA = strategy.getPathfinder().getPathETA(rechargePath, strategy.getMe().getEnergy());
            if(time > rechargePathETA && time < rechargePathETA+5){
                treeData.add(rechargePath);
                treeScore.add(new Score(world.getMaxRobotEnergy()));
            }
        }
        
        //Choose path with best score/length ratio.
        int max=0;
        for(int i=0; i<treeData.size(); i++){
            
            if(treeData.get(i).size()>1 && containsEnemy(treeData.get(i).get(treeData.get(i).size()-2))){
                continue;
            }
            int length = strategy.getPathfinder().getPathETA(treeData.get(i), strategy.getMe().getEnergy());            
            double value = (double) treeScore.get(i).getVal()/length;
            length = strategy.getPathfinder().getPathETA(treeData.get(max), strategy.getMe().getEnergy());            
            double maxValue = (double) treeScore.get(max).getVal()/length;
            
            if(value > maxValue){
                max = i;                
            }            
        }
        
        //End-game resting                
        //Recovering does not take place while in the LOS of enemy.
        int  recoverTime = (world.getMaxRobotEnergy()-me.getEnergy())/(world.getRestRobotEnergy()-1)+1;
        boolean tryRest = true;
        for(VirtualOpponent o : opponentList){
            if(isInLineOfSight(MyPos, o))   tryRest = false;
        }       
        //Compares end-game-resting result to buildtree-max result in case the time is up.
        if(tryRest && time<=recoverTime && Math.min(world.getMaxRobotEnergy(), me.getEnergy()+treeScore.get(max).getVal()) < Math.min(world.getMaxRobotEnergy(), me.getEnergy()+time*(world.getRestRobotEnergy()-1))){            
            retValue = null;            
        } else{            
            retValue = treeData.get(max);            
        }
        
        return retValue;
        
    }
    
    protected void buildTreeDefault(ArrayList<VirtualOpponent> opponentListParam, ArrayList<VirtualTreasure> lost, int depth, ArrayList<VirtualTreasure> treasureList, ArrayList<Position> currentPath, int currentScore, int currentTreasureIndex){
        
        World world = strategy.getWorld();
        PathFinder p = strategy.getPathfinder();
        Robot me = strategy.getMe();
        Position myPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(),strategy.getMe().getDirection());
        int time = world.getRoundsCount() - world.getRound();
        
        if(depth ==0){            
            treeScore.add(treeScore.size(), new Score(currentScore));
            treeData.add(treeData.size(), currentPath);
            return;
        }        
        
        for(int i=0; i<opponentListParam.size();i++){
            
            ArrayList<VirtualOpponent> nextOpponentList = new ArrayList();
            for(int j =0; j<opponentListParam.size(); j++){
                nextOpponentList.add((VirtualOpponent)opponentListParam.get(j).clone());
            }
            VirtualOpponent o = nextOpponentList.get(i);
            ArrayList<Position> nextPathParam=null;
            ArrayList<Position> path = new ArrayList();
            int value=0;
            Position currentPos;
            if(currentPath.size()>0){
                currentPos = currentPath.get(0);
            } else{
                currentPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(),strategy.getMe().getDirection());
            }
            
            if(isInLineOfSight(currentPos, o)){
                ArrayList<Position> pathSuffix;
                Position combatPosition = findCombatPosition(currentPos, o.getRobot());
                if(currentPos.equals(combatPosition)){
                    pathSuffix=new ArrayList();
                    path.add(combatPosition);
                } else{
                    pathSuffix = p.findPath(currentPos, combatPosition);
                    path.add((Position)pathSuffix.get(0).clone());
                }
                path.get(0).setShootAction(true);
                path.addAll((ArrayList<Position>)pathSuffix.clone());
                int hisE =o.getEnergy() - p.getPathETA(path, me.getEnergy())-1;
                if(hisE < 1)       hisE = world.getRestRobotEnergy();
                if(hisE < 3)       hisE = 0;
                value = hisE/2;
                o.setEnergy(Math.max(1,hisE - hisE/2));
                nextPathParam = path;
                nextPathParam.addAll((ArrayList<Position>) currentPath.clone());
                
            } else{
                Position target = new Position(o.getX(), o.getY());
                ArrayList<Position> pathSuffix = p.findPath(currentPos, target);                
                path.add((Position)pathSuffix.get(0).clone());
                path.get(0).setShootAction(true);
                path.addAll((ArrayList<Position>)pathSuffix.clone()); 
                int hisE = o.getEnergy() - p.getPathETA(pathSuffix, p.expectedEnergy(currentPath,me.getEnergy()))-1;
                if(hisE < 1)                hisE = world.getRestRobotEnergy();
                if(hisE < 3)           hisE = 0;                
                value = hisE/2;             
                o.setEnergy(Math.max(1,hisE - hisE/2));
                nextPathParam = path;
                nextPathParam.addAll((ArrayList<Position>) currentPath.clone());
            }
            
            int nextScoreParam = currentScore + value;
            ArrayList<VirtualTreasure> nextTreasureListParam = new ArrayList();
            ArrayList<VirtualTreasure> nextLost = (ArrayList<VirtualTreasure>) lost.clone();
            for(VirtualTreasure t : treasureList){
                if(currentPos.equals(t)){
                    continue;
                }
                VirtualTreasure newT = (VirtualTreasure) t.clone();
                newT.setPathFromPlayer(p.findOptimizedPath(nextPathParam.get(0), newT, p.expectedEnergy(nextPathParam, me.getEnergy() )));
                newT.setLifeTime(Math.round(newT.getLifeTime()-p.getPathETA(newT.getPathFromPlayer(), strategy.getMe().getEnergy() )*newT.getDeathSpeed()));
                
                if(newT.getLifeTime()>(p.getPathETA(newT.getPathFromPlayer(),p.expectedEnergy(nextPathParam, me.getEnergy()) ))){
                    nextTreasureListParam.add(newT);
                } else{
                    nextScoreParam = nextScoreParam - /*Math.round(Math.max(0.5f,newT.getDeathSpeed())**/Math.max(1,newT.getValue()/*/4)*/);//treasureList.size();
                    nextLost.add(newT);
                }                
            }

            buildTreeDefault(nextOpponentList, nextLost, depth-1, nextTreasureListParam, nextPathParam, nextScoreParam, -1);
        }
        
        for(int i=0; i<reachableTreasures.size(); i++){
            VirtualTreasure treasure = reachableTreasures.get(i);
            
            if(treasureList.contains(treasure)){
                
                int nextScoreParam = currentScore + treasure.getValue();
                
                ArrayList<Position> nextPathParam=null;
                if(currentTreasureIndex == -1){
                    Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(),strategy.getMe().getDirection());
                    if(currentPath.size()>0){
                        MyPos = currentPath.get(0);
                    }
                    
                    boolean doOptimization = true;
                    for(VirtualOpponent o : opponentList){
                        if(isInLineOfSight(treasure, o)){
                            doOptimization = false;
                            break;
                        }
                    }
                    if(treasure.getValue()<world.getMaxRobotEnergy()/4){
                        doOptimization =  false;
                    }
                    
                    if(doOptimization){
                        nextPathParam = strategy.getPathfinder().findOptimizedPath(MyPos, treasure, me.getEnergy());
                    } else{
                        nextPathParam = strategy.getPathfinder().findPath(MyPos, treasure);
                    }
                    
                    nextPathParam.addAll(currentPath);
                }
                
                else{
                    if(pathList.get(currentTreasureIndex).get(i)==null){
                        Position source = new Position(reachableTreasures.get(currentTreasureIndex), Direction.UP);
                        pathList.get(currentTreasureIndex).add(i, strategy.getPathfinder().findPath(source, treasure));
                    }
                    nextPathParam = (ArrayList<Position>) pathList.get(currentTreasureIndex).get(i).clone();
                    
                    boolean doOptimization = true;
                    for(VirtualOpponent o : opponentList){
                        if(isInLineOfSight(treasure, o)){
                            doOptimization = false;
                            break;
                        }
                    }
                    if(treasure.getValue()<world.getMaxRobotEnergy()/4){
                        doOptimization =  false;
                    }
                    
                    if(doOptimization){
                        nextPathParam = strategy.getPathfinder().optimizePath(nextPathParam, strategy.getPathfinder().expectedEnergy(currentPath, strategy.getMe().getEnergy()));
                    }
                    
                    
                    if(nextPathParam.size()>1)
                        nextPathParam.remove(nextPathParam.size()-1);
                    if(nextPathParam.size()>1)
                        nextPathParam.remove(nextPathParam.size()-1);
                    //if(nextPathParam.size()>1)
                    //   nextPathParam.remove(nextPathParam.size()-1);
                    
                    ArrayList<Position> pathJoint=strategy.getPathfinder().findPath(currentPath.get(0), nextPathParam.get(nextPathParam.size()-1));
                    pathJoint.remove(0);
                    nextPathParam.addAll(pathJoint);
                    
                    nextPathParam.addAll((ArrayList<Position>)currentPath.clone());
                }
                
                
                ArrayList<VirtualTreasure> nextTreasureListParam = new ArrayList();
                ArrayList<VirtualTreasure> nextLost = (ArrayList<VirtualTreasure>) lost.clone();
                for(VirtualTreasure t : treasureList){
                    
                    VirtualTreasure newT = (VirtualTreasure) t.clone();
                    
                    newT.setPathFromPlayer(p.findOptimizedPath(nextPathParam.get(0), newT, strategy.getPathfinder().expectedEnergy(nextPathParam, strategy.getMe().getEnergy() )));
                    
                    newT.setLifeTime(Math.round(newT.getLifeTime()-p.getPathETA(newT.getPathFromPlayer(), p.expectedEnergy(currentPath, me.getEnergy() ))*newT.getDeathSpeed()));
                    
                    if(!t.equals(treasure)){
                        if(newT.getLifeTime()>p.getPathETA(newT.getPathFromPlayer(),p.expectedEnergy(nextPathParam, strategy.getMe().getEnergy()) )){
                            nextTreasureListParam.add(newT);
                        } else{
                            nextScoreParam = nextScoreParam-/*Math.round(Math.max(0.5f,newT.getDeathSpeed())**/Math.max(1, newT.getValue()/*/4)*/);//treasureList.size();
                            nextLost.add(newT);
                        }
                    }
                }
                
                boolean tryRest = true;
                for(VirtualOpponent o : opponentList){
                    if(isInLineOfSight(myPos, o))   tryRest = false;
                }
                
                if(tryRest && strategy.getPathfinder().getPathETA(nextPathParam,strategy.getMe().getEnergy()) > time){
                    if(depth>1){
                        nextScoreParam = currentScore;
                    } else{
                        int restTime = time - p.getPathETA(nextPathParam,me.getEnergy());
                        nextScoreParam = currentScore+Math.min(world.getMaxRobotEnergy(), me.getEnergy()+time*(world.getRestRobotEnergy()-1));
                    }                    
                } 
                
                ArrayList<VirtualOpponent> nextOpponentListParam = (ArrayList<VirtualOpponent>) opponentListParam.clone();
                for(VirtualOpponent o : nextOpponentListParam){
                    o.setEnergy( Math.max(1, o.getEnergy() - p.getPathETA(nextPathParam, me.getEnergy())+p.getPathETA(currentPath, me.getEnergy())));
                }
                
                buildTreeDefault(nextOpponentListParam, nextLost, depth-1, nextTreasureListParam, nextPathParam, nextScoreParam, i);
              
            }
        }       
    }
    
    private ArrayList<Position> optimalPathGather(){
        
        Robot me = strategy.getMe();
        World world = strategy.getWorld();
        
        ArrayList<Position> initialPath =new ArrayList();
        Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(),strategy.getMe().getDirection());
        
        initialPath.add(MyPos);
        ArrayList<Position> retValue=initialPath;
        treeScore.clear();
        treeScore.trimToSize();
        treeData.clear();
        treeData.trimToSize();
        
        int time = world.getRoundsCount() - world.getRound();
        ArrayList<Position> rechargePath = findBestRecharge();
        int rechargePathETA = strategy.getPathfinder().getPathETA(rechargePath, strategy.getMe().getEnergy());
        
        if(time > rechargePathETA && time < rechargePathETA+5){
            treeData.add(rechargePath);
            treeScore.add(new Score(world.getMaxRobotEnergy()));
        }
        
        
        
        
        buildTreeGather(searchDepth-1, treasureList, initialPath, 0, -1);
        
        int max=0;
        for(int i=0; i<treeData.size(); i++){            
            int length =treeData.get(i).size();
            double value = (double) treeScore.get(i).getVal()/length;            
            length =treeData.get(max).size();
            double maxValue = (double) treeScore.get(max).getVal()/length;
            
            if(value > maxValue){
                max = i;
                length = strategy.getPathfinder().getPathETA(treeData.get(max), strategy.getMe().getEnergy());                
            }
        }
        
        
        int  recoverTime = (world.getMaxRobotEnergy()-me.getEnergy())/(world.getRestRobotEnergy()-1)+1;
        if(time<=recoverTime && treeScore.get(max).getVal() <= Math.min(world.getMaxRobotEnergy(), me.getEnergy()+time*(world.getRestRobotEnergy()-1))){            
            retValue = null;
        } else{
            retValue = treeData.get(max);            
        }
        
        return retValue;
    }
    
    protected void buildTreeGather(int depth, ArrayList<VirtualTreasure> treasureList, ArrayList<Position> currentPath, int currentScore, int currentTreasureIndex){
        
        World world = strategy.getWorld();
        Robot me = strategy.getMe();
        int time = world.getRoundsCount() - world.getRound();
        if(depth ==0 || treasureList.isEmpty()){
            
            treeScore.add(treeScore.size(), new Score(currentScore));
            treeData.add(treeData.size(), currentPath);
            return;
        }        
        
        for(int i=0; i<reachableTreasures.size(); i++){
            VirtualTreasure treasure = reachableTreasures.get(i);
            
            if(treasureList.contains(treasure)){
                
                int nextScoreParam = currentScore + treasure.getValue();
                
                ArrayList<Position> nextPathParam=null;
                if(currentTreasureIndex == -1){
                    Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(),strategy.getMe().getDirection());
                    nextPathParam = strategy.getPathfinder().findOptimizedPath(MyPos, treasure, strategy.getMe().getEnergy());
                }
                
                else{
                    if(pathList.get(currentTreasureIndex).get(i)==null){
                        Position source = new Position(reachableTreasures.get(currentTreasureIndex), Direction.UP);
                        pathList.get(currentTreasureIndex).add(i, strategy.getPathfinder().findPath(source, treasure));
                    }
                    nextPathParam = (ArrayList<Position>) pathList.get(currentTreasureIndex).get(i).clone();
                    
                    nextPathParam = strategy.getPathfinder().optimizePath(nextPathParam, strategy.getPathfinder().expectedEnergy(currentPath, strategy.getMe().getEnergy()));
                    if(nextPathParam.size()>1)
                        nextPathParam.remove(nextPathParam.size()-1);
                    if(nextPathParam.size()>1)
                        nextPathParam.remove(nextPathParam.size()-1);
                    if(nextPathParam.size()>1)
                        nextPathParam.remove(nextPathParam.size()-1);
                    
                    ArrayList<Position> pathJoint=strategy.getPathfinder().findPath(currentPath.get(0), nextPathParam.get(nextPathParam.size()-1));
                    pathJoint.remove(0);
                    nextPathParam.addAll(pathJoint);
                    
                    nextPathParam.addAll((ArrayList<Position>)currentPath.clone());
                }
                ArrayList<VirtualTreasure> nextTreasureListParam = (ArrayList<VirtualTreasure>) treasureList.clone();
                nextTreasureListParam.remove(treasure);
                
                if(strategy.getPathfinder().getPathETA(nextPathParam,strategy.getMe().getEnergy()) > time){
                    if(depth>1){
                        nextScoreParam = currentScore;
                    } else{
                        int restTime = time - strategy.getPathfinder().getPathETA(nextPathParam,strategy.getMe().getEnergy());
                        nextScoreParam = currentScore+Math.min(world.getMaxRobotEnergy(), me.getEnergy()+time*(world.getRestRobotEnergy()-1));
                    }                    
                }
                
                buildTreeGather(depth-1, nextTreasureListParam, nextPathParam, nextScoreParam, i);
            }
        }
    }
    
    private void initTreasureList(){
        World world = strategy.getWorld();
        ArrayList<VirtualTreasure> treasures = new ArrayList();
        treasures.clear();
        for(int i=0; i<world.getDimensionY();i++){
            for(int j=0; j<world.getDimensionX();j++){
                if(world.containsTreasure(j,i)){
                    VirtualTreasure t = new VirtualTreasure(j,i, world.getTreasureValue(j,i),strategy);
                    Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(), strategy.getMe().getDirection());
                    ArrayList<Position> path = strategy.getPathfinder().findPath(MyPos, t);
                    if(path != null){                        
                        reachableTreasures.add(reachableTreasures.size(), t);
                    }
                }
            }
        }
        for(int i=0; i<reachableTreasures.size(); i++){
            pathList.add(new ArrayList());
            for(int j=0; j<reachableTreasures.size(); j++){
                pathList.get(i).add(null);
            }
        }
    }
    
    private void initBatteryList(){
        World world = strategy.getWorld();
        for(int i=0; i<world.getDimensionY();i++){
            for(int j=0; j<world.getDimensionX();j++){
                if(world.containsBattery(j,i)){
                    Position b = new Position(j,i);
                    Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(), strategy.getMe().getDirection());
                    ArrayList<Position> path = strategy.getPathfinder().findPath(MyPos, b);
                    if(path != null){
                        reachableBatteries.add(reachableBatteries.size(), b);
                    }
                }
            }
        }
    }
    
    public void updateTreasureList(){
        World world = strategy.getWorld();
        treasureList.clear();
        
        for(VirtualTreasure t : reachableTreasures){
            if(world.containsTreasure(t.getX(), t.getY())){
                if(!containsEnemy(t.getX(), t.getY())){
                    treasureList.add(treasureList.size(), t);
                }
            }
        }
    }
    
    public void initOpponentList(){
        opponentList.clear();
        for(Robot r : strategy.getWorld().getRobots()){
            if(opponentList.size()>3){
                break;
            }
            if(!r.equals(strategy.getMe())){
                opponentList.add(new VirtualOpponent( strategy, r));
            }
        }
        while(opponentList.size()<3){
            opponentList.add(new VirtualOpponent(strategy, opponentList.get(0).getRobot()));
        }
    }
    
    public void filterTreasureList(){
        
        World world = strategy.getWorld();
        Robot me = strategy.getMe();
        PathFinder pathfinder= strategy.getPathfinder();
        ArrayList<VirtualTreasure> newTreasureList = new ArrayList();
        for(VirtualTreasure treasure : treasureList){
            if(newTreasureList.size()>10){
                int difX, difY;
                difX = Math.abs(treasure.getX() - me.getPositionX());
                difY = Math.abs(treasure.getY() - me.getPositionY());
                if(difX+difY > 10){
                    continue;
                }
                if(newTreasureList.size()>30){
                    continue;
                }
            }
            VirtualTreasure newT = (VirtualTreasure) treasure.clone();
            Position myPosition = new Position(me.getPositionX(), me.getPositionY(), me.getDirection());
            
            ArrayList<Position> path = pathfinder.findOptimizedPath(myPosition, newT, me.getEnergy());
            if(path == null){
                continue;
            }
            newT.setPathFromPlayer(path);
            
            for(int k=0; k<opponentList.size(); k++){
                
                path = strategy.getPathfinder().findOptimizedPath(opponentList.get(k), newT, opponentList.get(k).getRobot().getEnergy());
                
                newT.setPathFromOpponent(k, path);
            }
            newT.validateLifeTime();
            newT.validateDeathSpeed();
            
            if(newT.getLifeTime()+2>strategy.getPathfinder().getPathETA(newT.getPathFromPlayer(), me.getEnergy())){
                newTreasureList.add(newT);
            }
            
        }
        this.treasureList.clear();
        this.treasureList = newTreasureList;
    }
    
    public BotAction resolvePath(ArrayList<Position> path){
        
        if(path.size()<1){
            return null;
        }
 
        Robot me = strategy.getMe();
        World world = strategy.getWorld();
        if(path.get(path.size()-1).getX() !=me.getPositionX()){
            throw new IllegalArgumentException("corrupt DATA X "+path.get(path.size()-1).getX()+" "+me.getPositionX());
        }
        if(path.get(path.size()-1).getY() !=me.getPositionY()){
            throw new IllegalArgumentException("corrupt DATA Y"+path.get(path.size()-1).getY()+" "+me.getPositionY());
        }
        if(path.get(path.size()-1).getDirection() !=me.getDirection()){
            throw new IllegalArgumentException("corrupt DATA D"+path.get(path.size()-1).getDirection()+" "+me.getDirection());
        }
        if(path.get(path.size()-1).isShootAction()){
            return BotAction.SHOOT;
        }
        if(path.size()<2){
            return null;
        }
        if(path.get(path.size()-2).equals(path.get(path.size()-1))){
           
            if(world.containsBattery(me.getPositionX(), me.getPositionY())){
                return BotAction.RECHARGE;
            }
            return BotAction.TAKE;
        }
        Position nextPosition = path.get(path.size()-2);
        int xDif = Math.abs(nextPosition.getX() - me.getPositionX());
        int yDif = Math.abs(nextPosition.getY() - me.getPositionY());
        int dDif = (nextPosition.getDirection().ordinal() - me.getDirection().ordinal());
        dDif=(dDif==3?-1:dDif);
        dDif=(dDif==-3?1:dDif);
        
        if(dDif==-1 && xDif ==0 && yDif==0) return BotAction.LEFT;
        else if(dDif==1 && xDif ==0 && yDif==0) return BotAction.RIGHT;
        else if(dDif==0 && xDif==1 && yDif == 0) return BotAction.STEP;
        else if(dDif==0 &&xDif==0 && yDif == 1) return BotAction.STEP;
        //else if(dDif==2 || dDif==-2)return BotAction.SHOOT;
        else throw new IllegalArgumentException("corrupt path "+dDif+" "+xDif+" "+yDif);
    }
    
    public ArrayList<VirtualTreasure> getTreasureList(){
        return treasureList;
    }
    
    public ArrayList<VirtualTreasure> getReachableTreasureList(){
        return reachableTreasures;
    }
    
    public ArrayList<Position> getReachableBatteryList(){
        return reachableBatteries;
    }
    
    public ArrayList<VirtualOpponent> getOpponentList(){
        return opponentList;
    }
    
    public ChoBot getStrategy() {
        return this.strategy;
    }
    
    public boolean isGatherOnlyMode(){
        return gatherOnlyMode;
    }
        
    
    protected void finalize() throws Throwable {
    }
      
    private void checkGatherMode(){
        gatherOnlyMode=true;
        for(int k=0; k<opponentList.size(); k++){
            Position MyPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(), strategy.getMe().getDirection());
            ArrayList<Position> path = strategy.getPathfinder().findPath(opponentList.get(k), MyPos);
            if(path != null){
                gatherOnlyMode = false;
                return;
            }
        }
    }
    
    private ArrayList<Position> findBestRecharge(){
        Position myPos = new Position(strategy.getMe().getPositionX(), strategy.getMe().getPositionY(), strategy.getMe().getDirection());
        PathFinder p = strategy.getPathfinder();
        ArrayList<Position> retValue = null;
        int bestETA = Integer.MAX_VALUE;
        for(Position b : reachableBatteries){
            int myEnergy = strategy.getMe().getEnergy();
            ArrayList<Position> path = p.findPath(myPos, b);
            int eta = p.getPathETA(path, myEnergy);
            if(eta < bestETA){
                bestETA = eta;
                retValue = path;
            }
        }
        return retValue;
    }
        
    
    protected Position findCombatPosition(Position source, Robot target){
        if(source.getX() == target.getPositionX() && source.getY() == target.getPositionY()){
            return source;
        }
        if(source.getX() < target.getPositionX()){
            return new Position(source.getX(), source.getY(), Direction.RIGHT);
        } else if(source.getX() > target.getPositionX()){
            return new Position(source.getX(), source.getY(), Direction.LEFT);
        } else if(source.getY() < target.getPositionY()){
            return new Position(source.getX(), source.getY(), Direction.DOWN);
        } else {
            return new Position(source.getX(), source.getY(), Direction.UP);
        }
    }
    
    protected Position findCombatPosition(Robot source, Robot target){
        return findCombatPosition(new Position(source.getPositionX(), source.getPositionY(), source.getDirection()), target);
    }
    
    protected boolean isInLineOfSight(Position p1, Position p2){
        World world = this.strategy.getWorld();
        if(p1.getY() == p2.getY()){
            int signum = p1.getX() < p2.getX()? 1 : -1;
            int x = p1.getX();
            int y = p1.getY();
            int i =0;
            while(!world.containsWall(x+i*signum, y) && (x+i*signum != p2.getX())){
                i++;
            }
            if(x+i*signum == p2.getX()) return true;
            else return false;
        } else if(p1.getX() == p2.getX()){
            int signum = p1.getY() < p2.getY()? 1 : -1;
            int x = p1.getX();
            int y = p1.getY();
            int i =0;
            while(!world.containsWall(x, y+i*signum) && y+i*signum != p2.getY()){
                i++;
            }
            if(y+i*signum == p2.getY()) return true;
            else return false;
        } else return false;
    }
    
     
    public ArrayList<ArrayList<Position>> getTreeData(){
        return treeData;
    }
    
    public ArrayList<Score> getTreeScore(){
        return treeScore;
    }
    
    public boolean containsEnemy(int x, int y){
        World world = strategy.getWorld();
        Robot robot = world.getRobotOn(x, y);
        Robot me = strategy.getMe();
        if(robot == null){
            return false;
        }
        if(robot.getId() == me.getId()){
            return false;
        }
        return true;
    }
    public boolean containsEnemy(Position p){
        return containsEnemy(p.getX(), p.getY());
    }
    
    public void setSearchDepth(int d){
        if(d>0) this.searchDepth = d;
    }
    
    public int getSearchDepth(){
        return searchDepth;
    }
}
