/*
 * PathFinder.java
 *
 * Created on 31. bezen 2006, 15:44
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.chobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.World;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Set;

/**
 *
 * @author Andy1
 */
public class PathFinder{
    
    private ArrayList<Position> batteryList;
    private ChoBot strategy;
    
    public PathFinder(ChoBot strategy){
        
        this.strategy=strategy;
        batteryList = new ArrayList();
        World world = strategy.getWorld();
        
        updateBatteryList();
        
    }
    
    public ArrayList<Position> findPath(Position start, Position target){
        
        World world = strategy.getWorld();
        PriorityQueue<SearchNode> queue = new PriorityQueue();
        Set<SearchNode> visited = new HashSet();
        SearchNode root = new SearchNode(start,target, null);
        queue.offer(root);
        int i=0;
        while(!queue.isEmpty()&& !queue.peek().getValue().equals(target) &&  i<5000){
            i++;
            SearchNode step = queue.poll();            
            SearchNode newNode;
            
            newNode = step.spawnForward(strategy);
            if(newNode != null && !visited.contains(newNode)){
                queue.offer(newNode);
                visited.add(newNode);
            }
            
            newNode = step.spawnLeft();
            if(newNode != null && !visited.contains(newNode)){
                queue.offer(newNode);
                visited.add(newNode);
            }
            
            newNode = step.spawnRight();
            if(newNode != null && !visited.contains(newNode)){
                queue.offer(newNode);
                visited.add(newNode);
            }
            
            
        }
        
        ArrayList<Position> path = null;
        if(queue.isEmpty()){
            return path;
        }
        if(queue.peek().getValue().equals(target)){
            SearchNode trace = queue.peek();
            path = new ArrayList();
            while(trace!=null){
                path.add(trace.getValue());
                
                trace = trace.getParent();
            }
            
            
        }
        
        
        return path;
    }
       
    public ArrayList<Position> optimizePath(ArrayList<Position> path, int energy){
        if(path == null){
            return null;
        }        
        Position start = path.get(path.size()-1);
        World world = strategy.getWorld();
        ArrayList<Position> minPath = path;
        int minETA;
        if(strategy.getCPU().isGatherOnlyMode()){
            minETA = this.getPathETA(path, energy) - expectedEnergy(path, energy);
        }
        else{            
            minETA = this.getPathETA(path, energy) - Math.min(4,expectedEnergy(path, energy));
        }
        
        ArrayList<Position> batterySubList = reduceBatteryList(this.batteryList, path);
        Position resultBattery=null;        
        
        for(Position battery : batterySubList){
            
            ArrayList<Position> newPath1 = findPath(start, battery);
            if(newPath1 == null){
                continue;
            }
            ArrayList<Position> newPath2 = findPath(newPath1.get(0), path.get(0));
            int thisETA;
            if(strategy.getCPU().isGatherOnlyMode()){
                thisETA =this.getPathETA(newPath1, energy) + this.getPathETA(newPath2, world.getMaxRobotEnergy()) - expectedEnergy(newPath2, world.getMaxRobotEnergy());
            }
            else{                
                thisETA =this.getPathETA(newPath1, energy) + this.getPathETA(newPath2, world.getMaxRobotEnergy()) - Math.min(4, expectedEnergy(newPath2, world.getMaxRobotEnergy()));
            }
            if(thisETA < minETA){
                minETA = thisETA;
                newPath2.addAll(newPath1);
                minPath = newPath2;
                resultBattery = battery;
            }
            
        }
        return minPath;
    }
    
    public ArrayList<Position> findOptimizedPath(Position start, Position target, int energy){
        ArrayList<Position> retValue = this.findPath(start, target);
        retValue = this.optimizePath(retValue, energy);
        return retValue;
    }
    
    
    public int getPathETA(List<Position> path, int energy){
        if(path == null){
            return 0;
        }
        int spareEnergy = energy-1;
        int rest = strategy.getWorld().getRestRobotEnergy();
        int distance = path.size()-1;
        if(spareEnergy >= distance){
            return distance;
        }
        
        int cost=0;
        List<Position> path2 = path;
        for(int i=0; i<path.size(); i++){
            if(strategy.getWorld().containsBattery(path.get(i).getX(),path.get(i).getY())){
                if(i+1<path.size() && path.get(i).equals(path.get(i+1))){
                    path2=path.subList(0, i);
                    distance-=i+1;
                    cost = getPathETA(path2, strategy.getWorld().getMaxRobotEnergy());
                    break;
                }
            }
        } 
        
        while(spareEnergy < distance){
            spareEnergy+=rest-1;
            cost++;
        }
        return distance+cost;
    }
    
    public int expectedEnergy(ArrayList<Position> path, int initialEnergy){
        
        int spareEnergy = initialEnergy-1;
        List<Position> path2 = path;
        for(int i=0; i<path.size(); i++){
            if(strategy.getWorld().containsBattery(path.get(i).getX(),path.get(i).getY())){
                if(i+1<path.size() && path.get(i).equals(path.get(i+1))){
                    path2=path.subList(0, i);
                    spareEnergy = strategy.getWorld().getMaxRobotEnergy()-1;
                    break;
                }
            }
        }
        
        int rest = strategy.getWorld().getRestRobotEnergy();
        int distance = path2.size()-1;
        if(spareEnergy >= distance){
            return spareEnergy - distance + 1;
        }
        int cost=0;
        while(spareEnergy < distance){
            spareEnergy+=rest-1;
            cost++;
        }
        return spareEnergy - distance + 1;
    }
    
    public ArrayList<Position> reduceBatteryList(ArrayList<Position> batteryList, ArrayList<Position> path){
        Robot me = strategy.getMe();
        if(path.size()<2){
            return new ArrayList();
        }
        ArrayList<Position> newList = new ArrayList();
        int pathETA = getPathETA(path, me.getEnergy());
        
        for(Position battery : batteryList){
            
            int x1=Math.abs(path.get(0).getX()-battery.getX());
            int y1=Math.abs(path.get(0).getY()-battery.getY());
            int x2=Math.abs(path.get(path.size()-1).getX()-battery.getX());
            int y2=Math.abs(path.get(path.size()-1).getY()-battery.getY());
            
            if(x1+y1+x2+y2 < pathETA+2){
                if(newList.size()<5){
                    newList.add(battery);
                }
                
                
            }
        }
        return newList;
    }
    
    public void updateBatteryList(){
        
        this.batteryList = strategy.getCPU().getReachableBatteryList();
    }
    
    public ArrayList<Position> getBatteryList(){
        return this.batteryList;
    }
    
    
    public ChoBot getStrategy(){
        return strategy;
    }
      
}
