/*
 * SearchNode.java
 *
 * Created on 31. bezen 2006, 18:12
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.chobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.World;

/**
 *
 * @author Andy1
 */
public class SearchNode implements Comparable{
    
    private Position value;
    private SearchNode parent;
    private int cost;
    private Position target;
    
    
    /** Creates a new instance of SearchNode */
    
    public SearchNode(Position p, Position target, SearchNode parent) {
        this.value = p;
        this.parent = parent;
        this.target = target;
        if(parent!=null){
            this.cost = parent.getCost()+1;            
        }
        else{
            this.cost = 0;
        }
    }
        
    public SearchNode(int x, int y, Direction direction,Position target, SearchNode parent) {
        this( new Position(x,y, direction),target, parent);
    }
    
    public SearchNode(int x, int y, Position target, SearchNode parent) {
        this( new Position(x,y),target, parent);
    }
    


    public boolean equals(Object o) {
       if(o instanceof SearchNode){
                    SearchNode p = (SearchNode) o;                    
                    if(p.value.getX()==value.getX() 
                    && p.value.getY()==value.getY() 
                    && p.value.getDirection()==value.getDirection())
                        return true;
       }
       return false;
    }

    public int compareTo(Object o) {
        
       if(o instanceof SearchNode){      
           
           SearchNode node = (SearchNode) o;                      
           int x = Math.abs(this.getValue().getX()-target.getX());
           int y = Math.abs(this.getValue().getY()-target.getY());  
           int dirDif;
           if(target.getDirection()!=null){
                dirDif=Math.abs(value.getDirection().ordinal() - target.getDirection().ordinal());
           }
           else{
             dirDif=0;
           }
           double distance1 = this.cost+x+y+(dirDif==3?1:dirDif);           
           x = Math.abs(node.getValue().getX()-node.getTarget().getX());
           y = Math.abs(node.getValue().getY()-node.getTarget().getY());
           if(target.getDirection()!=null){
                dirDif=Math.abs(node.getValue().getDirection().ordinal() - node.getTarget().getDirection().ordinal());
           }
           else{
               dirDif=0;
           }
           
           double distance2 = node.cost+x+y+(dirDif==3?1:dirDif);
           
           if(distance1 < distance2){
                return -1;
           }
           else if (distance1 == distance2){
               if(this.cost>node.cost) return -1;
               if(this.cost<node.cost) return 1;
               return 0;
           }
           else {
                return 1;
           }
       }
       return 0;
       //throw new ClassCastException("SearchNode instance compared to incompatible class type.");
    }

    public int hashCode() {
        int retValue;
        
        retValue = this.value.hashCode();
        return retValue;
    }

    public int getCost() {
        return cost;
    }

    public Position getValue() {
        return this.value;
    }
    
    public Position getTarget() {
        return this.target;
    }
    
    public SearchNode spawnLeft(){
        int x=value.getX();
        int y=value.getY();
        switch(value.getDirection().ordinal()){
                    case 0:  return new SearchNode(x,y,Direction.DOWN,target, this);
                    case 1:  return new SearchNode(x,y,Direction.LEFT,target, this);
                    case 2:  return new SearchNode(x,y,Direction.UP,target, this);
                    case 3:  return new SearchNode(x,y,Direction.RIGHT,target, this);                
        }
        return null;
    }
    public SearchNode spawnRight(){
        int x=value.getX();
        int y=value.getY();
        switch(value.getDirection().ordinal()){
                    case 0:  return new SearchNode(x,y,Direction.UP,target, this);
                    case 1:  return new SearchNode(x,y,Direction.RIGHT,target, this);
                    case 2:  return new SearchNode(x,y,Direction.DOWN,target, this);
                    case 3:  return new SearchNode(x,y,Direction.LEFT,target, this);              
        }
        return null;
    }
    public SearchNode spawnForward(ChoBot strategy){
        
        int x=value.getX();
        int y=value.getY();
        Direction dir = value.getDirection();
        World world = strategy.getWorld();
        Robot me = strategy.getMe();
        
        //if(world.containsRobot(x,y)&&world.getRobotOn(x,y).getId()!=me.getId()){
        //    return null;
        //}
                
        switch(value.getDirection().ordinal()){
                    case 0:{
                        if(x-1<0 || x-1>=world.getDimensionX()) return null;
                        if(y<0 || y>=world.getDimensionY()) return null;
                        if(!world.containsWall(x-1,y)){
                            return new SearchNode(x-1, y, dir,target, this);
                        }   
                        return null;
                    }
                    case 1:{
                        if(x<0 || x>=world.getDimensionX()) return null;
                        if(y-1<0 || y-1>=world.getDimensionY()) return null;
                        if(!world.containsWall(x,y-1)){
                            return new SearchNode(x, y-1, dir,target, this);
                        }
                        return null;
                    }
                    case 2:{
                        if(x+1<0 || x+1>=world.getDimensionX()) return null;
                        if(y<0 || y>=world.getDimensionY()) return null;
                        if(!world.containsWall(x+1,y)){
                            return new SearchNode(x+1, y, dir,target, this);
                        }
                        return null;
                    }
                    case 3:{
                        if(x<0 || x>=world.getDimensionX()) return null;
                        if(y+1<0 || y+1>=world.getDimensionY()) return null;
                        if(!world.containsWall(x,y+1)){
                            return new SearchNode(x, y+1, dir,target, this);
                        }
                        return null;
                    }                    
                }
        return null;
    }
    
    

    SearchNode getParent() {
        return parent;
    }
   
 
}
