package cz.muni.fi.fibot.teststrategybot;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.PriorityQueue;
import java.util.Vector;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot.Direction;

public class TestStrategyBot implements Strategy {
	
	private World world;
	private Robot robot;
	private int mapX, mapY, myX, myY;
	
	// pouziva se v prioritni fronte pri vybirani neprojiteho stavu s nejmensi cenou 
	public final Comparator<Vector<Integer>> coordComp = 
		new Comparator<Vector<Integer>>() {
	    public int compare(Vector<Integer> pos1, Vector<Integer> pos2) {
	    	Integer c1 = costs.get(pos1), c2 = costs.get(pos2);
	    	if (c1 == null || c2 == null) return 0;
	    	return c1.compareTo(c2); 
	    }
	};
	
	
	private Vector<Integer> dirVec;
	/* x, y, dir 
	 * 0 - sever
	 * 1 - zapad
	 * 2 - jih
	 * 3 - vychod
	 */
	
	// seznam vsech uz projitych stavu
	private HashSet<Vector<Integer>> completed;
	// stavy, ktere se maji projit 
	private PriorityQueue<Vector<Integer>> discovered;
	// zatim nejlepsi ceny cest do urcitych stavu
	private HashMap<Vector<Integer>, Integer> costs;
	// predchudci stavu
	private HashMap<Vector<Integer>, Vector<Integer>> predcessors;

	public void setWorld(World world) {
		this.world = world;
		this.mapX = world.getDimensionX();
		this.mapY = world.getDimensionY();
	}

	public void setRobot(Robot robot) {
		this.robot = robot;
	}

	public void initialize() {} // nepouzivane

	// vrati smer jako vektor
	private static Vector<Integer> getDirVect(Direction direct) {
		Vector<Integer> d = new Vector<Integer>();
		if (direct == Direction.DOWN) {
			d.add(0, new Integer(0)); d.add(1, new Integer(1));
		} else if (direct == Direction.UP) {
			d.add(0, new Integer(0)); d.add(1, new Integer(-1));
		} else if (direct == Direction.LEFT) {
			d.add(0, new Integer(-1)); d.add(1, new Integer(0));
		} else {
			d.add(0, new Integer(1)); d.add(1, new Integer(0));
		}
		return d;
	}
	
	// vrati cislo pouzivane u stavu pro smer
	private int getDirAsNumber(Direction dir) {
		if (dir == Direction.UP) return 0;
		if (dir == Direction.RIGHT) return 1;
		if (dir == Direction.DOWN) return 2;
		if (dir == Direction.LEFT) return 3;
		return -1;
	}
	
	// tato funkce zkusi zjistit, jestli je ve smeru jeho natoceni nejaky robot
	private boolean isRobotInLine() {
		// posun se o jedna - netestujem pozici, kde stojime
		int x = myX+dirVec.get(0), y = myY+dirVec.get(1);
		// posouvej se dokud na neco nenarazis
		while (x >= 0 && y >= 0 && x < mapX && y < mapY) {
			// zed nebo okraj 
			if (world.containsWall(x,y)) return false;
			// nebo robota
			if (world.containsRobot(x,y)) return true;
			x += dirVec.get(0); y += dirVec.get(1);
		}
		return false;
	}
	
	// zrelaxuj jednu hranu
	private void relaxPosition(int x, int y, int d, Integer c, Vector<Integer> pred) {
		if (x >= 0 && y >= 0 && x < mapX && y < mapY) {
			// tudy to nepujde
			if (world.containsWall(x,y)) return;
			Vector<Integer> pos = new Vector<Integer>();
			pos.add(new Integer(x));
			pos.add(new Integer(y));
			pos.add(new Integer(d));
			// tato hrana neni treba prohledavat
			if (completed.contains(pos)) return;
			// mame lepsi cenu pro cestu do tohoto stavu ?
			if (discovered.contains(pos)) {
				if (costs.get(pos) > c) {
					costs.remove(pos);
					predcessors.remove(pos);
					predcessors.put(pos,pred);
					costs.put(pos, c);
				}
			// uplne novy stav 
			} else {
				discovered.offer(pos);
				predcessors.put(pos,pred);
				costs.put(pos, c);
			}
		}
	}
	
	// projdi vsechny sousedy
	private void discoverNeighbours(Vector<Integer> pos) {
		completed.add(pos);
		Integer cost = costs.get(pos);
		// zvol prochazena pole podle natoceni
		switch (pos.get(2)) {
			case 0:
				relaxPosition(pos.get(0), pos.get(1), 3, cost+1, pos);
				relaxPosition(pos.get(0), pos.get(1), 1, cost+1, pos);
				relaxPosition(pos.get(0), pos.get(1)-1, 0, cost+1, pos);
				break;
			case 1:
				relaxPosition(pos.get(0), pos.get(1), 0, cost+1, pos);
				relaxPosition(pos.get(0), pos.get(1), 2, cost+1, pos);
				relaxPosition(pos.get(0)+1, pos.get(1), 1, cost+1, pos);
				break;
			case 2:
				relaxPosition(pos.get(0), pos.get(1), 1, cost+1, pos);
				relaxPosition(pos.get(0), pos.get(1), 3, cost+1, pos);
				relaxPosition(pos.get(0), pos.get(1)+1, 2, cost+1, pos);
				break;
			case 3:
				relaxPosition(pos.get(0), pos.get(1), 2, cost+1, pos);
				relaxPosition(pos.get(0), pos.get(1), 0, cost+1, pos);
				relaxPosition(pos.get(0)-1, pos.get(1), 3, cost+1, pos);
				break;
		} 
	}
	
	private void initSearch() {
		// inicializuj vsechny promenne
		costs = new HashMap<Vector<Integer>, Integer>();
		predcessors = new HashMap<Vector<Integer>, Vector<Integer>>();
		completed = new HashSet<Vector<Integer>>();
		discovered = new PriorityQueue<Vector<Integer>>(50, coordComp);
		Vector<Integer> pos = new Vector<Integer>();
		pos.add(new Integer(myX));
		pos.add(new Integer(myY));
		pos.add(new Integer(getDirAsNumber(robot.getDirection())));
		// zname cenu cesty do vrcholu odkud vychazime
		costs.put(pos, new Integer(0));
		// predchudce neni
		predcessors.put(pos, null);
		// pridej k vrcholum pro prozkoumani
		discovered.add(pos);		
	}
	
	// prohledavej
	private Vector<Integer> doSearch() {
		initSearch();
		Vector<Integer> pos;
		// dokud mame nejake neprohledane sousedy
		while (!discovered.isEmpty()) {
			// hledej dal
			pos = discovered.peek();
			// pokud se ovsem nenarazi na poklad
			if (world.containsTreasure(pos.get(0),pos.get(1))) return pos;
			discoverNeighbours(pos);
			discovered.remove(pos);
		}
		return null;
	}
	
	// jaka tedy ma byt dalsi akce ? 
	// projde cestou od pokladu zpet k aktualni pozici a zjisti jaka je prvni 
	// akce na teto ceste
	private BotAction getNextAction(Vector<Integer> treasPos) {
		Vector<Integer> pos = treasPos, lastPos = pos, tpos;
		while ((tpos = predcessors.get(pos)) != null) {
			lastPos = pos;
			pos = tpos;
		}
		// pokud je stejne natoceni mezi poslednimi dvema stavy, tak lze udelat krok 
		if (pos.get(2).equals(lastPos.get(2))) return BotAction.STEP;
		else {
			// zjisti rozdil natoceni
			int dir = (pos.get(2) - lastPos.get(2));
			// otacej doprava
			if ((lastPos.get(2) == 0 && pos.get(2) == 3) || dir == -1)
				return BotAction.RIGHT;
			// nebo doleva
			return BotAction.LEFT;
		}
	}

	public BotAction decide() {
		// Mame dost energie ?
		if (this.world.getRestRobotEnergy() > 0 && this.robot.getEnergy() < 2
				&& this.robot.getEnergy() < this.world.getMaxRobotEnergy()) {
			return BotAction.REST;
		}
		this.dirVec = getDirVect(robot.getDirection());
		this.myX = robot.getPositionX();
		this.myY = robot.getPositionY();
		// mam pod sebou poklad ?
		if (world.containsTreasure(myX, myY)) return BotAction.TAKE;
		// je pred nami robot ?
		if (isRobotInLine()) return BotAction.SHOOT;
		Vector<Integer> nextTreas;
		// je na planu jeste nejaky poklad ?
		if ((nextTreas = doSearch()) == null) return BotAction.REST;
		// tak jdi k nemu
		return getNextAction(nextTreas);
	}

}

