/*
 * HamsterficialPsychology.java
 *
 * Created on 19. duben 2006, 20:42
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hamsterinside;

import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;

/**
 * Simultor nhodnho pohybu zrn v kekovi. 
 * @author bobo
 */
public class HamsterficialIntelligence {
    private List<Layer> layers;
    
    private Hamster hamster;
    
    private BatteryTaker bt;
    private Rater rt;
    
    private int rules;
    
    private List<LayerItem> starts = new ArrayList<LayerItem>();
    
    /**
     * Pi prohledvn pot spotebu energie a nutn odpovn
     */
    public final static int RULE_ENERGY     = 0x01;
    /**
     * Pi prohledvn zaznemenv krok potebn k sebrn pokladu a pidv polko do seznamu zrn
     */
    public final static int RULE_GRAIN      = 0x02;
    /**
     * Pi prohledvn testuje zda doel na baterku, zept se, jestli m odpovat a podle vsledku
     * pokrauje. Pid polko do seznamu baterek. Musm bt tak nastavno pravidlo RULE_ENERGY
     */
    public final static int RULE_ACTBATTERY = 0x04;
    /**
     * Pokud se dostane enegie na jedna a na polku se nachz baterka, tak ji sebere.
     * Pid polko do seznamu baterek. Mus bt tak nastaven pravidlo RULE_ENERGY
     */
    public final static int RULE_PASBATTERY = 0x08;
    /**
     * Pi prohledvan kontroluje zdi.
     */
    public final static int RULE_WALL       = 0x10;
    /**
     * Pi prohledvn kontroluje roboty
     */
    public final static int RULE_ROBOTS     = 0x20;
    /**
     * Upesuje chovn pravidla RULE_ACTBATTERY. A to tak, jestli m po prohledn se sebranou baterkou
     * nsledovat pekryt sti minulho prohledn, prohlednm neuvajc sebrn baterky.
     */
    public final static int RULE_OVERLAY    = 0x40;
    
    public final static int RULE_NOALTERNATIVE = 0x80;

    /** Stvo HamsterficialIntelligence */
    public HamsterficialIntelligence(Hamster nHamster) {
        hamster = nHamster;
        layers = new ArrayList<Layer>();
    }
    /**
     * Vytvo a vrat novou vrstvu
     */
    public Layer addLayer() {
        Layer result =  new Layer(this);
        layers.add(result);
        return result;
    }
    /**
     * Vytvo a vrt zadan poet vrstev
     */
    public List<Layer> addLayers(int count) {
        List<Layer> result = new ArrayList<Layer>();
        for (int l = 0; l < count; l++) {
            result.add(addLayer());
        }
        return result;
    }
    /**
     * Zavol metodu clear() zadan vrstvy
     */
    public void clearLayer(int index) {
        layers.get(index).clear();
    }
    /**
     * Zavol metodu clear() na vech vrstvch
     */
    public void clearLayers() {
        for (Layer l: layers) {
            l.clear();
        }
    }
    /**
     * Odebere vrstvu. Nen nijak dsledn. Neustle lze s vrtvou pracovat, jen HI si ji napamatuje.
     */
    public void removeLayer(Layer value) {
        layers.remove(value);
    }
    /**
     * Zavol vechny clearNco() metody
     */
    public void clear() {
        clearLayers();
        clearStarts();
    }
    /**
     * Nastav pravidlo pro prohledn. Mon pravidla jsou tam nkde nahoe. Zanaj na RULE_
     */
    public void setRules(int value) {
        rules = value;
    }
    /**
     * Nastav tdu, kter bude rozhodovat, zda vzt baterku nebo ne.
     */
    public void setBatteryTaker(BatteryTaker value) {
        bt = value;
    }
    /**
     * Pid polko do seznamu vchozch polek prohledvn
     */
    public void addStart(LayerItem value) {
        if (value != null) starts.add(value);
    }
    /**
     * Vymae seznam vchozch polek
     */
    public void clearStarts() {
        starts.clear();
    }
    /**
     * Spust prohledvn. Zane na polcch zadanch metodou addStart.
     */
    public void findWays(int howFar) {
        LayerItem searchs[] = new LayerItem[starts.size()];
        for (int i = 0; i < starts.size(); i++) searchs[i] = starts.get(i);
        clearStarts();
        findWays(howFar, searchs);
    }
    /**
     * Spust prohledn na zadanch polcch. Ignoruje polka zadan metodou addStart.
     * Vechna polka musm nastavena steps, direction a energy, jinak se nepidj do prohledn.
     */
    public void findWays(int howFar, LayerItem... searchs) {
        ItemClasses itemClasses = new ItemClasses(howFar);
        for (LayerItem li: searchs) {
            if (li != null) itemClasses.addStart(li);
        }
        itemClasses.theHeuristic();
    }
    /**
     * Nastav tdu, kter se pouje k vpotu hodnocen vchozch polek
     */
    public void setRater(Rater value) {
        rt = value;
    }
    /**
     * Ohodnot zadan polko. Projde jak samo polko, tak i ppadn alternativy. Jako cena se pouije cena zadanho polka
     */
    public void rateFrom(LayerItem li) {
        while (li != null) {
            rt.rate(li.getStart(), li);
            li = li.getAlternative();
        }
    }
    /**
     * Ohodnot zadan polko. Projde jak samo polko, tak i ppadn alternativy. Zadanou cenu nastav jako cenu zadanho polka
     */
    public void rateFrom(LayerItem li, int price) {
        li.price = price;
        rateFrom(li);
    }
    
    /**
     * Uren tzpu kontrolnho vstupu
     */
    public enum Print {STEPS, PRICE, ENERGY};
    
    /**
     * Vype zadan hodnoty polek ve vrstv
     */
    protected final void printLayer(Layer lay, Print... what) {
      /*  String row = "";
        for (Print p: what) {
            row+= "    ";
            for (int w = 0; w < hamster.getWorld().getDimensionY(); w++) {
                String si = w + " ";
                row+= (si.length() < 3) ? " " + si : si;
            }
        }
        System.out.println(row);
        
        for (int h = 0; h < hamster.getWorld().getDimensionY(); h++) {
            row = "";
            for (Print p: what) {
                String s = " "  + h + " ";
                s = (s.length() < 4) ? " " + s : s;
                for (int w = 0; w < hamster.getWorld().getDimensionY(); w++) {
                    String si;
                    switch (p) {
                        case STEPS:
                            si = (lay.get(w, h).getSteps() < Integer.MAX_VALUE) ? lay.get(w, h).getSteps() + " " :
                                ((hamster.getWorld().containsWall(w, h)) ? " # " : " . ");
                            break;
                        case PRICE:
                            si = (lay.get(w, h).price != 0) ? lay.get(w, h).price + "" : " . ";
                            if (si.length() < 2) si = "0" + si;
                            break;
                        case ENERGY:
                            si = (lay.get(w, h).getEnergy() > 0) ? lay.get(w, h).getEnergy() + " " :
                                ((hamster.getWorld().containsWall(w, h)) ? " # " : " . ");
                            break;
                        default: si = "   ";
                    }
                    if (si.length() < 3) si = " " + si;
                    s+= si;
                }
                row+= s;
            }
            System.out.println(row);
        }*/
    }
    /**
     * Vrt keka, ke ktermu pat HI
     */
    public Hamster getHamster() {
        return hamster;
    }
    /**
     * Vrt vrstvu podle indexu
     */
    public Layer getLayer(int index) {
        return layers.get(index);
    }
     
    private class ItemClasses extends ArrayList<Set<LayerItem>> {
        private int howFar;
        private int iRules;
        
        public ItemClasses(int nHowFar) {
            howFar = nHowFar;
            iRules = rules;
        }
        
        private void theHeuristic() {
            Vector v = new Vector(0, 0);
            for (int c = 0; (c < howFar) && (c < this.size()) ; c++) {
                for (LayerItem li: this.get(c)) {
                    LayerItem l = li;
                    while (l != null) {
                        v = l.getDirection();
                        checkLayerItem(l, v, 1);
                        
                        v.turnLeft();
                        checkLayerItem(l, v, 2);
                        
                        v.turnBack();
                        checkLayerItem(l, v, 2);
                        
                        v.turnRight();
                        checkLayerItem(l, v, 3);
                        l = (!rule(RULE_NOALTERNATIVE)) ? l.getAlternative() : null;
                    }
                }
            }
        }
        
        private void checkLayerItem(LayerItem back, Vector v, int add) {
            int x = back.getX() + v.x;
            int y = back.getY() + v.y;
            
            if (hamster.isInWorld(x, y)) {
                if (testFree(x, y)) {
                    LayerItem li = back.getNeighbour(x, y);
                    LayerItem newli = li.clone();
                    newli.assign(back);
                    newli.setDirection(v);
                    newli.incSteps(add);
                    
                    if (rule(RULE_ENERGY)) {
                        //tahy aby se dotsl kam chce
                        for (int i = 0; i < add; i++) {
                            checkEnergy(newli, back);
                            newli.decEnergy();
                        }
                        checkEnergy(newli, li);
                    }
                    //poklad
                    if (rule(RULE_GRAIN) && hamster.getWorld().containsTreasure(x, y)) {
                        newli.incSteps();
                        newli.addGrains(li);
                        if (rule(RULE_ENERGY)) {
                            newli.decEnergy();
                            checkEnergy(newli, li);
                        }
                    }
                    boolean over = true;
                    if (rule(RULE_ENERGY | RULE_ACTBATTERY) && hamster.getWorld().containsBattery(x, y)) {
                        if ((li.getSteps() > newli.getSteps() + 1) && (newli.getSteps() + 1 < howFar)) {
                            if (bt.shallRecharge(newli)) {
                                this.testMax(li);
                                li.assign(newli);
                                li.incSteps();
                                li.setEnergy(hamster.getWorld().getMaxRobotEnergy());
                                li.addBattery(li);
                                li.addUsedBattery(li);
                                findWays(howFar, li);
                                over = rule(RULE_OVERLAY);
                                if (!over) this.addLayerItem(li);
                            }
                        }
                        newli.addBattery(li);
                    }
                    if (over) {
                        if ((li.getSteps() > newli.getSteps()) &&  (newli.getSteps() < howFar)) {
                            testMax(li);
                            li.assign(newli);
                            this.addLayerItem(li);
                        } else if (li.getSteps() == newli.getSteps()) {
                            LayerItem altr = li.addAlternative();
                            altr.assign(newli);
                        }
                    }
                }
            }
        }
        
        private void addLayerItem(LayerItem li) {
            while (size() <= li.getSteps()) {
                add(new HashSet<LayerItem>());
            }
            get(li.getSteps()).add(li);
        }
        
        private void testMax(LayerItem li) {
            if (li.getSteps() < Integer.MAX_VALUE) {
                if (this.size() > li.getSteps()) {
                    this.get(li.getSteps()).remove(li);
                }
                li.clearAlternative();
            }
        }
        
        private boolean testFree(int x, int y) {
            boolean result = true;
            result&= !(rule(RULE_WALL) && hamster.getWorld().containsWall(x, y));
            result&= !(rule(RULE_ROBOTS) && hamster.getWorld().containsRobot(x, y));
            return result;
        }
        
        private void checkEnergy(LayerItem newli, LayerItem li) {
            if (newli.getEnergy() <= 1) {
                if (rule(RULE_PASBATTERY) && hamster.getWorld().containsBattery(li.getX(), li.getY())) {
                    newli.setEnergy(hamster.getWorld().getMaxRobotEnergy());
                    newli.addBattery(li);
                    newli.addUsedBattery(li);
                } else {
                    newli.setEnergy(hamster.getWorld().getRestRobotEnergy());
                }
                newli.incSteps();
            }
        }
        
        private boolean rule(int value) {
            return (iRules & value) != 0;
        }
        
        private void addStart(LayerItem li) {
            if ((li.getSteps() < Integer.MAX_VALUE) && (li.getEnergy() > 0) &&
                    ((li.getDirection().x != 0) || (li.getDirection().y != 0))) {
                li.createStart();
                li.clearGrains();
                li.createBatteries();
                addLayerItem(li);
            }
        }
    }
    
    /**
     * Rozhrn pro implementaci metody pro rozhodovn, zda sebrat baterku
     */
    public interface BatteryTaker {
        /**
         * Dostane, u kterho polka se rozhoduje, zda vzt baterku. 
         * Toto polka je kop pvodn polka ve vrstv s aktualizovanmi daji.
         */ 
        public boolean shallRecharge(LayerItem li);
    }
    
    /**
     * Rozhran pro ohodnocen polka.
     */
    public interface Rater {
        /**
         * Metoda hodnocen. Dostane hodnocen polko a k nmu vchoz polko
         */
        public void rate(LayerItem start, LayerItem li);
    }
}
