/*
 * Layer.java
 *
 * Created on 19. duben 2006, 21:00
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hamsterinside;

import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;

import cz.muni.fi.fibot.Robot;

/**
 * Obsahuje dvourozmrn pole polek. 
 * Spravuje seznamy polek s baterkami a zrnm. Dle jete seznam start, nebo-li 
 * polek odkud se zaalo s prohledvnm
 * @author bobo
 */
public class Layer {
    
    private LayerItem items[][];
    
    private Set<List<LayerItem>> setOfBatteries;
    private Set<List<LayerItem>> setOfGrains;
    private List<LayerItem> starts;
    
    private HamsterficialIntelligence hi;
    
    /**
     * Vytvo novou vrstvu
     */
    public Layer(HamsterficialIntelligence nHI) {
        hi = nHI;
        cz.muni.fi.fibot.World world = hi.getHamster().getWorld();
        items = new LayerItem[world.getDimensionX()][world.getDimensionY()];
        for (int w = 0; w < world.getDimensionX(); w++) {
            for (int h = 0; h < world.getDimensionY(); h++) {
                items[w][h] = new LayerItem(new Point(w, h), this);
            }
        }
        setOfBatteries = new HashSet<List<LayerItem>>();
        setOfGrains = new HashSet<List<LayerItem>>();
        starts = new ArrayList<LayerItem>();
    }
    
    /**
     * Smae celou vrstvu. Vol vechny ostatn clearNco metody
     */
    public void clear() {
        clearSetOfBatteries();
        clearSetOfGrains();
        clearItems();
        clearStarts();
    }
    /**
     * Zavol metodu clear vech polek patcch do vrstvy
     */
    public void clearItems() {
        for (LayerItem[] coll: items) {
            for (LayerItem li: coll) {
                li.clear();
            }
        }
    }
    /**
     * Vymae seznam seznam baterek
     */
    public void clearSetOfBatteries() {
        setOfBatteries.clear();
    }
    /**
     * Vymae seznam seznam zrn
     */
    public void clearSetOfGrains() {
        setOfGrains.clear();
    }
    /**
     * Vymae seznam start vrstvy
     */
    public void clearStarts() {
        starts.clear();
    }
    /**
     * Pd do zadanho sezanamu baterku a zkontroluje, jest-li ji stejn seznam baterek neexistuje. Kdy ano, vrt ji existujc.
     **/
    public List<LayerItem> addBattery(List<LayerItem> bl, LayerItem battery) {
        List<LayerItem> result = new ArrayList<LayerItem>();
        if (bl != null) result.addAll(bl);
        if (battery != null) result.add(battery);
      /*  for (List<LayerItem> l: setOfBatteries) {
            if (l.equals(result)) {
                return l;
            }
        }*/
        setOfBatteries.add(result);
        return result;
    }
    
    /**
     * Pd do zadanho sezanamu poklad a zkontroluje, jest-li ji stejn seznam zrn neexistuje. Kdy ano, vrt ji existujc.
     **/
    public List<LayerItem> addGrain(List<LayerItem> gl, LayerItem grain) {
        List<LayerItem> result = new ArrayList<LayerItem>();
        if (gl != null) result.addAll(gl);
        if (grain != null) result.add(grain);
     /*   for (List<LayerItem> l: setOfGrains) {
            if (l.equals(result)) {
                return l;
            }
        }*/
        setOfGrains.add(result);
        return result;
    }
    
    /**
     * Vrt polko se stejnou souadnc jakou m zadan robot
     */
    public LayerItem get(Robot value) {
        return items[value.getPositionX()][value.getPositionY()];
    }
    /**
     * Vrt polko o zadan souadnici
     */
    public LayerItem get(Point value) {
        return items[value.x][value.y];
    }
    /**
     * Vrt polko o zadanch souadnicch
     */
    public LayerItem get(int x, int y) {
        return items[x][y];
    }
    /**
     * Vrt HamsterficialIntelligence, kter vytvoila tadytuhlencdtuto vrstvu
     */
    public HamsterficialIntelligence getHI() {
        return hi;
    }
    /** 
     * Nastav polku, na kterm je zadan robot, energi a smr zadanho robota
     */
    public void assignRobot(Robot value) {
        get(value).assignRobot(value);
    }
    /**
     * Vrt seznam vech start ve vrstv
     */
    public List<LayerItem> getStarts() {
        return starts;
    }
}
