/*
 * LayerItem.java
 *
 * Created on 19. duben 2006, 20:58
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hamsterinside;

import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.HashSet;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;

/**
 * Jedno polko vrstvy - Layer.
 * Pamatuje si svou pozici, vrstvu, do kter pat, smr pchodu na polko,
 * vzdlenost v krocch od vchozho polka, vchoz polko - start, energii,
 * kterou bude mt a dojde z vchozho polka, hodnotu polka.
 * Pokud se lze na polko dostat jinou cestou se stejnou energi, pak si tamhlenctamtamtu
 * cestu pamatuje v alternative. V batteries si uchovv seznam baterek, kter m po cest
 * z vchozho polka a v usedBatteries jen ty, kter cestou pouil.
 * Grains je skoro stejn jak batteries, jen si zaznamenv poklady msto baterek.
 * @author bobo
 */
public class LayerItem {
    
    private Point pos;
    private Layer layer;

    private Vector direction;
    private int steps;
    private int energy;
    private LayerItem start;
    private LayerItem alternative;
    private List<LayerItem> batteries;
    private List<LayerItem> usedBatteries;
    private List<LayerItem> grains;
    
    /**
     * Ohodnocen polka, v prbhu hledn cest se nemn.
     */
    public int price;

    /**
     * Vytvo polko. Souadnici si uchov
     */
    public LayerItem(Point nPos, Layer nLayer) {
        pos = nPos;
        layer = nLayer;
        direction = new Vector(0, 0);
        steps = Integer.MAX_VALUE;
        price = 0;
        energy = 0;
    }
    /**
     * Vymae informace o vchozm polku
     **/
    public void clearStart() {
        start = null;
    }
    /**
     * Vymae informace o baterkch a pouitch baterkch
     */
    public void clearBatteries() {
        batteries = null;
        usedBatteries = null;
    }
    /**
     * Vymae informace o pokladech
     */
    public void clearGrains() {
        grains = null;
    }
    /**
     * Nastav vchoz hodnoty pro steps, price a energy
     */
    public void  clearValues() {
        steps = Integer.MAX_VALUE;
        price = 0;
        energy = 0;
    }
    /**
     * Smae vechny alternativn cesty k polku
     */
    public void clearAlternative() {
        if(alternative != null) alternative.clearAlternative();
        alternative = null;
    }
    /**
     * Vymae vechny informace o polku. Vol vechny ostatn clearNco metody
     **/
    public void clear() {
        clearValues();
        clearStart();
        clearBatteries();
        clearGrains();
        clearAlternative();
    }
    /**
     * Zkopruje informace nkter informace ze zadanho polka. Kopruje se steps, energy a direction
     */
    public void weakAssign(LayerItem value) {
        direction.setDir(value.direction);
        steps = value.steps;
        energy = value.energy;
    }
    /**
     * Zkopruje informace nkter informace ze zadanho polka. kopruje se steps, energy, direction,
     * start, batteries, usedBatteries, grains
     */
    public void assign(LayerItem value) {
        weakAssign(value);
        start = value.start;
        batteries = value.batteries;
        usedBatteries = value.usedBatteries;
        grains = value.grains;
    }
    /**
     * Kdy polko nem pizazeno dn start, tak se polko prohls za start.
     */
    public void createStart() {
        if (start == null) {
            start = this;
            layer.getStarts().add(this);
        }
    }
    /**
     * Zalo seznam baterek a pouitch baterek, pokud jet dn neexistuje
     */
    public void createBatteries() {
        if (batteries == null) batteries = layer.addBattery(null, null);
        if (usedBatteries == null) usedBatteries = layer.addBattery(null, null);
    }
    /**
     * Pid polko do seznamu polek s baterkami
     */
    public void addBattery(LayerItem value) {
        batteries = layer.addBattery(batteries, value);
    }
    /**
     * Pid polko do seznamu polek s pouitmi baterkami
     */
    public void addUsedBattery(LayerItem value) {
        usedBatteries = layer.addBattery(usedBatteries, value);
    }
    /**
     * Zalo seznam poklad, pokud jet neexistuje.
     */
    public void createGrains() {
        if (grains == null) grains = layer.addGrain(null, null);
    }
    /**
     * Pid polko do seznamu polek se zrnm
     */
    public void addGrains(LayerItem value) {
        grains = layer.addGrain(grains, value);
    }
    /**
     * Pid k polko alternatovn polko
     */
    public LayerItem addAlternative() {
        if (alternative != null) return alternative.addAlternative();
        else {
            this.alternative = this.clone();
            return this.alternative;
        }
    }
    /**
     * Vrti x pozici polka
     */
    public int getX() {
        return pos.x;
    }
    /**
     * Vrt y pozici polka
     */
    public int getY() {
        return pos.y;
    }
    /**
     * Vrt vrstvu, do kter polko pat
     */
    public Layer getLayer() {
        return layer;
    }
    /**
     * Vrt polko, kter m stejnou pozici a vrstvu
     */
    public LayerItem clone() {
        LayerItem result = new LayerItem(this.pos, this.layer);
        return result;
    }
    /**
     * Vrt polko ze stejn vrstvy na souadnicch x a y
     */
    public LayerItem getNeighbour(int x, int y) {
        return layer.get(x, y);
    }
    /**
     * Vrt polko ze stejn vrstvy na souadnicch robota
     */
    public LayerItem getNeighbour(Robot value) {
        return layer.get(value);
    }
    /**
     * Vrt polko ze stejn vrstvy na souadinici
     */
    public LayerItem getNeighbour(Point value) {
        return layer.get(value);
    }
    /**
     * Vrt vedlej polko v zadanm smru
     */
    public LayerItem getNeighbour(Vector value) {
        int x = pos.x + value.x;
        int y = pos.y + value.y;
        return layer.getHI().getHamster().isInWorld(x, y) ? layer.get(x, y) : null;
    }
    /**
     * Vrt vedlej polko ve smru polka otoenm podle zadanho smru.
     */
    public LayerItem getNeighbour(Direction value) {
        Vector v = getDirection();
        v.applyDirection(value);
        return getNeighbour(v);
    }
    /**
     * Vrt polko na stejn pozici v zadan vrstv
     */
    public LayerItem inLayer(Layer value) {
        return value.get(pos);
    }
    /**
     * Vrt alternativn polko
     */
    public LayerItem getAlternative() {
        return this.alternative;
    }
    /**
     * Vrt vchoz polko - start
     */
    public LayerItem getStart() {
        return this.start;
    }
    /**
     * Vrt list baterek po cest v poad od startu
     */
    public List<LayerItem> getBatteries() {
        return java.util.Collections.unmodifiableList(batteries);
    }
    /**
     * Vrt list pouitch baterek v poad od startu
     */
    public List<LayerItem> getUsedBatteries() {
        return java.util.Collections.unmodifiableList(usedBatteries);
    }
    /**
     * Vrt list poklad po cest v poad od startu
     */
    public List<LayerItem> getGrains() {
        return java.util.Collections.unmodifiableList(grains);
    }
    /**
     * Vrt vdlenost od startu
     */
    public int getSteps() {
        return this.steps;
    }
    /**
     * Zv o zadanou hodnotu poet krok
     */
    public void incSteps(int value) {
        this.steps+= value;
    }
    /**
     * Zv o 1 poet krok
     */
    public void incSteps() {
        this.steps++;
    }
    /**
     * Vrt zbvajc energi po cest ze startu
     */
    public int getEnergy() {
        return this.energy;
    }
    /**
     * Nastav zbvajc energii
     */
    public void setEnergy(int value) {
        this.energy = value;
    }
    /**
     * Sn o 1 zbvajc energii
     */
    public void decEnergy() {
        energy--;
    }
    /**
     * Vrt nov vektor se stejnm smrem jako m polko
     */
    public Vector getDirection() {
        return new Vector(this.direction);
    }
    /**
     * Nastav smru na stejn jako m zadan vektor
     */
    public void setDirection(Vector value) {
        direction.setDir(value);
    }
    /**
     * Nastav smr podle zadanho smru
     */
    public void setDirection(Direction value) {
        direction.setDirection(value);
    }
    /**
     * Nastav smr a energy podle zadanho robota
     */
    public void assignRobot(Robot value) {
        setDirection(value.getDirection());
        energy = value.getEnergy();
        steps = 0; 
    }
    /**
     * Nastav poet krok do vchozho polka
     */
    public void setSteps(int value) {
        steps = value;
    }
    /**
     * Vrt nov Point se stejnmi souadnicemi jako m polko
     */ 
    public Point getPos() {
        return new Point(pos);
    }
    
    public boolean equals(Object value) {
        if (value.getClass() == this.getClass()) {
            LayerItem li = (LayerItem) value;
            if ((li.layer == this.layer) && (li.pos.equals(this.pos))) {
                return true;
            }
        } return false;
    }
    /**
     * Zjisti, zda ma policko stejnou pozici
     */
    public boolean samePos(LayerItem value) {
        return samePos(value.pos);
    }
    public boolean samePos(Point value) {
        return pos.equals(value);
    }
    public boolean samePos(int x, int y) {
        return samePos(new Point(x, y));
    }
    public boolean samePos(Robot value) {
        return samePos(new Point(value));
    }
}
