/*
 * TuringHamster.java
 *
 * Created on 20. duben 2006, 12:18
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hamsterinside;

import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy;

import java.util.Random;
import java.util.Set;
import java.util.HashSet;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Stack;

/**
 * Turingv keek, nebo-li keek se dvma zsobnky na zrn.
 * @author bobo
 */
public class TuringHamster implements Strategy, Hamster {
    protected World world;
    protected Robot robot;
    public final Random random = new Random();
    
    protected int svet[][];
    protected static final int POLE_CLEAR     = 0;
    protected static final int POLE_WALL      = 0x01;
    protected static final int POLE_BAT       = 0x2;
    protected static final int POLE_TRES      = 0x4;
    protected static final int POLE_ROBOT     = 0x8;
    protected static final int POLE_HARD      = 0x10;
    protected static final int POLE_VALUE     = 0x7E0;
    protected static final int POLE_ID        = 0x7800;
    protected static final int POLE_DANGER    = 0x8000;
    protected static final int POLE_LOST      = 0x10000;
    protected static final int TO_VALUE       = 64;
    protected static final int TO_ID          = 4096;
    
    
    protected final static int RULE_ENERGY     = HamsterficialIntelligence.RULE_ENERGY;
    protected final static int RULE_GRAIN      = HamsterficialIntelligence.RULE_GRAIN;
    protected final static int RULE_ACTBATTERY = HamsterficialIntelligence.RULE_ACTBATTERY;
    protected final static int RULE_PASBATTERY = HamsterficialIntelligence.RULE_PASBATTERY;
    protected final static int RULE_WALL       = HamsterficialIntelligence.RULE_WALL;
    protected final static int RULE_ROBOTS     = HamsterficialIntelligence.RULE_ROBOTS;
    protected final static int RULE_OVERLAY    = HamsterficialIntelligence.RULE_OVERLAY;
    protected final static int RULE_NOALTERNATIVE = HamsterficialIntelligence.RULE_NOALTERNATIVE;
    
    /**
     * Seznam veho zrn
     */
    protected Set<Grain> grain = new HashSet<Grain>();
    /**
     * Seznam vech robot
     */
    protected List<Robot> roboti;
    /**
     * Roboti, kte mohou po kekovi stlet
     */
    protected Set<Robot> dangers = new HashSet<Robot>();
    /**
     * Roboti, kte minul kolo po kekovi stelili
     */
    protected Set<Robot> attackers = new HashSet<Robot>();
    /**
     * Roboti, kte by mohli pt kolo stelit po kekovi
     */
    protected Set<Robot> possibleDangers = new HashSet<Robot>();
    /**
     * Roboti, pe vedle nebo za kekem
     */
    protected Map<Direction, List<Robot>> targets = new HashMap<Direction, List<Robot>>();
    /**
     * Hodnocen jednotlivch tah
     */
    protected Map<BotAction, Integer> actionsPrices = new HashMap<BotAction, Integer>();
    
    protected HamsterficialIntelligence hi;
    /**
     * Lev lcn torba
     */
    protected Stack<Point> leftCheekPouch = new Stack<Point>();
    /**
     * Prav lcn torba
     */
    protected Stack<Point> rigthCheekPouch = new Stack<Point>();
    
    /** Porod TuringHamster */
    public TuringHamster() {
        hi = new HamsterficialIntelligence(this);
        targets.put(Direction.DOWN, new ArrayList<Robot>());
        targets.put(Direction.LEFT, new ArrayList<Robot>());
        targets.put(Direction.RIGHT, new ArrayList<Robot>());
        targets.put(Direction.UP, new ArrayList<Robot>());
        
    }
    
    public final World getWorld() {
        return world;
    }
    
    public final void setWorld(World nWorld) {
        this.world = nWorld;
    }
    
    public final Robot getRobot() {
        return robot;
    }
    
    public final void setRobot(Robot r) {
        this.robot = r;
    }
    
    public final BotAction decide() {
        BotAction result;
        try {
            beginStep();
            result = hi();
            endStep();
        } catch (Exception e) {
            result = BotAction.REST;
        }
        return result;
    }
    
    public void initialize() {
        //pocatecni priprava sveta
        svet = new int[world.getDimensionX()][world.getDimensionY()];
        for (int h = 0; h < world.getDimensionY(); h++) {
            for (int w = 0; w < world.getDimensionY(); w++) {
                svet[w][h] = POLE_CLEAR;
                if (world.containsWall(w, h)) svet[w][h] = POLE_WALL | POLE_HARD;
                if (world.containsBattery(w, h)) {
                    svet[w][h] = POLE_BAT;
                }
                if (world.containsTreasure(w, h)) {
                    svet[w][h] = POLE_TRES | (world.getTreasureValue(w, h) * TO_VALUE);
                    grain.add(new Grain(w ,h, world.getTreasureValue(w, h)));
                }
            }
        }
        roboti = world.getRobots();
    }
    
    protected BotAction hi() {
        return BotAction.STEP;
    }
    /**
     * Zjist, zda je polko jet ve svt
     */
    public final boolean isInWorld(int x, int y) {
        return ((x >= 0) && (y >= 0) && (x < world.getDimensionX()) && (y < world.getDimensionY()));
    }
    /**
     * Jestli mus robot toto kolo odpovat.
     */
    protected final boolean mustRest(Robot r) {
        return (!isLastRound()) ? r.getEnergy() == 1 : false;
    }
    /**
     * Jest-li keek mus toto kolo odpovat
     */
    protected final boolean mustRest() {
        return mustRest(robot);
    }
    /**
     * Zda u je posledn kolo
     */
    protected final boolean isLastRound() {
        return world.getRound() == world.getRoundsCount() - 1;
    }
    /**
     * Poet zbvajcch kol
     */
    protected final int remainRounds() {
        return world.getRoundsCount() - world.getRound();
    }
    /**
     * Vrt nejhodnotnj akci z actionsPrices
     */
    protected final BotAction resolveBestAction() {
        BotAction best = BotAction.REST;
        int max = 1;
        for (BotAction ba: BotAction.values()) {
            if (actionsPrices.get(ba) > max) {
                best = ba;
                max = actionsPrices.get(best);
            } else if ((actionsPrices.get(ba) == max) && (random.nextInt(2) == 0)) {
                best = ba;
                max = actionsPrices.get(best);
            }
        }
        // System.out.println(best);
        return best;
    }
    
// nastavuje nebezpecne policka a roboty do sveta
    private void beginStep() {
        int x, y;
        Vector v;
        
        dangers.clear();
        attackers.clear();
        possibleDangers.clear();
        for (List<Robot> t: targets.values()) t.clear();
        
        for (Robot r: roboti) {
            x = r.getPositionX();
            y = r.getPositionY();
            if (r.getLastAction() == BotAction.TAKE) removeGrain(x, y);
            svet[x][y]|= POLE_ROBOT | POLE_HARD | r.getId() * TO_ID;
        }
        
        //nebezpeci strelby, nastavuje, dokud nenajde zed, robota v ceste ignoruje
        for (Robot r: roboti) {
            if ((r.getId() != robot.getId()) && (r.getEnergy() > 0)) {
                v = new Vector(r.getDirection());
                v.turnLeft();
                for (int i = 0; i < 3; i++) {
                    if ((i == 1) || !mustRest(r)) {
                        x = r.getPositionX() + v.x;
                        y = r.getPositionY() + v.y;
                        if ((i == 1) && (!mustRest(r) && isInWorld(x, y))) {
                            svet[x][y]|= POLE_LOST;
                        }
                        while (isInWorld(x, y) && world.containsWall(x, y)) {
                            svet[x][y]|= POLE_DANGER;
                            if ((x == robot.getPositionX()) && (y == robot.getPositionY())) {
                                if ((i == 1) && (!mustRest(r))) {
                                    if (r.getLastAction() == BotAction.SHOOT) {
                                        attackers.add(r);
                                    } else dangers.add(r);
                                } else {
                                    possibleDangers.add(r);
                                }
                            }
                            x+= v.x;
                            y+= v.y;
                        }
                    }
                    v.turnRight();
                }
            }
        }
        
        v = new Vector(robot.getDirection());
        v.turnLeft();
        for (Direction d: Direction.values()) {
            x = robot.getPositionX() + v.x;
            y = robot.getPositionY() + v.y;
            while (isInWorld(x, y) && !world.containsWall(x, y)) {
                if (world.containsRobot(x, y)) targets.get(d).add(world.getRobotOn(x, y));
                x+= v.x;
                y+= v.y;
            }
            v.turnRight();
        }
        
        for (BotAction ba: BotAction.values()) {
            actionsPrices.put(ba, 0);
        }
    }
    
//rusi nebezpecne policka a roboty ze sveta
    private void endStep() {
        int x, y;
        Vector v;
        
        for (Robot r: roboti) {
            x = r.getPositionX();
            y = r.getPositionY();
            svet[x][y]&= ~(POLE_ROBOT | POLE_HARD | POLE_ID);
        }
        
        for (Robot r: roboti) {
            if ((r.getId() != robot.getId()) && (r.getEnergy() > 0)) {
                v = new Vector(r.getDirection());
                v.turnLeft();
                for (int i = 0; i < 3; i++) {
                    x = r.getPositionX() + v.x;
                    y = r.getPositionY() + v.y;
                    if ((i == 1) && (!mustRest(r) && isInWorld(x, y))) {
                        svet[x][y]&= ~POLE_LOST;
                    }
                    while (isInWorld(x, y) && world.containsWall(x, y)) {
                        svet[x][y]&= ~POLE_DANGER;
                        x+= v.x;
                        y+= v.y;
                    }
                    v.turnRight();
                }
            }
        }
    }
    
    private void removeGrain(int x, int y) {
        for (Grain g: grain) {
            if ((g.x == x) && (g.y == y)) {
                grain.remove(g);
                break;
            }
        }
        svet[x][y]&= ~(POLE_TRES | POLE_VALUE);
    }
    /**
     * Test zda na poli s kreckem je jezek, teda baterka
     */
    protected final boolean containsBattery() {
        return world.containsBattery(robot.getPositionX(), robot.getPositionY());
    }
    /**
     * test zda na poli s kreckem je zrni
     */
    protected final boolean containsTreasure() {
        return world.containsTreasure(robot.getPositionX(), robot.getPositionY());
    }
    /**
     * Vrati kolik zrni moze krecek sezrat
     */
    protected int getTreasureValue() {
        return world.getTreasureValue(robot.getPositionX(), robot.getPositionY());
    }
    
//vrati policko sveta relativne k pozici a smeru robota v zadanem smeru
    protected final int scan(Direction dir, int howFar) {
        howFar = (howFar >= 0) ? howFar : 0;
        Vector v =  new Vector(robot.getDirection());
        v.applyDirection(dir);
        int x = robot.getPositionX() + v.x * howFar;
        int y = robot.getPositionY() + v.y * howFar;
        if (isInWorld(x, y)) {
            return svet[x][y];
        }  return POLE_WALL | POLE_HARD;
    }
    
    protected final int scanUp(int howFar) {
        return scan(Direction.UP, howFar);
    }
    
    protected final int scanDown(int howFar) {
        return scan(Direction.DOWN, howFar);
    }
    
    protected final int scanLeft(int howFar) {
        return scan(Direction.LEFT, howFar);
    }
    
    protected final int scanRight(int howFar) {
        return scan(Direction.RIGHT, howFar);
    }
    
    
}
