
/*
 * Hrobot.java
 *
 * Created on 15. duben 2006, 19:05
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hrobot;

import java.util.Random;
import java.lang.Math;
import java.util.List;
import java.lang.Integer;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Queue;
import java.lang.RuntimeException;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.hrobot.Coords;
import cz.muni.fi.fibot.hrobot.CoordsAndDirection;
import cz.muni.fi.fibot.hrobot.CoordsAndValue;
import cz.muni.fi.fibot.hrobot.Reachability;
import cz.muni.fi.fibot.hrobot.HrobotCommon;
import cz.muni.fi.fibot.hrobot.TakeTactic;
import cz.muni.fi.fibot.hrobot.RestTactic;
import cz.muni.fi.fibot.hrobot.HuntTactic;

/**
 * Trida co implementuje strategii robota
 *
 * @author Pavel
 */
public class Hrobot implements Strategy{
    
private World world;
private Robot robot;
private HrobotCommon common;

private TakeTactic take_tactic;
private RestTactic rest_tactic;
private HuntTactic hunt_tactic;
private boolean timeout_test;

static final int   MAX_ENERGY_TO_IGNORE_ATTACK = 10;

///// nemenne parametry sveta nactene pri inicializaci /////
private int width;             //sirka sveta
private int height;            //vyska sveta
private int resting_energy;    //energie ziskana odpocivanim
private int max_energy;        //maximalni enegie, kterou robot muze mit
private int round_count;       //pocet kol

///// promenne a typy pro vlastni strategii /////
private enum GeneralStrategy { TAKING, //primarni duraz na poklady
                               HUNTING, //primarni duraz na strelbu
                               RESTING //hlavne mit na konci vsechnu energii, snazit se ostatnim ubrat na energii
                             };
                             

private GeneralStrategy current_strategy;
private boolean first_decision;

///// Metody tridy: /////

/**
 * konstruktor
 */
public Hrobot() {
    timeout_test = false;
}

/**
 * povinna metoda na predani instance sveta
 */
public void setWorld(World w) {
 this.world = w;
}

/**
 * povinna metoda na predani instance robota
 */
public void setRobot(Robot r) {
 this.robot = r;
}


/** Metoda, ktera analyzuje protivnikovy tahy. Podle odhadu, co ma protivnik
 *  v planu (sebrat nejaky poklad, nabit se v baterii, ulovit me) lze udelat
 *  nejakou vychytavku. Zpusob reakce jeste neni presne promyslen.
 */
private void analyze_adversary_actions() {

}

/**
 * Strategie pro naprosto nouzove situace
 */
private Strategy.BotAction play_emergency_strategy(Strategy.BotAction chosen_by_stategy) {
    Strategy.BotAction result = chosen_by_stategy;
    
    boolean is_under_attack = false;
    boolean future_under_attack = false;
    boolean next_place_under_attack = false;
    boolean[] attack_from_direction = new boolean[4]; //TODO: vyporadat se s celnimi utoky

    /* vypocitani utoku */
    is_under_attack = common.under_attack(robot.getPositionX(), robot.getPositionY(), attack_from_direction);
    future_under_attack = common.future_under_attack(robot.getPositionX(), robot.getPositionY(), attack_from_direction);
    
    
    if (!is_under_attack && !future_under_attack && current_strategy == GeneralStrategy.TAKING &&
            result == Strategy.BotAction.STEP && robot.getEnergy()>MAX_ENERGY_TO_IGNORE_ATTACK)
    {
        CoordsAndDirection pos = new CoordsAndDirection(robot.getPositionX(), robot.getPositionY(), common.direction_to_int(robot.getDirection()));
        CoordsAndDirection next_pos = common.move_in_direction_from_coords(pos);
        next_place_under_attack = common.under_attack(next_pos.x, next_pos.y, attack_from_direction);
        if (next_place_under_attack)
        {
            result = Strategy.BotAction.TAKE;
//            System.out.println("CEKAM!!!");
        }
    }
    

    /* Ukol vysoke urovne - musi se zachranit stuj co stuj pri energii 1: */
    if (robot.getEnergy()==1)
        if (world.containsBattery(robot.getPositionX(),robot.getPositionY()))
        {
            /*nechceme utocicimu nahrat tim, ze se dobijem, aby si z toho mohl udelat body*/
            if (is_under_attack && max_energy>MAX_ENERGY_TO_IGNORE_ATTACK && resting_energy>=4)
                result = Strategy.BotAction.REST;
            else
                result = Strategy.BotAction.RECHARGE;
        }
        else result = Strategy.BotAction.REST;
    
    return result;
}

/**
 * Hraj zvolenou strategii. Podle strategie ulozene v current_strategy
 * se rozhodne, kterou z metod pro vypocet dalsiho tahu zavola.
 */
private Strategy.BotAction play_current_strategy() {
    Strategy.BotAction result;
    
    if (current_strategy == GeneralStrategy.TAKING)
        result = take_tactic.getNextAction();
    else if (current_strategy == GeneralStrategy.HUNTING)
        result = hunt_tactic.getNextAction();
    else if (current_strategy == GeneralStrategy.RESTING)
        result = rest_tactic.getNextAction();
    else
        throw new RuntimeException("Neznamy typ strategie v promenne current_strategy");
    
    //pokud nazna, ze je v nouzove situaci (ma fakt malo energie), tak se to
    //pokusi zachranit. Jinak jen zkopiruje result do sebe sama.
    result = play_emergency_strategy(result);
    
    return result;
}

/**
 * Rozhodovaci metoda. Volana z vnejsku - vraci dalsi akci, kterou robot vykona.
 * Odtud se hierarchicky vola vetsina ostatnich (soukromych) metod teto tridy.
 */
public Strategy.BotAction decide() {
 Strategy.BotAction result;
 try {

 if (timeout_test) common.set_timedout(true);

  timeout_test = true;
 
 if (!first_decision) refresh_remaining_rounds(); //na zacatku inicializovano spravne, neni treba menit
 refresh_lists();
 reachability_from_all();
 if (!first_decision) analyze_adversary_actions(); //na zacatku jeste nikdo nic neudelal, neni co analyzovat
 choose_strategy();
 
 result = play_current_strategy();
 
 
 timeout_test = false;

 } catch(Throwable t) {
 result = Strategy.BotAction.STEP;
 ////System.out.println("Vyskocila vyjimka: "+t.getMessage());
 }

 if (first_decision) first_decision = false;
 
 
 
 return result;
}

/**
 * Zkopiruje parametry, ktere jsou pro cely souboj konstantni.
 * Vola se pri inicializaci.
 */
private void copy_constant_parameters() {
 resting_energy = world.getRestRobotEnergy();
 max_energy = world.getMaxRobotEnergy();
 width = world.getDimensionX();
 height = world.getDimensionY();
 round_count = world.getRoundsCount();
  
 common = new HrobotCommon(world, robot);
}

/**
 * Obnovi listy pozic robotu, pokladu a baterii. U pokladu si poznaci k pozici
 * i hodnotu.
 */
private void refresh_lists() {
    List<Coords> battery_list = new LinkedList<Coords>();
    List<Coords> robot_list = new LinkedList<Coords>();
    List<CoordsAndValue> treasure_list = new LinkedList<CoordsAndValue>();
    for (int x=0; x!=width; ++x)
        for (int y=0; y!=height; ++y)
        {   /* diva se po baterkach, robotech a pokladech a jejich souradnice
             * dava do seznamu */
            if (world.containsBattery(x,y))
                battery_list.add(new Coords(x,y));
            if (world.containsRobot(x,y))
                robot_list.add(new CoordsAndDirection(x,y,
                                      common.direction_to_int(world.getRobotOn(x,y).getDirection())));
            if (world.containsTreasure(x,y))
                treasure_list.add(new CoordsAndValue(x,y,world.getTreasureValue(x,y)));
        }
    ///// Kopirovani listu do poli: /////
    Coords[] aux_battery_coords = new Coords[battery_list.size()];
    battery_list.toArray(aux_battery_coords);
    common.set_battery_coords_field(aux_battery_coords);
    CoordsAndDirection[] aux_robot_coords = new CoordsAndDirection[robot_list.size()];
    robot_list.toArray(aux_robot_coords);
    common.set_robot_coords_field(aux_robot_coords);
    CoordsAndValue[] aux_treasure_coords = new CoordsAndValue[treasure_list.size()];
    treasure_list.toArray(aux_treasure_coords);
    common.set_treasure_coords_field(aux_treasure_coords);
    common.initialize_reachabilities();
    hunt_tactic.setHrobotCommon(common);
    rest_tactic.setHrobotCommon(common);
    take_tactic.setHrobotCommon(common);
}

/**
 * Obnovi hodnotu promenne remaining_rounds udavajici pocet kol zbyvajicich do
 * konce. Vola se na zacatku metody decide()
 */
private void refresh_remaining_rounds() {
    common.refresh_remaining_rounds();
}

/**
 * Vypocita vzdalenosti ze vsech vyznamnych bodu (roboti, baterie, poklady)
 */
private void reachability_from_all() {
    //reachability z robota
    //=> dosazitelne poklady a baterie
    //zaplavovy algoritmus (graf 50x50x4)
    common.get_reachability_from_own_robot().reachability_from_robot(robot, common.get_remaining_rounds());
    
    //reachability z protivniku
    //=> dosazitelne poklady a baterie
    //zaplavovy algoritmus (graf 50x50x4)
    for (int i=0; i!=common.get_adversary_count(); ++i)
        common.get_reachability_from_adversary(i).reachability_from_robot(common.get_adversary(i), common.get_remaining_rounds());
}

/**
 * Na zaklade udaju predvypocitanych v ruznych lokalnich promennych tridy
 * urci obecnou strategii, kterou se robot bude ridit
 */
private void choose_strategy() {
    //spusteni vypoctu jednotlivych strategii
    take_tactic.execute();
    hunt_tactic.execute();
    rest_tactic.execute();
    
    //ulozeni profitu do pom. promennych
    int take_profit = take_tactic.getNormalizedProfit();
    int hunt_profit = hunt_tactic.getNormalizedProfit();
    int rest_profit = rest_tactic.getNormalizedProfit();
    
    //jednoduchy vyber maxima
    if (take_profit > hunt_profit)
    {
        current_strategy = GeneralStrategy.TAKING;
//        System.out.println("TAKING");
    }
    else if (rest_profit > take_profit)
    {
        current_strategy = GeneralStrategy.RESTING;
//        System.out.println("RESTING");
    }
    else
    {
        current_strategy = GeneralStrategy.HUNTING;
//        System.out.println("HUNTING");
    }
}

/**
 * Inicializace. Pouze nacteni nejakych prvnich dat z mapy
 */
public void initialize() {
 ///// Pocatecni analyza situace: /////

 copy_constant_parameters();
 
 take_tactic = new TakeTactic();
 rest_tactic = new RestTactic();
 hunt_tactic = new HuntTactic();
 
 first_decision = true;
}


/**
 * Heuristikou vypocte co nejvyhodnejsi cestu po pokladech.
 */
private int compute_walk_around_treasures() {
    return 0;
}




}
