/*
 * HrobotCommon.java
 *
 * Created on 20. květen 2006, 23:50
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hrobot;

import java.util.List;
import java.util.ListIterator;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.hrobot.Coords;
import cz.muni.fi.fibot.hrobot.CoordsAndDirection;
import cz.muni.fi.fibot.hrobot.CoordsAndValue;

/**
 * Obsahuje metody pro ruzne konverze ci manipulace nezavisle na strategii 
 *
 * @author p
 */
public class HrobotCommon {
    private World world;
    private Robot robot;
    private Coords[] battery_coords; //souradnice baterii
    private CoordsAndDirection[] robot_coords;   //souradnice robotu
    private CoordsAndValue[] treasure_coords;//souradnice pokladu
    private Reachability my_reach;           //reachability z meho robota
    private Reachability[] adversary_reach;  //reachability z protivniku
    private Robot[] adversaries;     //protivnici
    private int[] robot_to_adv_index;
    private int remaining_rounds;
    private boolean timedout;

    /** Creates a new instance of HrobotCommon */
    public HrobotCommon(World world_arg, Robot robot_arg) {
        timedout = false;
        world = world_arg;
        robot = robot_arg;
        remaining_rounds = world.getRoundsCount();
        List<Robot> robot_list = world.getRobots();
        adversaries = new Robot[robot_list.size()-1];
        robot_to_adv_index = new int[robot_list.size()];
        adversary_reach = new SimpleReachability[robot_list.size()-1];
        ListIterator<Robot> i = robot_list.listIterator();
        int index = 0;
        Robot r;
        int j=0;
        while (i.hasNext())
        {
            r = i.next();
            if (r!=robot)
            {
                adversaries[index] = r;
                robot_to_adv_index[j] = index;
                index++;
            }
            else robot_to_adv_index[j] = Integer.MAX_VALUE;
            j++;
        }
        if (index!=adversaries.length)
            throw new RuntimeException("Posledni index neodpovida poctu");

    }
    
    public int adv_index(int robot_index)
    { return robot_to_adv_index[robot_index]; }
    
    public void set_timedout(boolean b)
    { timedout = b; }
    
    public boolean get_timedout()
    { return timedout; }
    
    public void refresh_remaining_rounds()
    {
        remaining_rounds = world.getRoundsCount() - world.getRound();        
    }
    
    public int get_remaining_rounds()
    {
        return remaining_rounds;
    }
    
    
    public void initialize_reachabilities()
    {
        for (int i=0; i!=adversary_reach.length; ++i)
        {
           adversary_reach[i] = new SimpleReachability(this);
        }
        my_reach = new SimpleReachability(this);
    }
    
    /**
     * Vrati robota protivnika s indexem index
     */        
    public Robot get_adversary(int index)
    {
        return adversaries[index];
    }
    
    public Reachability get_reachability_from_own_robot()
    {
        return my_reach;
    }
    
    public void set_reachability_from_own_robot(Reachability reach)
    {
        my_reach = reach;
    }
    
    public Reachability get_reachability_from_adversary(int index)
    {
        return adversary_reach[index];
    }
    
    public void set_reachability_from_adversary(int index, Reachability reach)
    {
        adversary_reach[index] = reach;
    }
    
    public int get_adversary_count()
    {
        return adversary_reach.length;
    }
    
    
    
    public void set_battery_coords_field(Coords[] battery_coords_arg)
    {
        battery_coords = battery_coords_arg;
    }
    
    public int get_battery_count()
    {
        return battery_coords.length;
    }
    
    public Coords get_battery_coords(int index)
    {
        return battery_coords[index];
    }
    
    public void set_robot_coords_field(CoordsAndDirection[] robot_coords_arg)
    {
        robot_coords = robot_coords_arg;
    }

    public int get_robot_count()
    {
        return robot_coords.length;
    }

    public CoordsAndDirection get_robot_coords(int index)
    {
        return robot_coords[index];
    }

    
    public void set_treasure_coords_field(CoordsAndValue[] treasure_coords_arg)
    {
        treasure_coords = treasure_coords_arg;
    }
    
    public int get_treasure_count()
    {
        return treasure_coords.length;
    }
    
    public CoordsAndValue get_treasure_coords(int index)
    {
        return treasure_coords[index];
    }
    
    public World get_world()
    {
        return world;
    }

    public Robot get_robot()
    {
        return robot;
    }

    /**
     * Vraci true prave kdyz je dane misto jeste v mape (tedy ne treba na pozici [-1,5]).
     */
    public boolean is_in_map_bounds(Coords coords) {
        if (coords.x<0 || coords.x>=world.getDimensionX() ||
            coords.y<0 || coords.y>=world.getDimensionY())
            return false;
        else
            return true;
    }


    public boolean future_under_attack(int x, int y, boolean[] attack_from_direction) {
        if (attack_from_direction.length<4)
            throw new RuntimeException("Prilis kratke pole attack_from_direction (musi mit aspon 4 prvky)");
        boolean is_under_attack = false;

        for (int dir = 0; dir < 4; ++dir)
        {
            CoordsAndDirection cad = new CoordsAndDirection(x,y,dir);
            cad = move_in_direction_from_coords(cad);
            boolean behind_another_robot=false;
            while (is_in_map_bounds(cad) && !world.containsWall(cad.x,cad.y)
                   && !behind_another_robot)
            {
                if (world.containsRobot(cad.x, cad.y) && (cad.x!= robot.getPositionX() || cad.y !=robot.getPositionY()))
                    if (direction_to_int(world.getRobotOn(cad.x,cad.y).getDirection())!=dir)
                    {
                        attack_from_direction[dir] = true;
                        is_under_attack = true;
                    }
                    else behind_another_robot = true;
                cad = move_in_direction_from_coords(cad);
            }
        }
        return is_under_attack;
    }


    public boolean under_attack(int x, int y, boolean[] attack_from_direction) {
        if (attack_from_direction.length<4)
            throw new RuntimeException("Prilis kratke pole attack_from_direction (musi mit aspon 4 prvky)");
        boolean is_under_attack = false;

        for (int dir = 0; dir < 4; ++dir)
        {
            CoordsAndDirection cad = new CoordsAndDirection(x,y,dir);
            cad = move_in_direction_from_coords(cad);
            boolean behind_another_robot=false;
            while (is_in_map_bounds(cad) && !world.containsWall(cad.x,cad.y)
                   && !behind_another_robot)
            {
                if (world.containsRobot(cad.x, cad.y) && (cad.x!= robot.getPositionX() || cad.y !=robot.getPositionY()))
                    if (((direction_to_int(world.getRobotOn(cad.x,cad.y).getDirection())+2)%4)==dir)
                    {
                        attack_from_direction[dir] = true;
                        is_under_attack = true;
                    }
                    else behind_another_robot = true;
                cad = move_in_direction_from_coords(cad);
            }
        }
        return is_under_attack;
    }


    /**
     * konveruje enum type Direction do pro indexovani a vypocty vhodnejsiho integeru
     */
    public int direction_to_int(Robot.Direction direction) {
        switch (direction) {
            case LEFT: return 0;
            case UP: return 1;
            case RIGHT: return 2;
            case DOWN: return 3;
            default: throw new RuntimeException("Neznamy typ smeru.");
        }

    }

    /**
     * Spocita misto a smer po jednom kroku dopredu z daneho mista
     */
    public CoordsAndDirection move_in_direction_from_coords(CoordsAndDirection cad) {
        switch (cad.direction) {
            case 0: return new CoordsAndDirection(cad.x-1, cad.y  , cad.direction);
            case 1: return new CoordsAndDirection(cad.x  , cad.y-1, cad.direction);
            case 2: return new CoordsAndDirection(cad.x+1, cad.y  , cad.direction);
            case 3: return new CoordsAndDirection(cad.x  , cad.y+1, cad.direction);
            default: throw new RuntimeException("Neznamy typ smeru daneho cislem.");
        }
    }


    /**
     * Vrati smer robota po otoceni vpravo (po smeru hodinovych rucicek).
     */
    public int rotate_direction_cw(int direction) {
        return ((direction+1)%4);
    }

    /**
     * Vrati smer robota po otoceni vlevo (proti smeru hodinovych rucicek)
     */
    public int rotate_direction_ccw(int direction) {
        return ((direction+3)%4);
    }

    
    
    /**
     * vypocita minimum ze zadaneho pole integeru
     */
    public int minimum(int[] array_of_integers) {
        int result = Integer.MAX_VALUE;

        for (int element : array_of_integers)
            if (element<result) result = element;

        return result;
    }

    /**
     * zjisti index minimalniho prvku z pole intu
     */
    public int minimal(int[] array_of_integers) {
        int result = 0;

        if (array_of_integers.length==0) return Integer.MAX_VALUE;
        else
        {
            for (int i=1; i<array_of_integers.length; ++i)
                if (array_of_integers[i]<array_of_integers[result]) result = i;

            return result;
        }
    }

    /**
     * zjisti index minimalniho prvku z pole bytu
     */
    public int minimal(byte[] array_of_bytes) {
        int result = 0;

        if (array_of_bytes.length==0) return Byte.MAX_VALUE;
        else
        {
            for (int i=1; i<array_of_bytes.length; ++i)
                if (array_of_bytes[i]<array_of_bytes[result]) result = i;

            return result;
        }
    }

    /**
     * vypocita minimum ze zadaneho pole integeru
     */
    public byte unsharp_minimum(byte[] array_of_bytes) {
        byte result = Byte.MAX_VALUE;

        for (byte element : array_of_bytes)
            if (element<=result) result = element;

        return result;
    }

    /**
     * zjisti index minimalniho prvku z pole bytu
     */
    public int unsharp_minimal(byte[] array_of_bytes) {
        int result = 0;

        if (array_of_bytes.length==0) return Byte.MAX_VALUE;
        else
        {
            for (int i=1; i<array_of_bytes.length; ++i)
                if (array_of_bytes[i]<=array_of_bytes[result]) result = i;

            return result;
        }
    }

}
