package cz.muni.fi.fibot.hrobot;


import cz.muni.fi.fibot.*;
import static cz.muni.fi.fibot.Robot.*;
import java.util.*;


public class HuntTactic implements Tactic {

    private HrobotCommon common;
    private Robot myself;
    private World world;
    private int[][] robotPlaceCountForApproriateStep;
    private int normalizedProfit;
    private Strategy.BotAction firstAction = Strategy.BotAction.RECHARGE;


    public void setHrobotCommon(HrobotCommon cmn)
    {
        common = cmn;
        myself = common.get_robot();
        world = common.get_world();
    }

    public void execute() {
        computeDistances();
    }

    public Strategy.BotAction getNextAction() {
        return firstAction;
    }

    public int getNormalizedProfit() {
        return normalizedProfit;
    }

    //--------------------------------------------------------------------------

/*    private void computeDistances() {
        for (int i=0; i<world.getRobots().size(); i++) {
            Robot targetRobot = world.getRobots().get(i);
            if (targetRobot.equals(myself))
                continue;

            // Rozdil mezi natocenimi jednotlivych robotu (mujRobot vs. targetRobot).
            // Interval -4 az 4.
            int advantage = countOrientationAdvantage(myself, targetRobot);



        }
    } */

    static class CandidateField {
        int scoreSum = 0;
        int x = -1;
        int y = -1;
        int direction = -1;
    }


    private void computeDistances() {
        List<Robot> robots = new ArrayList<Robot>(common.get_world().getRobots());
        robots.set(robots.indexOf(myself), null);
        Map<Robot, CandidateField> bestChoices = new HashMap<Robot, CandidateField>(robots.size());
        for (Robot r : robots) {
            if (r!=null)
                bestChoices.put(r, new CandidateField());
        }
        final int maxStepCount = Math.min(8, common.get_remaining_rounds());

        // Vypocitej pocty policek pro kazdeho robota, na ktere se dostane v maxStepCount tazich
        robotPlaceCountForApproriateStep = new int[robots.size()][maxStepCount+1];
        for (int i = 0; i < robots.size(); i++) {
            if (robots.get(i) == null)
                continue;
            countRobotPlaces(i, maxStepCount);
        }

/*        int directSum = testDirectShoot(robots);
        if (directSum > 0) {
            firstAction = Strategy.BotAction.SHOOT;
            normalizedProfit = directSum * 1024;
            return;
        }*/

        Reachability myselfReach = common.get_reachability_from_own_robot();
        int currentStep = 0;
        while ((currentStep <= maxStepCount) && (robots.size() > 0)) {
            for (int robotIndex = 0; robotIndex<robots.size(); robotIndex++) {
                if (robots.get(robotIndex) == null)
                    continue;
                for (int x=0; x<world.getDimensionX(); x++) {
                    for (int y=0; y<world.getDimensionY(); y++) {
                        for (int direction = 0; direction<4; direction++) {
                            if (myselfReach.get_distance(x, y, direction) == currentStep) {
                                int sum = sumRobotHits(
                                        x,
                                        y,
                                        direction,
                                        currentStep,
                                        robotIndex
                                );
                                int robotTotalFieldCountForThisStep = robotPlaceCountForApproriateStep[robotIndex][currentStep];
                                if (sum>0) {
                                    sum = sum * 1024 / robotTotalFieldCountForThisStep;
                                    CandidateField field = new CandidateField();
                                    field.x = x;
                                    field.y = y;
                                    field.direction = direction;
                                    field.scoreSum = sum;
                                    bestChoices.put(robots.get(robotIndex), field);
                                    robots.set(robotIndex, null);
                                }
                            }
                        }
                    }
                }
            }

            currentStep++;
        }

        // Find maximum in profits
        normalizedProfit = 0;
        Robot bestRobot = null;
        for (Robot currentRobot : bestChoices.keySet()) {
            if (currentRobot==null) continue;
            CandidateField field = bestChoices.get(currentRobot);
            //System.out.println("Kandidat: "+currentRobot.getName()+" score: "+field.scoreSum);
            if (field.scoreSum > normalizedProfit) {
                normalizedProfit = field.scoreSum;
                firstAction = myselfReach.get_first_action(field.x, field.y, field.direction);
                bestRobot = currentRobot;
                if (myself.getPositionX()==field.x && myself.getPositionY()==field.y && myself.getDirection().ordinal()==field.direction) {
                    firstAction = Strategy.BotAction.SHOOT;
                }
            }
        }
        if (bestRobot!=null) {
            //System.out.println("Jdeme po: "+bestRobot.getName());
        } else {
            //System.out.println("Nejdeme po nikom");
        }
    }

    //- - - - - - - -

    /* Pocita, kolik policek
    */
    private void countRobotPlaces(int robotIndex, int maxStepCount) {
        Reachability reachstruct = common.get_reachability_from_adversary(common.adv_index(robotIndex));
        for (int direction = 0; direction<4; direction++) {
            for (int x=0; x<world.getDimensionX(); x++) {
                for (int y=0; y<world.getDimensionY(); y++) {
                    int step = reachstruct.get_distance(x, y, direction);
                    if (step <= maxStepCount) {
                        robotPlaceCountForApproriateStep[robotIndex][step]++;
                    }
                }
            }
        }
    }

    //- - - - - - - -

    /*
    private int testDirectShoot(List<Robot> robots) {
        int sum = 0;
        for (int i = 0; i<robots.size(); i++) {
            if (robots.get(i) != null) {
                System.out.println("testDirectShoot pro robota "+i);
                sum = sumRobotHits(myself.getPositionX(), myself.getPositionY(), myself.getDirection().ordinal(), 0, i);
            }
        }
        return sum;
    } */

    private int sumRobotHits(int x, int y, int direction, int currentStep, int robotIndex) {
        if (direction==0)   //Direction.LEFT
            return goThroughMazeHorizontal(x-1, y, -1, -1, currentStep, robotIndex);
        if (direction==1)   //Direction.UP
            return goThroughMazeVertical(x, y-1, -1, -1, currentStep, robotIndex);
        if (direction==2)   //Direction.RIGHT
            return goThroughMazeHorizontal(x+1, y, world.getDimensionX(), 1, currentStep, robotIndex);
        if (direction==3)   //Direction.DOWN
            return goThroughMazeVertical(x, y+1, world.getDimensionY(), 1, currentStep, robotIndex);
        throw new IllegalStateException("Direction neni ani jedna z LEFT, UP, RIGHT ani DOWN.");
    }

    private int goThroughMazeVertical(int x, int y, int lastY, int incrementY, int currentStep, int robotIndex) {
        int total = 0;
        int robotEnergy = world.getRobots().get(robotIndex).getEnergy()/2;
        Reachability reaches = common.get_reachability_from_adversary(common.adv_index(robotIndex));
        for (; y!=lastY; y+=incrementY) {
            if (world.containsWall(x, y))
                break;
            int dist = getMinimumDistance(reaches, x, y);
            if (dist==currentStep) {
                 total += robotEnergy;
            }
        }
        return total;
    }

    private int goThroughMazeHorizontal(int x, int y, int lastX, int incrementX, int currentStep, int robotIndex) {
        int total = 0;
        int robotEnergy = world.getRobots().get(robotIndex).getEnergy()/2;
        Reachability reaches = common.get_reachability_from_adversary(common.adv_index(robotIndex));
        for (; x!=lastX; x+=incrementX) {
            if (world.containsWall(x, y))
                break;
            int dist = getMinimumDistance(reaches, x, y);
            if (dist==currentStep) {
                 total += robotEnergy;
            }
        }
        return total;
    }

    private int getMinimumDistance(Reachability reaches, int x, int y) {
        int min = reaches.get_distance(x, y, 0);
        int second = reaches.get_distance(x, y, 1);
        min = min <= second ? min : second;
        second = reaches.get_distance(x, y, 2);
        min = min <= second ? min : second;
        second = reaches.get_distance(x, y, 3);
        min = min <= second ? min : second;

        return min;
    }

    //--------------------------------------------------------------------------

    private int countOrientationAdvantage(Robot first, Robot second) {
        int result = 0;
        Robot.Direction attackDirection = decideAttackDirection(first, second);

        // Nejprve si pricteme body, pokud First stoji spravne
        if ((attackDirection==Direction.LEFT) || (attackDirection==Direction.RIGHT))
            if ((first.getDirection()==Direction.UP)
                    || (first.getDirection()==Direction.DOWN))
                result -= 1;
        if (attackDirection==Direction.LEFT)
            if ((first.getDirection()==Direction.LEFT) )
                result += 2;
            else if ((first.getDirection()==Direction.RIGHT))
                result -= 2;
        if (attackDirection==Direction.RIGHT)
            if ((first.getDirection()==Direction.LEFT) )
                result -= 2;
            else if ((first.getDirection()==Direction.RIGHT))
                result += 2;

        // Pote si odecteme body, pokud Second stoji nebezpecne
        if ((attackDirection==Direction.LEFT) || (attackDirection==Direction.RIGHT))
            if ((second.getDirection()==Direction.UP)
                    || (second.getDirection()==Direction.DOWN))
                result += 1;
        if (attackDirection==Direction.LEFT)
            if ((second.getDirection()==Direction.LEFT) )
                result += 2;
            else if ((second.getDirection()==Direction.RIGHT))
                result -= 2;
        if (attackDirection==Direction.RIGHT)
            if ((second.getDirection()==Direction.LEFT) )
                result -= 2;
            else if ((second.getDirection()==Direction.RIGHT))
                result += 2;

        return result;
    }


    private Direction decideAttackDirection(Robot first, Robot second) {
        int roughX = first.getPositionX() - second.getPositionX();
        int absX = Math.abs(roughX);
        int roughY = first.getPositionY() - second.getPositionY();
        int absY = Math.abs(roughY);

        Robot.Direction attackDirection;
        if (absX <= absY)
            if (roughX < 0)
                attackDirection = Direction.RIGHT;
            else
                attackDirection = Direction.LEFT;
        else
            if (roughY < 0)
                attackDirection = Direction.DOWN;
            else
                attackDirection = Direction.UP;

        return attackDirection;
    }

}
