/*
 * Tento soubor obsahuje trudu Reachability
 */

package cz.muni.fi.fibot.hrobot;


import java.util.LinkedList;
import java.util.Queue;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.hrobot.Coords;
import cz.muni.fi.fibot.hrobot.CoordsAndDirection;
import cz.muni.fi.fibot.hrobot.CoordsAndValue;
import cz.muni.fi.fibot.hrobot.Reachability;

/**
 * Trida pro ochovani vydleku reachability na World
 */
public class SimpleReachability implements Reachability {   
    private class DistanceMovementAndDirection {
        public int dist;
        public Strategy.BotAction move;
        public int dir;
    }

    private int stored_x;
    private int stored_y;
    private byte[][][] projected_map;
    private Strategy.BotAction[][][] projected_first_act;
    private DistanceMovementAndDirection[] battery_info;
    private DistanceMovementAndDirection[] treasure_info;
    private DistanceMovementAndDirection[] robot_info;
    private HrobotCommon common;
    private World world;
    
    private void fill_array(DistanceMovementAndDirection[] myarray){
        for (int i=0; i!=myarray.length; ++i) myarray[i] = new DistanceMovementAndDirection();
    }
    
    public SimpleReachability(HrobotCommon common_arg)
    {
        common = common_arg;
        world = common.get_world();
        battery_info = new DistanceMovementAndDirection[common.get_battery_count()];
        fill_array(battery_info);
        treasure_info = new DistanceMovementAndDirection[common.get_treasure_count()];
        fill_array(treasure_info);
        robot_info = new DistanceMovementAndDirection[common.get_robot_count()];
        fill_array(robot_info);
        projected_map = new byte[world.getDimensionX()][world.getDimensionY()][4];
        projected_first_act = new Strategy.BotAction[world.getDimensionX()][world.getDimensionY()][4];
    }
    
    public int get_from_x() { return stored_x; }
    public int get_from_y() { return stored_y; }
    
    public void set_battery_distance(int index, int d) { battery_info[index].dist = d; }
    public void set_battery_direction(int index, int d) { battery_info[index].dir = d; }
    public void set_battery_movement(int index, Strategy.BotAction m) { battery_info[index].move = m; }
    public int get_battery_distance(int index) { return battery_info[index].dist; }
    public int get_battery_direction(int index) { return battery_info[index].dir; }
    public Strategy.BotAction get_battery_movement(int index) { return battery_info[index].move; }
    public int get_battery_count() { return battery_info.length; }
    
    public void set_treasure_distance(int index, int d) { treasure_info[index].dist = d; }
    public void set_treasure_direction(int index, int d) { treasure_info[index].dir = d; }
    public void set_treasure_movement(int index, Strategy.BotAction m) { treasure_info[index].move = m; }
    public int get_treasure_distance(int index) { return treasure_info[index].dist; }
    public int get_treasure_direction(int index) { return treasure_info[index].dir; }
    public Strategy.BotAction get_treasure_movement(int index) { return treasure_info[index].move; }
    public int get_treasure_count() { return treasure_info.length; }
   
    public void set_robot_distance(int index, int d) { robot_info[index].dist = d; }
    public void set_robot_direction(int index, int d) { robot_info[index].dir = d; }
    public void set_robot_movement(int index, Strategy.BotAction m) { robot_info[index].move = m; }
    public int get_robot_distance(int index) { return robot_info[index].dist; }
    public int get_robot_direction(int index) { return robot_info[index].dir; }
    public Strategy.BotAction get_robot_movement(int index) { return robot_info[index].move; }
    public int get_robot_count() { return robot_info.length; }
    
    /**
     * Vrati nejkratsi vzdalenost do daneho mista
     */
    public int get_distance(int x, int y, int direction)
    { return projected_map[x][y][direction]; }
    
    /**
     * Vrati prvni akci na nejkratsi ceste do daneho mista
     */
    public Strategy.BotAction get_first_action(int x, int y, int direction)
    { return projected_first_act[x][y][direction]; }
    
    /**
     * Otestuje, zda byla nalezena kratsi cesta (delka dana pomoci arg. value)
     * do mista place a pripadne zaznaci do mapy vzdalenosti map.
     */
    private boolean update_distance(byte[][][] map, Strategy.BotAction[][][] first_act,
                                    CoordsAndDirection place, byte value,
                                    Strategy.BotAction from_first_act) {
        if (map[place.x][place.y][place.direction]>value) {
            map[place.x][place.y][place.direction]=value;
            first_act[place.x][place.y][place.direction]=from_first_act;
            return true;
        }
        else return false;
    }

    /**
     * Provede pruzkum vzdalenosti z bodu [x,y] ve mape
     */
    public void reachability_from(int x, int y, int direction, int bound) {
        stored_x = x;
        stored_y = y;
        byte[][][] map = new byte[world.getDimensionX()][world.getDimensionY()][4];  //mapa vzdalenosti
        Strategy.BotAction[][][] first_act = new Strategy.BotAction[world.getDimensionX()][world.getDimensionY()][4]; //mapa prvich akci

        //inicializace mapy na sama "nekonecna""
        for (int i=0; i!=map.length; ++i)
            for (int j=0; j!=map[i].length; ++j)
                for (int k=0; k!=map[i][j].length; ++k)
                {
                    map[i][j][k] = Byte.MAX_VALUE;
                    first_act[i][j][k] = Strategy.BotAction.NOTHING;
                }

        map[x][y][direction] = 0;

        Queue<CoordsAndDirection> queue = new LinkedList<CoordsAndDirection>();

        //predpocitam si prvni tri nasledniky - abych pak mohl pocitat mapu prvnich akci
        CoordsAndDirection initial_position = new CoordsAndDirection(x,y,direction);
        CoordsAndDirection step_succ = common.move_in_direction_from_coords(initial_position);
        CoordsAndDirection left_succ = new CoordsAndDirection(x,y,common.rotate_direction_ccw(direction));
        CoordsAndDirection right_succ = new CoordsAndDirection(x,y,common.rotate_direction_cw(direction));

        if (common.is_in_map_bounds(step_succ) && !world.containsWall(step_succ.x,step_succ.y))
        {
            map[step_succ.x][step_succ.y][step_succ.direction] = 1;
            first_act[step_succ.x][step_succ.y][step_succ.direction] = Strategy.BotAction.STEP;
            queue.add(step_succ);
        }

        map[left_succ.x][left_succ.y][left_succ.direction] = 1;
        first_act[left_succ.x][left_succ.y][left_succ.direction] = Strategy.BotAction.LEFT;
        queue.add(left_succ);

        map[right_succ.x][right_succ.y][right_succ.direction] = 1;
        first_act[right_succ.x][right_succ.y][right_succ.direction] = Strategy.BotAction.RIGHT;    
        queue.add(right_succ);

        CoordsAndDirection from; /* pomocna promenna na uchovavani aktualne zpracovavane pozice*/
        CoordsAndDirection step_to, rotate_cw, rotate_ccw; /* pomocne promenne na uchovani 3 nasledniku pozice from */
        byte next_distance; /* pomocna promenna na uchovani prirazovane vzdalenosti */
        Strategy.BotAction current_first_act;

        while (queue.size()>0)
        {
            from = queue.remove();
            /* kazda pozice charakterizovana 2 souradnicemi a otocenim muze mit nejvyse 3
             * (a nejmene 2) nasledniky: krok vpred, otoceni vlevo (CCW), otoceni pravo (CW)*/
            step_to = common.move_in_direction_from_coords(from);
            rotate_cw = new CoordsAndDirection(from.x, from.y,
                                               common.rotate_direction_cw(from.direction));
            rotate_ccw = new CoordsAndDirection(from.x, from.y,
                                                common.rotate_direction_ccw(from.direction));
            next_distance = (byte)(map[from.x][from.y][from.direction]+1);
            current_first_act = first_act[from.x][from.y][from.direction];

            if (common.is_in_map_bounds(step_to) && !world.containsWall(step_to.x,step_to.y))
                if (update_distance(map,first_act,step_to,next_distance,current_first_act) &&
                        next_distance<bound)
                    queue.add(step_to);
            if (!world.containsWall(rotate_cw.x,rotate_cw.y))
                if (update_distance(map,first_act,rotate_cw,next_distance,current_first_act) &&
                        next_distance<bound)
                    queue.add(rotate_cw);        
            if (!world.containsWall(rotate_ccw.x,rotate_ccw.y))
                if (update_distance(map,first_act,rotate_ccw,next_distance,current_first_act) &&
                        next_distance<bound)
                    queue.add(rotate_ccw);

        }

        /* zapis vysledku prohledavani */
        int min_index;
        for (int i=0; i!=common.get_battery_count(); ++i)
        {
            min_index = common.minimal(map[common.get_battery_coords(i).x][common.get_battery_coords(i).y]);
            set_battery_distance(i,map[common.get_battery_coords(i).x][common.get_battery_coords(i).y][min_index]);
            set_battery_direction(i,min_index);
            set_battery_movement(i,first_act[common.get_battery_coords(i).x][common.get_battery_coords(i).y][min_index]);
        }
        for (int i=0; i!=common.get_robot_count(); ++i)
        {
            min_index = common.minimal(map[common.get_robot_coords(i).x][common.get_robot_coords(i).y]);
            set_robot_distance(i,map[common.get_robot_coords(i).x][common.get_robot_coords(i).y][min_index]);
            set_robot_direction(i,min_index);
            set_robot_movement(i,first_act[common.get_robot_coords(i).x][common.get_robot_coords(i).y][min_index]);
        }
        for (int i=0; i!=common.get_treasure_count(); ++i)
        {
            min_index = common.minimal(map[common.get_treasure_coords(i).x][common.get_treasure_coords(i).y]);
            set_treasure_distance(i,map[common.get_treasure_coords(i).x][common.get_treasure_coords(i).y][min_index]);
            set_treasure_direction(i,min_index);
            set_treasure_movement(i,first_act[common.get_treasure_coords(i).x][common.get_treasure_coords(i).y][min_index]);
        }

        for (int i=0; i!=map.length; ++i)
            for (int j=0; j!=map[i].length; ++j)
                for (int k=0; k!=map[i][j].length; ++k)
                {
                    projected_map[i][j][k] = map[i][j][k];
                    projected_first_act[i][j][k] = first_act[i][j][k];
                }
    }

    /**
     * Provede pruzkum vzdalenosti z robota r
     */
    public void reachability_from_robot(Robot r, int bound) {
        reachability_from(r.getPositionX(), r.getPositionY(),
                          common.direction_to_int(r.getDirection()), bound);
    }

}

