/*
 * TakeTactic.java
 *
 * Created on 21. květen 2006, 17:44
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.hrobot;

import java.util.List;
import java.util.LinkedList;
import cz.muni.fi.fibot.hrobot.Tactic;
import cz.muni.fi.fibot.hrobot.HrobotCommon;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Strategy;
        

/**
 *
 * @author p
 */
public class TakeTactic implements Tactic {

    private class ChosenTreasure {
        public int x;
        public int y;
        public int direction;
        public int id;
        public int value;
        public int distance;
    }


    static final float FOREIGN_TREASURE_PESIMISM = 0.5f;
    static final float COMPETITION_TREASURE_PESIMISM = 0.95f;
    static final float DISTANCE_DIFFERENCE_DISADVANTAGE = 0.9f;
    static final float MEET_BATTERY_OPIMISM = 0.6f;
    static final int   MULTIPLY_TO_SHOT = 4;

    private HrobotCommon common;
    private int normalized_profit;
    private List<ChosenTreasure> chosen_treasure_path;
    private Strategy.BotAction first_bot_action; //prvni akce na chosen_treasure_path
    private boolean going_through_battery;       //jestli prvni ace vede k baterii
    private int aimed_battery_id;                //ID baterie, kam mirim
    

    
    /** Creates a new instance of TakeTactic */
    public TakeTactic() {

    }
    
    public void setHrobotCommon(HrobotCommon common_arg)
    {
        common = common_arg;
    }

    public void execute()
    {
        ChosenTreasure chosen_treasure;
        int result = 0;
        chosen_treasure_path = new LinkedList<ChosenTreasure>();

        Robot robot = common.get_robot();
        int aux_remaining_rounds = common.get_remaining_rounds();
        int aux_robot_energy = robot.getEnergy();
        int needed_resting_rounds = 0;
        int x = robot.getPositionX(), y = robot.getPositionY();
        int direction = common.direction_to_int(robot.getDirection());
        int treasure_count = 0;
        boolean[] is_chosen = new boolean[common.get_treasure_count()];
        boolean compute_first_bot_action = true;
        boolean no_proper_treasure = false;

        while (aux_remaining_rounds>0 && !no_proper_treasure)
        {
            chosen_treasure = choose_next_treasure(x,y,direction,aux_remaining_rounds,
                                                   is_chosen,compute_first_bot_action);
            compute_first_bot_action = false; //pocitej first_bot_action jen poprve
            if (chosen_treasure.id != Integer.MAX_VALUE)
            {
                chosen_treasure_path.add(chosen_treasure);
                result += chosen_treasure.value;
                aux_remaining_rounds -= chosen_treasure.distance;
                aux_robot_energy -= chosen_treasure.distance;
                if (aux_robot_energy<0)
                {
                    int resting_energy = common.get_world().getRestRobotEnergy();
                    needed_resting_rounds = (int)((float)((-aux_robot_energy)/resting_energy)*MEET_BATTERY_OPIMISM)+1;
                    aux_remaining_rounds -= needed_resting_rounds;
                    aux_robot_energy += needed_resting_rounds*resting_energy;
                }
            }
            else no_proper_treasure = true;
        }
        
        int maybe_near = 0;
        
        if (chosen_treasure_path.size()>=1)
        {
            ChosenTreasure first_chosen = chosen_treasure_path.get(0);
            maybe_near = first_chosen.value * 1024 / (first_chosen.distance+1);
        }
        
        result *= 1024;
        result /= common.get_remaining_rounds();
        
        if (result>maybe_near) normalized_profit = result;
        else normalized_profit = maybe_near;
    }

    /**
     * zjisti, kteri roboti maji nejbliz k jakym pokladum
     */
    private void find_nearest_robots(int[] closest, int[] same_count,
                                     Reachability xy_reach, boolean[] is_chosen)
    {
        int adversary_distance; 
        //urceni nejblizsich robotu k pokladum
        for (int i=0; i!=xy_reach.get_treasure_count(); ++i)
        {
            if (!is_chosen[i]) //jiz jednou vybrane neuvazujem
            {
                int min_dist = xy_reach.get_treasure_distance(i);
                closest[i] = common.get_robot().getId();
                same_count[i] = 1;

                //projdou se souperi a zjisti se, zda nektery neni blize
                for (int j=0; j!=common.get_adversary_count(); ++j)
                {
                    adversary_distance = common.get_reachability_from_adversary(j).get_treasure_distance(i);
                    if (adversary_distance<min_dist)
                    {
                        closest[i] = common.get_adversary(j).getId();
                        min_dist = adversary_distance;
                        same_count[i] = 1;
                    }
                    //pri stejnem poctu se zaznamena, ze je to takovy pripad
                    else if (adversary_distance==min_dist) same_count[i] += 1;

                }
            }

        }
    }

    private ChosenTreasure choose_next_treasure(int x, int y, int direction,
                                                int remaining_rounds,
                                                boolean[] is_chosen,
                                                boolean compute_first_bot_action) {
        ChosenTreasure result = new ChosenTreasure();
        Reachability xy_reach;
        if (compute_first_bot_action && !common.get_timedout()) xy_reach = new UltraReachability(common);
        else xy_reach = new SimpleReachability(common);
        xy_reach.reachability_from(x,y,direction, remaining_rounds);

        //ID robotu, kteri jsou nejbliz
        int[] closest = new int[xy_reach.get_treasure_count()];
        //pocet robotu se stejnou vzdalenosti:
        int[] same_count = new int[xy_reach.get_treasure_count()];


        find_nearest_robots(closest, same_count, xy_reach, is_chosen);

        //maximalni hodnota pokladu, kte kteremu jsou nejbliz protivnici
        int max_reachable_foreign_value = 0;
        int max_reachable_foreign_index = Integer.MAX_VALUE;

        //poklad maximalni hodnoty takovy, ze je k nemu robot s protivniky stejne daleko
        int max_competition_reachable_value = 0;
        int max_competition_reachable_index = Integer.MAX_VALUE;
        int max_competition_reachable_distance = Byte.MAX_VALUE;

        //pokladu maximalni hodnoty takovy, ze je k nemu robot nejbliz ze vsech robotu
        int max_my_own_reachable_value = 0;
        int max_my_own_reachable_index = Integer.MAX_VALUE;
        int max_my_own_reachable_distance = Byte.MAX_VALUE;

        //soucet pokladu, ke kterym je robot nejbliz ze vsech robotu
        int sum_my_own_reachable = 0;

        //nejblizsi, ke kteremu je robot nejbliz ze vsech robotu
        int nearest_my_own_reachable_value = 0;
        int nearest_my_own_reachable_index = Integer.MAX_VALUE;
        int nearest_my_own_reachable_distance = Byte.MAX_VALUE;
        
        //vybere poklady z vyse uvedenymi charkteristikami
        for (int i=0; i!=xy_reach.get_treasure_count(); ++i)
        {
            if (!is_chosen[i]) //jiz jednou vybrane neuvazujem
            {
                int treasure_distance = xy_reach.get_treasure_distance(i);
                if (treasure_distance!=Byte.MAX_VALUE)
                {
                    int treasure_value = common.get_treasure_coords(i).value;
                    if (max_reachable_foreign_value<treasure_value &&
                            closest[i]!=common.get_robot().getId())
                    {
                        max_reachable_foreign_value = treasure_value;
                        max_reachable_foreign_index = i;
                    }

                    if (closest[i]==common.get_robot().getId())
                    {
                        if (same_count[i]>1)
                            if (max_competition_reachable_value<treasure_value ||
                                   (max_competition_reachable_value==treasure_value &&
                                        treasure_distance<max_competition_reachable_distance))
                            {
                                max_competition_reachable_distance = treasure_distance;
                                max_competition_reachable_value = treasure_value;
                                max_competition_reachable_index = i;
                            }
                        if (same_count[i]==1)
                        {
                            sum_my_own_reachable += treasure_value;

                            //pokud je poklad vetsi nebo je stejne velky, ale bliz...
                            if (max_my_own_reachable_value<treasure_value ||
                                    (max_my_own_reachable_value==treasure_value &&
                                        treasure_distance<max_my_own_reachable_distance))
                            {
                                max_my_own_reachable_distance = treasure_distance;
                                max_my_own_reachable_value = treasure_value;
                                max_my_own_reachable_index = i;
                            }

                            //pokud je bliz nebo je stejne blizko, ale vyssi hodnoty,...
                            if (treasure_distance<nearest_my_own_reachable_distance ||
                                    (treasure_distance==nearest_my_own_reachable_distance &&
                                        treasure_value > nearest_my_own_reachable_value))
                            {
                                nearest_my_own_reachable_distance = treasure_distance;
                                nearest_my_own_reachable_value = treasure_value;
                                nearest_my_own_reachable_index = i;
                            }
                        }
                    }
                }
            }
        }

        //vybere vetsi z max_reachable_foreign a max_competition_reachable, s "max_my_own"
        //se to porovna dale.
        result.id = (((float)(max_reachable_foreign_value)*FOREIGN_TREASURE_PESIMISM) >
                     ((float)(max_competition_reachable_value)*COMPETITION_TREASURE_PESIMISM)) ?
                       max_reachable_foreign_index : max_competition_reachable_index;

        //pokud maximalni hodnata pokladu, ke kterym jsem nejbliz je vetsi nez ty,
        //ke kterym jsem stejne daleko nebo dal (znevyhodneny pesimismem, zda se vyplati tam
        //chodit), tak vyberu ten nejvetsi lokalni poklad
        if (max_my_own_reachable_value>((float)(max_reachable_foreign_value)*FOREIGN_TREASURE_PESIMISM) &&
                max_my_own_reachable_value>((float)(max_competition_reachable_value)*COMPETITION_TREASURE_PESIMISM))
        {
            result.id = max_my_own_reachable_index;
        }

        //pokud sumarni hodnota v mych vlastnich presahuje maximalni jinou dosazitenou, tak se
        //nebudu hnat zbytecne za jinou a zkusim posbirat ty vlastni
        if (sum_my_own_reachable > (float)(max_reachable_foreign_value)*FOREIGN_TREASURE_PESIMISM &&
                sum_my_own_reachable > (float)(max_competition_reachable_value)*COMPETITION_TREASURE_PESIMISM)
            result.id = max_my_own_reachable_index;

        if (result.id != Integer.MAX_VALUE && nearest_my_own_reachable_distance != Byte.MAX_VALUE)
        {
            int distance_diff = xy_reach.get_treasure_distance(result.id)-nearest_my_own_reachable_distance;
            //pokud je ta s nejvetsi hodnotou moc daleko, tak radsi pujdu k te blizsi
            if (distance_diff > 0 &&
                nearest_my_own_reachable_value >=
                   (float)(common.get_treasure_coords(result.id).value)*
                        disadvantage_farer_treasure(distance_diff))
                result.id = nearest_my_own_reachable_index;
        }

        //Uz jsme si vybrali ten poklad - dopocitame nektere hodnoty do result
        if (result.id != Integer.MAX_VALUE)
        {
            result.x = common.get_treasure_coords(result.id).x;
            result.y = common.get_treasure_coords(result.id).y;
            result.value = common.get_treasure_coords(result.id).value;
            result.distance = xy_reach.get_treasure_distance(result.id);
            result.direction = xy_reach.get_treasure_direction(result.id);
        }

        if (compute_first_bot_action)
        {
            if (result.id != Integer.MAX_VALUE) //pokud byl vybran nejaky poklad
            {
                //Propocitani, zda nebude lepsi jit pres baterii
//                going_through_battery = false;
//                int resting_energy = common.get_world().getRestRobotEnergy();
//                int remaining_robot_energy = common.get_robot().getEnergy()
//                                             - xy_reach.get_treasure_distance(result.id);
//                int needed_resting_rounds = 0;
//                if (remaining_robot_energy<0)
//                   needed_resting_rounds = (int)(Math.ceil((float)(-remaining_robot_energy)/(float)(resting_energy)));
//
//                int sum_of_needed_rounds = xy_reach.get_treasure_distance(result.id) + needed_resting_rounds;
//                int sum_of_needed_rounds_for_battery;
//                int minimal_sum_for_battery = Integer.MAX_VALUE;
//
//                Reachability from_battery_reach = new Reachability(common);
//
//                for (int i=0; i!=xy_reach.get_battery_count(); ++i)
//                {
//                    if (xy_reach.get_battery_distance(i)<sum_of_needed_rounds &&
//                        xy_reach.get_battery_distance(i)<common.get_robot().getEnergy())
//                    {
//                        from_battery_reach.reachability_from(common.get_battery_coords(i).x,
//                                                             common.get_battery_coords(i).y,
//                                                             xy_reach.get_battery_direction(i),
//                                                             remaining_rounds);
//
//                        sum_of_needed_rounds_for_battery =
//                                from_battery_reach.get_treasure_distance(result.id)+xy_reach.get_battery_distance(i)+1;
//
//                        if (sum_of_needed_rounds_for_battery <= sum_of_needed_rounds &&
//                            sum_of_needed_rounds_for_battery <= minimal_sum_for_battery)
//                        {
//                            going_through_battery = true;
//                            aimed_battery_id = result.id;
//                            minimal_sum_for_battery = sum_of_needed_rounds_for_battery;
//                        }
//
//
//                    }
//                }
//
//
//                /*akce pohybu*/
//                if (going_through_battery)
//                    /* pokud se jde pres baterii, vezme se prvni akce vedouci k baterii*/
//                    first_bot_action = xy_reach.get_battery_movement(aimed_battery_id);
//                else
//                    /* pokud se jde primo k pokladu, vezme se prvni akce vedouci k vybranemu pokladu */
                    first_bot_action = xy_reach.get_treasure_movement(result.id);

                World world = common.get_world();
                Robot robot = common.get_robot();
                
                /*lokalni akce (TAKE, RECHARGE, REST, SHOOT) maji prednost pred pohybem*/
                if (world.containsTreasure(robot.getPositionX(),robot.getPositionY()))
                    /* pokud je to ten poklad, za kterym jdeme, tak se udela TAKE */
                    if (common.get_treasure_coords(result.id).x == robot.getPositionX() && 
                        common.get_treasure_coords(result.id).y == robot.getPositionY())
                    {
                        first_bot_action = Strategy.BotAction.TAKE;
                    }

//                if (world.containsBattery(robot.getPositionX(),robot.getPositionY()))
//                    if (going_through_battery &&
//                           common.get_battery_coords(aimed_battery_id).x == robot.getPositionX() &&
//                           common.get_battery_coords(aimed_battery_id).y == robot.getPositionY())
//                    {
//                        first_bot_action = Strategy.BotAction.RECHARGE;
//                    }
            
            } //konec IF vetve pro pripad uspesneho vybrani pokladu
            //kdyz poklad neni, tato strategie nema co delat
            else first_bot_action = Strategy.BotAction.NOTHING;

//            /*rozhodnuti, zda vystreli, kdyz ho ma na musce*/
//            if (result.value < MULTIPLY_TO_SHOT*points_by_shot(common.get_robot().getPositionX(),
//                                                               common.get_robot().getPositionY(),
//                                                                common.direction_to_int(common.get_robot().getDirection())))
//                first_bot_action = Strategy.BotAction.SHOOT;


        }


        if (result.id != Integer.MAX_VALUE)
            is_chosen[result.id] = true; //ano, je vybrano - ulozi se pro pristi volani funkce

        return result;
    }

    /**
     * Pro dany rozdil ve vzdalenostech od nejakeho mista vypocte celkove znevyhodneni
     * vzdalenejsiho
     */
    public float disadvantage_farer_treasure(int distance_difference)
    {
        float disadvantage = 1;
        for (int i=0; i!=distance_difference; ++i)
            disadvantage *= DISTANCE_DIFFERENCE_DISADVANTAGE;

        return disadvantage;
    }

    private int points_by_shot(int x, int y, int dir) {
        CoordsAndDirection cad = new CoordsAndDirection(x,y,dir);
        cad = common.move_in_direction_from_coords(cad);

        while (common.is_in_map_bounds(cad) && !common.get_world().containsWall(cad.x,cad.y))
        {
            if (common.get_world().containsRobot(cad.x, cad.y)) return (common.get_world().getRobotOn(cad.x,cad.y).getEnergy()/2);
            cad = common.move_in_direction_from_coords(cad);
        }

        return 0;
    }

    
    public Strategy.BotAction getNextAction()
    {
        return first_bot_action;
    }
    
    
    public int getNormalizedProfit()
    {
        return normalized_profit;
    }
    
}
