/*
 * karelWorld.java
 *
 * Created on 11. duben 2006, 14:47
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.karel;

import cz.muni.fi.fibot.*;
import java.util.ArrayList;
import java.util.List;


/**
 *
 * @author xdurovec
 */
class KarelWorld {
    
    public static final short NORTH = 2;
    public static final short WEST = 4;
    public static final short EAST = 8;
    public static final short SOUTH = 16;
    
    private short[][][][][] karelWorld;
    private World world;
    
    public List<Point> poklady = new ArrayList<Point>();
    public List<Point> baterky = new ArrayList<Point>();
    
    public KarelWorld(World world) {
      this.world = world;
      // iniciace Karlova svta
      //System.out.println("Ped new pole");
      karelWorld = new short[world.getDimensionX()][world.getDimensionY()][world.getDimensionX()][world.getDimensionY()][2];
      //System.out.println("Po new pole");
      // spotn vzdlenost ze vech do vech buek
      for (int i = 0; i < world.getDimensionX(); i++)
        for (int j = 0; j < world.getDimensionY(); j++) {
          if (world.containsTreasure(i, j))
            poklady.add(new Point(i, j));
          else if (world.containsBattery(i, j))
            baterky.add(new Point(i, j));          
          spocitejVzdalenostiZBodu(i, j);
        }
    }
    
    public List<Point> getPoklady() {
      List<Point> poklady = new ArrayList<Point>();
        for (int j = 0; j < world.getDimensionY(); j++) {
      for (int i = 0; i < world.getDimensionX(); i++)
          if (world.containsTreasure(i, j))
            poklady.add(new Point(i, j));
        }
      return poklady;
    }
    
    /**
     * Do pole <code>karelWorld</code> so ulo vzdlenosti ze zadanho bodu.
     */
    public void spocitejVzdalenostiZBodu(int x, int y) {
      // iniciace vchozho bodu
      karelWorld[x][y][x][y][0] = 0;
      karelWorld[x][y][x][y][1] = 1;
      // zaloen fronty
      List<Point> fronta = new ArrayList<Point>();
      // vloen potench polek
      setAndAddBunku(x, y, x - 1, y, EAST, (short) 2, fronta);  // zpad
      setAndAddBunku(x, y, x, y - 1, SOUTH, (short) 2, fronta);  // sever
      setAndAddBunku(x, y, x + 1, y, WEST, (short) 2, fronta);  // vchod
      setAndAddBunku(x, y, x, y + 1, NORTH, (short) 2, fronta);  // jih
      // prochzen fronty
      while (!fronta.isEmpty()) {
        // zskme prvn buku
        Point bod = fronta.remove(0);
        int pStrana = karelWorld[x][y][bod.x][bod.y][0];
        int vzdalenost = karelWorld[x][y][bod.x][bod.y][1];
        if (vzdalenost + 1 >= world.getRound()) {
          int zapad = 0, sever = 0, vychod = 0, jih = 0;
          // uren vzdlenosti do sousedn buky podle vchozho smru
          switch (pStrana) {
            case WEST:
              sever = 2; vychod = 1; jih = 2;
              break;
            case NORTH:
              vychod = 2; jih = 1; zapad = 2;
              break;
            case EAST:
              jih = 2; zapad = 1; sever = 2;
              break;
            case SOUTH:
              zapad = 2; sever = 1; vychod = 2;
          }
          // vloen sousednch buek
          if (zapad > 0)  // zpad
            setAndAddBunku(x, y, bod.x - 1, bod.y, EAST, (short) (vzdalenost + zapad), fronta);
          if (sever > 0)  // sever
            setAndAddBunku(x, y, bod.x, bod.y - 1, SOUTH, (short) (vzdalenost + sever), fronta);
          if (vychod > 0)  // vchod
            setAndAddBunku(x, y, bod.x + 1, bod.y, WEST, (short) (vzdalenost + vychod), fronta);
          if (jih > 0)  // jih
            setAndAddBunku(x, y, bod.x, bod.y + 1, NORTH, (short) (vzdalenost + jih), fronta);
        }
      }
    }
    
    /**
     * Pokud zadan buka existuje, ulo se vzdlenost do pole a buka se
     * vlo do fronty, pokud v poli je vt vzdlenost ne ta nov.
     */
    private void setAndAddBunku(int zx, int zy, int x, int y, short vStrana, short vzdalenost, List<Point> l) {
      // pokud buka existuje
      if (x >= 0 && x < world.getDimensionX() && y >= 0 && y < world.getDimensionY()) {
        // pokud tam je ze, uloit do vzdlenosti maximum
        if (world.containsWall(x, y)) {
          karelWorld[zx][zy][x][y][1] = Short.MAX_VALUE;
        } else {
          // pokud nov vzdlenost je men ne ta pvodn, nebo jsme to jet neprochzeli
          if (karelWorld[zx][zy][x][y][1] > vzdalenost || karelWorld[zx][zy][x][y][1] == 0) {
            karelWorld[zx][zy][x][y][0] = vStrana;
            karelWorld[zx][zy][x][y][1] = vzdalenost;
            l.add(new Point(x, y));
          }
        }
      }
    }
    
    /**
     * Vrac vzdlenost z bodu do bodu.
     */
    public int getVzdalenost(int zx, int zy, int x, int y) {
      return karelWorld[zx][zy][x][y][1];
    }
    
    /**
     * Vrac cestu v seznamu z bodu do bodu.
     */
    public List<Point3D> getPath(int zx, int zy, int x, int y) {
      List<Point3D> cesta = new ArrayList<Point3D>();
      // prochzen, dokud nejme na potku
      Point3D p = new Point3D(x, y, 0);
      while (karelWorld[zx][zy][p.x][p.y][1] > 1) {
        int zSmer = karelWorld[zx][zy][p.x][p.y][0];
        // vloen smru, kterm se dostaneme na dan polko
        p.z = zSmer;
        cesta.add(p);   
        switch (zSmer) {
          case WEST:
            p = new Point3D(p.x - 1, p.y, 0);
            break;
          case NORTH:
            p = new Point3D(p.x, p.y - 1, 0);
            break;
          case EAST:
            p = new Point3D(p.x + 1, p.y, 0);
            break;
          case SOUTH:
            p = new Point3D(p.x, p.y + 1, 0);
        }
      }
      return cesta;
    }
    
    public void vypisPole(int x, int y) {
        for (int j = 0; j < world.getDimensionY(); j++) {
      for (int i = 0; i < world.getDimensionX(); i++) {
          if (karelWorld[x][y][i][j][1] == Short.MAX_VALUE || karelWorld[x][y][i][j][1] == 0)
            System.out.print("  x");
          else System.out.print(" " + ((karelWorld[x][y][i][j][1] < 10)?" ":"") + (karelWorld[x][y][i][j][1]-1));
        }
        System.out.println();
      }
    }
    
    public void vypisCestu(List<Point3D> cesta) {
      for (int i = 0; i < cesta.size(); i++) {
        Point3D p = cesta.get(i);
        System.out.println("x = " + p.x + ", y = " + p.y + ", pchod z: " + getSmerName(p.z));
      }
    }
    
    public String getSmerName(int smer) {
        switch (smer) {
          case WEST:
            return "WEST";
          case NORTH:
            return "NORTH";
          case EAST:
            return "EAST";
          case SOUTH:
            return "SOUTH";
        }
      return "";
    }
    
    // JaD
   public int shootValue(int hor, int ver, Robot.Direction direction){
        byte vertical = 0;
        byte horizontal = 0;
        switch (direction){
            case UP: 
                vertical = -1; break;
            case DOWN: 
                vertical = 1; break;
            case LEFT: 
                horizontal = -1; break;
            case RIGHT: 
                horizontal = 1; break;
        }   
             
       int h = hor;
       int v = ver;
       h += horizontal;
       v += vertical;    
       while (h >= 0 && h < world.getDimensionX() && v >= 0 && v < world.getDimensionY()){
        if (world.containsRobot(h,v)) return (int) (world.getRobotOn(h,v).getEnergy() / 2);
        if (world.containsWall(h,v)) return 0;           
        h += horizontal;
        v += vertical;    
       }
       return 0;
    }
  }
