package cz.muni.fi.fibot.lennonbot;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Stack;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot.Direction;

public class LennonBot implements Strategy {
    
    private enum Worth {
        TREASURE, BATTERY
    }
    
    private static final int MAX_X = 51;
    private static final int MAX_Y = 51;
    
    private World world;
    private Robot robot;
    /**
     * matice [][] - pruchodna mista maji 0; trvale nepruchodna -1
     */
    private int wallMap[][];
    /**
     * matice [][] oblasti s nejlepsimi poklady
     * - zdi -1
     * - poklady 1
     * - baterky 2
     * - volna mista 0
     */
    private int worthestArea[][];
    /**
     * (3d) matice nejkratsich cest z mista, ktere bylo argumentem naposledy volane
     * metody shortestPath, directionInt - 0=LEFT, 1=UP, 2=RIGHT, 3=DOWN
     * WALL=9999
     */
    private int shortPathMap[][][];
    private HashSet<Node> predecessor[][][];  //X x Y x Direction
    // private Node predecessor[][][];
    /**
     * mna uzlu jake vlastnosti?
     */
    private HashSet<Node> queueQ;
    private Stack<Strategy.BotAction> actions; //posloupnost nasledujicich akci
    private List<Node> batteries;
    private List<Node> treasures;
    private List<Robot> robotsOnMap;//seznam zivych cizich robotu
    private Node worthestACenter;
    
    public LennonBot() {
        //weKnowWorths = false;
    }
    
    public void initialize() {
        worthestArea = new int[world.getDimensionX()][world.getDimensionY()];
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                worthestArea[i][j]=-2;
            }
        }
        wallMap = new int[MAX_X][MAX_Y];
        shortPathMap = new int[MAX_X][MAX_Y][4];
        predecessor = new HashSet [MAX_X][MAX_Y][4];
        queueQ = new HashSet<Node>(1000);
        actions = new Stack<Strategy.BotAction>();
        worthestACenter= new Node(-1,-1,-1);
    }
    
    public void setWorld(World w) {
        this.world = w;
    }
    
    public void setRobot(Robot r) {
        this.robot = r;
    }
    
    private String toStringShortPathMap() {
        StringBuffer strbuf = new StringBuffer("shortPathMap:\n");
        for (int j=-1; j<world.getDimensionY(); j++) { // radky
            strbuf.append(j+"\t");
            for (int i=0; i<world.getDimensionX(); i++) { // sloupce
                if (j==-1) {
                    strbuf.append(i+"\t");
                    continue;
                }
                if (i>0)
                    strbuf.append("\t");
                if (shortPathMap[i][j][0]==9999)
                    strbuf.append("WALL");
                else for (int k=0; k<=3; k++) {
                    if (k==0)
                        strbuf.append("(");
                    strbuf.append(shortPathMap[i][j][k]);
                    if (k!=3)
                        strbuf.append(",");
                    else
                        strbuf.append(")");
                }
            }
            strbuf.append("\n");
        }
        strbuf.append("\n");
        return strbuf.toString();
    }
    
    private String toStringWall() {
        StringBuffer strbuf = new StringBuffer("walls:\n");
        for (int i = 0; i < MAX_Y; i++) {
            for (int j = 0; j < MAX_X; j++) {
                if (wallMap[j][i]==-1) strbuf.append("#");
                else strbuf.append(" ");
            }
            strbuf.append("\n");
        }
        
        return strbuf.toString();
    }
    private String toStringWorthestArea() {
        StringBuffer strbuf = new StringBuffer("worthestArea:\n ");
        for (int i = 0; i < world.getDimensionY(); i++) {
            strbuf.append(i%10);
        }
        strbuf.append("\n");
        for (int i = 0; i < world.getDimensionY(); i++) {
            strbuf.append(i%10);
            for (int j = 0; j < world.getDimensionX(); j++) {
                
                if (worthestArea[j][i]==-2) strbuf.append(":");
                else if (worthestArea[j][i]==-1) strbuf.append("#");
                else if (worthestArea[j][i]==1) strbuf.append("$");
                else if (worthestArea[j][i]==2) strbuf.append("b");
                else strbuf.append(" ");
            }
            strbuf.append("\n");
        }
        
        return strbuf.toString();
    }
    
// po zmene datove struktury predecessor neni funkcni
    private String toStringPredecessor() {
        StringBuffer strbuf = new StringBuffer("Predecessor:\n");
        for (int j=-1; j<world.getDimensionY(); j++) { // radky
            strbuf.append(j+"\t");
            for (int i=0; i<world.getDimensionX(); i++) { // sloupce
                if (j==-1) {
                    strbuf.append(i+"\t");
                    continue;
                }
                if (i>0)
                    strbuf.append("\t");
                for (int k=0; k<=3; k++) {
                    if (k==0)
                        strbuf.append("[");
                    strbuf.append((predecessor[i][j][k]==null)?"-":predecessor[i][j][k].toString());
                    if (k!=3)
                        strbuf.append(";");
                    else
                        strbuf.append("]");
                }
            }
            strbuf.append("\n");
        }
        strbuf.append("\n");
        return strbuf.toString();
    }
    
    
    /**
     * Jaky se ma provest nasledujici tah.
     */
    public Strategy.BotAction decide() {
        //System.out.println("actions pred: "+actions.size());
        //long startTime = System.nanoTime(); //casova znacka zacatku v nanosekundach
        actions.clear();
        int shootPoints = 0; //body za strelbu
        Robot bestTarget;
        Node nearAreaNode = new Node(-1,-1,-1);
        
        
        // reseni zalozitosti posledniho kola
        if ((world.getRoundsCount()-world.getRound())==1) {
            for (Robot r:robotsOnMap) {
                if ((canShootInOne(r))&&((r.getEnergy()/2)>shootPoints))
                    shootPoints=r.getEnergy()/2;
            }
            if (world.containsBattery(robot.getPositionX(), robot.getPositionY())
            &&world.getMaxRobotEnergy()>shootPoints)
                return Strategy.BotAction.RECHARGE;
            if (world.containsTreasure(robot.getPositionX(), robot.getPositionY())
            &&(world.getTreasureValue(robot.getPositionX(), robot.getPositionY())>
                    world.getRestRobotEnergy())
                    &&world.getTreasureValue(robot.getPositionX(), robot.getPositionY())
                    >shootPoints)
                return Strategy.BotAction.TAKE;
            
            if (world.getRestRobotEnergy()>shootPoints)
                return Strategy.BotAction.REST;
            
            return Strategy.BotAction.SHOOT;
        }
        
       
        // pokud ma robot energii jen 1 a neni posledni kolo
        if (robot.getEnergy() == 1) {
            if (world.containsBattery(robot.getPositionX(), robot.getPositionY())
            &&(!isUnderAttack())) {
                return Strategy.BotAction.RECHARGE;
            }
            return Strategy.BotAction.REST;
        }
        
        
        
        //upraveno, kdy se ma vzit baterka
        if ((world.containsBattery(robot.getPositionX(),robot.getPositionY()))
        && (robot.getEnergy()<(world.getMaxRobotEnergy()-world.getRestRobotEnergy()))
        &&(!isUnderAttack()))
            return BotAction.RECHARGE;
        
        shortestPath(robot.getPositionX(), robot.getPositionY(), robot.getDirection());
        
        //long afterShortestPathTime=System.nanoTime(); //cas po provedeni shortestPath()
        //System.out.println("cas od zacatku za provedeni shortestPath(): " + (Double.toString((afterShortestPathTime - startTime)/(double)1000000)) + " ms");
        robotsOnMap = getRobotsFromMap();
        updateWorthestArea();
        
        
        //vyhodnoceni strelby
        shootPoints =0;
        bestTarget = null;
        if (!robotsOnMap.isEmpty()) {
            bestTarget = robotsOnMap.get(0);
            shootPoints = getShootPoints(bestTarget);
        }
        for (Robot r : robotsOnMap) {
            if (getShootPoints(r)>shootPoints) {
                bestTarget = r;
                shootPoints=getShootPoints(r);
            }
        } //konec vyhodnoceni strelby
        
        // vem poklad kdyz na nem stojis (pokud neni lepsi strilet
        if (world.containsTreasure(robot.getPositionX(), robot.getPositionY()))
            if (world.getTreasureValue(robot.getPositionX(),robot.getPositionY())<
                shootPoints) {
            shoot(bestTarget); return actions.pop();} else
                return Strategy.BotAction.TAKE;
        
        getNodesContaining();
        
        //uprava konce hry - kdyz chyby do konce ZBYTEK (definovany v getZbytek())
        if (getZbytek()>world.getRoundsCount()-world.getRound()) {
            if (!isUnderAttack()) {
                if (canShootInZero(bestTarget)&&shootPoints>world.getRestRobotEnergy())
                    return Strategy.BotAction.SHOOT;
                return Strategy.BotAction.REST;
            }
            if (isUnderAttack()) shoot(bestTarget);
        }
        
        
        //oblast neni prazdna
        if (!isEmptyWorthestArea()) {
            if (!inWorthestArea()) {
                if ((getAreaValue(worthestACenter)<shootPoints)&&
                        !robotsOnMap.isEmpty())
                    shoot(bestTarget);
                else {
                    if (shortPathMap[worthestACenter.getX()][worthestACenter.getY()][worthestACenter.getDir()]
                            !=9999)
                        getActions(worthestACenter.getX(),worthestACenter.getY(),
                                intToDir(worthestACenter.getDir()));
                    else {
                        worthestACenter= new Node(-1,-1,-1);
                        resetWorthestArea();
                    }
                }
                
                
            }
            if (inWorthestArea()){
                nearAreaNode=getNearestInWorthArea();
                if ((shootPoints>((world.getTreasureValue(nearAreaNode.getX(),nearAreaNode.getY())*10)/
                        getTimeCost(nearAreaNode.getX(),nearAreaNode.getY(),nearAreaNode.getDir())))
                        &&!robotsOnMap.isEmpty())
                    shoot(bestTarget);
                else getActions(nearAreaNode.getX(),nearAreaNode.getY(),
                        intToDir(nearAreaNode.getDir()));
            }
        }
        
        //oblast je prazdna
        if (isEmptyWorthestArea()) {
            if (treasures.isEmpty()) {
                hunt();
            } else {
                worthestACenter=getWorthestTreasure();
                if ((getAreaValue(worthestACenter)<shootPoints)&&
                        !robotsOnMap.isEmpty())
                    shoot(bestTarget);
                else {
                    setWorthestArea(worthestACenter);
                    getActions(worthestACenter.getX(),worthestACenter.getY(),
                            intToDir(worthestACenter.getDir()));
                }
            }
            
            //System.out.println(toStringWorthestArea());
            //System.out.println("actions za: "+actions.size());
        }
        if (actions.empty())
            return Strategy.BotAction.REST;
        else
            return actions.pop();
        
    }
    
    /**
     * Upravena metoda shortestPath().
     * - zmeni matici shortPathMap
     * - pripravi k pouziti metodu getActions
     *
     *
     * @param fromX
     * @param fromY
     * @param fromDir
     */
    private void shortestPath(int fromX, int fromY, Robot.Direction fromDir) {
        
        // init of wallMap + shortPathMap
        for (int i=0; i<MAX_X; i++) {
            for (int j=0; j<MAX_Y; j++) {
                if (i<world.getDimensionX()&&j<world.getDimensionY()) {
                    if (world.containsWall(i,j))
                        wallMap[i][j]=-1;
                    else wallMap[i][j]=0;
                } else wallMap[i][j]=-1;
                
                for (int k=0; k<4;k++) {
                    shortPathMap[i][j][k]= 9999;
                }
            }
        }
        shortPathMap[fromX][fromY][directToInt(fromDir)]=0;
        //OPRAVENO jako zdi musime brat i roboty, kteri nemaji 0 energie, nedokaze to pres ne projit
        for (Robot ciziRobot:world.getRobots()){
            if (!ciziRobot.equals(robot))
                wallMap[ciziRobot.getPositionX()][ciziRobot.getPositionY()]=-1;
        }
        //vypisWallMap();
        // mno tady jsem nevedela...
        if (wallMap[fromX][fromY]!=-1)
            dijkstra();
        
    }
    
    /**
     * modifikovany dijkstruv algoritmus
     */
    private void dijkstra() {
        HashSet<Node> doneNodes = new HashSet<Node>();
        Node debugNode;
        Node tempNode;
        
        //init
        doneNodes.clear();
        for (int i=0; i<world.getDimensionX(); i++) {
            for (int j=0; j<world.getDimensionY(); j++) {
                if (!(world.containsWall(i,j))) {
                    queueQ.add(new Node(i,j,0));
                    queueQ.add(new Node(i,j,1));
                    queueQ.add(new Node(i,j,2));
                    queueQ.add(new Node(i,j,3));
                }
            }
        }
        
        // inicializace predecessoru, ktera v puvodni verzi (single-shortest-path) nabyla
        for (int i=0; i<world.getDimensionX(); i++) {
            for (int j=0; j<world.getDimensionY(); j++) {
                if (!(world.containsWall(i,j))) {
                    predecessor[i][j][0]= new HashSet<Node> (4, (float) 0.9);
                    predecessor[i][j][1]= new HashSet<Node> (4, (float) 0.9);
                    predecessor[i][j][2]= new HashSet<Node> (4, (float) 0.9);
                    predecessor[i][j][3]= new HashSet<Node> (4, (float) 0.9);
                }
            }
        }
        // end of init
        while (!queueQ.isEmpty()) {
            tempNode = extractLeast();
            if (tempNode.getX()==-1) break;
            doneNodes.add(new Node(tempNode.getX(),tempNode.getY(), tempNode.getDir()));
            //for vsechny sousedy, tj. vzdy max 3 (go, turn left, turn right)
            switch (tempNode.getDir()) {
                //relaxe direct left
                case 0: {if (tempNode.getX()!=0) { //GO - vlevo neco je
                    if (wallMap[tempNode.getX()-1][tempNode.getY()]==0) { //vlevo neni zed
                        if ((shortPathMap[tempNode.getX()-1][tempNode.getY()][0]>=0&&
                                shortPathMap[tempNode.getX()-1][tempNode.getY()][0]>=
                                shortPathMap[tempNode.getX()][tempNode.getY()][0]+1)||
                                shortPathMap[tempNode.getX()-1][tempNode.getY()][0]==-1) {
                            shortPathMap[tempNode.getX()-1][tempNode.getY()][0]=
                                    shortPathMap[tempNode.getX()][tempNode.getY()][0]+1;
                            //  predecessor[tempNode.getX()-1][tempNode.getY()][0] =
                            //          new Node(tempNode.getX(),tempNode.getY(),0);
                            predecessor[tempNode.getX()-1][tempNode.getY()][0].add(
                                    new Node(tempNode.getX(),tempNode.getY(),0));
                        }
                    }
                }//turn right
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][1]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][1]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][0]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][1]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][1]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][0]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][1] =
                    //        new Node(tempNode.getX(),tempNode.getY(),0);
                    predecessor[tempNode.getX()][tempNode.getY()][1].add(
                            new Node(tempNode.getX(),tempNode.getY(),0));
                }
                //turn left
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][3]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][3]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][0]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][3]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][3]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][0]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][3] =
                    //        new Node(tempNode.getX(),tempNode.getY(),0);
                    predecessor[tempNode.getX()][tempNode.getY()][3].add(
                            new Node(tempNode.getX(),tempNode.getY(),0));
                }
                break;
                }
                
                case 1: {if (tempNode.getY()!=0) { //GO - nahore neco je
                    if (wallMap[tempNode.getX()][tempNode.getY()-1]==0) { //nahore neni zed
                        if ((shortPathMap[tempNode.getX()][tempNode.getY()-1][1]>=0&&
                                shortPathMap[tempNode.getX()][tempNode.getY()-1][1]>=
                                shortPathMap[tempNode.getX()][tempNode.getY()][1]+1)||
                                shortPathMap[tempNode.getX()][tempNode.getY()-1][1]==-1) {
                            shortPathMap[tempNode.getX()][tempNode.getY()-1][1]=
                                    shortPathMap[tempNode.getX()][tempNode.getY()][1]+1;
                            // predecessor[tempNode.getX()][tempNode.getY()-1][1] =
                            //        new Node(tempNode.getX(),tempNode.getY(),1);
                            predecessor[tempNode.getX()][tempNode.getY()-1][1].add(
                                    new Node(tempNode.getX(),tempNode.getY(),1));
                        }
                    }
                    
                }//turn right
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][2]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][2]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][1]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][2]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][2]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][1]+1;
                    // predecessor[tempNode.getX()][tempNode.getY()][2] =
                    //         new Node(tempNode.getX(),tempNode.getY(),1);
                    predecessor[tempNode.getX()][tempNode.getY()][2].add(
                            new Node(tempNode.getX(),tempNode.getY(),1));
                }
                //turn left
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][0]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][0]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][1]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][0]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][0]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][1]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][0] =
                    //        new Node(tempNode.getX(),tempNode.getY(),1);
                    predecessor[tempNode.getX()][tempNode.getY()][0].add(
                            new Node(tempNode.getX(),tempNode.getY(),1));
                }
                break;
                }
                case 2: {if (tempNode.getX()!=world.getDimensionX()-1) { //GO - vpravo neco je
                    if (wallMap[tempNode.getX()+1][tempNode.getY()]==0) { //vlevo neni zed
                        if ((shortPathMap[tempNode.getX()+1][tempNode.getY()][2]>=0&&
                                shortPathMap[tempNode.getX()+1][tempNode.getY()][2]>=
                                shortPathMap[tempNode.getX()][tempNode.getY()][2]+1)||
                                shortPathMap[tempNode.getX()+1][tempNode.getY()][2]==-1) {
                            shortPathMap[tempNode.getX()+1][tempNode.getY()][2]=
                                    shortPathMap[tempNode.getX()][tempNode.getY()][2]+1;
                            // predecessor[tempNode.getX()+1][tempNode.getY()][2] =
                            //         new Node(tempNode.getX(),tempNode.getY(),2);
                            predecessor[tempNode.getX()+1][tempNode.getY()][2].add(
                                    new Node(tempNode.getX(),tempNode.getY(),2));
                        }
                    }
                }//turn right
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][3]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][3]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][2]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][3]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][3]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][2]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][3] =
                    //        new Node(tempNode.getX(),tempNode.getY(),2);
                    predecessor[tempNode.getX()][tempNode.getY()][3].add(
                            new Node(tempNode.getX(),tempNode.getY(),2));
                }
                //turn left
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][1]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][1]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][2]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][1]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][1]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][2]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][1] =
                    //        new Node(tempNode.getX(),tempNode.getY(),2);
                    predecessor[tempNode.getX()][tempNode.getY()][1].add(
                            new Node(tempNode.getX(),tempNode.getY(),2));
                }
                break;
                }
                
                case 3: {if (tempNode.getY()!=world.getDimensionY()-1) { //GO - dole neco je
                    if (wallMap[tempNode.getX()][tempNode.getY()+1]==0) { //dole neni zed
                        if ((shortPathMap[tempNode.getX()][tempNode.getY()+1][3]>=0&&
                                shortPathMap[tempNode.getX()][tempNode.getY()+1][3]>=
                                shortPathMap[tempNode.getX()][tempNode.getY()][3]+1)||
                                shortPathMap[tempNode.getX()][tempNode.getY()+1][3]==-1) {
                            shortPathMap[tempNode.getX()][tempNode.getY()+1][3]=
                                    shortPathMap[tempNode.getX()][tempNode.getY()][3]+1;
                            //predecessor[tempNode.getX()][tempNode.getY()+1][3] =
                            //        new Node(tempNode.getX(),tempNode.getY(),3);
                            predecessor[tempNode.getX()][tempNode.getY()+1][3].add(
                                    new Node(tempNode.getX(),tempNode.getY(),3));
                        }
                    }
                    
                }//turn right
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][0]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][0]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][3]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][0]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][0]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][3]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][0] =
                    //        new Node(tempNode.getX(),tempNode.getY(),3);
                    predecessor[tempNode.getX()][tempNode.getY()][0].add(
                            new Node(tempNode.getX(),tempNode.getY(),3));
                }
                //turn left
                if ((shortPathMap[tempNode.getX()][tempNode.getY()][2]>=0&&
                        shortPathMap[tempNode.getX()][tempNode.getY()][2]>=
                        shortPathMap[tempNode.getX()][tempNode.getY()][3]+1)||
                        shortPathMap[tempNode.getX()][tempNode.getY()][2]==-1) {
                    shortPathMap[tempNode.getX()][tempNode.getY()][2]=
                            shortPathMap[tempNode.getX()][tempNode.getY()][3]+1;
                    //predecessor[tempNode.getX()][tempNode.getY()][2] =
                    //        new Node(tempNode.getX(),tempNode.getY(),3);
                    predecessor[tempNode.getX()][tempNode.getY()][2].add(
                            new Node(tempNode.getX(),tempNode.getY(),3));
                }
                break;
                }
            }
            
        }
        
    }
    
    
    private int directToInt(Robot.Direction direction)  {
        switch (direction) {
            case LEFT: return 0;
            case UP: return 1;
            case RIGHT: return 2;
            case DOWN: return 3;
        }
        return -1;
    }
    
    private Robot.Direction intToDir(int dir) {
        switch (dir) {
            case 0: return Robot.Direction.LEFT;
            case 1: return Robot.Direction.UP;
            case 2: return Robot.Direction.RIGHT;
            case 3: return Robot.Direction.DOWN;
        }
        return null;
    }
    
    /**
     * vybere (a odstrani) z queueQ uzel s nejm hodnotou v shortPathMap
     */
    private Node extractLeast() {
        Node leastNode= new Node(-1,-1,-1);
        Node temp;
        int leastValue = 9999;
        //  System.err.println("queue "+queueQ.size());
        
        for (Iterator<Node> i=queueQ.iterator();i.hasNext();) {
            temp = i.next();
            if (shortPathMap[temp.getX()][temp.getY()][temp.getDir()]<leastValue) {
                leastValue= shortPathMap[temp.getX()][temp.getY()][temp.getDir()];
                leastNode = new Node(temp.getX(),temp.getY(),temp.getDir());
            }
            if ((shortPathMap[temp.getX()][temp.getY()][temp.getDir()]==leastValue)
            &&(world.containsTreasure(temp.getX(), temp.getY()))) {
                leastValue= shortPathMap[temp.getX()][temp.getY()][temp.getDir()];
                leastNode = new Node(temp.getX(),temp.getY(),temp.getDir());
            }
            
        }
        if (leastValue==9999)
            queueQ.clear();
        else
            queueQ.remove(leastNode);
        
        return leastNode;
    }
    
    
    /**
     * Do actions naflaka posloupnost instrukci, ktere vedou z mista,
     *  pro ktere bylo naposledy spusteno shortestPath, do mista urceneho
     *  argumenty.
     * - lze zavolat jen po shortestPath
     * - z mista, z ktereho byla naposled volana shortestPath, pripravi instrukce
     * potrebne k dosazeni TO. Instrukce jsou ulozeny v zasobniku 'actions' - staci je
     * popovat.
     * @param toX
     * @param toY
     * @param toDir
     */
    private void getActions(int toX, int toY, Robot.Direction toDir) {
        //   int x = toX;
        //   int y = toY;
        int dir = directToInt(toDir);
        //    int predX=-1;
        //    int predY=-1;
        int predDir=-1;
        Node[] bestPath = getBestPath(toX, toY, dir);
        
        for (int i=1; i<bestPath.length; i++) {
            //       predX = bestPath[i].getX();
            //       predY = bestPath[i].getY();
            predDir = bestPath[i].getDir();
            
            if (predDir==dir) actions.push(Strategy.BotAction.STEP);
            if (predDir==(dir+1)%4) actions.push(Strategy.BotAction.LEFT);
            if (predDir==(dir+3)%4) actions.push(Strategy.BotAction.RIGHT);
            //       x=predX;
            //       y=predY;
            dir=predDir;
        }
        return;
    }
    
    private Node[] getBestPath(int toX, int toY, int toDir) {
        HashSet<Node[]> allPaths = new HashSet<Node[]>(5);
        HashSet<Node[]> tempPathsSet = new HashSet<Node[]>();
        int pathLength = shortPathMap[toX][toY][toDir];
        int x = toX;
        int y = toY;
        int dir = toDir;
        //actions = new Stack<StrategyBotAction>();
        Node[] aux;
        int maxTreasureValue = 0; //hodnota nejbohatsi cesty
        int auxTreasureValue = 0; //pomocna hodnota pro zjisteni nejboh. cesty
        Node lastTreasure; //aby se nepocitaly 2x poklady pro stejne pole a jiny dir
        if (pathLength<=0||pathLength==9999) return null;
        
        aux = new Node[pathLength+1];
        aux[0]=new Node(toX, toY, toDir);
        allPaths.add(aux);
        //generovani vsech nejkratsich cest
        for (int i=1; i<=pathLength; i++) {
            for (Node[] path:allPaths){
                x=path[i-1].getX();
                y=path[i-1].getY();
                dir=path[i-1].getDir();
                for (Node n: predecessor[x][y][dir]) {
                    aux = path.clone();
                    aux[i]=n;
                    tempPathsSet.add(aux);
                }
                
            }
            allPaths= (HashSet<Node[]>) tempPathsSet.clone();
            tempPathsSet.clear();
        }
        
        //vyber nejlepsi ze vsech
        int auxBatteries =0;
        int maxBatteries =0;
        aux = null;
        for (Node[] path:allPaths) {
            auxTreasureValue = 0;
            lastTreasure = new Node(-1, -1, -1);
            for (int i=0; i<pathLength+1; i++) {
                if ((world.containsTreasure(path[i].getX(), path[i].getY()))
                &&(path[i].getX()!=lastTreasure.getX()&&path[i].getY()!=lastTreasure.getY())) {
                    auxTreasureValue += world.getTreasureValue(path[i].getX(), path[i].getY());
                }
                if ((world.containsBattery(path[i].getX(), path[i].getY()))
                &&(path[i].getX()!=lastTreasure.getX()&&path[i].getY()!=lastTreasure.getY())) {
                    auxBatteries += 1;
                }
            }
            if (auxTreasureValue>maxTreasureValue) {
                maxTreasureValue = auxTreasureValue;
                maxBatteries = auxBatteries;
                aux = path.clone();
            } else
                if ((auxTreasureValue==maxTreasureValue)&&(auxBatteries>maxBatteries)) {
                maxBatteries = auxBatteries;
                aux = path.clone();
                }
            
            if (aux==null) aux= path.clone();
        }
        
        return aux;
    }
    
    private int getPathValue(Node[] nodeArray) {
        int sum = 0;
        for (int i=0; i<nodeArray.length; i++) {
            if (world.containsTreasure(nodeArray[i].getX(), nodeArray[i].getY())) {
                sum += world.getTreasureValue(nodeArray[i].getX(), nodeArray[i].getY());
            }
        }
        
        return sum;
    }
    
    //Vraci pocet baterek na ceste
    private int getPathBatteries(Node[] nodeArray) {
        int sum = 0;
        for (int i=0; i<nodeArray.length; i++) {
            if (world.containsBattery(nodeArray[i].getX(), nodeArray[i].getY())) {
                sum += 1;
            }
        }
        
        return sum;
    }
    
    
    /**
     * Naplni actions kroky, ktere povedou k nejblizsimu robotovi.
     * kdyz neni nejblizsi robot (sme na mape sami) nedela nic
     *
     *musi byt volano po shoot()
     */
    private void hunt() {
        Node nearest;
        
        nearest = getShootNode();
        if (nearest.getX()==-1||nearest.getY()==-1||nearest.getDir()==-1) {
            actions.push(Strategy.BotAction.REST);return;}
        //System.out.println("shootPoint: "+nearest.toString());
        if (shortPathMap[nearest.getX()][nearest.getY()][nearest.getDir()]==0)
            actions.push(Strategy.BotAction.SHOOT);
        else getActions(nearest.getX(),nearest.getY(),intToDir(nearest.getDir()));
    }
    
    // vrati uzel, ze ktereho jde na nekoho strilet
    private Node getShootNode() {
        HashSet<Node> shootPoints = new HashSet<Node>();
        Robot nearest = null;
        int x =0;
        int y =0;
        boolean notWall = true;
        Node nearestPoint = new Node(-1, -1, -1);
        int nearestValue=9999;
        
        //napln kandidatske pozice
        
        for (Robot r:robotsOnMap) {
            //nahoru
            x=r.getPositionX();
            y=r.getPositionY();
            notWall=true;
            
            while (notWall) {
                if (!(y==r.getPositionY())) {
                    if (world.containsWall(x,y)) notWall=false; else
                        shootPoints.add(new Node(x,y,3));
                }
                if (y==0) notWall=false; else y--;
            }
            
            //dolu
            x=r.getPositionX();
            y=r.getPositionY();
            notWall=true;
            while (notWall) {
                if (!(y==r.getPositionY())) {
                    if (world.containsWall(x,y)) notWall=false; else
                        shootPoints.add(new Node(x,y,1));
                }
                if (y+1==world.getDimensionY()) notWall=false; else y++;
            }
            
            // vlevo
            x=r.getPositionX();
            y=r.getPositionY();
            notWall=true;
            while (notWall) {
                if (!(x==r.getPositionX())) {
                    if (world.containsWall(x,y)) notWall=false; else
                        shootPoints.add(new Node(x,y,2));
                }
                if (x==0) notWall=false; else x--;
            }
            
            //vpravo
            x=r.getPositionX();
            y=r.getPositionY();
            notWall=true;
            while (notWall) {
                if (!(x==r.getPositionX())) {
                    if (world.containsWall(x,y)) notWall=false; else
                        shootPoints.add(new Node(x,y,0));
                }
                if (x+1==world.getDimensionX()) notWall=false; else x++;
            }
            
        }
        
        //vyber nejblizsiho z kandidatu
        for (Node n:shootPoints) {
            if (nearestValue>shortPathMap[n.getX()][n.getY()][n.getDir()]) {
                nearestValue=shortPathMap[n.getX()][n.getY()][n.getDir()];
                nearestPoint=n;
            }
        }
        shootPoints.clear();
        
        return nearestPoint;
    }
    
    
    
    
    /* @return true - prave kdyz na nej jde strilet a ho mame ho pred sebou;
     */
    
    
    private boolean canShootInZero(Robot r) {
        boolean canShoot=false;
        if (isInSameRow(r)&&(r.getPositionX()<robot.getPositionX())&&(
                robot.getDirection()==Direction.LEFT)) canShoot=true;
        if (isInSameRow(r)&&(r.getPositionX()>robot.getPositionX())&&(
                robot.getDirection()==Direction.RIGHT)) canShoot=true;
        if (isInSameCol(r)&&(r.getPositionY()<robot.getPositionY())&&(
                robot.getDirection()==Direction.UP)) canShoot=true;
        if (isInSameCol(r)&&(r.getPositionY()>robot.getPositionY())&&(
                robot.getDirection()==Direction.DOWN)) canShoot=true;
        return canShoot;
    }
    
/* @return true - prave kdyz na nej jde strilet a ho mame na boku;
 */
    
    
    private boolean canShootInOne(Robot r) {
        boolean canShoot=false;
        if (isInSameCol(r)&&((robot.getDirection()==Direction.LEFT)||
                (robot.getDirection()==Direction.RIGHT))) canShoot=true;
        if (isInSameRow(r)&&((robot.getDirection()==Direction.UP)||
                (robot.getDirection()==Direction.DOWN))) canShoot=true;
        return canShoot;
    }
    
   /* @return true - prave kdyz na nej jde strilet a ho mame za zady;
    */
    
    
    private boolean canShootInTwo(Robot r) {
        boolean canShoot=false;
        if (isInSameRow(r)&&(r.getPositionX()<robot.getPositionX())&&(
                robot.getDirection()==Direction.RIGHT)) canShoot=true;
        if (isInSameRow(r)&&(r.getPositionX()>robot.getPositionX())&&(
                robot.getDirection()==Direction.LEFT)) canShoot=true;
        if (isInSameCol(r)&&(r.getPositionY()<robot.getPositionY())&&(
                robot.getDirection()==Direction.DOWN)) canShoot=true;
        if (isInSameCol(r)&&(r.getPositionY()>robot.getPositionY())&&(
                robot.getDirection()==Direction.UP)) canShoot=true;
        
        return canShoot;
    }
    
    
    /*@ return body za strelbu prepoctene na kolo - jednotky: body*10/kolo
     *
     */
    
    private int getShootPoints(Robot r) {
        boolean canShoot = false;
        int shootPoints = 0;
        if (canShootInZero(r)) {
            shootPoints=(r.getEnergy()*10)/2;}
        if (canShootInOne(r)) {
            if (r.getEnergy()==1) {
                if (world.containsBattery(r.getPositionX(),r.getPositionY()))
                    shootPoints=(world.getMaxRobotEnergy()*10)/4;
                else shootPoints =(world.getRestRobotEnergy()*10)/4;} else {
                if (world.containsBattery(r.getPositionX(),r.getPositionY())
                &&(r.getLastAction()!=Strategy.BotAction.RECHARGE))
                    shootPoints=(world.getMaxRobotEnergy()*10)/4;
                else
                    shootPoints =((r.getEnergy()-1)*10)/4;
                }
        }
        if (canShootInTwo(r)) {
            if (r.getEnergy()>3) {
                if (r.getEnergy()==2) {
                    if (world.containsBattery(r.getPositionX(),r.getPositionY()))
                        shootPoints=((world.getMaxRobotEnergy()-1)*10)/6;
                    else shootPoints =(world.getRestRobotEnergy()*10)/6;}
                if (r.getEnergy()==1) {
                    if (world.containsBattery(r.getPositionX(),r.getPositionY()))
                        shootPoints=((world.getMaxRobotEnergy()-1)*10)/6;
                    else shootPoints =((world.getRestRobotEnergy()-1)*10)/6;
                }
            } else {
                if (world.containsBattery(r.getPositionX(),r.getPositionY())
                &&(r.getLastAction()!=Strategy.BotAction.RECHARGE))
                    shootPoints=((world.getMaxRobotEnergy()-2)*10)/6;
                else
                    shootPoints =((r.getEnergy()-2)*10)/6;
            }
            
        }
        if (shootPoints<0) shootPoints=0;
        return shootPoints;
    }
    
    
    /**
     * Priprava strileni.
     * Naplni actions podle pozice nepritele. (zatim na strileni a otaceni)
     */
    
    private void shoot(Robot r) {
        if (canShootInZero(r))  actions.push(Strategy.BotAction.SHOOT);
        if (canShootInOne(r)) {
            if ((robot.getDirection()==Direction.LEFT&&r.getPositionY()<robot.getPositionY())
            ||(robot.getDirection()==Direction.UP&&r.getPositionX()>robot.getPositionX())
            ||(robot.getDirection()==Direction.RIGHT&&r.getPositionY()>robot.getPositionY())
            ||(robot.getDirection()==Direction.DOWN&&r.getPositionX()<robot.getPositionX()))
                actions.push(Strategy.BotAction.RIGHT); else
                    actions.push(Strategy.BotAction.LEFT);
        }
        if (canShootInTwo(r)) {
            for (Robot r2:robotsOnMap) {
                if (canShootInOne(r2)) {
                    shoot(r2);
                    return;
                }
                actions.push(Strategy.BotAction.RIGHT);}
        }
        
        
        return;
        
        
    }
    
    
    
    /**
     *  metoda pro urceni zda je robot ve stejnem sloupci jako my
     *  a neni mezi nami zed
     *  UPRAVENO - kdyz je mezi nami jiny robot tak toho za nim to nebere
     */
    private boolean isInSameCol(Robot r) {
        int enemyX = r.getPositionX();
        int enemyY = r.getPositionY();
        int ourX = robot.getPositionX();
        int ourY = robot.getPositionY();
        
        //if (r.getEnergy()==0) return false;
        if (enemyX!=ourX) return false;
        if (enemyX==ourX) { //stejny sloupec
            if (enemyY>ourY) { //pod nami
                for (int i = ourY; i < enemyY; i++) {
                    if (world.containsWall(ourX,i)) return false;
                    if (world.containsRobot(ourX,i)&&(i!=ourY)) return false;
                }
            }
            if (enemyY<ourY) { //nad nami
                for (int j = ourY; j > enemyY; j--) {
                    if (world.containsWall(ourX,j)) return false;
                    if (world.containsRobot(ourX,j)&&(j!=ourY)) return false;
                }
            }
        }
        return true;
    }
    /**
     *  metoda pro urceni zda je robot ve stejnem radku jako my
     *  a nejsou mezi nami zdi
     */
    private boolean isInSameRow(Robot r) {
        int enemyX = r.getPositionX();
        int enemyY = r.getPositionY();
        int ourX = robot.getPositionX();
        int ourY = robot.getPositionY();
        
        //if (r.getEnergy()==0) return false;
        if (enemyY!=ourY) return false;
        if (enemyY==ourY) { //stejny radek
            if (enemyX>ourX) { // vpravo od nas
                for (int k = ourX; k < enemyX; k++) {
                    if (world.containsWall(k,ourY)) return false;
                    if (world.containsRobot(k,ourY)&&(k!=ourX)) return false;
                }
            }
            if (enemyX<ourX) { //vlevo od nas
                for (int l = ourX; l > enemyX; l--) {
                    if (world.containsWall(l,ourY)) return false;
                    if (world.containsRobot(l,ourY)&&(l!=ourX)) return false;
                }
            }
        }
        return true;
    }
    
    
    
    /**
     * Vrati pocet kroku k urcenemu mistu.
     *
     */
    private int getTimeCost(int x, int y, int dir) {
        if ((dir<0)||(dir>3)) dir=directToInt(robot.getDirection());//FIX
        
        int pocetKolOdpocinku,temp,timeCost;
        if ((x==robot.getPositionX())&&(y==robot.getPositionY())&&
                (intToDir(dir)==robot.getDirection())) return 0;
        temp = shortPathMap[x][y][dir]-robot.getEnergy()-1;//ERR
        pocetKolOdpocinku=(temp>0)?temp/world.getRestRobotEnergy():0;
        timeCost=pocetKolOdpocinku+shortPathMap[x][y][dir];
        return timeCost;
    }
    
    /**
     * Najde nejblizsiho robota.
     * pokud neni na mape dalsi robot,vrati null
     * aktualne neni pouzivana
     */
    private Robot getTimeNearestBot() {
        //System.out.println("number of bots on map: "+world.getRobots().size());
        if (robotsOnMap.isEmpty()) return null;
        Robot nBot = robotsOnMap.get(0);
        
        //the nearest bot, for start the first bot in list
        
        for (Robot r : robotsOnMap) {
            //if (r.getEnergy()==0) continue;
            if (getTimeCost(r.getPositionX(),r.getPositionY(),directToInt(r.getDirection()))
            < getTimeCost(nBot.getPositionX(),nBot.getPositionY(),directToInt(nBot.getDirection()))) {
                nBot=r;
            }
            //DROBNE VYLEPSENI - asi bude vyhodnejsi pri vyberu dvou stejne vzdalenych robotu jit strilet
            //na toho ktery ma vice E
            if (getTimeCost(r.getPositionX(),r.getPositionY(),directToInt(r.getDirection()))
            == getTimeCost(nBot.getPositionX(),nBot.getPositionY(),directToInt(nBot.getDirection()))
            && r.getEnergy()>nBot.getEnergy()) {
                nBot=r;
            }
        }
        return nBot;
    }
    /**
     *  do seznamu ulozi roboty, kteri nejsou my a nemaji nulovou E
     */
    public List<Robot> getRobotsFromMap() {
        List<Robot> robotsOnMap = new ArrayList<Robot>();
        for (Robot r : world.getRobots()) {
            if (!r.equals(robot) && (r.getEnergy()!=0)) {
                robotsOnMap.add(r);
            }
        }
        return robotsOnMap;
    }
    
    
    /**
     * Naplni actions pro sber zlata.
     * UPRAVENO: pridana podminka inWorthestArea(0 - kdyz jsme ve worthest area,
     * sbirame poklady z worthest area, kdyz tam nejsme, hleda se nova worthest area
     * aktualne neni pouzivana
     *
     */
    private void pickUp() {
        Node nearest = null;
        
        if (!isEmptyWorthestArea() && inWorthestArea()) {
            //System.out.println("bude se hledat v oblasti 3x3");
            nearest = getNearestWorth(Worth.TREASURE);
        } else {
            //System.out.println("bude se hledat na celem planu");
            if (isEmptyWorthestArea()) resetWorthestArea();
            nearest = getWorthestTreasure();//getNearestWorth(Worth.TREASURE);
        }
        //System.out.println("Pick up treasure:"+nearest+", areaValue is "+getAreaValue(nearest));
        getActions(nearest.getX(), nearest.getY(), intToDir(nearest.getDir()));
    }
    
    private Node getNearestInWorthArea() {
        int nearest=9999;
        Node nearestNode = new Node(-1, -1, -1);
        if (isEmptyWorthestArea()) return nearestNode;
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                if (worthestArea[i][j]==1) {
                    if (nearest>shortPathMap[i][j][getWorthDir(i,j)]){
                        nearest=shortPathMap[i][j][getWorthDir(i,j)];
                        nearestNode = new Node(i,j,getWorthDir(i,j));
                    }
                }
            }
        }
        if (nearest==9999) {
            resetWorthestArea();
            worthestACenter = new Node(-1,-1,-1);
        }
        return nearestNode;
    }
    
    /**
     * Zjisti pozici uzlu, ktery obsahuje nejblizsi zlato/baterku.
     * @param worthType - zlato nebo baterka?
     * @return node - uzel s patricnym objektem
     * ZMENA - vraci nejlepsi z nejblizsich, proto MUSI byt spusten az PO
     * shortestPath()
     *
     *
     */
    private Node getNearestWorth(Worth w) {
        //if (!this.weKnowWorths)
        //    getNodesContaining();
        
        if ((w==Worth.BATTERY && batteries.isEmpty()) || (w==Worth.TREASURE && treasures.isEmpty()))
            return null;
        List<Node> worthList = ((w==Worth.TREASURE)?this.treasures:this.batteries);
        Node nearestWorthNode = worthList.get(0);
        for (Node n:worthList) {
            if (getTimeCost(n.getX(), n.getY(), n.getDir()) < getTimeCost(nearestWorthNode.getX(), nearestWorthNode.getY(), nearestWorthNode.getDir()))
                nearestWorthNode = n;
        }
        
        return nearestWorthNode;
    }
    
    /**
     * Zjisti pozici uzlu, ktery obsahuje zlato s nejvetsi hodnotou.
     * Pokud ma neco stejnou hodnotu, porovna vzdalenosti a vrati blizsi.
     * DODELAVKA - Pokud jsou 2 poklady se stejnou hodnotou stejne vzdaleny,
     * pujde k tomu jehoz cesta je bohatsi na poklady.
     * @return
     */
    private Node getWorthestTreasure() {
        if (treasures.isEmpty())
            return null;
        Node worthest = treasures.get(0);
        //int maxValue = world.getTreasureValue(worthest.getX(),worthest.getY());
        int maxValue = getAreaValue(worthest); //
        HashSet<Node> candidatesSet = new HashSet<Node>(5); //mnozina casove nejblizsich&&nejhodnotjsich pokladu
        //sber uzlu do candidateSet
        for (int i=0; i<4; i++) {
            if (shortPathMap[worthest.getX()][worthest.getY()][i]==shortPathMap[worthest.getX()][worthest.getY()][worthest.getDir()])
                candidatesSet.add(new Node(worthest.getX(), worthest.getY(), i));
        }
        
        // konec inicializace, zacina se plnit candidatesSet
        for (Node n:treasures) {
            if //(world.getTreasureValue(n.getX(), n.getY()) > maxValue)
                    (getAreaValue(n) > maxValue) {
                candidatesSet.clear();
                worthest = n;
                //maxValue=world.getTreasureValue(n.getX(), n.getY());
                maxValue=getAreaValue(n);
                for (int i=0; i<4; i++) {
                    if (shortPathMap[worthest.getX()][worthest.getY()][i]==shortPathMap[worthest.getX()][worthest.getY()][worthest.getDir()])
                        candidatesSet.add(new Node(worthest.getX(), worthest.getY(), i));
                }
            } else
                if // ((world.getTreasureValue(n.getX(), n.getY()) == maxValue)
                    ((getAreaValue(n)==maxValue)
                    &&getTimeCost(n.getX(), n.getY(), n.getDir()) <= getTimeCost(worthest.getX(), worthest.getY(), worthest.getDir())) {
                if (getTimeCost(n.getX(), n.getY(), n.getDir()) < getTimeCost(worthest.getX(), worthest.getY(), worthest.getDir())) {
                    candidatesSet.clear();
                    worthest = n;
                    for (int i=0; i<4; i++) {
                        if (shortPathMap[worthest.getX()][worthest.getY()][i]==shortPathMap[worthest.getX()][worthest.getY()][worthest.getDir()])
                            candidatesSet.add(new Node(worthest.getX(), worthest.getY(), i));
                    }
                } else {
                    for (int i=0; i<4; i++) {
                        if (shortPathMap[worthest.getX()][worthest.getY()][i]==shortPathMap[worthest.getX()][worthest.getY()][worthest.getDir()])
                            candidatesSet.add(new Node(worthest.getX(), worthest.getY(), i));
                    }
                }
                
                }
        }
        
        // vyber nejlepsiho
        
        
        
        int tempBattery = 0;
        int maxBattery = 0;
        int tempValue=0;
        maxValue=0;
        worthest = new Node(-1, -1,-1);
        
        for (Node n:candidatesSet) {
            tempValue=getPathValue(getBestPath(n.getX(), n.getY(), n.getDir()));
            
            if (tempValue>maxValue) {
                maxValue=tempValue;
                maxBattery=getPathBatteries(getBestPath(n.getX(), n.getY(), n.getDir()));;
                worthest=n;
            } else
                if (tempValue==maxValue) {
                tempBattery=getPathBatteries(getBestPath(n.getX(), n.getY(), n.getDir()));
                if (tempBattery>maxBattery) worthest=n;
                
                }
            
        }
        
        
        //UPRAVENO
        //System.out.println("worthest:"+worthest);
        return worthest;
    }
    
   /* Vraci hodnotu lokality, ve ktere se vyskytuje uzel v prostred
    * NAVRH - udelat vedlejsi nadrazenejsi strategii, ktera nastavi flag boolean minorTask
    * a docasne zmeni chovani prave v dobe, kdy se robot nachazi v zone s poklady
    */
    
    private int getAreaValue(Node n) {
        int sum = 0;
        int tCount=0;
        if (world.containsTreasure(n.getX(), n.getY())) { //stred
            sum=world.getTreasureValue(n.getX(), n.getY());
            tCount++;
        }
        if (n.getX()>0) {if (world.containsTreasure(n.getX()-1, n.getY())) { //leve
            sum+= world.getTreasureValue(n.getX()-1, n.getY());
            tCount++;
        }
        }
        if (n.getX()<world.getDimensionX()-1) {if (world.containsTreasure(n.getX()+1, n.getY())) {//prave
            sum+= world.getTreasureValue(n.getX()+1, n.getY());
            tCount++;}
        }
        if (n.getY()>0)  {if (world.containsTreasure(n.getX(), n.getY()-1)) {//horni
            sum+= world.getTreasureValue(n.getX(), n.getY()-1);
            tCount++;}
        }
        if (n.getY()<world.getDimensionY()-1) {if (world.containsTreasure(n.getX(), n.getY()+1)) {//dolni
            sum+= world.getTreasureValue(n.getX(), n.getY()+1);
            tCount++;}
        }
        if ((n.getX()>0)&&(n.getY()>0)&&((wallMap[n.getX()-1][n.getY()]==0)||(wallMap[n.getX()][n.getY()-1]==0))) {//leve horni
            if (world.containsTreasure(n.getX()-1, n.getY()-1)) {
                sum+= world.getTreasureValue(n.getX()-1, n.getY()-1);
                tCount++;}
        }
        if ((n.getX()>0)&&(n.getY()<world.getDimensionY()-1)&&((wallMap[n.getX()-1][n.getY()]==0)||(wallMap[n.getX()][n.getY()+1]==0))) {//leve dolni
            if (world.containsTreasure(n.getX()-1, n.getY()+1)) {
                sum+= world.getTreasureValue(n.getX()-1, n.getY()+1);
                tCount++;}
        }
        if ((n.getX()<world.getDimensionX()-1)&&(n.getY()>0)&&((wallMap[n.getX()+1][n.getY()]==0)||(wallMap[n.getX()][n.getY()-1]==0))) {//prave horni
            if (world.containsTreasure(n.getX()+1, n.getY()-1)) {
                sum+= world.getTreasureValue(n.getX()+1, n.getY()-1);
                tCount++;}
        }
        if ((n.getX()<world.getDimensionX()-1)&&(n.getY()<world.getDimensionY()-1)&&((wallMap[n.getX()+1][n.getY()]==0)||(wallMap[n.getX()][n.getY()+1]==0))) {
            if (world.containsTreasure(n.getX()+1, n.getY()+1)) {//prave dolni
                sum+= world.getTreasureValue(n.getX()+1, n.getY()+1);
                tCount++;}
        }
        //zahadna konstanta 3 - ma simulovat vzdalenost mezi poklady v Area
        // tCount-1, aby se 2x nepocital prostredni
        return (sum*10)/(getTimeCost(n.getX(),n.getY(),n.getDir())+3*(tCount-1));
    }
    
    /**
     *  nastavi pole worthestArea[][]tak,ze vsechna mista krome dane oblasti
     *  jsou zdi, mista nedosazitelna ze stredu oblasti pouze pomoci poli oblasti
     *  jsou zdi a jsou zde vyznaceny poklady,roboti a baterky v dane oblasti
     *  (robot je bran jako zed.)
     */
    public void setWorthestArea(Node n) {
        resetWorthestArea();
        
        for (int i = n.getX()-1; i <= n.getX()+1; i++) {
            for (int j = n.getY()-1; j <= n.getY()+1; j++) {
                if (i>=0 && i<world.getDimensionX() && j>=0 && j<world.getDimensionY()) {
                    if (!world.containsWall(i,j)) {
                        worthestArea[i][j] = 0;
                    } else {
                        worthestArea[i][j] = -1;
                    }
                    
                    if (world.containsTreasure(i,j)) {
                        worthestArea[i][j]=1;
                    }
                    if (world.containsBattery(i,j)) {
                        worthestArea[i][j]=2;
                    }
                    if (world.containsRobot(i,j) && robot.getPositionX()!=i && robot.getPositionY()!=j) {
                        worthestArea[i][j]=-1;
                    }
                }
            }
        }
        //zruseni nedostupnych poli ve ctverci
        //leve horni
        if ((n.getX()>0)&&(n.getY()>0)&& (n.getX()<world.getDimensionX()-1) && (n.getY()<world.getDimensionY()-1)
        &&(wallMap[n.getX()-1][n.getY()]==-1)&&(wallMap[n.getX()][n.getY()-1]==-1)) {
            worthestArea[n.getX()-1][n.getY()-1]=-1;
        }
        //prave horni
        if ((n.getX()>0)&&(n.getY()>0)&& (n.getX()<world.getDimensionX()-1) && (n.getY()<world.getDimensionY()-1)
        &&(wallMap[n.getX()+1][n.getY()]==-1)&&(wallMap[n.getX()][n.getY()-1]==-1)) {
            worthestArea[n.getX()+1][n.getY()-1]=-1;
        }
        //leve dolni
        if ((n.getX()>0)&&(n.getY()>0)&& (n.getX()<world.getDimensionX()-1) && (n.getY()<world.getDimensionY()-1)
        &&(wallMap[n.getX()-1][n.getY()]==-1)&&(wallMap[n.getX()][n.getY()+1]==-1)) {
            worthestArea[n.getX()-1][n.getY()+1]=-1;
        }
        //prave dolni
        if ((n.getX()>0)&&(n.getY()>0)&& (n.getX()<world.getDimensionX()-1) && (n.getY()<world.getDimensionY()-1)
        &&(wallMap[n.getX()+1][n.getY()]==-1)&&(wallMap[n.getX()][n.getY()+1]==-1)) {
            worthestArea[n.getX()+1][n.getY()+1]=-1;
        }
    }
    /**
     * @return true if we're in worthestArea, false otherwise
     */
    public boolean inWorthestArea() {
        if (worthestArea[robot.getPositionX()][robot.getPositionY()]!=-2) return true;
        return false;
    }
    
    /**
     *  @return true if sum of all treasures in worthestArea is zero, false otherwise
     */
    public boolean isEmptyWorthestArea() {
        int sum=0;
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                if (worthestArea[i][j]==1) {
                    sum+=world.getTreasureValue(i,j);
                }
            }
        }
        return (sum==0)?true:false;
    }
    /**
     * updatuje worthestArea[][], melo by byt volano kazde kolo.
     * poklady, ke kterym se nedostane (malo casu), se automaticky vymazou
     * stejne tak se vymazou poklady, na kterych uz stoji jiny robot
     */
    public void updateWorthestArea() {
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                if (worthestArea[i][j]==1) {
                    if (!world.containsTreasure(i,j)) worthestArea[i][j]=0;
                    if ((world.containsRobot(i,j)&&(robot.getPositionX()!=i||robot.getPositionY()!=j)))
                        worthestArea[i][j]=0;
                    else{if(world.containsTreasure(i,j)&&shortPathMap[i][j][0]==9999) //stalo se, ze robot byl odstaveny od mista z pokladem v Area (vlevo zed, jeden nad, jeden pod, jeden vpravo, poklad vpravoNahore
                        worthestArea[i][j]=0; else
                            if(world.containsTreasure(i,j)&&
                            (getTimeCost(i,j,getWorthDir(i,j))>world.getRoundsCount()-world.getRound()))
                                worthestArea[i][j]=0;
                    }
                }
                if ((worthestArea[i][j]==0)&&world.containsTreasure(i,j)&&
                        shortPathMap[i][j][0]!=9999) worthestArea[i][j]=1;
                if (world.containsTreasure(i,j)&&!containsForeignRobot(i,j)&&
                        worthestArea[i][j]==-1) worthestArea[i][j]=1;
            }
        }
        if (isEmptyWorthestArea()) {
            resetWorthestArea();
            worthestACenter = new Node(-1,-1,-1);
        }
    }
    
    /**
     * resetuje mapu worthestArea[][]
     */
    public void resetWorthestArea() {
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                worthestArea[i][j]=-2;
            }
        }
    }
    
    
    
    public boolean containsForeignRobot(int x, int y) {
        
        return (world.containsRobot(x,y)&&(robot.getPositionX()!=x||robot.getPositionY()!=y));
    }
    
    /**
     *  vraci misto nejbohatsiho pokladu z dane oblasti 3x3
     */
    public Node getWorthestTreasureInArea() {
        //System.out.println("in getWorthestTreasureInArea() method");
        int xStart = robot.getPositionX()-1;
        int xStop = robot.getPositionX()+1;
        int yStart = robot.getPositionY()-1;
        int yStop = robot.getPositionY()+1;
        Node worthest = null;
        int worthestValue = 0;
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                if (world.containsTreasure(i,j) && world.getTreasureValue(i,j)>worthestValue ) {
                    worthestValue = world.getTreasureValue(i,j);
                    worthest = new Node(i,j,0);
                }
            }
        }
        
        return worthest;
    }
    
    /**
     * Zjisti vsechny uzly, ktere obsahuji zlato/baterku a ulozi je do seznamu baterek/zlata.
     * PRIDANO - VASEK - puvodne posilal robota i k pokladum, ktere jsou nedosazitelne,
     * takove poklady (ani baterky) tedy nebudou pridavany do ArrayListu
     * - tim vznika nutnost volat tuto metodu az po ShortestPath
     * PRIDANO - VASEK - pocita jen s poklady, ke kterym je schopny dojit
     *
     */
    private void getNodesContaining() {
        batteries = new ArrayList<Node>();
        treasures = new ArrayList<Node>();
        
        for (int x=0; x < world.getDimensionX(); x++) {
            for (int y=0; y < world.getDimensionY(); y++) {
                if (world.containsBattery(x, y)&&(shortPathMap[x][y][0]!=9999)&&
                        (getTimeCost(x,y, getWorthDir(x,y))>=(world.getRoundsCount()-world.getRound()+1)))
                    batteries.add(new Node(x, y, getWorthDir(x, y)));
                if (world.containsTreasure(x, y)&&(shortPathMap[x][y][0]!=9999)&&
                        (getTimeCost(x,y, getWorthDir(x,y))<=(world.getRoundsCount()-world.getRound()+1)))
                    treasures.add(new Node(x, y, getWorthDir(x, y)));
            }
        }
        //this.weKnowWorths = true;
        
    }
/*HashSet<Node> candidatesSet = new HashSet<Node>(5); //mnozina casove nejblizsich&&nejhodnotjsich pokladu
        //sber uzlu do candidateSet
        for (int i=0; i<4; i++) {
            if (shortPathMap[worthest.getX()][worthest.getY()][i]==shortPathMap[worthest.getX()][worthest.getY()][worthest.getDir()])
                candidatesSet.add(new Node(worthest.getX(), worthest.getY(), i));
 */
    
    /**
     * Zjisti, z ktereho je smeru baterka nejbliz
     * @param x
     * @param y
     * @return vysledny smer robota v zaveru cesty do uzlu
     * s nejkratsi a nejbohatsi cestou
     */
    private int getWorthDir(int x, int y) {
        int bestDir=0;
        int tempValue=0;
        int bestValue=0;
        int shortest=9999;
        
        if (x==robot.getPositionX()&&y==robot.getPositionY()) return 0;
        for (int i = 0; i < 4; i++) {
            tempValue = getPathValue(getBestPath(x,y,i));
            if (shortPathMap[x][y][i]<shortest) {
                bestDir=i;
                shortest=shortPathMap[x][y][i];
                bestValue=tempValue;
            } else {
                if (shortPathMap[x][y][i]==shortest) {
                    if (tempValue>bestValue) {
                        bestDir=i;
                        bestValue=tempValue;
                    } else {
                        if ((tempValue==bestValue)&&(getPathBatteries(getBestPath(x,y,i))
                        >getPathBatteries(getBestPath(x,y,bestDir)))) {
                            bestDir=i;
                        }
                    }
                }
            }
        }
        return bestDir;
    }
    /**
     *  @return pocet kol do konce,ktere je treba odpocivat, abychom na konci meli fullE
     */
    public int getZbytek() {
        if ((world.getMaxRobotEnergy()-robot.getEnergy())%world.getRestRobotEnergy()==0)
            return (world.getMaxRobotEnergy()-robot.getEnergy())/world.getRestRobotEnergy();
        return ((world.getMaxRobotEnergy()-robot.getEnergy())/world.getRestRobotEnergy())+1;
    }
    
    /**
     *  @return pocet kroku potrebnych k dobyti u nejblizsi baterky
     */
    public int getNearestBattDistance() {
        Node nb = getNearestWorth(Worth.BATTERY);
        if (nb==null) return 9999;
        int batteryDistance = getTimeCost(nb.getX(),nb.getY(),getWorthDir(nb.getX(),nb.getY()));
        return batteryDistance+1;
    }
    
    /**
     *  return true if we are under attack, else otherwise
     */
    public boolean isUnderAttack() {
        
        boolean isOnUp, isOnDown, isOnLeft, isOnRight;
        for (Robot r : robotsOnMap) {
            isOnUp = robot.getPositionY()>r.getPositionY();
            isOnDown = robot.getPositionY()<r.getPositionY();
            isOnLeft = robot.getPositionX()>r.getPositionX();
            isOnRight = robot.getPositionX()<r.getPositionX();
            if (r.getLastAction()==Strategy.BotAction.SHOOT) {
                if  ((isInSameRow(r) && isOnLeft && directToInt(r.getDirection())==2)
                ||(isInSameRow(r) && isOnRight && directToInt(r.getDirection())==0)
                ||(isInSameCol(r) && isOnUp && directToInt(r.getDirection())==3)
                ||(isInSameCol(r) && isOnDown && directToInt(r.getDirection())==1)) {
                    //System.out.println("underattack");
                    return true;
                }
            }
        }
        return false;
    }
    
    
    /**
     *
     * dir - 0=LEFT, 1=UP, 2=RIGHT, 3=DOWN
     */
    class Node {
        private int x;
        private int y;
        private int dir;
        
        public Node(int x, int y, int dir) {
            this.x = x;
            this.y = y;
            this.dir = dir;
        }
        
        // nemenit tuto metodu, na jejim presnem tvaru je zavisla jina fce
        public String toString() {
            return (x+"x"+y+"y"+dir+"d"); }
        
        public boolean equals(Object o){
            return ((o instanceof Node)&&
                    ((Node)o).x==this.x&&
                    ((Node)o).y==this.y&&
                    ((Node)o).dir==this.dir);
        }
        public int hashCode() {
            return (this.dir+this.y*100+this.x*10000);
        }
        public int getX() {return x;}
        public int getY() {return y;}
        public int getDir() {return dir;}
        
    }
    
}