package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.number5.Number5Square.SquareType;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;

import java.util.List;

public class Number5 implements Strategy {
	
	/* Verejne rozhrani world
	 * pouzitelne metody:
	 * - world.getDimensionX() - rozmer pole
	 * - world.getDimensionY() - rozmer pole
	 * - world.getRestRobotEnergy()
	 * - world.getMaxRobotEnergy()
	 */
	protected World world = null;
	
	/*
	 * objekt naseho robota (stejny objekt jako v poli robotu robots)
	 */
	protected Number5Robot me = null;
	 
	/* toto pole bude stejne po celou dobu hry, tedy se lze spolehat na to,
	 *	ze jeden robot bude mit stejny index v tomto poli po celou hru
	 */
	protected Number5Robot robots[] = null;
	
	/* 
	 * cela mapa - zapisovat do ni smi jenom jadro a predzpracovani
	 */
	protected Number5Square gameMap[][] = null;
        
        protected Number5Square special[] = null;
	
	/*
	 * aktualni idealni uroven energie - smi upravovat jen jadro a STRATEGIE U MOCI
	 * pri zmene strategie u moci je tato hodnota znovu nastavena metodou resetIdealEnergy
	 */
	protected int idealEnergyLevel = -1;
	
	/* rozhrani objektu na hledani cest */
	protected Number5PathFinder path = null;
	 
        protected boolean debug = false;
	
	public Number5(){
		run = new Number5Run(this);
	}
	
	public void setWorld(World w) {
		world = w;
	}
	
	public void setRobot(Robot robot) {
		initMe = robot;
	}
        
        public void initialize(){
            int x, y, treasure;
            SquareType type;

            resetIdealEnergy();

            // vytvori pole robotu
            List<Robot> bots = world.getRobots();
            robots = new Number5Robot[bots.size()];

            for(x = 0; x < bots.size(); x++){
                    robots[x] = new Number5Robot(bots.get(x), x, this);
            }

            // najdi v seznamu robotu sam sebe
            for (Number5Robot r : robots){
                    if(r.robot == initMe) me = r;
            }

            // vytvori hraci pole
            int width = world.getDimensionX();
            int height = world.getDimensionY();
            gameMap = new Number5Square[width][height];
            
            int specialCount = 0;
            
            // vytvor pole specialnich policek
            for(x = 0; x < width; x++){
                    for(y = 0; y < height; y++){
                            if(world.containsBattery(x, y) || world.getTreasureValue(x, y) > 0) specialCount++;
                    }
            }
            special = new Number5Square[specialCount];
            specialCount = 0;
            
            for(x = 0; x < width; x++){
                    for(y = 0; y < height; y++){

                            if(world.containsWall(x, y)) type = SquareType.WALL;
                            else if(world.containsBattery(x, y)) type = SquareType.BATTERY;
                            else type = SquareType.HALL;

                            treasure = world.getTreasureValue(x, y);

                            gameMap[x][y] = new Number5Square(x, y, type, treasure, bots.size(), special.length, me.index);
                            
                            if(type == SquareType.BATTERY || treasure > 0){
                                special[specialCount] = gameMap[x][y];
                                specialCount++;
                            }
                    }
            }

            // uloz pozice robotu
            for(Number5Robot r : robots){
                    gameMap[r.state.positionX][r.state.positionY].robot = r;
            }

            preprocessor = new Number5Preprocessor(this);
            path = new Number5PathFinder(this);
            
            preprocessor.preprocessSpecial();
        }
	
	public BotAction decide() {
		
		int x, y;
		BotAction action = null;
		int maxEnergy = world.getMaxRobotEnergy();
		
		// zavolej akce robotu
		for(Number5Robot r : robots){
			
			// zjisti, co udelal
			if(r.state.positionX != r.robot.getPositionX() || r.state.positionY != r.robot.getPositionY()) {
				action = BotAction.STEP;
				// ze stareho policka robot zmizel
				gameMap[r.state.positionX][r.state.positionY].robot = null;
				// na novem se objevil
				gameMap[r.robot.getPositionX()][r.robot.getPositionY()].robot = r;
				
			// odtud muzeme predpokladat, ze se robot nepohnul
			} else if(r.robot.getDirection() != r.state.direction) action = getDirectionDiff(r.state.direction, r.robot.getDirection());
			// nabil energii
			else if(r.state.energy < r.robot.getEnergy()){
				if(gameMap[r.state.positionX][r.state.positionY].type == SquareType.BATTERY 
					&& r.robot.getEnergy() == world.getMaxRobotEnergy()) action = BotAction.RECHARGE;
				else action = BotAction.REST;
			
			// sebral poklad
			} else if(gameMap[r.state.positionX][r.state.positionY].treasure > world.getTreasureValue(r.state.positionX, r.state.positionY)){
				action = BotAction.TAKE;
				// aktualizuj hodnotu pokladu na policku
				gameMap[r.state.positionX][r.state.positionY].treasure = world.getTreasureValue(r.state.positionX, r.state.positionY);
				
			// uz nezbyva nic jineho, nez ze vystrelil
			} else action = BotAction.SHOOT;
			
			// proved samotnou akci
			r.step(action);
		}
		
		// proved predzpracovani
		preprocessor.preprocess();
		return run.step();
	}
	
	/*
	 * vypocita (podle parametru hry), jaka je idealni hladina energie, a ulozi do idealEnergyLevel
	 */
	 public void resetIdealEnergy(){
		 idealEnergyLevel = (int)((float) world.getMaxRobotEnergy() * 0.3);
		 if(idealEnergyLevel < 10) idealEnergyLevel = Math.min(10, world.getMaxRobotEnergy());
	 }
	
	/*
	* vraci dane policko nebo null pokud je policko mimo rozsah hraciho pole
	*/
	public Number5Square getSquare(int x, int y){
		try {
			return gameMap[x][y];
		} catch (ArrayIndexOutOfBoundsException e) {
			return null;
		}
	}
        
        public Number5Square getSquare(Number5Robot robot){
            return gameMap[robot.state.positionX][robot.state.positionY];
        }
        
        public Number5Square getNextSquare(Number5Square square, Direction direction){
            switch(direction){
                case LEFT: return getSquare(square.x - 1 , square.y);
                case UP: return getSquare(square.x, square.y - 1);
                case RIGHT: return getSquare(square.x + 1, square.y);
                default: return getSquare(square.x, square.y + 1);
            }
        }
	
	public BotAction getDirectionDiff(Direction old, Direction now){
		if(old == now) return null;
		else if((now == Direction.LEFT && old == Direction.UP) || (now == Direction.DOWN && old == Direction.LEFT)  || (now == Direction.RIGHT && old == Direction.DOWN)  || (now == Direction.UP && old == Direction.RIGHT)) return BotAction.LEFT;
		else if((now == Direction.LEFT && old == Direction.DOWN) || (now == Direction.DOWN && old == Direction.RIGHT)  || (now == Direction.RIGHT && old == Direction.UP)  || (now == Direction.UP && old == Direction.LEFT)) return BotAction.RIGHT;
		else return null;
	}
        
        public Direction getNextDirection(Direction start, BotAction action){
            if((start == Direction.LEFT && action == BotAction.RIGHT) || (start == Direction.RIGHT && action == BotAction.LEFT)) return Direction.UP;
            else if((start == Direction.DOWN && action == BotAction.RIGHT) || (start == Direction.UP && action == BotAction.LEFT)) return Direction.LEFT;
            else if((start == Direction.RIGHT && action == BotAction.RIGHT) || (start == Direction.LEFT && action == BotAction.LEFT)) return Direction.DOWN;
            else if((start == Direction.UP && action == BotAction.RIGHT) || (start == Direction.DOWN && action == BotAction.LEFT)) return Direction.RIGHT;
            else return start;
        }
	
	
	/**************************************************** PRIVATE *************************************************/
			
	/* predzpracovani */
	private Number5Preprocessor preprocessor = null;
	private Number5Run run = null;
			
	private Robot initMe;
}
