package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.Robot.*;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.number5.Number5Square.*;
import java.util.LinkedList;

public class Number5Charger implements Number5Strategy {
    
    Number5 core;
    
    public void setCore(Number5 core){
        this.core = core;
    }
    
    
    /*********************************** VOLEBNI FUNKCE **********************************/
    
    public Number5Proposal getMainProposal(){
        battery = getRechargePlace();
        if(battery == null) return Number5Proposal.NO_PLAN;
        else return new Number5Proposal(core.world.getRoundsCount() - core.world.getRound() + 1, core.world.getMaxRobotEnergy() - core.me.state.energy, "Na zaver bych se sel dobit, ziskam tak spoustu bodu.");
    }
    
    public Number5Proposal getInterruptProposal(int stepsLimit, boolean interruptSelf){
        recharged = false;
        
        // fake points we can achieve to get highest possible priority
        if(core.me.state.energy == 1) return new Number5Proposal(1, 100, Number5Run.HIGH_PRIORITY, "Potrebuju dobit!");
        //else if(core.getSquare(core.me).type == SquareType.BATTERY && core.me.state.energy < core.idealEnergyLevel) return new Number5Proposal(1, (core.world.getMaxRobotEnergy() - core.me.state.energy - 1)/2, Number5Run.MEDIUM_PRIORITY, "Jsem na baterii, proc se nedobit?");
        else return Number5Proposal.NO_PLAN;
    }
    
    /**************************** FUNKCE PRO STRATEGII U MOCI **************************/
    
    public void manage(Number5Run run){
        /*int reserve;
        
        // SPOCITEJ REZERVU
        int stepsToRest = 0;
        int energy = core.world.getRestRobotEnergy() - 1;
        while(energy > 1) {
            stepsToRest++;
            // realita neni tak kruta, dejme do toho nahodu
            if(Math.random()*2 > 1) energy = (int)Math.ceil(energy/2) - 1;
            else energy -= 1;
        }
        
        int maxEnergy = 0;
        for(Number5Robot r : core.robots){
            if(r != core.me && r.state.energy > maxEnergy) maxEnergy = r.state.energy;
        }
        
        // nepritel
        int his = battery.minimum.getMinimumStepsToStand();
        // ... musi dobijet
        his += Math.max(0, Math.ceil((double)(his - maxEnergy) / core.world.getRestRobotEnergy()));
        
        // ja...
        int mine = battery.me.getMinimumStepsToStand();
        // ...nejinak (ale po mne muze i nekdo strilet a ja chci mit jistotu, ze to stihnu)
        mine += Math.max(0, Math.ceil((double)(mine - core.me.state.energy) / stepsToRest));
        
        int remaining = core.world.getRoundsCount() - core.world.getRound() + 1;
        reserve = remaining - (mine + 1);
        
        // o dobijeni se kdyztak postaram sam, mozna jde fakt o cas
        run.setInterruptable(reserve, Number5Run.HIGH_PRIORITY);*/
        run.setInterruptable(0, Number5Run.HIGH_PRIORITY);
    }
    
    public Strategy.BotAction decide(){
        int remaining = core.world.getRoundsCount() - core.world.getRound() + 1;
        
        // jsem na baterii
        if(core.getSquare(core.me) == battery){
            if(remaining > 2 && core.me.state.energy > 1) return BotAction.TAKE;
            else if(remaining > 2 && core.me.state.energy == 1) return BotAction.REST;
            else return BotAction.RECHARGE;
        // potrebuju nutne dobit
        } else if(core.me.state.energy == 1) return BotAction.REST;
        // jdu k ni
        else {
            LinkedList<BotAction> actions = core.path.findPath(core.getSquare(core.me), battery, core.me.state.direction, battery.me.getBestReachingDirection(), core.me.state.energy, 1000);
            return actions.get(0);
        }
    }
    
    /**************************** FUNKCE PRERUSUJICI STRATEGIE **************************/
    
    public void manageInterrupt(Number5Run run){
        if(recharged){
            run.resign();
            return;
        }
    }
    
    public Strategy.BotAction decideInterrupt(){
        recharged = true;
        
        //if(core.getSquare(core.me).type == SquareType.BATTERY) return BotAction.RECHARGE;
        //else
        return BotAction.REST;
    }
    
    /**************************** PRIVATE **************************/
    
    boolean recharged = false;
    Number5Square battery = null;
    
    private Number5Square getRechargePlace(){
        int remaining = core.world.getRoundsCount() - core.world.getRound();
        //System.out.println("R:"+remaining);
        int highestEnergy = 0;
        int restEnergy = core.world.getRestRobotEnergy();
        int maxEnergy = core.world.getMaxRobotEnergy();
        
        int his = 0;
        int mine = 0;
        
        int fastest = 10000;
        double bestChance = 0;
        Number5Square battery = null;
        
        for(Number5Robot r : core.robots){
            if(r != core.me && r.state.energy > highestEnergy) highestEnergy = r.state.energy;
        }
        
        
        for(Number5Square[] col : core.gameMap){
            for(Number5Square s : col){
                if(s.type == SquareType.BATTERY){
                    // pokud mame vubec priblizne sanci se tam dostat driv nez cizi robot
                    if(s.me.getMinimumStepsToStand() < s.minimum.getMinimumStepsToStand() + 5){
                        //System.out.println(s);
                        // nepritel
                        his = s.minimum.getMinimumStepsToStand();
                        // ... musi dobijet
                        his += Math.max(0, Math.ceil((double)(his - highestEnergy) / restEnergy));
                        
                        
                        // ja...
                        Number5State state = core.me.state.clone();
                        Number5Square position;
                        LinkedList<BotAction> actions = core.path.findPath(core.getSquare(core.me), s, core.me.state.direction, s.me.getBestReachingDirection(), core.me.state.energy, 1000);
                            
                        mine = 0;
                        int reserve;
                        double chance = 1;
                        boolean hit;
                        
                        while(state.positionX != s.x || state.positionY != s.y){
                            
                            position = core.getSquare(state.positionX, state.positionY);
                            
                            // gets hit
                            hit = false;
                            if(position.minimum.stepsToShoot <= mine){
                                reserve = mine - position.minimum.stepsToShoot;
                                reserve = 10 - reserve;
                                if(Math.random()*14 > reserve) hit = true;
                            }
                            
                            // robot stands in his way
                            if(position.minimum.getMinimumStepsToStand() <= mine){
                                reserve = mine - position.minimum.getMinimumStepsToStand();
                                chance *= 0.95;
                                while(reserve-- > 0) chance *= 0.999;
                            }
                            
                            mine++;
                            
                            //if(state.energy == 1) state.simulate(BotAction.REST, restEnergy, maxEnergy, hit);
                            //else 
                            state.simulate(actions.poll(), restEnergy, maxEnergy, hit);
                        }
                        chance *= ((double) his / (double)mine);
                        //System.out.println(s + " chance: " + chance + " mine: " + mine + " (" + remaining + ") " + his);
                        //System.out.println(mine);
                        if((mine + 1) <= remaining && chance > bestChance){
                            bestChance = chance;
                            fastest = mine + 1;
                            battery = s;
                        }
                    }
                }
            }
        }
        
        if(battery == null) return null;
        //else if(fastest + 1 < remaining) return null;
        // vetsi rezerva
        else if(fastest + 5 < remaining) return null;
        return battery;
    }
}
