package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.Robot.*;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.number5.Number5Square.*;
import java.util.LinkedList;

public class Number5GoldDigger implements Number5Strategy {
    
    Number5 core;
    
    public void setCore(Number5 core){
        this.core = core;
    }
    
    
    /*********************************** VOLEBNI FUNKCE **********************************/
    
    public Number5Proposal getMainProposal(){
        treasure = findBestTreasure();
        
        if(treasure != null){
            endingDirection = treasure.me.getBestReachingDirection();
            return new Number5Proposal(treasure.me.getStepsToStand(endingDirection)+1, treasure.treasure, "Jdu k nejvyhodnejsimu pokladu ("+treasure+")");
        } else return Number5Proposal.NO_PLAN;
    }
    
    public Number5Proposal getInterruptProposal(int stepsLimit, boolean interruptSelf){
        if(core.getSquare(core.me) != treasure && core.getSquare(core.me).treasure > 0) return new Number5Proposal(1, core.getSquare(core.me).treasure, Number5Run.HIGH_PRIORITY, "Jsem na pokladu, tak bych ho vzal...");
        else return Number5Proposal.NO_PLAN;
    }
    
    /**************************** FUNKCE PRO STRATEGII U MOCI **************************/
    
    public void manage(Number5Run run){
        // nekdo (ja) uz poklad sebral
        if(treasure == null || treasure.treasure == 0){
            run.resign();
            return;
        }
        
        stepsReserve = countStepsReserve();
        run.setInterruptable(stepsReserve, Number5Run.MEDIUM_PRIORITY);
    }
    
    public Strategy.BotAction decide(){
        // vezmi poklad
        if(core.getSquare(core.me) == treasure) return BotAction.TAKE;
        else {
            LinkedList<BotAction> actions = core.path.findPath(
                    core.gameMap[core.me.state.positionX][core.me.state.positionY],
                    treasure,
                    core.me.state.direction,
                    treasure.me.getBestReachingDirection(), 
                    core.me.state.energy, 
                    1000); // no time limit
            return actions.get(0);
        }
    }
    
    /**************************** FUNKCE PRERUSUJICI STRATEGIE **************************/
    
    public void manageInterrupt(Number5Run run){
        // uz jsem poklad sebral
        if(core.getSquare(core.me).treasure == 0){
            run.resign();
            return;
        }
    }
    
    public Strategy.BotAction decideInterrupt(){
        return BotAction.TAKE;
    }
    
    /**************************** PRIVATE **************************/
    
    boolean recharged = false;
    Number5Square treasure = null;
    int stepsReserve;
    Direction endingDirection;
    
    // vrati policko s pokladem nebo null
    private Number5Square findBestTreasure(){
        Number5Square best = null;
        float bestRatio = 0;
        float ratio;
        
        for(Number5Square[] col : core.gameMap){
            for(Number5Square s : col){
                if(s.treasure > 0){
                    ratio = getTreasureRatio(s);
                    if(best == null || ratio > bestRatio){
                        best = s;
                        bestRatio = ratio;
                    }
                }
            }
        }
        
        if(bestRatio > 0) return best;
        else return null;
    }
    
    // vraci cislo odhadujici vyhodnost nejblizsiho pokladu nebo 0, pokud tam jiny robot muze byt rychleji nez my
    private float getTreasureRatio(Number5Square s){
        int reserve = s.minimum.getMinimumStepsToStand() - s.me.getMinimumStepsToStand();
        double treshhold = Math.floor((double)s.minimum.getMinimumStepsToStand()/5);
        // nebudeme tam prvni
        if(reserve <= -treshhold) return 0;
        // rezerva vetsi nez 10 uz neprida pokladu na zajimavosti
        if(reserve > 10) reserve = 10;
        
        return ((float)(s.treasure + (float)reserve/3) / (s.me.getMinimumStepsToStand() + 1));
    }
    
    // spocita, kolik mame tahu k dobru, abychom si poklad jistili
    private int countStepsReserve(){
        if(treasure == null) return 0;
        
        // nestaci tam dojit stejne jako druhy robot, musime tam byt o tah driv
        return treasure.minimum.getMinimumStepsToStand() - treasure.me.getMinimumStepsToStand() - 1;
    }
}
