package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.Robot.*;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import java.util.LinkedList;

public class Number5Killer implements Number5Strategy {
    
    Number5 core;
    
    private static double WEAKNESS_TRESHHOLD = 0.4;
    
    /**
     * interface pro jednotlive strategie naseho robota
     */
    
    /**
     * je zavolano tesne po konstruktoru, da strategii odkaz na jadro robota
     * @param core - jadro robota, pres ktere pozdeji strategie komunikuje s okolnim svetem
     */
    public void setCore(Number5 core){
        this.core = core;
    }
    
    
    /*********************************** VOLEBNI FUNKCE **********************************/
    
    /**
     * funkce, kterou jadro zavola pri hlavnich volbach
     * @return plan teto strategie v podobe objektu Number5Proposal - viz jeho popis
     * 		nesmi byt null, pokud se nechce zapojit do voleb, musi vratit navrh se ziskem 0 bodu
     */
    public Number5Proposal getMainProposal(){
        hit = false;
        
        Number5Robot previous = closestRobot;
        
        findClosest();
        updateClosest();
        
        int energy = closestRobot.state.energy;
        //System.out.println("Odhad: "+energy+" "+stepsToClosest);
        energy = Math.max(core.world.getRestRobotEnergy()/2, energy-stepsToClosest);
        
        int gain = (int)Math.floor((energy)/2);
        
        // probability
        gain = (int)((double)gain * probability);
        
        // we shoot again
        if(previous == closestRobot){
            gain /= 2;
        }
        return new Number5Proposal(stepsToClosest, gain, "jdu po nejblizsim robotovi ("+closestRobot+" z "+target+")");
    }
    
    /**
     * funkce, kterou jadro zavola pri volbe preruseni
     * @param stepsLimit - maximalni pocet kroku na preruseni
     * @return plan teto strategie v podobe objektu Number5Proposal - viz jeho popis
     * 		nesmi byt null, pokud se nechce zapojit do voleb, musi vratit navrh se ziskem 0 bodu
     */
    public Number5Proposal getInterruptProposal(int stepsLimit, boolean interruptSelf){
        Number5Robot atSight = core.me.robotAtSight();
        int gain = 0 ;
        if(atSight != null) gain = (int)Math.floor((atSight.state.energy-1)/2);
        
        if(atSight != null && atSight != closestRobot && gain > 0) return new Number5Proposal(1, gain, Number5Run.HIGH_PRIORITY, "Strilim!!!");
        else return Number5Proposal.NO_PLAN;
    }
    
    
    /**************************** FUNKCE PRO STRATEGII U MOCI **************************/
    
    /**
     * jadro se pta, jake ma strategie plany, tedy jestli se nevzdava moci, jestli se smi prerusit, atd.
     * podle toho, jak strategie upravi objekt run, po teto funkci muzou probehnout hlavni volby a/nebo volby preruseni
     * @param run - objekt, kterym strategie spravuje preruseni, atd.
     */
    public void manage(Number5Run run){
        updateClosest();
        
        // too many steps performed
        if(stepsMade + stepsToClosest > 5*initialSteps){
            run.resign();
            return;
        }
        // too weak opponent
        if(hit && closestRobot.state.energy < (double)core.world.getMaxRobotEnergy()*WEAKNESS_TRESHHOLD){
            run.resign();
            return;
        }
        
        run.setInterruptable(interruptLimit, Number5Run.HIGH_PRIORITY);
        
        interruptLimit--;
    }
    
    /**
     * jadro se pta na samotny tah, ktery strategie v tomto kole provede
     * @return jedna z nadefinovanych akci (LEFT, RIGHT, STEP, SHOOT, TAKE, REST, RECHARGE)
     */
    public Strategy.BotAction decide(){
        interruptLimit = 3;
        if(core.me.robotAtSight() == closestRobot){
            hit = true;
            return BotAction.SHOOT;
        } else if(core.getSquare(core.me) == target && core.me.state.direction == shootDirection) {
            return BotAction.TAKE;
        } else {
            //System.out.println("!!!!!!!!!!!!!!Jdu k nemu");
            LinkedList<BotAction> actions = core.path.findPath(core.getSquare(core.me), target, core.me.state.direction, shootDirection, core.me.state.energy, 1000);
            //if(actions.size() > 0)
            return actions.get(0);
            //else return BotAction.REST;
        }
    }
    
    /**************************** FUNKCE PRERUSUJICI STRATEGIE **************************/
    
    /**
     * dve analogicke funkce, ktere jadro vola, pokud je tato strategie prave PRERUSUJICI STRATEGII
     */
    public void manageInterrupt(Number5Run run){
    }
    
    public Strategy.BotAction decideInterrupt(){
        return BotAction.SHOOT;
    }
    
    /**************************** PRIVATE **************************/
    
    int stepsToClosest;
    private Number5Robot closestRobot;
    private Number5Square target;
    private Direction shootDirection;
    private int stepsMade;
    private int initialSteps;
    
    double probability;
    
    boolean hit = false;
    int interruptLimit = 3;
    
    void updateClosest(){
        probability = 1;
        // policko, kde stoji souper
        Number5Square s = core.getSquare(closestRobot);
        // odkud na nej mam strilet?
        target = s.me.shootPosition;
        // a s jakym natocenim mam strilet?
        shootDirection = s.me.shootDirection;
        // jak dlouho to bude trvat?
        stepsToClosest = target.me.getStepsToStand(shootDirection);
        
        //System.out.println("_______________ ptam se na stav po "+stepsToClosest+" krocich");
        
        // nepotrebujem do predpovedi zahrnout zaverecnou strelbu
        int steps = stepsToClosest - 1;
        if(steps > 0){
            Number5Prediction p = closestRobot.predict(steps);

            while((p == p.DONT_KNOW || p.probability == 0) && steps > 1){
                if(steps > 6) steps /= 2;
                else steps--;

                p = closestRobot.predict(steps);
            }

            /*System.out.println("PREDPOVED pro " + closestRobot + "____________");
            System.out.println(p);
            System.out.println("-----------------------");*/

            //if(p != p.DONT_KNOW && p.probability > 10){

            if(p != p.DONT_KNOW){

                // tak jeste jednou
                s = p.position;
                // odkud na nej mam strilet?
                target = s.me.shootPosition;
                // a s jakym natocenim mam strilet?
                shootDirection = s.me.shootDirection;
                // jak dlouho to bude trvat?
                stepsToClosest = Math.max(steps+1, target.me.getStepsToStand(shootDirection));
                probability = Math.sqrt((double)p.probability/100);
            }
        }
        
        // a s jakym natocenim mam strilet?
        shootDirection = s.me.shootDirection;
    }
    
    void findClosest(){
        stepsToClosest = 10000;
        double ratio = 0;
        double ratioThis;
        stepsMade = 0;
        Number5Square s;
        
        for(Number5Robot r : core.robots){
            s = core.gameMap[r.state.positionX][r.state.positionY];
            ratioThis = (double)(r.state.energy - s.me.stepsToShoot) / s.me.stepsToShoot;
            if(ratioThis > ratio && r != core.me){
                stepsToClosest = s.me.stepsToShoot;
                ratio = ratioThis;
                closestRobot = r;
                target = s.me.shootPosition;
                shootDirection = s.me.shootDirection;
            }
        }
        initialSteps = stepsToClosest;
    }
}
