package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.number5.Number5Square.*;
import cz.muni.fi.fibot.Robot.*;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import java.util.LinkedList;

public class Number5MachineGunner implements Number5Strategy {
    
    Number5 core;
    
    // objekt informaci ke kazdemu (cizimu) robotu
    protected class Number5MachineGunnerData {
        // jestli je zapojen do strilecky, predkopladame, ze ano
        //boolean active = true;
        // kdyz se presunuje, kam asi?
        //Number5Position target = null;
        Number5Robot victim = null;
        //boolean moving = false;
    }
    // objekt informaci ke kazdemu (cizimu) robotu
    public class Number5MachineGunnerPosition {
        
        Number5Square place = null;
        Direction direction = Direction.LEFT;
        int wait = 0;
        
        int steps = 0;
        int gain = 0;
        
        public Number5MachineGunnerPosition(Number5Square place, Direction direction, int wait){
            this.place = place;
            this.direction = direction;
            this.wait = wait;
        }
        
        public void set(Number5Square place, Direction direction, int wait){
            this.place = place;
            this.direction = direction;
            this.wait = wait;
        }
        
        public boolean betterThan(Number5MachineGunnerPosition that){
            if(that == null) return true;
            if(((double)(this.gain)/this.steps) > ((double)(that.gain)/that.steps)) return true;
            else return false;
        }
        public Number5MachineGunnerPosition clone(){
            Number5MachineGunnerPosition clone = new Number5MachineGunnerPosition(this.place, this.direction, this.wait);
            clone.steps = this.steps;
            clone.gain = this.gain;
            return clone;
        }
        public String toString(){
            return "Pozice: " + place + ", " + direction + ", wait: " + wait + "  " + gain + " / " + steps + " = " + (double)gain/steps;
        }
    }
    
    protected int STEPS = 30;
    protected int CYCLE = 10;    
    
    /**
     * interface pro jednotlive strategie naseho robota
     */
    
    /**
     * je zavolano tesne po konstruktoru, da strategii odkaz na jadro robota
     * @param core - jadro robota, pres ktere pozdeji strategie komunikuje s okolnim svetem
     */
    public void setCore(Number5 core){
        this.core = core;
    }
    
    
    /*********************************** VOLEBNI FUNKCE **********************************/
    
    /**
     * funkce, kterou jadro zavola pri hlavnich volbach
     * @return plan teto strategie v podobe objektu Number5Proposal - viz jeho popis
     * 		nesmi byt null, pokud se nechce zapojit do voleb, musi vratit navrh se ziskem 0 bodu
     */
    public Number5Proposal getMainProposal(){
        hit = false;
        atPlace = false;
        if(core.world.getRestRobotEnergy() <= 4) return Number5Proposal.NO_PLAN;
        
        STEPS = core.world.getRestRobotEnergy()*2;
        
        traceRobots();
        
        bestDeal = getBestPosition(core.me);
        
        if(bestDeal != null) return new Number5Proposal(bestDeal.steps, bestDeal.gain, "jdu strilet z mista " + bestDeal.place + " smerem " + bestDeal.direction + " delsi dobu...");
        else return Number5Proposal.NO_PLAN;
    }
    
    /**
     * funkce, kterou jadro zavola pri volbe preruseni
     * @param stepsLimit - maximalni pocet kroku na preruseni
     * @return plan teto strategie v podobe objektu Number5Proposal - viz jeho popis
     * 		nesmi byt null, pokud se nechce zapojit do voleb, musi vratit navrh se ziskem 0 bodu
     */
    public Number5Proposal getInterruptProposal(int stepsLimit, boolean interruptSelf){
        return Number5Proposal.NO_PLAN;
    }
    
    
    /**************************** FUNKCE PRO STRATEGII U MOCI **************************/
    
    /**
     * jadro se pta, jake ma strategie plany, tedy jestli se nevzdava moci, jestli se smi prerusit, atd.
     * podle toho, jak strategie upravi objekt run, po teto funkci muzou probehnout hlavni volby a/nebo volby preruseni
     * @param run - objekt, kterym strategie spravuje preruseni, atd.
     */
    public void manage(Number5Run run){
        // we have shot already
        if(hit){
            run.resign();
            return;
        }
        
        traceRobots();
        
        if(bestDeal.place != core.getSquare(core.me)){
            
            bestDeal = getBestPosition(core.me);
            if(bestDeal == null){
                run.resign();
                return;
            }
        } else atPlace = true;
        
        if(core.getSquare(core.me) == bestDeal.place) run.setInterruptable(0, Number5Run.HIGH_PRIORITY);
        else run.setInterruptable(5, Number5Run.MEDIUM_PRIORITY);
    }
    
    /**
     * jadro se pta na samotny tah, ktery strategie v tomto kole provede
     * @return jedna z nadefinovanych akci (LEFT, RIGHT, STEP, SHOOT, TAKE, REST, RECHARGE)
     */
    public Strategy.BotAction decide(){
        // jdu na misto cinu
        if(core.me.state.energy == 1){
            //System.out.println("Nemam stavu");
            bestDeal.wait--;
            return BotAction.REST;
        }
        
        if(bestDeal.place != core.getSquare(core.me) || bestDeal.direction != core.me.state.direction){
            LinkedList<BotAction> actions = core.path.findPath(core.getSquare(core.me), bestDeal.place, core.me.state.direction, bestDeal.direction, core.me.state.energy, 1000);
            bestDeal.wait--;
            if(actions.size() > 0) return actions.get(0);
            else return BotAction.REST;
        } else {
            if(bestDeal.wait > 0){
                bestDeal.wait--;
                return BotAction.TAKE;
            } else {
                hit = true;
                return BotAction.SHOOT;
            }
        }
    }
    
    /**************************** FUNKCE PRERUSUJICI STRATEGIE **************************/
    
    /**
     * dve analogicke funkce, ktere jadro vola, pokud je tato strategie prave PRERUSUJICI STRATEGII
     */
    public void manageInterrupt(Number5Run run){
    }
    public Strategy.BotAction decideInterrupt(){
        return BotAction.REST;
    }
    
    /**************************** PRIVATE **************************/
    
    private Number5MachineGunnerPosition bestDeal;
    
    private boolean hit = false;
    private boolean atPlace = false;
    
    private boolean inited = false;
    
    Number5MachineGunnerPosition computeGain(Number5Robot robot, Number5MachineGunnerPosition position){
        Number5Robot atSight;
        int loss;
        int gain = 0;
        int realSteps = 0;
        int moves;
        
        int steps = Math.min(STEPS, core.world.getRoundsCount() - core.world.getRound() + 1 - position.wait);
        if(steps <= 0){
            position.steps = 1;
            position.gain = 0;
            return position;
        }
        
        Number5State[] states = new Number5State[core.robots.length];
        for(Number5Robot r : core.robots) states[r.index] = r.state.clone();
        states[robot.index].positionX = states[robot.index].positionY = -1;
        
        for(int x = -position.wait; x < steps; x++){
            //System.out.println("Tah: "+x);
            // vrat robota na cilovou pozici
            if(x == 0){
                states[robot.index].positionX = position.place.x;
                states[robot.index].positionY = position.place.y;
                states[robot.index].direction = position.direction;
            }
            
            // ztrata energie
            for(Number5Robot r : core.robots){
                states[r.index].energy--;
                //System.out.println("Energie["+r.index+"]: "+states[r.index].energy);
            }
            // strileni
            for(Number5Robot r : core.robots){
                // cizi robot
                if(r != robot){
                    if(states[r.index].energy > 0 && (atSight = r.robotAtSight(states)) != null) {
                        loss = (int)Math.floor(states[atSight.index].energy / 2);
                        //System.out.println("Robot["+r.index+"] ubere ["+atSight.index+"] "+loss+" bodu");
                        states[atSight.index].energy -= loss;
                    }
                    // nas
                } else if(x >= 0){
                    if(states[r.index].energy > 0 && (atSight = r.robotAtSight(states)) != null) {
                        loss = (int)Math.floor(states[atSight.index].energy / 2);
                        states[atSight.index].energy -= loss;
                        
                        int i = 0;
                        for(Number5State state : atSight.history){
                            if(state.action == BotAction.STEP) gain *= 0.9;
                            if(i++ >= CYCLE) break;
                        }
                        gain += loss;
                        //System.out.println("Ja uberu ["+atSight.index+"] "+loss+" bodu");
                    }
                }
            }
            // dobijeni
            for(Number5Robot r : core.robots){
                if(states[r.index].energy == 0){
                    states[r.index].energy += core.world.getRestRobotEnergy();
                    if(r != robot && x < 0) x--;
                }
            }
            realSteps++;
        }
        
        position.steps = realSteps;
        position.gain = gain;
        return position;
    }
    
    Number5MachineGunnerPosition getBestPosition(Number5Robot robot){
        Number5MachineGunnerPosition position = new Number5MachineGunnerPosition(core.getSquare(robot), robot.state.direction, 0);
        //System.out.println("+++++++++++++++++++++++++++++++++++++++");
        //System.out.println(position);
        //System.out.println("---------------------------------------");
        Number5MachineGunnerPosition bestPosition = null;
        
        int movingLimit = 6;
        
        if(goodPosition(robot, position)){
            do {
                position = computeGain(robot, position);
                //System.out.println(position);
                if(position.betterThan(bestPosition)) bestPosition = position.clone();
                position.wait++;
            } while(position.wait < CYCLE);
            
        } else movingLimit = 15;
        
        //System.out.println("---------------------------------------");
        
        for(Number5Square[] col : core.gameMap){
            for(Number5Square s : col){
                if(s.robots[robot.index].getMinimumStepsToStand() > movingLimit || s.type == SquareType.WALL) continue;
                
                for(Direction d : Direction.values()){
                    int initialWait = s.robots[robot.index].getStepsToStand(d);
                    position.set(s, d, initialWait);
                    if(goodPosition(robot, position)){
                        do {
                            position = computeGain(robot, position);
                            //System.out.println(position);
                            if(position.betterThan(bestPosition)) bestPosition = position.clone();
                            position.wait++;
                        } while(position.wait - initialWait < CYCLE);
                    }
                }
            }
        }
        //System.out.println("***********************************");
        // might be null
        return bestPosition;
    }
    
    boolean goodPosition(Number5Robot robot, Number5MachineGunnerPosition position){
        Number5Robot target = robot.robotAtSight(position.place, position.direction);
        
        // there is no target to shoot at
        if(target == null) return false;
        
        // aren't there too many vultures for this poor robot?
        int shootingAtMe = 0;
        for(Number5Robot r : core.robots){
            Number5MachineGunnerData data = (Number5MachineGunnerData) r.data.get("MG");
            if(r != robot && data.victim == target) return false;
            // if robot is a victim of r && robot doesn't intend to move
            else if(data.victim == robot && position.place == core.getSquare(robot)) shootingAtMe++;
        }
        
        // OR am I not just a victim?
        if(shootingAtMe > 1) return false;
        
        // well, it's OK
        return true;
    }
    
    void initRobots(){
        for(Number5Robot r : core.robots){
            r.data.put("MG", (Object) new Number5MachineGunnerData());
        }
        inited = true;
    }
    
    void traceRobots(){
        if(!inited) initRobots();
        else {
            for(Number5Robot r : core.robots){
                Number5MachineGunnerData data = (Number5MachineGunnerData) r.data.get("MG");
                BotAction last = r.history.getFirst().action;
                switch(last){
                    // shoots!
                    case SHOOT: data.victim = r.robotAtSight();
                    break;
                    // just needs energy, we can't tell anything from this information
                    case RECHARGE:
                    case REST: break;
                    // he stopped shooting!
                    default: data.victim = null;
                }
            }
        }
    }
}

