/*
 * Number5MushroomPicker.java
 *
 * Created on 21. květen 2006, 22:36
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.number5;


import cz.muni.fi.fibot.Robot.*;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.number5.Number5Square.*;
import java.util.LinkedList;

public class Number5MushroomPicker implements Number5Strategy {
    
     Number5 core;
    
    public void setCore(Number5 core){
        this.core = core;
    }
    
    /*********************************** VOLEBNI FUNKCE **********************************/
    
    public Number5Proposal getMainProposal(){
        findBestTreasure();
        finished = false;
        
        if(treasure != null && steps < core.world.getRoundsCount() - core.world.getRound()){
            return new Number5Proposal(steps, points, "Zacinam cestu za pokladem ("+treasure+")");
        } else return Number5Proposal.NO_PLAN;
    }
    
    public Number5Proposal getInterruptProposal(int stepsLimit, boolean interruptSelf){
        if(core.getSquare(core.me) != treasure && core.getSquare(core.me).treasure > 0) return new Number5Proposal(1, core.getSquare(core.me).treasure, Number5Run.HIGH_PRIORITY, "Jsem na pokladu, tak bych ho vzal...");
        else return Number5Proposal.NO_PLAN;
    }
    
    /**************************** FUNKCE PRO STRATEGII U MOCI **************************/
    
    public void manage(Number5Run run){
        boolean reached = (core.getSquare(core.me) == treasure);
        // nekdo (ja) uz poklad sebral nebo uz jsem to provedl
        if(treasure == null || (treasure.type != SquareType.BATTERY && treasure.treasure == 0) || finished){
            run.resign();
            return;
        }
        
        
        stepsReserve = countStepsReserve();
        if(reached) stepsReserve = 0;
        run.setInterruptable(stepsReserve, Number5Run.MEDIUM_PRIORITY);
    }
    
    public Strategy.BotAction decide(){
        // vezmi poklad
        if(core.getSquare(core.me) == treasure){
            if(core.me.state.energy == 1 && treasure.treasure > 0) return BotAction.REST;
            else {
                finished = true;
                if(treasure.treasure > 0) return BotAction.TAKE;
                else return BotAction.RECHARGE;
            }
        }
        else {
            LinkedList<BotAction> actions = core.path.findPath(
                    core.gameMap[core.me.state.positionX][core.me.state.positionY],
                    treasure,
                    core.me.state.direction,
                    treasure.me.getBestReachingDirection(), 
                    core.me.state.energy, 
                    1000); // no time limit
            return actions.get(0);
        }
    }
    
    /**************************** FUNKCE PRERUSUJICI STRATEGIE **************************/
    
    public void manageInterrupt(Number5Run run){
        // uz jsem poklad sebral
        if(core.getSquare(core.me).treasure == 0){
            run.resign();
            return;
        }
    }
    
    public Strategy.BotAction decideInterrupt(){
        return BotAction.TAKE;
    }
    
    /**************************** PRIVATE **************************/
    
    Number5Square treasure = null;
    int steps = 0;
    int points = 0;
    boolean finished = false;
    
    int stepsReserve;
    Direction endingDirection;
    
    // vrati policko s pokladem nebo null
    private void findBestTreasure(){
        //System.out.println("Start");
        int lengths[][] = new int[core.special.length][core.special.length];
        
        /*if(core.special.length > 20){
            java.util.Arrays.sort(core.special);
        }*/
        
        int treasures = 0;
        
        for(int x = 0; x < core.special.length; x++){
            if(core.special[x].treasure > 0) treasures++;
            for(int y = 0; y < core.special.length; y++){
                lengths[x][y] = core.special[y].special[x].getMinimumStepsToStand();
            }
        }
        
        if(treasures == 0){
            treasure = null;
            return;
        }
        
        Number5GoldPath paths[] = new Number5GoldPath[core.special.length];
        
        for(int x = 0; x < core.special.length; x++){
            // I'm standing here or this place has no value any more or there are too many places
            if(core.getSquare(core.me) == core.special[x]
               || (core.special[x].treasure == 0 && core.special[x].type != SquareType.BATTERY)
               ) paths[x] = null;
            // it's too dangerous to recharge or someone stands on the place of interest which is close
            else if((core.special[x].type == SquareType.BATTERY 
                    && core.special[x].me.getMinimumStepsToStand() > core.special[x].minimum.stepsToShoot)
                    || core.special[x].me.getMinimumStepsToStand() == 1 && core.special[x].robot != null) {
                paths[x] = Number5GoldPath.CANT_GO;
            } else {
                // base
                paths[x] = new Number5GoldPath(core.getSquare(core.me), core.me.state.direction, core.me.state.energy);
                int length = fillWithRests(core.special[x].me.getMinimumStepsToStand(), core.me.state.energy, (core.special[x].type == SquareType.BATTERY));
                paths[x] = new Number5GoldPath(paths[x], core.special[x], core.special[x].me.getBestReachingDirection(), 
                                        countGain(core.special[x], length), core);
            }
        }
        
        Number5GoldPath absolutelyBest = null;
        
        // steps
        for(int step = 0; step < Math.min(20, core.special.length - 1); step++){
            
            Number5GoldPath newPaths[] = new Number5GoldPath[paths.length];
            
            // for each path (i.e. special path target)
            for(int x = 0; x < paths.length; x++){
                 
                // find the best (steps-1)-prefix
                Number5GoldPath best = null;
                int bestLength, bestPoints, bestIndex;
                bestPoints = bestLength = bestIndex = 0;
                int length, points;
                for(int y = 0; y < paths.length; y++){
                    // this target is not valid or has been already visited
                    if(paths[y] == null || paths[y] == Number5GoldPath.CANT_GO 
                            || paths[y].path.contains(core.special[x])) continue;
                    
                    length = paths[y].length + fillWithRests(lengths[y][x], paths[y].energy, (core.special[x].type == SquareType.BATTERY));
                    points = paths[y].points + countGain(core.special[x], length);
                    
                    if(best == null || ((double)points/length) > (double)bestPoints/bestLength){
                        best = paths[y];
                        bestLength = length;
                        bestPoints = points;
                        bestIndex = y;
                    }
                }
                
                // neni zadna cesta, ktera by uz nevedla pres tento bod
                if(best == null){
                    if(paths[x] != null && paths[x].betterThan(absolutelyBest, core.world.getRoundsCount() - core.world.getRound())) absolutelyBest = paths[x];
                    newPaths[x] = null;
                    
                } else {
                    if(step < 1) newPaths[x] = new Number5GoldPath(best, core.special[x], core.special[x].special[bestIndex].getBestReachingDirection(), bestPoints, core);
                    else newPaths[x] = new Number5GoldPath(best, core.special[x], core.special[x].special[bestIndex].getBestReachingDirection(), bestLength, bestPoints, core);
                }
            }
            
            paths = newPaths;
        }
        
        // find the best
        for(int x = 0; x < paths.length; x++){
            if(paths[x] != null && paths[x].betterThan(absolutelyBest, core.world.getRoundsCount() - core.world.getRound())) absolutelyBest = paths[x];
        }
        
        // 0 - me, 1 - first treasure
        treasure = absolutelyBest.path.get(1);
        this.points = absolutelyBest.bestPoints;
        this.steps = absolutelyBest.bestLength;
        //System.out.println("End");
    }
    
    // spocita, kolik mame tahu k dobru, abychom si poklad jistili
    private int countStepsReserve(){
        if(treasure == null) return 0;
        
        // nestaci tam dojit stejne jako druhy robot, musime tam byt o tah driv
        return treasure.minimum.getMinimumStepsToStand() - treasure.me.getMinimumStepsToStand() - 1;
    }
    
    private int fillWithRests(int length, int energy, boolean isBattery){
        int bonus = 1;
        int remains = length - energy + 1;
        
        if(remains > 0) {
            bonus += remains / (core.world.getRestRobotEnergy() - 1);
            // final rest
            if(!isBattery && (remains % (core.world.getRestRobotEnergy() - 1) == 0)) bonus++;
        }
        // final activity
        bonus++;
        return length + bonus;
    }
    
    private int countGain(Number5Square target, int length){
        if(target.type == SquareType.BATTERY) return 0;
        else if(target.treasure == 0) return 0;
        else {
            int reserve = target.minimum.getMinimumStepsToStand() - length;
            if(reserve > 0) return target.treasure;
            else {
                double t = target.treasure;
                while(reserve < 0){
                    t *= 0.92;
                    reserve++;
                }
                return (int)t;
            }
        }
    }
}

// objekt informaci ke kazdemu (cizimu) robotu
class Number5GoldPath {

    LinkedList<Number5Square> path;
    Direction direction;
    int energy;

    int length;
    int points;

    int bestLength;
    int bestPoints;

    protected static Number5GoldPath CANT_GO = new Number5GoldPath(null, Direction.LEFT, 0);

    public Number5GoldPath(Number5Square start, Direction direction, int energy){
        path = new LinkedList<Number5Square>();
        path.addLast(start);

        this.direction = direction;
        this.energy = energy;

        bestLength = this.length = 0;
        bestPoints =this.points = 0;
    }
    public Number5GoldPath(Number5GoldPath base, Number5Square target, Direction bestDirection, int points, Number5 core){
        path = new LinkedList<Number5Square>();
        for(Number5Square s : base.path){
            path.addLast(s);
        }
        path.addLast(target);

        LinkedList<BotAction> join = core.path.findPath(base.path.getLast(), target, base.direction, bestDirection, base.energy, 1000);
        length = base.length + join.size() + 1;
        this.points = points;

        this.direction = base.direction;
        this.energy = base.energy;

        bestLength = base.bestLength;
        bestPoints = base.bestPoints;

        // simulate
        for(BotAction a : join){
            direction = core.getNextDirection(direction, a);
            energy--;
            if(a == BotAction.REST) energy += core.world.getRestRobotEnergy();
        }

        if(target.type == SquareType.BATTERY) energy = core.world.getMaxRobotEnergy();

        if(bestLength == 0 || (((double)points/length) > ((double)bestPoints/bestLength) 
            && length <= (core.world.getRoundsCount() - core.world.getRound()))){
            bestLength = length;
            bestPoints = points;
        }
    }
    
    public Number5GoldPath(Number5GoldPath base, Number5Square target, Direction bestDirection, int length, int points, Number5 core){
        path = new LinkedList<Number5Square>();
        for(Number5Square s : base.path){
            path.addLast(s);
        }
        path.addLast(target);

        this.length = length;
        this.points = points;

        this.direction = bestDirection;

        bestLength = base.bestLength;
        bestPoints = base.bestPoints;

        if(target.type == SquareType.BATTERY) energy = core.world.getMaxRobotEnergy();
        // middle value
        else energy = core.world.getRestRobotEnergy()/2;

        if(bestLength == 0 || ((double)points/length) > ((double)bestPoints/bestLength)){
            bestLength = length;
            bestPoints = points;
        }
    }

    public boolean betterThan(Number5GoldPath other, int remaining){
        if(other == null || other.bestLength == 0) return true;
        else if(other.bestLength > remaining) return true;
        else if(bestLength > remaining) return false;
        else if(bestLength == 0) return false;
        else if((double)bestPoints/bestLength > (double)other.bestPoints/other.bestLength) return true;
        else return false;
    }
}
