package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.number5.Number5Square.SquareType;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;

import java.util.LinkedList;

public class Number5PathFinder {
        /**
         * Soukroma trida reprezentujici vrchol v grafu pri prohledavani
         */
        protected class Number5PathVertex {
                // do kterych uzlu se dostane po provedeni akci STEP,LEFT,RIGHT
                public Number5PathVertex vertices[] = null;                                
                // zpetny ukazatel na vrchol jez je soucasti minimalni cesty
                public LinkedList<Number5PathVertex> back = null;
                public LinkedList<BotAction> backAction = null;                
                // pro zjisteni minimalni vzdalenosti
                public int distance = -1;
                // pro zjisteni, jestli je distance a back validni
                public boolean flag = true;
                // odkaz na ctverec, v nemz tento vrchol grafu lezi;
                public Number5Square square = null;
                
                public Number5PathVertex(){
                        vertices = new Number5PathVertex[3];
                        backAction = new LinkedList<BotAction>();
                        back = new LinkedList<Number5PathVertex>();
                }
                
                public void setVertex(Number5PathVertex step, Number5PathVertex left, Number5PathVertex right, Number5Square square) {
                        this.vertices[0] = step;
                        this.vertices[1] = left;
                        this.vertices[2] = right;
                        this.square = square;
                }                              
                /*public void setDistance(int dist){
                        distance = dist;
                }
                public void setBack(Number5PathVertex back){
                        this.back = back;
                }
                public void setFlag(boolean flag){
                        this.flag = flag;
                }*/
        }

	/**
         * jadro robota pro ktereho se budou hledat cesty
         */
         protected Number5 core = null;        
         
         /**
          * graf hraciho pole, pro kazdy ctverec [x][y] jsou ctyri vrcholy [0..3] podle smeru ve kterych
          * na ctverci robot stoji
          */
         protected Number5PathVertex[][][] gameMapGraph = null;
         protected int graphWidth = -1;
         protected int graphHeight = -1;
         // pred kazdy zjistenim cesty se pathFlag zneguje aby bylo videt, ze atributy distance a back nejsou validni
         protected boolean pathFlag = true;
         
         /**
	 * konstruktor
	 * @param core - jadro robota, nevim, jestli to bude potrebovat, ale co kdyby
	 */
	 protected Number5PathFinder(Number5 core){
                this.core = core;                 
                buildGameMapGraph ();
	 }
         
         /**
          * Okomentovat
          */
         public Number5PathVertex getVertex(int x, int y, int dir){
                if (x < 0 || x >= graphWidth || y < 0 || y >= graphHeight || dir < 0 || dir > 3) return null;
                else return gameMapGraph[x][y][dir];
         }
	       
         private void buildGameMapGraph (){
                // vytvori graf herniho pole
		int x, y, i;              
                Number5Square square = null;
                
                graphWidth = core.world.getDimensionX();
		graphHeight = core.world.getDimensionY();
                
                gameMapGraph = new Number5PathVertex[graphWidth][graphHeight][4];
                
		for(x = 0; x < graphWidth; x++){
			for(y = 0; y < graphHeight; y++){                                
				if (core.getSquare(x,y).type == SquareType.WALL) {
                                        for (i = 0; i < 4; i++) gameMapGraph[x][y][i] = null;                                        
                                }
                                else {
                                        for (i = 0; i < 4; i++) gameMapGraph[x][y][i] = new Number5PathVertex();                                        
                                }
			}
		}
                
                for(x = 0; x < graphWidth; x++){
			for(y = 0; y < graphHeight; y++){                                
				if (core.getSquare(x,y).type != SquareType.WALL) {
                                        square = core.getSquare(x,y);
                                        gameMapGraph[x][y][0].setVertex(getVertex(x-1,y,0),
                                                                        getVertex(x,y,3),
                                                                        getVertex(x,y,1),
                                                                        square                                                                        
                                        );
                                        gameMapGraph[x][y][1].setVertex(getVertex(x,y-1,1),
                                                                        getVertex(x,y,0),
                                                                        getVertex(x,y,2),
                                                                        square                                                                        
                                        );
                                        gameMapGraph[x][y][2].setVertex(getVertex(x+1,y,2),
                                                                        getVertex(x,y,1),
                                                                        getVertex(x,y,3),
                                                                        square                                                                        
                                        );
                                        gameMapGraph[x][y][3].setVertex(getVertex(x,y+1,3),
                                                                        getVertex(x,y,2),
                                                                        getVertex(x,y,0),
                                                                        square                                                                        
                                        );
                                }                                
			}
		}
         }
         
         /**
          * 
          */
         protected LinkedList<BotAction> findPath(Number5Square start, Number5Square end, Direction startDirection, Direction endDirection, int startEnergy, int stepLimit){
                LinkedList<Number5PathVertex> vertexQueue = new LinkedList<Number5PathVertex>();
                LinkedList<BotAction> actionQueue = new LinkedList<BotAction>();
                
                int startDir = 0;
                switch (startDirection) {
                    case LEFT:
                        startDir = 0;
                        break;
                    case UP:
                        startDir = 1;
                        break;
                    case RIGHT:
                        startDir = 2;
                        break;
                    case DOWN:
                        startDir = 3;
                        break;
                }
                int endDir = 0;
                switch (endDirection) {
                    case LEFT:
                        endDir = 0;
                        break;
                    case UP:
                        endDir = 1;
                        break;
                    case RIGHT:
                        endDir = 2;
                        break;
                    case DOWN:
                        endDir = 3;
                        break;
                }
                Number5PathVertex vertex = getVertex(start.x,start.y,startDir);
                Number5PathVertex endVertex = getVertex(end.x,end.y,endDir);
                vertex.distance = 0;
                vertex.back.clear();
                pathFlag = !pathFlag;
                vertex.flag = pathFlag;                
                
                vertexQueue.add(vertex);
                
                int vertexMinDist = -1;
                int i;
		BotAction[] poleAkci = {BotAction.STEP,BotAction.LEFT,BotAction.RIGHT};
                
                while (vertexQueue.size() > 0 /*&& vertexMinDist == -1*/) { // bud hleda prvni nejkratsi cestu nebo vsechny
                        vertex = vertexQueue.poll();
                        if (vertex.distance <= stepLimit) {
                                if (vertex == endVertex) {
                                        vertexMinDist = vertex.distance;
                                }
                                else if (vertex.distance < stepLimit) for (i = 0; i < 3; i++){ 
                                        if (vertex.vertices[i] != null && 
                                                    (vertex.vertices[i].flag != pathFlag || 
                                                        (vertex.vertices[i].flag == pathFlag && vertex.vertices[i].distance >= vertex.distance + 1)
                                                    )
                                                ) {
                                                if (vertex.vertices[i].flag != pathFlag ) {
                                                        // pridat podminku pritomnosti robotu
                                                        vertex.vertices[i].flag = pathFlag;
                                                        vertex.vertices[i].distance = vertex.distance + 1;
                                                        if(vertex.vertices[i].back == null){
                                                            int pokus = 1;
                                                        }
                                                        vertex.vertices[i].back.clear();
                                                        vertex.vertices[i].back.add(vertex);
                                                        vertex.vertices[i].backAction.clear();
                                                        vertex.vertices[i].backAction.add(poleAkci[i]);
                                                        vertexQueue.add(vertex.vertices[i]);
                                                }
                                                else {                                                    
                                                        vertex.vertices[i].back.add(vertex);
                                                }
                                        }
                                }
                        }                        
                }                                
                
                if (endVertex.flag == pathFlag) {
                    int minRechargeCount = 0;
                    if (endVertex.distance >= startEnergy) minRechargeCount = (endVertex.distance - startEnergy)/(core.world.getRestRobotEnergy()-1)+1;
                    if (endVertex.distance + minRechargeCount > stepLimit) return null;
                    else {
                        vertex = endVertex;
                        while (vertex.back != null && vertex.back.size() != 0) {                            
                            actionQueue.addFirst(vertex.backAction.peek());
                            vertex = vertex.back.peek();
                        }
                        int energy = startEnergy;
                        for (i = 0; i < actionQueue.size(); i++ ) {                            
			    if (energy<3) actionQueue.add(i,BotAction.REST);
                            energy += core.world.getRestRobotEnergy();
                            energy--;
                        }
                        return actionQueue;
                        /*vertexQueue.add(endVertex);
                        while (vertexQueue.size() > 0) {
                            vertex = vertexQueue.peek();
                            
                        }
                        actionQueue.offer();*/
                    }                                            
                }
                else return null;
         }
         
         /**
	 * hledani idealni cesty (nevymysli cestu pres baterie - to musi vymyslet strategie)
	 * @param start - startovni policko
	 * @param end - koncove policko
	 * @param startRobot.Direction - pocatecni natoceni robota
	 * @param endRobot.Direction - koncove natoceni robota
	 * @param startEnergy - energie na zacatku cesty
	 * @param stepLimit - maximalni pocet tahu cesty
	 * @return sezam tahu (prvni tah cesty je v tomto seznamu jako prvni) nebo null pokud neexistuje
	 */
	 protected LinkedList<BotAction> findPath(Number5Square start, Number5Square end, Direction startDirection, Direction endDirection, int stepLimit){
		 
                return null;
	 }
}
	
