package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.number5.Number5Square.*;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.LinkedList;
import java.util.HashMap;

public class Number5Robot {
    
    /******************************* SMI UPRAVOVAT JENOM SAM ROBOT *************************/
    /**
     * poradi robota v seznamu core.robots
     */
    int index;
    protected Number5 core;
    
    protected Number5State state;
    
    protected Robot robot;
    
    /**
     * prvni v seznamu je posledni akce
     * nemusi obsahovat vsechny akce za celou hru - mel by ale obsahovat rozumny pocet
     */
    protected LinkedList<Number5State> history;
    
    
    /******************************** VEREJNA DATA ******************************************/
    
    /**
     * - objekt, kam si kdokoliv muze ukladat jakakoliv data k tomuto robotu pro pozdejsi vyuziti
     * - kazdy, kdo s timto objektem pracuje, si musi dat pozor, aby String, pod kterym si svoje data ulozi byl dost unikatni
     */
    protected HashMap<String, Object> data;
    
    
    /******************************** VEREJNE METODY ******************************************/
    
    /**
     * konstruktor
     * @param robot - robot z verejneho rozhrani, prislusici tomuto robotu
     * @param core - jadro robota, nevim, jestli to bude potrebovat, ale co kdyby
     * @param index - poradi robota, musi ulozit do this.index
     */
    protected Number5Robot(Robot robot, int index, Number5 core){
        this.index = index;
        this.robot = robot;
        this.core = core;
        
        this.state = new Number5State(robot.getPositionX(), robot.getPositionY(), robot.getEnergy(),
                robot.getDirection(), robot.getPoints());
        
        this.history = new LinkedList<Number5State>();
        this.data = new HashMap<String, Object>();
    }
    
    /**
     * predikce pro tohoto robota (nemelo by byt volano na objektu naseho robota)
     * @param steps - po kolika tazich vestime pozici robota
     * @return predikce
     */
    protected Number5Prediction predict(int steps){
        /*predikce na moc kroku dopredu nema smysl*/
        if(steps > 3)
            return(Number5Prediction.DONT_KNOW);
        
        /*kratsi zapis souradnic*/
        int x = robot.getPositionX();
        int y = robot.getPositionY();
        Number5Square position = core.getSquare(x, y);
        
        /*  neni robot na nule?*/
        if(robot.getEnergy() == 0)
            return(new Number5Prediction(100,position,robot.getDirection(),0));
        
        /*neptame se na blbosti?*/
        if(steps == 0)
            return(new Number5Prediction(100,position,robot.getDirection(),robot.getEnergy()));
        
        /*tak a ted se jde pocitat*/
        int prob = 100; //prozatim pravdepodobnost, behem vypoctu zmenime
        Number5Prediction vysledek;
        
        /*mame-li historii, pocitame nejpravdepodobnejsi cestu vzhledem k dosavadnimu smeru*/
        if(history.size()>0){
            
            /*najdi nejake cile*/
            LinkedList<Number5Square> cile = new LinkedList<Number5Square>();
           
            
            //TADY JE ASI MISTO, KDE SE DA ROZHODNOUT O TOM, JAK DLOUHO BUDE TRVAT VYPOCET
            int maxVzdalCil = 15;
            
            for(int i = 1; i < maxVzdalCil; i++){
                //pridej do seznamu cilu vsechny baterie a poklady ve "vzdalenosti" i od startu
                int xt, yt;
                Number5Square cil;
                for(xt = x - i; xt <= x + i; xt++){ //vrchni a spodni hrana ctverce ve vzdalenosti i
                    yt = y - i;
                    if(core.getSquare(xt, yt) != null){
                        if(core.gameMap[xt][yt].type == SquareType.BATTERY){
                            cile.addLast(core.gameMap[xt][yt]);
                        }else if(core.gameMap[xt][yt].treasure > 0){
                            cile.addLast(core.gameMap[xt][yt]);
                        }
                    }
                    yt = y + i;
                    if(core.getSquare(xt, yt) != null){
                        if(core.gameMap[xt][yt].type == SquareType.BATTERY){
                            cile.addLast(core.gameMap[xt][yt]);
                        }else if(core.gameMap[xt][yt].treasure > 0){
                            cile.addLast(core.gameMap[xt][yt]);
                        }
                    }
                }
                for(yt = y - i; yt <= y + i; yt++){
                    xt = x - i + 1;
                    if(core.getSquare(xt, yt) != null){
                        if(core.gameMap[xt][yt].type == SquareType.BATTERY){
                            cile.addLast(core.gameMap[xt][yt]);
                        }else if(core.gameMap[xt][yt].treasure > 0){
                            cile.addLast(core.gameMap[xt][yt]);
                        }
                    }
                    xt = x + i - 1;
                    if(core.getSquare(xt, yt) != null){
                        if(core.gameMap[xt][yt].type == SquareType.BATTERY){
                            cile.addLast(core.gameMap[xt][yt]);
                        }else if(core.gameMap[xt][yt].treasure > 0){
                            cile.addLast(core.gameMap[xt][yt]);
                        }
                    }
                }
                
                if(cile.size()>20)
                    break;
            }
            
            /*vytvor seznam ctvercu v nedavne historii / souradnice jsou mezi 0..49 */
                /*souradnice x a y zakodovane do int cisla ve tvaru xxyy
                 */
            LinkedList<Number5Square> ctverceHist = new LinkedList<Number5Square>();
            
            int hloubka = Math.min(history.size(),10);
                        
            for(int i = 0; i < hloubka; i++){
                int xc = history.get(i).positionX;
                int yc = history.get(i).positionY;
                //int souradnice = 100 * xc + yc; //stacilo by urcit ctverec jako (x,y)
                Number5Square item = core.getSquare(xc, yc);
                if(!ctverceHist.contains(item)){
                    ctverceHist.addLast(item); 
                }
            }
            
                /*postupne prochazej cesty ze ctverce v historii do cilu,
                 *vyber nejlepsi cestu vzhledem k vlastnosti:
                 *velikost pruniku mnoziny poli, pres ktera vede navrh. cesta a mnoziny
                 *navstiv. poli v historii
                 *(ctverce bliz k aktualni pozici by mohly mit vetsi vahu)
                 */
            LinkedList<BotAction> cesta, cestaPath, cestaTemp;
            cesta = new LinkedList<BotAction>();
            cestaTemp = new LinkedList<BotAction>();
            cestaPath = new LinkedList<BotAction>();
            
            prob = 0;
            /*cesty a pocty poli spolecnych s historii*/
            for(int i = 0; i < cile.size(); i++){
                //spocitat si hodnoty typu Square - odkud a kam, ziskat startDirection, vyzkouset vsechny endDirection,  ziskat energy z historie
                Number5Square sourad = ctverceHist.get(0);
                Number5State startStav = history.get(hloubka - 1);
                Number5Square odkud = core.getSquare(startStav.positionX, startStav.positionY);
                int limit = startStav.energy;
                
                //spocitame nejkratsi cestu, ze vsech ctyr natoceni
                for(Direction s : Direction.values()){
                    cestaTemp = core.path.findPath(odkud,cile.get(i),startStav.direction, s, state.energy, limit);
                    if(cestaTemp != null && !cestaTemp.isEmpty()&&( (cestaTemp.size() < cestaPath.size())||(cestaPath.size() == 0) ))
                        cestaPath = cestaTemp;
                }
                
                //pocet ctvercu v pruniku Pathfinderem naplanovane cesty do i.teho cile a Historie
                int pocetStejnychCtvercu = 0;
                //pravdepodobnost, ze se robot vydal zrovna do i.teho cile nastav. na 0, zvysi se pri prochazeni cesty
                int probPath = 0;
                //nastavime pocatecni souradnice a natoceni a projdeme Pathfinderem vypocitanou cestu
                int xt = startStav.positionX;
                int yt = startStav.positionY;
                Direction dt = startStav.direction;
                
                for(int j = 0; j < hloubka && j < cestaPath.size(); j++){
                        /* prochazet cestu cestaPath, vypocitat souradnice ctvercu, kudy vede min. cesta do cile,
                         * a zkoumat, zda jsou v ctverceHist*/
                    switch(cestaPath.get(j)){
                        case LEFT:
                            switch(dt){
                                case DOWN:
                                    dt = Direction.RIGHT;
                                    break;
                                case LEFT:
                                    dt = Direction.DOWN;
                                    break;
                                case RIGHT:
                                    dt = Direction.UP;
                                    break;
                                case UP:
                                    dt = Direction.LEFT;
                                    break;
                            }
                            break;
                        case RIGHT:
                            switch(dt){
                                case DOWN:
                                    dt = Direction.LEFT;
                                    break;
                                case LEFT:
                                    dt = Direction.UP;
                                    break;
                                case RIGHT:
                                    dt = Direction.DOWN;
                                    break;
                                case UP:
                                    dt = Direction.RIGHT;
                                    break;
                            }
                            break;
                        case STEP:
                            switch(dt){
                                case DOWN:
                                    yt++;
                                    break;
                                case LEFT:
                                    xt--;
                                    break;
                                case RIGHT:
                                    xt++;
                                    break;
                                case UP:
                                    yt--;
                                    break;
                            }
                            //je-li ctverec, na ktery jsme slapli, v historii, pripocitame
                            Number5Square itemt = core.getSquare(xt, yt);
                            if(ctverceHist.contains(itemt)){
                                pocetStejnychCtvercu++;
                            }
                            if(itemt.type == SquareType.WALL)
                                pocetStejnychCtvercu = 0;
                            break;
                        default:
                            //nezmeni se ani xt ani yt ani dt, tak nas to nezajima
                    }
                    
                    probPath = 100 * pocetStejnychCtvercu / (cestaPath.size());
                }
                
                if(probPath > prob){
                    prob = probPath;
                    cesta = cestaPath;
                }
            }
            
            if(cesta.isEmpty())
                prob = 0;
            /*nejpravdepodobnejsi cesta... cesta, pravdepodobnost ... prob*/
            /*"zdrzovani se" - poklady, baterie, strelba*/
            
                /*najdeme na vyhledane ceste pozici a spocitame energii po steps krocich
                 *public Number5State(int positionX, int positionY, int energy, Direction direction, int points){
                 */
            int xt = x;
            int yt = y;
            Direction dt = state.direction;
            int et = state.energy;
            int pt = state.points;
            int zdrzeni = 0;//zdrzeni na zacatku cesty kvuli strelbe apod, pokud je to vyhodne
            
            //vypocet zdrzeni
            if((robotAtSight()!=null)||((core.gameMap[x][y]).treasure > 0)){
                zdrzeni++;
                et--;
            }else if(((core.gameMap[x][y]).type == SquareType.BATTERY)&&((history.getLast().positionX != x)||(history.getLast().positionY != y))){
                zdrzeni++;
                et = core.world.getMaxRobotEnergy();
            }
                
            for(int r = 0; (r < (steps - zdrzeni)) && (r < cesta.size()); r++){
                switch(cesta.get(r)){
                    case LEFT:
                        switch(dt){
                            case DOWN:
                                dt = Direction.RIGHT;
                                break;
                            case LEFT:
                                dt = Direction.DOWN;
                                break;
                            case RIGHT:
                                dt = Direction.UP;
                                break;
                            case UP:
                                dt = Direction.LEFT;
                                break;
                        }
                        break;
                    case RIGHT:
                        switch(dt){
                            case DOWN:
                                dt = Direction.LEFT;
                                break;
                            case LEFT:
                                dt = Direction.UP;
                                break;
                            case RIGHT:
                                dt = Direction.DOWN;
                                break;
                            case UP:
                                dt = Direction.RIGHT;
                                break;
                        }
                        break;
                    case STEP:
                        switch(dt){
                            case DOWN:
                                yt++;
                                break;
                            case LEFT:
                                xt--;
                                break;
                            case RIGHT:
                                xt++;
                                break;
                            case UP:
                                yt--;
                                break;
                        }
                    case RECHARGE:
                        et = core.world.getMaxRobotEnergy();
                        break;
                    case TAKE:
                        pt += core.world.getTreasureValue(xt,yt);
                        break;
                    case REST:
                        et += core.world.getRestRobotEnergy();
                        break;
                    case SHOOT:
                        break;
                }
                if((cesta.get(r) != BotAction.REST) && (cesta.get(r) != BotAction.RECHARGE))
                    et--;
                
            }
            position = core.getSquare(xt, yt);
            if(position == null || position.type == SquareType.WALL) return Number5Prediction.DONT_KNOW;
            return (new Number5Prediction(prob,position,dt,et));
            
            /*pokud nemame k dispozici historii*/
        }else{
            if(steps == 1){
                /*pokud robot stoji na baterii a nestal tam uz v minulem kroku provede RECHARGE*/
                 if(((core.gameMap[x][y]).type == SquareType.BATTERY)&&((history.getLast().positionX != x)||(history.getLast().positionY != y))){
                     prob = 100 - 100*robot.getEnergy()/core.world.getMaxRobotEnergy();
                     return (new Number5Prediction(prob,position,robot.getDirection(),core.world.getMaxRobotEnergy()));
                 /*strelba a sbirani pokladu*/
                 }else if((robotAtSight()!=null)||((core.gameMap[x][y]).treasure > 0)){
                     prob = 100;    
                     return(new Number5Prediction(prob,position,robot.getDirection(),(robot.getEnergy() - 1)));
                 }else return(Number5Prediction.DONT_KNOW);
            }else{
                return(Number5Prediction.DONT_KNOW);
            }
            
        }
    }
    
    /**
     * - vola jadro po te, co robot tahne
     * - ve chvili volani funkce jsou udaje robota (pozice, enrgie, body...) nastavene na aktualni hodnotu
     * @param action - co robot pod dohledem prave provedl
     * @return predikce
     */
    protected void step(BotAction action){
        //prida akci do historie
        state.action = action;
        history.addFirst(state);
        
        // aktualizuj udaje
        this.state = new Number5State(robot.getPositionX(), robot.getPositionY(), robot.getEnergy(),
                robot.getDirection(), robot.getPoints());
    }
    
    /**
     * - vrati objekt robota, ktery je na dostrel nebo null, pokud zadny takovy robot neexistuje
     */
    protected Number5Robot robotAtSight(){
        Number5Square s = core.getSquare(this);
        while((s = core.getNextSquare(s, state.direction)) != null){
            if(s.type == SquareType.WALL) return null;
            else if(s.robot != null) return s.robot;
        }
        return null;
    }
    /**
     * - vrati objekt robota, ktery je na dostrel z mista position nebo null, pokud zadny takovy robot neexistuje
     */
    protected Number5Robot robotAtSight(Number5Square s, Direction direction){
        while((s = core.getNextSquare(s, direction)) != null){
            if(s.type == SquareType.WALL) return null;
            else if(s.robot != null) return s.robot;
        }
        return null;
    }
    protected Number5Robot robotAtSight(Number5State states[]){
        Number5State me = states[index];
        Number5Square s = core.getSquare(me.positionX, me.positionY);
        while((s = core.getNextSquare(s, me.direction)) != null){
            if(s.type == SquareType.WALL) return null;
            else {
                int x = 0;
                for(Number5State state : states){
                    if(state.positionX == s.x && state.positionY == s.y && state != me) return core.robots[x];
                    x++;
                }
            }
        }
        return null;
    }
    
    public String toString(){
        return robot.getName();
    }
}