package cz.muni.fi.fibot.number5;

import cz.muni.fi.fibot.Strategy;

import java.util.LinkedList;

public class Number5Run {
    
    public static final int PRIORITY_NOT_SET = 0;
    public static final int LOW_PRIORITY = 1;
    public static final int MEDIUM_PRIORITY = 2;
    public static final int HIGH_PRIORITY = 3;
    
    Number5 core;
    
    /**
     * - ridici strategie zavola, pokud se chce vzdat rizeni (nebo rizeni preruseni)
     * - smi volat jak STRATEGIE U MOCI, tak i PRERUSUJICI STRATEGIE
     */
    protected void resign(){
        resigned = true;
    }
    
    /**
     * - smi volat jenom STRATEGIE U MOCI
     * - strategie u moci zavola, kdyz chce zpatky moc :)
     *   (strategie bude ridit robota hned v tomto kole)
     */
    protected void breakInterrupt(){
        if(interrupting == null) return;
        setInterruptable(0, MEDIUM_PRIORITY);
    }
    
    
    /**
     * - smi volat jenom STRATEGIE U MOCI
     * - strategie u moci timto nastavi, jestli smi byt prerusitelna
     * - muze zavolat i ve stavu preruseni - pokud prehodnoti svuj postoj
     * 	- jadro pak muze v dusledku tohoto nastaveni vzit prerusujici strategii rizeni
     * @param stepsLimit - na kolik dalsich tahu vcetne tohoto je ochotna vzdat se preruseni
     * @param priorityLimit - preruseni musi mit nastavenou takovou nebo vyssi prioritu,
     *				aby mohlo byt zvoleno/pokracovat ve vykonavani
     *		- jedna z hodnot Number5Manager.LOW_PRIORITY, ...
     */
    protected void setInterruptable(int stepsLimit, int priorityLimit){
        if(calling != directive) return;
        interruptionStepsLimit = stepsLimit;
        interruptionPriorityLimit = priorityLimit;
    }
    
    /**
     * - smi volat jenom PRERUSUJICI STRATEGIE (tedy jenom ve stavu preruseni)
     * @param neededSteps - kolik bude jeste potreba kroku ke splneni cile
     * @param priority - priorita tohoto preruseni (jedna z hodnot Number5Manager.LOW_PRIORITY, ...)
     */
    protected void updateInterrupt(int neededSteps, int priority){
        if(calling != interrupting) return;
        interruptionNeededSteps = neededSteps;
        interruptionPriority = priority;
    }
    
    
    public Number5Run(Number5 core){
        this.core = core;
        
        Number5Strategy add;
        
        strategies = new LinkedList<Number5Strategy>();
        
        // killer
        add = new Number5Killer();
        add.setCore(core);
        strategies.add(add);
        
        // charger
        add = new Number5Charger();
        add.setCore(core);
        strategies.add(add);
        
        // gold digger
        //add = new Number5GoldDigger();
        add = new Number5MushroomPicker();
        add.setCore(core);
        strategies.add(add);
        
        // machine gunner
        add = new Number5MachineGunner();
        add.setCore(core);
        strategies.add(add);
        
        /* pridavej strategie */
    }
    
    public Strategy.BotAction step(){
        try {
        // neni zadna strategie :)
        if(strategies.size() == 0) return Strategy.BotAction.REST;
        
        // zacatek hry - neni zadna ridici strategie
        if(directive == null) directive = chooseDirective();
        
        // jak to ted vidi ridici strategie?
        calling = directive;
        directive.manage(this);
        
        // vzdala se moci - cyklus proto
        while (hasResigned()){
            Number5Strategy temp = chooseDirective();
            // cykli volba a vzdani se rizeni !!
            /*if(temp == directive){
                System.out.println("Cyklim - REST");
                return Strategy.BotAction.REST;
            }
            else*/ directive = temp;
            calling = directive;
            directive.manage(this);
        }
        
        // nezakazala preruseni??
        if(interruptionStepsLimit == 0) interrupting = null;
        
        // je zrovna preruseni?
        if(interrupting != null){
            calling = interrupting;
            interrupting.manageInterrupt(this);
            
            // nevzdala se moci
            if(hasResigned()) interrupting = null;
            // nema vetsi pozadavky nebo nizsi prioritu, nez jsou moznosti??
            else if(interruptionStepsLimit < interruptionNeededSteps || interruptionPriority < interruptionPriorityLimit) interrupting = null;
        }
        
        // neni zrovna preruseni - nove volby
        if(interrupting == null && interruptionStepsLimit > 0) interrupting = chooseInterrupting();
        
        // SAMOTNE ROZHODNUTI
        Strategy.BotAction a;
        if(interrupting != null) a = interrupting.decideInterrupt();
        else a = directive.decide();
        //if(core.debug) System.out.println("Krok "+core.world.getRound()+": "+a);
        return a;
        } catch (Exception e){
            return Strategy.BotAction.REST;
        }
    }
    
    /**************************************************** PRIVATE *************************************************/
    
    /* vsechny nase strategie */
    private LinkedList<Number5Strategy> strategies = null;
    private Number5Strategy directive = null;
    private Number5Strategy interrupting = null;
    private Number5Strategy calling = null;
    
    private Number5Proposal directiveProposal = null;
    
    private int interruptionStepsLimit = 0;
    private int interruptionPriorityLimit = MEDIUM_PRIORITY;
    private int interruptionNeededSteps = 0;
    private int interruptionPriority = LOW_PRIORITY;
    private boolean resigned = false;
    
    private boolean hasResigned(){
        if(resigned) {
            resigned = false;
            return true;
        }
        return false;
    }
    
    private Number5Strategy chooseDirective(){
        int x;
        
        // muzu pocitat s tim ze pocet je nenulovy
        int strategiesCount = strategies.size();
        
        Number5Proposal props[] = new Number5Proposal[strategiesCount];
        int best = 0;
        
        for(x = 0; x < strategiesCount; x++){
            props[x] = strategies.get(x).getMainProposal();
        }
        
        for(x = 1; x < strategiesCount; x++){
            if(props[x].compareTo(props[best]) > 0) best = x;
        }
        
        if(core.debug){
            System.out.println("---------------------------");
            for(x = 0; x < strategiesCount; x++){
                if(props[x] != Number5Proposal.NO_PLAN){
                    if(x == best) System.out.println(" * " + props[x]);
                    else System.out.println("   " + props[x]);
                }
            }
        }
        //System.out.println("Vitez hlavnich voleb: "+props[best].description);
        directiveProposal = props[best];
        return strategies.get(best);
    }
    
    private Number5Strategy chooseInterrupting(){
        int x;
        
        // muzu pocitat s tim ze pocet je nenulovy
        int strategiesCount = strategies.size();
        
        Number5Proposal props[] = new Number5Proposal[strategiesCount];
        int best = -1;
        
        Number5Strategy strategy;
        for(x = 0; x < strategiesCount; x++){
            strategy = strategies.get(x);
            props[x] = strategy.getInterruptProposal(interruptionStepsLimit, strategy == directive);
        }
        
        for(x = 0; x < strategiesCount; x++){
            if(props[x].priority >= interruptionPriorityLimit
                    && (best == -1 || props[x].compareTo(props[best]) > 0)
                    && props[x].compareTo(directiveProposal) > 0){
                best = x;
            }
        }
        
        if(best == -1) return null;
        else {
            if(core.debug){
                System.out.println("---------------------------");
                for(x = 0; x < strategiesCount; x++){
                    if(props[x] != Number5Proposal.NO_PLAN){
                        if(x == best) System.out.println("|| * " + props[x]);
                        else System.out.println("||   " + props[x]);
                    }
                }
            }
            //System.out.println("Vitez preruseni: "+props[best].description);
            return strategies.get(best);
        }
    }
}

