/*
 * AvailabilityPlan.java
 *
 * Created on 7. bezen 2006, 21:21
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.pikachu;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 * This class determines the space where the robot can get and in what round
 *
 * This does not cover the "battery" - only steps + turns + rest
 * @author Administrator
 */
public class AvailabilityMap{// implements Cloneable {
    
    /** plan */
    protected FieldDescriptor map[][];
    
    private int dimensionX;
    
    private int dimensionY;
    
    /** the field on which the robot stands at the beginning */
    protected FieldDescriptor startingField = null;
    
    /** the priority queue of the unprocessed fields - see Dijkstra */
    //private PriorityQueue<FieldDescriptor> queue;
    private Set<FieldDescriptor> queue;
    
    /** the world pointer */
    private World world;
    
    /* clone */
    /*public Object clone() {
        AvailabilityMap cloMap = new AvailabilityMap();
        cloMap.dimensionX = this.dimensionX;
        cloMap.dimensionY = this.dimensionY;
        cloMap.world = this.world;
        cloMap.queue = this.queue.clone();
        for (int i=0; i < dimensionX; i++) {
            for (int j=0; j < dimensionY; j++) {
                if (this.map[i][j] == null)
                    cloMap.map[i][j] = null;
                else {
                    cloMap.map[i][j] = (FieldDescriptor) map[i][j].clone();
                }
            }
        }
        // set the predecessor fields now
        for (int i=0; i < dimensionX; i++) {
            for (int j=0; j < dimensionY; j++) {
                if ((this.map[i][j] != null) && (this.map[i][j].predecessorField != null))
                    cloMap.map[i][j].predecessorField =
                            cloMap.map[map[i][j].predecessorField.x][map[i][j].predecessorField.y];
            }
        }
        return cloMap;
    }*/
    
    /** Creates a new empty instance of AvailabilityPlan */
    public AvailabilityMap() {
        map = new FieldDescriptor[dimensionX][dimensionY];
    }
    
    /** Creates a new instance of AvailabilityPlan */
    public AvailabilityMap(World world) {
        dimensionX = world.getDimensionX();
        dimensionY = world.getDimensionY();
        this.world = world;
        queue = new HashSet<FieldDescriptor>(dimensionX * dimensionY);
        map = new FieldDescriptor[dimensionX][dimensionY];
        
        for (int i=0; i < dimensionX; i++) {
            for (int j=0; j < dimensionY; j++) {
                if (world.containsWall(i, j))
                    map[i][j] = null;
                else {
                    map[i][j] = new FieldDescriptor(i,j);
                }
            }
        }
        // initialize the pointers
        /*for (int i=0; i < dimensionX; i++) {
            for (int j=0; j < dimensionY; j++) {
                if (map[i][j] != null)
                    initPointers(map[i][j]);
            }
        } */
    }
    
    /** generates the map of available positions. Basically use the Dijkstra algorithm.
     */
    public AvailabilityMap fillAvailabilityMap(Robot me) {
        return fillAvailabilityMap(me, Integer.MAX_VALUE);
    }
    public AvailabilityMap fillAvailabilityMap(Robot me, int maxSteps) {
        return fillAvailabilityMap(new Position(me.getPositionX(), me.getPositionY()), me.getEnergy(),
                me.getDirection(), maxSteps);
    }
    /** generates the map of available positions. Basically use the Dijkstra algorithm.
     * @param maxSteps maximal number of steps to generate the map to (e.g. max steps
     * of the whole algorithm).
     */
    public AvailabilityMap fillAvailabilityMap(Position position, int energy, Robot.Direction direction, int maxSteps) {
        resetMap();
        // init the startup field - the robot is standing on it at the moment
        startingField = map[position.x][position.y];
        startingField.energy = energy;
        startingField.round = 0;
        startingField.direction = direction;
        
        int steps = 0;
        FieldDescriptor tmpField = null;
        Robot.Direction tmpDirection = null;
        
        FieldDescriptor currentField = getSmallestElement(queue);//queue.poll();
        
        while ((steps < maxSteps) && (!queue.isEmpty())) {
            
            for (int i=0; i < 4; i++) {
                // 0 = step forward, 1 = step right, 2 = turn, turn + step, 3 = turn left + step
                tmpDirection = Robot.Direction.values()[(currentField.direction.ordinal() + i) % 4];
                tmpField = currentField.doStep(tmpDirection);
                
                // determine the number of steps to get to that direction
                int relaxSteps = i+1;
                if (i == 3) relaxSteps = 2;
                
                // do relax = improve the # of steps to get to tmpField
                if ((tmpField != null) && (queue.contains(tmpField))) {
                    // if the relaxation worked - resort the queue
                    tmpField.relax(currentField, tmpDirection, relaxSteps);
                }
            } // the end of moving into 4 directions from this point
            
            currentField = getSmallestElement(queue);//queue.poll();
            steps = currentField.round;
        }
        
        return this;
    }
    
    /** given an availibility map for a battery, this method finds the minimums
     * for each of the position.
     * @param stepsToBattery steps to get to the battery
     * @param overPositions a list o positions (batteries) to go over
     */
    public AvailabilityMap findMin(FieldDescriptor batteryMap[][], int stepsToBattery, List<Position> batPositions) {
        Position theBattery = batPositions.get(batPositions.size()-1);
        for (int i=0; i < dimensionX; i++) {
            for (int j=0; j < dimensionY; j++) {
                if ((map[i][j] != null) && (!((theBattery.x == i) && (theBattery.y == j)))) {
                    map[i][j].findMin(batteryMap[i][j], stepsToBattery, batPositions);
                }
            }
        }
        
        return this;
    }
    
    /** return the distance to the given position
     * @throws IllegalArgumentException if out of the map or filled with wall
     */
    public int getDistance(Position pos) throws IllegalArgumentException {
        return getDistance(pos.x, pos.y);
    }
    /** return the distance to the given position
     * @throws IllegalArgumentException if out of the map or filled with wall
     */
    public int getDistance(int x, int y) throws IllegalArgumentException {
        if (! areCoordinates(x,y) || map[x][y] == null)
            throw new IllegalArgumentException();
        return map[x][y].round;
    }
    /** return the direction that I will have on this position
     * @throws IllegalArgumentException if out of the map or filled with wall
     */
    public Robot.Direction getDirection(Position pos) throws IllegalArgumentException {
        return getDirection(pos.x, pos.y);
    }
    public Robot.Direction getDirection(int x, int y) throws IllegalArgumentException {
        if (! areCoordinates(x,y) || map[x][y] == null)
            throw new IllegalArgumentException();
        return map[x][y].direction;
    }
    
    public List<Position> getBatPositions(Position pos) throws IllegalArgumentException {
        return getBatPositions(pos.x,pos.y);
    }
    public List<Position> getBatPositions(int x, int y) throws IllegalArgumentException {
        if (! areCoordinates(x,y) || map[x][y] == null)
            throw new IllegalArgumentException();
        return map[x][y].batPositions;
    }
    
    /** return the chain of actions necessary to get to the given position
     * EXCLUDING THE NECESSARY "REST" STEPS
     *@throws IllegalArgumentException if out of the map or filled with wall
     */
    public ChainOfActions getChainOfActions(Position pos) throws IllegalArgumentException {
        return getChainOfActions(pos.x, pos.y);
    }
    public ChainOfActions getChainOfActions(int x, int y) throws IllegalArgumentException {
        if (! areCoordinates(x,y) || map[x][y] == null)
            throw new IllegalArgumentException();
        ChainOfActions chain = new ChainOfActions(new Position(x,y));
        FieldDescriptor currentField = map[x][y];
        //System.out.println(this);
        //System.out.println(currentField);
        while (! currentField.equals(startingField)) {
            if (currentField.predecessorField == null)
                throw new IllegalStateException("The map is not initilized properly to find the track to position ("+x+","+y+")");
            chain.pushAction(BotAction.STEP);
            if (currentField.direction != currentField.predecessorField.direction) {
                int dirDiff = (currentField.predecessorField.direction.ordinal() - currentField.direction.ordinal() + 4) % 4;
                switch (dirDiff) {
                    case 1:
                        chain.pushAction(BotAction.LEFT);
                        break;
                    case 2:
                        chain.pushAction(BotAction.LEFT);
                        chain.pushAction(BotAction.LEFT);
                        break;
                    case 3:
                        chain.pushAction(BotAction.RIGHT);
                        break;
                    default:
                        throw new IllegalStateException("The map is not initilized properly - inconsistency during the path finding");//: "+currentField.direction+", "+currentField.predecessorField.direction+", diff: "+dirDiff);
                }
            }
            currentField = currentField.predecessorField;
        }
        //System.out.println("Planned actions: "+chain);
        
        return chain;
    }
    
    /** supplementary method - given a Set of fields find, remove and return the one with lowest "round" */
    private FieldDescriptor getSmallestElement(Set<FieldDescriptor> queue) {
        if (queue.isEmpty())
            return null;
        FieldDescriptor retVal = null;
        for (FieldDescriptor field:queue) {
            if ((retVal == null) || (field.compareTo(retVal) < 0))
                retVal = field;
        }
        queue.remove(retVal);
        return retVal;
    }
    
    /** supplementary method - returns true if the given coordinates are relevant in this world */
    private boolean areCoordinates(int x, int y) {
        return (x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0);
    }
    
    /** resets the values of all the field and put them into the queue */
    private void resetMap() {
        queue.clear();
        for (int i=0; i < dimensionX; i++) {
            for (int j=0; j < dimensionY; j++) {
                if (map[i][j] != null) {
                    map[i][j].resetField();
                    queue.add(map[i][j]);
                }
            }
        }
    }
    
    /* supplementary method - init the pointers up, down, left, right of the given field */
    /*private void initPointers(FieldDescriptor field) {
        int x = field.x;
        int y = field.y;
        if (areCoordinates(x+1, y))
            field.right = map[x+1][y];
        if (areCoordinates(x-1, y))
            field.left = map[x-1][y];
        if (areCoordinates(x, y+1))
            field.down = map[x][y+1];
        if (areCoordinates(x, y-1))
            field.up = map[x][y-1];
    }*/
    
    
    /** descriptor of the AvailibilityMap field -
     *     - the energy of the robot when it gets there
     *     - number of rounds to get into the state
     *     - the predecessor field to get there from
     */
    protected class FieldDescriptor implements Comparable {
        /** the X coordinate */
        protected int x;
        
        /** the Y coordinate */
        protected int y;
        
        /** the round in which I can get on this field */
        protected int round = Integer.MAX_VALUE;
        
        /** the energy that I would have when getting here */
        protected int energy = 0;
        
        /** direction into which the robot would be turned */
        protected Robot.Direction direction = Robot.Direction.UP;
        
        /** the predecessor field */
        protected FieldDescriptor predecessorField = null;
        
        /** if filled, these battery positions have to be passed over to get here
         *
         * if the field does not have the predecessorField filled, it
         * gotta have filled this field */
        public List<Position> batPositions = null;
        
        /** the constructor */
        protected FieldDescriptor(int x, int y) {
            this.x = x;
            this.y = y;
        }
        
        /** get pointers to fields "up", "down", left, right. If border or wall in that direction - return null */
        protected FieldDescriptor up() {
            return (areCoordinates(x, y-1))?(map[x][y-1]):null;
        }
        protected FieldDescriptor down() {
            return (areCoordinates(x, y+1))?(map[x][y+1]):null;
        }
        protected FieldDescriptor left() {
            return (areCoordinates(x-1, y))?(map[x-1][y]):null;
        }
        protected FieldDescriptor right() {
            return (areCoordinates(x+1, y))?(map[x+1][y]):null;
        }
        
        /** reset the values to the initial values */
        protected void resetField() {
            round = Integer.MAX_VALUE;
            energy = 0;
            predecessorField = null;
            batPositions = null;
        }
        
        /** relax in the sense of Dijkstra */
        protected boolean relax(FieldDescriptor source, Robot.Direction direction, int steps) {
            int targetEnergy = source.energy - steps;
            while (targetEnergy < 1) {
                steps++;
                targetEnergy += (world.getRestRobotEnergy() - 1);
            }
            
            if (this.round > source.round + steps) {
                this.round = source.round + steps;
                this.energy = targetEnergy;
                this.predecessorField = source;
                this.direction = direction;
                return true;
            }
            return false;
        }
        
        /** compare with the other field + additional steps and get the minimum of these two
         * This method do not preserve the possibility of building the path by "predecessorField"
         */
        protected boolean findMin(FieldDescriptor alternative, int stepsToBattery, List<Position> batPositions) {
            if (this.round >= alternative.round + stepsToBattery) {
                this.round = alternative.round + stepsToBattery;
                this.energy = alternative.energy;
                this.direction = direction;
                //this.predecessorField = null;
                this.batPositions = batPositions;
                return true;
            }
            return false;
        }
        
        /** supplementary method - given a direction return the field in given direction
         * or "null" if wall is in that diretion */
        protected FieldDescriptor doStep(Robot.Direction direction) {
            switch (direction) {
                case UP: return up();
                case DOWN: return down();
                case LEFT: return left();
                case RIGHT: return right();
            }
            return null;
        }
        
        /* clone the Field - the "predecessorField" is set to null */
        /*public Object clone() {
            FieldDescriptor clo = new FieldDescriptor(x,y);
            clo.round = round;
            clo.energy = energy;
            clo.direction = direction;
        }*/
        
        /** comparator according to the round to get here */
        public int compareTo(Object o) {
            if (o instanceof FieldDescriptor) {
                FieldDescriptor fd = (FieldDescriptor) o;
                return ((Integer) round).compareTo(fd.round);
            }
            return 0;
        }
        
        /** equeals method - do not use compareTo - how to do that? */
        public boolean equals(Object obj) {
            if (!( obj instanceof FieldDescriptor)) {
                return false;
            }
            return (obj.hashCode() == this.hashCode());
        }
        
        public String toString() {
            return "Field: x="+this.x+", y="+this.y+", round: "+this.round+", energy: "+this.energy+", dir: "+this.direction
                    +", pred: \n\t"+predecessorField;
        }
        
        /** toString method - formating to 4x4 fields */
        public String[] toStrings() {
            if (batPositions != null) {
                String[] retVal0 = {"BBBB", format(round, 4), "("+format(energy, 2) + ")", "BBBB"};
                return retVal0;
            }
            switch (direction) {
                case UP:
                    String[] retVal1 = {"  ^ ", "  | ", format(round, 4), "("+format(energy, 2) + ")"};
                    return retVal1;
                case DOWN:
                    String[] retVal2 = {format(round, 4), "("+format(energy, 2) + ")", "  | ", "  V "};
                    return retVal2;
                case LEFT:
                    String[] retVal3 = {format(round, 4), "("+format(energy, 2) + ")", "<-  ", "    "};
                    return retVal3;
                case RIGHT:
                    String[] retVal4 = {format(round, 4), "("+format(energy, 2) + ")", "  ->", "    "};
                    return retVal4;
            }
            String[] retVal5 = {"    ", "    ", "    ", "    "};
            return retVal5;
        }
        private String format(int number, int digits) {
            String num = ""+number;
            if (num.length() > digits)
                return num.substring(num.length()-digits);
            return "          ".substring(num.length()+10-digits) + num;
        }
    } // end of FieldDescriptor
    
    /** to string */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        String line0, line1 = "", line2 = "", line3 = "", line4 = "";
        String fieldStrings[];
        for (int i=0; i < dimensionY; i++) {
            line0 = ""; line1 = ""; line2 = ""; line3 = ""; line4 = "";
            for (int j=0; j < dimensionX; j++) {
                line0 += "-----";
                if (map[j][i] != null) {
                    fieldStrings = map[j][i].toStrings();
                    line1 += "|"+fieldStrings[0];
                    line2 += "|"+fieldStrings[1];
                    line3 += "|"+fieldStrings[2];
                    line4 += "|"+fieldStrings[3];
                } else {
                    line1 += "|####";
                    line2 += "|####";
                    line3 += "|####";
                    line4 += "|####";
                }
            }
            buf.append(line0+"|\n");
            buf.append(line1+"|\n");
            buf.append(line2+"|\n");
            buf.append(line3+"|\n");
            buf.append(line4+"|\n");
        }
        //buf.append("-----");
        return buf.toString();
    }
    
}
