/*
 * ChainOfActions.java
 *
 * Created on 15. bezen 2006, 22:04
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.pikachu;

import cz.muni.fi.fibot.Strategy.BotAction;
import java.util.ArrayList;
import java.util.List;

/**
 * This class encapsulates the prepared chain of actions that the robot is going to take
 * (if no unexpected event appeared)
 *
 * TODO: add some inteligence (no its just a wrapper for List<BotAction>)
 * 
 * @author Administrator
 */
public class ChainOfActions {
  
    /** the own chain of actions */
    private List<BotAction> chain;
    
    /** what will I gain by this chain of actions: in the easy case: treasure - #of steps */
    protected int actionsGain = 0;
    
    /** target field to be reached by this chain (if any) */
    private Position targetField = null;    
    /** getter */
    public Position getTargetField() { return targetField; }
    /** setter */
    public void setTargetField(Position targetField) { this.targetField = targetField; }
    
    /** Creates a new instance of ChainOfActions */
    public ChainOfActions() {
        chain = new ArrayList<BotAction>();
    }

    /** Creates a new instance of ChainOfActions */
    public ChainOfActions(Position position) {
        chain = new ArrayList<BotAction>();
        setTargetField(position);
    }
    
    /** Creates a new instance of ChainOfActions */
    public ChainOfActions(List<BotAction> chain) {
        this.chain = chain;
    }
    
    /** Adds an action to the end of the chain 
     * @return the chain itself
     */
    protected ChainOfActions addAction(BotAction action) {
        chain.add(action);
        return this;
    }

    /** Adds a chain of actions to the end of the chain 
     * @return the chain itself
     */
    protected ChainOfActions addActions(ChainOfActions actions) {
        for (BotAction action:actions.chain)
            chain.add(action);
        return this;
    }
    
    /** Adds an action to the BEGINNING of the chain 
     * @return the chain itself
     */
    protected ChainOfActions pushAction(BotAction action) {
        chain.add(0, action);
        return this;
    }
    
    /** returns the count of the actions 
     * 0 if empty
     * -1 if not initialized (should not happen)
     */
    public int getActionsCount() {
        if (chain == null)
            return -1;
        return chain.size();
    }
    
    /** return (do not remove) the next action to be taken 
     * @throws IllegalStateException if chain empty 
     */
    public BotAction seeNextAction() throws IllegalStateException {
        if (chain.isEmpty())
            throw new IllegalStateException();
        return chain.get(0);
    }

    /** return and remove the next action to be taken 
     * @throws IllegalStateException if chain empty 
     */
    public BotAction popNextAction() throws IllegalStateException {
        if (chain.isEmpty())
            throw new IllegalStateException();
        return chain.remove(0);
    }
    
    /** clear the list */
    protected void clear() {
        chain.clear();
    }
    
    /** to string */
    public String toString() {
        if (chain.isEmpty())
            return "Empty chain of actions";
        return chain.toString();
      
    }
}
