/*
 * Pikachu.java
 *
 * Created on 15. bezen 2006, 20:55
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package cz.muni.fi.fibot.pikachu;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Random;
/**
 *
 * @author Administrator
 */
public class Pikachu implements Strategy {
    
    private int FLAG_IWASSHOOT = 0x00000001;
    private int FLAG_ENEMYINSHOOTLINE = 0x00000002;
    
    
    // GLOBAL WEIGHTS FOR SUBACTIONS GOALS
    private double WEIGHT_ATTACK_GAINS = 1;
    private double WEIGHT_DEFENSE_GAINS = 1;
    private double WEIGHT_TREASURE_GAINS = 1;
    
    // WEIGHTS SPECIFIC FOR ATTACK SUBACTIONS GOALS
    private double WEIGHT_ATTACK_1STEP =        10;
    private double WEIGHT_ATTACK_2STEP =        0.5;
    private double WEIGHT_ATTACK_3STEP =        0.25;
    private double WEIGHT_ATTACK_DISTANTENEMY = 1;
    
    
    /** actual World */
    private World world;
    
    /** the robot to decide for */
    private Robot robot;
    
    /** the map showing in how many steps I can get where */
    private AvailabilityMap map = null;
    
    /** the list of maps for each battery - precompute the availibility from all batteries (expecting
     * the bot recharges here)
     */
    private Map<Position,AvailabilityMap> batteryMaps = null;
    
    /** map of AvailibilityMap  for each of other robots
     */
    private Map<Robot,AvailabilityMap> robotMaps = null;
    
    /** list of battery positions */
    List<Position> batteries = null;
    
    /** my current coordinates (don't want to find them out through the Robot everytime) */
    private Position myPosition = null;
    
    /** remember the last value of energy in order to detect being shot */
    private int lastEnergy = 0;
    
    Random generator = null; 
    
    /** Creates a new instance of GeneralStrategy */
    public Pikachu() {
    }
    
    public void setWorld(World world) {
        this.world = world;
    }
    
    public void setRobot(Robot robot) {
        this.robot = robot;
    }
    
    public void initialize() {
        // init the map
//        //System.out.println("inicializace");
        this.map = new AvailabilityMap(world);
        batteries = Tools.findFields(world, Tools.FieldType.BATERY);
        batteryMaps = new HashMap<Position,AvailabilityMap>(batteries.size());
        
        AvailabilityMap batteryMap = null;
        for (Position position:batteries) {
            for (Direction direction : Robot.Direction.values()) {
                batteryMap = new AvailabilityMap(world);
                batteryMap.fillAvailabilityMap(position, world.getMaxRobotEnergy(),
                        direction, Integer.MAX_VALUE);
                Position pos = new Position(position.x, position.y, direction);
                batteryMaps.put(pos,batteryMap);
//                //System.out.println("putting: "+pos);
            }
        }
        // prepare the availibility maps for all other robots
        this.robotMaps = new HashMap<Robot,AvailabilityMap>(3);
        for (Robot rob : world.getRobots()) {
            if (!robot.equals(rob)) {
                robotMaps.put(rob, new AvailabilityMap(world));
            }
        }
        
        
    }
    
    private class ActionsGain {
        public  double stepGain = 0;
        public  double turnClockwiseGain = 0;
        public  double turnCounterClockwiseGain = 0;
        public  double shootGain = 0;
        public  double takeGain = 0;
        public  double rechargeGain = 0;
        public  double restGain = 0;
        
        public void multiplyByWeight(double weight) {
            stepGain *= weight;
            turnClockwiseGain *= weight;
            turnCounterClockwiseGain *= weight;
            shootGain *= weight;
            takeGain *= weight;
            rechargeGain *= weight;
            restGain *= weight;
        }
        
        public BotAction getBestAction() {
            if (isMax(stepGain)) return BotAction.STEP;
            if (isMax(turnClockwiseGain)) return BotAction.RIGHT;
            if (isMax(turnCounterClockwiseGain)) return BotAction.LEFT;
            if (isMax(shootGain)) return BotAction.SHOOT;
            if (isMax(takeGain)) return BotAction.TAKE;
            if (isMax(rechargeGain)) return BotAction.RECHARGE;
            if (isMax(restGain)) return BotAction.REST;
            
            return BotAction.REST;
        }
        private boolean isMax(double value) {
            return ((value >= stepGain) && (value >= shootGain) && (value >= turnClockwiseGain) &&
                    (value >= turnCounterClockwiseGain) && (value >= takeGain) && (value >= rechargeGain) && (value >= restGain));
        }
        public String toString() {
            return "Step: "+stepGain+", turnCW: "+turnClockwiseGain+", turnCCW: "+turnCounterClockwiseGain + ", shoot: "+shootGain
                    +", take: "+takeGain+", recharge: "+rechargeGain+", rest: "+restGain;
        }
        public void clear() {
            stepGain = 0;
            turnClockwiseGain = 0;
            turnCounterClockwiseGain = 0;
            shootGain = 0;
            takeGain = 0;
            rechargeGain = 0;
            restGain = 0;
        }
    }
    
    public Strategy.BotAction decide() {
        //
        //  DECISION PROCESS:
        //  1. PERFORM LIFE CRITICAL ACTIONS (REST IF energy == 1)
        //  2. EVALUTE INDEPENDENTLY ATTACK, DEFENSE AND TREASURE SUBACTIONS
        //     => EXPECTED GAIN FOR EACH FROM ACTIONS (step, turnClockwise, turnCounterClockwise, shoot)
        //  3. SUM EXPECTED GAIN FOR EACH ACTION BASED ON GAINS FROM SUBACTIONS
        //  4. CHOOSE ACTION WITH BEST SCORE
        
        try {
            myPosition = new Position(robot.getPositionX(), robot.getPositionY(), robot.getDirection());
            //System.out.println("My position: " + myPosition);
            generator = new Random();

            
            // ESTIMATED GAINS FROM EACH SUBACTION
            ActionsGain defenseGain = new ActionsGain();
            ActionsGain attackGain = new ActionsGain();
            ActionsGain treasureGain = new ActionsGain();
            ActionsGain totalGain = new ActionsGain();
            
            // ***************************
            // LIFE CRITICAL ACTIONS
            // ***************************
            // if energy very low - rest otherwise not being able to continue
            if ((robot.getEnergy() == 1) && (! world.containsBattery(myPosition.x, myPosition.y))) {
                //System.out.println("My action: " + BotAction.REST);
                return BotAction.REST;
            }
            
            // ***************************
            // DEFENSE STRATEGY SUBACTIONS
            // ***************************
            //
            try {
                defenseGain = getDefenseSubactionsGain();
            } catch (Throwable e) {
                e.printStackTrace();
                defenseGain.clear();
            }
            /**/
            // ***************************
            // ATTACK STRATEGY SUBACTIONS
            // ***************************
            //
            try {
                attackGain = getAttackSubactionsGain();
            } catch (Throwable e) {
                e.printStackTrace();
                attackGain.clear();
            }
            /**/
            // ***************************
            // TREASURE STRATEGY SUBACTIONS
            // ***************************
            //
            try {
                treasureGain = getTreasureSubactionsGain();
            } catch (Throwable e) {
                e.printStackTrace();
                treasureGain.clear();
            }
            
            
            
            // ***************************
            // CHOOSE RESULT ACTION
            // ***************************
            //
            defenseGain.multiplyByWeight(WEIGHT_DEFENSE_GAINS);
            attackGain.multiplyByWeight(WEIGHT_ATTACK_GAINS);
            treasureGain.multiplyByWeight(WEIGHT_TREASURE_GAINS);
            
            totalGain.stepGain = defenseGain.stepGain + attackGain.stepGain + treasureGain.stepGain;
            totalGain.shootGain = defenseGain.shootGain + attackGain.shootGain + treasureGain.shootGain;
            totalGain.turnClockwiseGain = defenseGain.turnClockwiseGain + attackGain.turnClockwiseGain + treasureGain.turnClockwiseGain;
            totalGain.turnCounterClockwiseGain = defenseGain.turnCounterClockwiseGain + attackGain.turnCounterClockwiseGain + treasureGain.turnCounterClockwiseGain;
            totalGain.takeGain = defenseGain.takeGain + attackGain.takeGain + treasureGain.takeGain; 
            totalGain.rechargeGain = defenseGain.rechargeGain + attackGain.rechargeGain + treasureGain.rechargeGain; 
            totalGain.restGain = defenseGain.restGain + attackGain.restGain + treasureGain.restGain; 
            
            // CHECK FOR INVALID OR DEADLOCK ACTIONS
            Position newPos = performStep(myPosition);
            try {
                if (containsWallSafe(newPos.x, newPos.y)) {
                    // STEP TO WALL DETECTED, TURN CLOCWISE INSTEAD;
                    //System.out.println("    POSSIBLE STEP TO WALL DETECTED");
//                    totalGain.stepGain = 0;
                }
            } catch (Throwable e) {
                e.printStackTrace();
                totalGain.stepGain = 0;
            }
            
            
            //System.out.println("Gains: "+totalGain);
            //System.out.println("   Defense Gains: "+defenseGain);
            //System.out.println("   Attack Gains: "+attackGain);
            //System.out.println("   Treasure Gains: "+treasureGain);
            
            // CHOOSE ACTION WITH HIGHEST EXPECTED GAIN
            BotAction action = totalGain.getBestAction();
            
            //System.out.println("My action: " + action);
            return action;
            
        } catch (Throwable e) {
            e.printStackTrace();
        }
        
        // perform random move action otherwise
        BotAction action = BotAction.REST;
        switch(generator.nextInt(4)+1) {
            case 1: action = BotAction.LEFT; break;
            case 2: action = BotAction.RIGHT; break;
            case 3: action = BotAction.SHOOT; break;
            case 4: action = BotAction.STEP; break;
        }
        //System.out.println("Exception detected, random action performed: " + action);
        return action;
    }
    
    private ActionsGain getDefenseSubactionsGain() {
        ActionsGain gains = new ActionsGain();
        
        // TODO: myslim, ze VALOR rikal, ze tuhle cast udela. Pokud by to nevyslo, tak lze pouzit tuto velmi jednoduchou strategii
        
        // CHECK IF I WAS SHOOT
        if (lastEnergy - 1 > robot.getEnergy()) {
            //System.out.println("I HAVE BEEN SHOT");
        }
        lastEnergy = robot.getEnergy();
        
        List<Robot> robots = getEnemyRobots(world.getRobots());
        // DEFENSE STRATEGY: COMPUTE EXPECTED LOSS OF ENERGY FOR EACH POSSIBLE ACTION
        // 1. POSITION IS NOT CHANGED => IF AT LEAST ONE ENEMY ROBOT  IS IN DIRECTION TO ME, THAN EXPECTED LOSS
        for (Robot enemy:robots) {
            Position enemyPos = getPosition(enemy);
            if (computeDirectRobotDistance(enemyPos, myPosition) != -1) {
                // ENEMY CAN SHOOT ME
                gains.rechargeGain = gains.restGain = gains.shootGain = gains.takeGain = gains.rechargeGain = - (int) (robot.getEnergy() / 2);
            }
        }
        
        // 2. POSITION IS CHANGED => IF AT LEAST ONE ENEMY ROBOT IS IN DIRECTION TO ME, THAN EXPECTED LOSS
        for (Robot enemy:robots) {
            Position enemyPos = getPosition(enemy);
            Position newPos = performStep(myPosition);
            if (computeDirectRobotDistance(enemyPos, newPos) != -1) {
                // ENEMY CAN SHOOT ME
                gains.stepGain = - (int) (robot.getEnergy() / 2);
            }
        }
        
        // 3. DIRECTION CHANGED => EVALUTE IF I WILL BE SHOOTED IF I WILL STEP NEXT TIME, TAKE WITH 1/2
        // TURN CLOCKWISE
        Position newPos = performTurn(myPosition, true);
        newPos = performStep(newPos);
        for (Robot enemy:robots) {
            Position enemyPos = getPosition(enemy);
            if (computeDirectRobotDistance(enemyPos, newPos) != -1) {
                // ENEMY CAN SHOOT ME
                gains.turnClockwiseGain = - (int) ((robot.getEnergy() / 2) / 2);
            }
        }
        // TURN COUNTER-CLOCKWISE
        newPos = performTurn(myPosition, false);
        newPos = performStep(newPos);
        for (Robot enemy:robots) {
            Position enemyPos = getPosition(enemy);
            if (computeDirectRobotDistance(enemyPos, newPos) != -1) {
                // ENEMY CAN SHOOT ME
                gains.turnCounterClockwiseGain = - (int) ((robot.getEnergy() / 2) / 2);
            }
        }
        
        
        /*
         * byl jsem strelen, miri na me nekdo nebo na me muze v dalsim kroku mirit?
         * pokud ano - obrana:
         *   stojim k nemu bokem (a muzu udelat krok) -> krok nebo otocit k nemu?
         *   pokud stojim celem - stihnu utect? pokud ne, tak bojovat? porovnat nase energie?
         *   stojim zady - no, to je otazka... :-)
         *
         */
        
        return gains;
    }
    
    private ActionsGain getAttackSubactionsGain() {
        ActionsGain gains = new ActionsGain();
        
        
        // IDEA: COMPUTE ALL ROBOTS, THAT ARE SHOOTABLE WITHIN 3 ROUNDS
        // WEIGHT BY PROBABILITY, THAT SUCH SITUATION OCCURE
        // SHOOTABLE FOR SURE (100%), 50% and 25% (DEPENDS ON ENEMY POSITION AND ACTUAL DIRECTION)
        // WALLS ARE ASSUMED
        // AT FIRST, CHECK DIRECT NEIGHBOUR (== POSITIONS GIVEN BY CROSS WITH CENTER IN CURRENT ROBOT POSITION
        // AND SECOND HLAF-CROSS WITH CENTER AT NEXT POSSIBLE ROBOT POSITION)
        // IF NO SHOOTABLE ROBOTS ARE FOUND, THEN FIND NEAREST ROBOTS AND TRY TO GO TO GO TO THE POSITION
        // THAT ALLOW TO SHOOT THIS ROBOT
        
        Position    newPos;
        Position    enemyPosition;
        List<Robot> shootableRobots;
        
        // ONE ROUND
        // ACTION == SHOOT
        //
        gains.shootGain = 0;
        int shootDist = Integer.MAX_VALUE;
        int enemyDist = 0;
        shootableRobots = getShootableRobots(myPosition, getEnemyRobots(world.getRobots()));
        // FIND CLOSEST ENEMY IN MY DIRECTION
        for (Robot enemy:shootableRobots) {
            enemyPosition = getPosition(enemy);
            enemyDist = computeDirectRobotDistance(myPosition, enemyPosition);
            // CHECK IF ACTUAL ENEMY IS CLOSER THAN PREVIOUS (IF ANY)
            if (shootDist > enemyDist) {
                gains.shootGain = (int) (enemy.getEnergy() / 2);
                shootDist = enemyDist;
            }
        }
        gains.shootGain *= WEIGHT_ATTACK_1STEP;
        
        // TWO ROUNDS
        // ACTION == TURN + SHOOT
        // ENEMY IN SAME BI-DIRECTION AFTER MY 90 DEGREE TURN WILL BE SHOOTABLE FOR SURE
        // ENEMY IN OPPOSITE BI-DIRECTION AFTER MY 90 DEGREE TURN WILL BE SHOOTABLE FOR 50%
        
        // TURN 90 DEGREE CLOCWISE
        gains.turnClockwiseGain = 0;
        newPos = performTurn(myPosition, true);
        shootableRobots = getShootableRobots(newPos, getEnemyRobots(world.getRobots()));
        // SUM ENERGY GAIN FROM ALL SHOOTABLE ROBOTS, TAKE 1/2 FOR OPPOSITE BI-DIRECTION
        // NOTE: IMPRECISION IS PRESENT, ONLY FIRST WILL BE HIT
        for (Robot enemy:shootableRobots) {
            if (isBiDirection(newPos.direction, enemy.getDirection())) {
                gains.turnClockwiseGain += (int) (enemy.getEnergy() / 2);
            } else {
                // 50% CHANGE THAN ENEMY WILL BE AWAY
                gains.turnClockwiseGain += ((int) (enemy.getEnergy() / 2) / 2);
            }
        }
        gains.turnClockwiseGain *= WEIGHT_ATTACK_2STEP;
        
        // TURN 90 DEGREE COUNTER-CLOCWISE
        gains.turnCounterClockwiseGain = 0;
        newPos = performTurn(myPosition, false);
        shootableRobots = getShootableRobots(newPos, getEnemyRobots(world.getRobots()));
        // SUM ENERGY GAIN FROM ALL SHOOTABLE ROBOTS, TAKE 1/2 FOR OPPOSITE BI-DIRECTION
        // NOTE: IMPRECISION IS PRESENT, ONLY FIRST ENEMY WILL BE HIT
        for (Robot enemy:shootableRobots) {
            if (isBiDirection(newPos.direction, enemy.getDirection())) {
                gains.turnCounterClockwiseGain += (int) (enemy.getEnergy() / 2);
            } else {
                // 50% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/2
                gains.turnCounterClockwiseGain += (int) ((enemy.getEnergy() / 2) / 2);
            }
        }
        gains.turnCounterClockwiseGain *= WEIGHT_ATTACK_2STEP;
        
        // THREE ROUNDS
        // ACTION == STEP + TURN + SHOOT
        // ENEMY IN SAME BI-DIRECTION AFTER MY STEP AND 90 DEGREE TURN WILL BE SHOOTABLE FOR 50%
        // ENEMY IN OPPOSITE BI-DIRECTION AFTER MY STEP AND 90 DEGREE TURN WILL BE SHOOTABLE FOR 25%
        gains.stepGain = 0;
        
        // STEP + TURN CLOCWISE
        newPos = performStep(myPosition);
        newPos = performTurn(newPos, true);
        shootableRobots = getShootableRobots(newPos, getEnemyRobots(world.getRobots()));
        // SUM ENERGY GAIN FROM ALL SHOOTABLE ROBOTS, TAKE 1/2 FOR OPPOSITE BI-DIRECTION
        // NOTE: SOME IMPRECISION IS PRESENT, ONLY FIRST WILL BE HIT
        for (Robot enemy:shootableRobots) {
            if (isBiDirection(newPos.direction, enemy.getDirection())) {
                // 50% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/2
                gains.stepGain += ((int) (enemy.getEnergy() / 2) / 2);
            } else {
                // 75% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/4
                gains.stepGain += ((int) (enemy.getEnergy() / 2) / 4);
            }
        }
        
        // STEP + TURN COUNTER-CLOCWISE
        newPos = performStep(myPosition);
        newPos = performTurn(newPos, false);
        shootableRobots = getShootableRobots(newPos, getEnemyRobots(world.getRobots()));
        // SUM ENERGY GAIN FROM ALL SHOOTABLE ROBOTS, TAKE 1/2 FOR OPPOSITE BI-DIRECTION
        // NOTE: SOME IMPRECISION IS PRESENT, ONLY FIRST WILL BE HIT
        for (Robot enemy:shootableRobots) {
            if (isBiDirection(newPos.direction, enemy.getDirection())) {
                // 50% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/2
                gains.stepGain += ((int) (enemy.getEnergy() / 2) / 2);
            } else {
                // 75% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/4
                gains.stepGain += ((int) (enemy.getEnergy() / 2) / 4);
            }
        }
        
        // ACTION == TURN CLOCKWISE + TURN CLOCKWISE + SHOOT
        newPos = performTurn(myPosition, true);
        newPos = performTurn(newPos, true);
        shootableRobots = getShootableRobots(newPos, getEnemyRobots(world.getRobots()));
        // SUM ENERGY GAIN FROM ALL SHOOTABLE ROBOTS, TAKE 1/2 FOR OPPOSITE BI-DIRECTION
        // NOTE: SOME IMPRECISION IS PRESENT, ONLY FIRST WILL BE HIT
        for (Robot enemy:shootableRobots) {
            if (isBiDirection(newPos.direction, enemy.getDirection())) {
                // 50% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/2
                gains.stepGain += ((int) (enemy.getEnergy() / 2) / 2);
            } else {
                // 75% CHANGE THAN ENEMY WILL BE AWAY => EXPECTED GAIN IS 1/4
                gains.stepGain += ((int) (enemy.getEnergy() / 2) / 4);
            }
        }
        
        gains.stepGain *= WEIGHT_ATTACK_3STEP;
        
        
        //
        //  NO CLOSE ENEMY => STEP TO THE QUARTER WITH MOST ENEMIES
        //
        if (gains.shootGain == 0 && gains.turnClockwiseGain == 0 && gains.turnCounterClockwiseGain == 0 && gains.stepGain == 0) {
            double q1Sum = 0;
            double q2Sum = 0;
            double q3Sum = 0;
            double q4Sum = 0;
            int distance = 0;
            List<Robot> robots;
            
            robots = getEnemyRobots(world.getRobots());
            for (Robot enemy:robots) {
                distance = getDistance(myPosition, enemy);
                // COMPUTE EXPECTED GAIN FROM DISTANT ENEMY BASED ON HIS ENERGY AND DISTANCE
                switch (getQuarter(myPosition, enemy)) {
                    case 1: q1Sum += ((double) 1/(distance*distance)) * enemy.getEnergy();  break;
                    case 2: q2Sum += ((double) 1/(distance*distance)) * enemy.getEnergy();  break;
                    case 3: q3Sum += ((double) 1/(distance*distance)) * enemy.getEnergy();  break;
                    case 4: q4Sum += ((double) 1/(distance*distance)) * enemy.getEnergy();  break;
                }
            }
            
            q1Sum *= WEIGHT_ATTACK_DISTANTENEMY;
            q2Sum *= WEIGHT_ATTACK_DISTANTENEMY;
            q3Sum *= WEIGHT_ATTACK_DISTANTENEMY;
            q4Sum *= WEIGHT_ATTACK_DISTANTENEMY;
            
            // SET turnClockwiseGain, turnCounterClockwiseGain AND stepGain BASED ON
            // NUMBER OF OTHER ROBOTS IN ITS ENERGY GIVEN QUARTER
            switch (myPosition.direction) {
                case LEFT: {
                    // Q2 and Q3 ARE ACCESSIBLE BY step ACTION
                    gains.stepGain += q2Sum + q3Sum;
                    // Q2 and Q1 ARE ACCESSIBLE BY turnClockwise ACTION
                    gains.turnClockwiseGain += q1Sum + q2Sum;
                    // Q3 and Q4 ARE ACCESSIBLE BY turnCounterClockwise ACTION
                    gains.turnCounterClockwiseGain += q3Sum + q4Sum;
                    break;
                }
                case RIGHT: {
                    // Q1 and Q4 ARE ACCESSIBLE BY step ACTION
                    gains.stepGain += q1Sum + q4Sum;
                    // Q3 and Q4 ARE ACCESSIBLE BY turnClockwise ACTION
                    gains.turnClockwiseGain += q3Sum + q4Sum;
                    // Q2 and Q1 ARE ACCESSIBLE BY turnCounterClockwise ACTION
                    gains.turnCounterClockwiseGain += q1Sum + q2Sum;
                    break;
                }
                case UP: {
                    // Q1 and Q2 ARE ACCESSIBLE BY step ACTION
                    gains.stepGain += q1Sum + q2Sum;
                    // Q1 and Q4 ARE ACCESSIBLE BY turnClockwise ACTION
                    gains.turnClockwiseGain += q1Sum + q4Sum;
                    // Q2 and Q3 ARE ACCESSIBLE BY turnCounterClockwise ACTION
                    gains.turnCounterClockwiseGain += q2Sum + q3Sum;
                    break;
                }
                case DOWN: {
                    // Q3 and Q4 ARE ACCESSIBLE BY step ACTION
                    gains.stepGain += q3Sum + q4Sum;
                    // Q2 and Q3 ARE ACCESSIBLE BY turnClockwise ACTION
                    gains.turnClockwiseGain += q2Sum + q3Sum;
                    // Q1 and Q4 ARE ACCESSIBLE BY turnCounterClockwise ACTION
                    gains.turnCounterClockwiseGain += q1Sum + q4Sum;
                    break;
                }
            }
            
        }
        
        
        // TODO: switch into some fighting mode
        // e.g. - aim to a battery, wait until someone gets there AND RECHARGES and then remove him :-)
        
        return gains;
    }
    
    private ActionsGain getTreasureSubactionsGain() {
        ActionsGain gains = new ActionsGain();
        
        // get the Treasure if standing above it
        if (world.containsTreasure(myPosition.x, myPosition.y) && world.getTreasureValue(myPosition.x, myPosition.y) > 1) {
            ////System.out.println("Taking treasure");
            //#return BotAction.TAKE;
            gains.takeGain = world.getTreasureValue(myPosition.x, myPosition.y);
            return gains;
        }
        
        // TODO: this condition is not sufficient because if I "turn" after rechanging I'd  recharge again
        // TODO: add test, whether someone is aiming at me
        // btw - the "recharge" action is evaluated as the last one
        if (world.containsBattery(myPosition.x, myPosition.y) && (robot.getEnergy() < world.getMaxRobotEnergy()-1)) {
            //#return BotAction.RECHARGE;
            gains.rechargeGain = world.getMaxRobotEnergy() - robot.getEnergy();
        }
        
        // fill in the map showing in how many steps I can get where - including batteries
        findClosestWays(robot, this.map);
        
        // TODO: filter out the treasures to which some other robot can get earlier
        // find the closest and most advantageous cluster of treasures
        List<Position> treasures = Tools.findFields(world, Tools.FieldType.TREASURE);
        if (! treasures.isEmpty()) {
            // find the closest (nuber of actions ignoring the bateries) and most valuable treasures
            //     treasure_value - # of steps    (TODO: too simple)
            // TODO: ignoring the neighborhood of the treasure (or steps to be done after)
            
            // first count the closest ways for all other robots as well
            for (Map.Entry<Robot,AvailabilityMap> entry : robotMaps.entrySet()) {
                findClosestWays(entry.getKey(), entry.getValue());
            }
            
            // now - filter out all treasures that are reachable more easily by other robot
            // do not filter those, to which some other robot's way is less then 3 steps closer
            Position treasure = null;
            for (Iterator<Position> iterator = treasures.iterator(); iterator.hasNext(); ) {
                treasure = iterator.next();
                for (AvailabilityMap robMap : robotMaps.values()) {
                    if ( Math.floor(map.getDistance(treasure) * 0.8)  >= robMap.getDistance(treasure)) {
                        iterator.remove();
                        break;
                    }
                }
            }
            // no treasure is reachable (or closer then from other bots)
            if (treasures.isEmpty())
                return gains;
            
            //iterate over the list of treasures and count the values for given directions
            BotAction action;
            List<Position> bats;
            for (Position trea:treasures) {
                // if the shortest path to this treasure doesn't lead over some batery(ies)
                bats = map.getBatPositions(trea);
                if (bats == null) {
                    action = map.getChainOfActions(trea).popNextAction();
                } else {
                    action = map.getChainOfActions(bats.get(0)).popNextAction();
                }
                switch (action) {
                    case STEP:
                        gains.stepGain += Math.max(1, world.getTreasureValue(trea.x, trea.y) - (map.getDistance(trea) + 1)/2);
                        break;
                    case RIGHT:
                        gains.turnClockwiseGain += Math.max(1, world.getTreasureValue(trea.x, trea.y) - (map.getDistance(trea)+ 1)/2);
                        break;
                    case LEFT:
                        gains.turnCounterClockwiseGain += Math.max(1, world.getTreasureValue(trea.x, trea.y) - (map.getDistance(trea)+ 1)/2);
                        break;
                }
            }
        }
        
        return gains;
    }
    
    /** this is a method that finds a pre-computed map for given battery and direction
     * I thought, that the "ArrayMap implements Map" will do this but it doesn't because
     * it does not use "Object.equals()" but uses hashcodes of the objects (keys) to compare them...
     */
    private AvailabilityMap findBattery(Position pos) throws IllegalArgumentException {
        for (Map.Entry<Position,AvailabilityMap> entry: batteryMaps.entrySet()) {
            if (entry.getKey().equals(pos))
                return entry.getValue();
        }
        throw new IllegalArgumentException("this battery is not stored in the pre-computed map of batteries...");
    }
    
    /** this method first simply uses the dijkstra algorithm from the
     * AvailabilityMap and then applies the possibility of going over batteries */
    private AvailabilityMap findClosestWays(Robot currentRobot, AvailabilityMap robotMap) throws IllegalArgumentException {
        robotMap.fillAvailabilityMap(currentRobot);//, world.getRoundsCount()-world.getRound());
        /*if (this.robot.equals(currentRobot)) {
            //System.out.println("Map of robot "+robot.getPositionX()+","+robot.getPositionY()+ "(no batteries)");
            //System.out.println(robotMap.toString()+"\n");
            //System.out.flush();
        }*/
        Set<Position> consideredBatteries = new HashSet<Position>(batteries.size());
        // the batteries have to be iterated in the order of distance from the position
        // the distance can change during the modifications
        Position battery = null;
        do {
            int minDistance = Integer.MAX_VALUE;
            battery = null;
            for (Position bat : batteries) {
                if (consideredBatteries.contains(bat))
                    continue;
                if (robotMap.getDistance(bat) <= minDistance) {
                    minDistance = robotMap.getDistance(bat);
                    battery = bat;
                }
            }
            if (battery != null) {
                consideredBatteries.add(battery);
                //TODO
                Direction batDirection = robotMap.getDirection(battery);
                ////System.out.println("searching for: "+new Position(battery.x,battery.y,batDirection));
                AvailabilityMap batMap = findBattery(new Position(battery.x,battery.y,batDirection));
                List<Position> batPositions = robotMap.getBatPositions(battery.x, battery.y);
                if (batPositions == null || batPositions.isEmpty())
                    batPositions = new ArrayList<Position>(1);
                batPositions.add(battery);
                if (! battery.equals(new Position(currentRobot.getPositionX(), currentRobot.getPositionY())))
                    robotMap.findMin(batMap.map,robotMap.getDistance(battery),batPositions);
            }
        } while (battery != null);
        /*if (this.robot.equals(currentRobot)) {
            //System.out.println("Map of robot "+robot.getPositionX()+","+robot.getPositionY()+". With batteries: ");
            //System.out.println(robotMap.toString()+"\n");
            //System.out.flush();
        }*/
        return robotMap;
    }
    
    private List<Robot> getEnemyRobots(List<Robot> robots) {
        List<Robot> enemyRobots = new ArrayList<Robot>();
        Position    enemyPos;

        for (Robot enemy:robots) {
            if (enemy.getId() != robot.getId()) {
                enemyRobots.add(enemy);
            }
        }
        
        return enemyRobots;
    }    
    private List<Robot> getShootableRobots(Position shooterPos, List<Robot> robots) {
        // OBTAIN ROBOTS, DECIDE WHEATHER IS ANYONE IN SHOOTABLE POSITION
        int     shootGain = 0;
        int     shootDist = Integer.MAX_VALUE;
        int     enemyDist = 0;
        Position    enemyPos;
        List<Robot> shootableRobots = new ArrayList<Robot>();
        for (Robot enemy:robots) {
            enemyPos = getPosition(enemy);
            
            enemyDist = computeDirectRobotDistance(shooterPos, enemyPos);
            if (enemyDist == -1) {
                // NOT SHOOTABLE
            } else {
                // SHOOTABLE, INSERT INTO SHOOTABLE LIST
                shootableRobots.add(enemy);
            }
        }
        return shootableRobots;
    }
    
    
    // RETURN NON-ZERO VALUE, IF ROBOT1 CAN DIRECTLY SHOOT ROBOT2, -1 OTHERWISE
    // WALLS ARE ASSUMED
    private int computeDirectRobotDistance(Position robot1, Position robot2) {
        int     dist = -1;
        // TODO: ASSUME WALLS
        if (isWallDirectBetween(robot1, robot2)) {
            // WALL BETWEEN ROBOTS, NO DIRECT DISTANCE POSSIBLE
        }
        else {
            // NO WALLS BETWEEN
            switch(robot1.direction) {
                case LEFT: {
                    if ((robot2.y == robot1.y) && (robot2.x < robot1.x)) {
                        dist = robot1.x - robot2.x;
                    }
                    break;
                }
                case UP: {
                    if ((robot2.x == robot1.x) && (robot2.y < robot1.y)) {
                        dist = robot1.x - robot2.y;
                    }
                    break;
                }
                case RIGHT: {
                    if ((robot2.y == robot1.y) && (robot2.x > robot1.x)) {
                        dist = robot2.x - robot1.x;
                    }
                    break;
                }
                case DOWN: {
                    if ((robot2.x == robot1.x) && (robot2.y > robot1.y)) {
                        dist = robot2.y - robot1.y;
                    }
                    break;
                }
            }
        }
        return dist;
    }
    
    private boolean containsWallSafe(int x, int y) {
        boolean containWall = true;
        if (x >= 0 && x < world.getDimensionX()) {
            if (y >= 0 && y < world.getDimensionY()) {
                containWall = world.containsWall(x, y);        
            }
        }
        return containWall;
    }
    private boolean isWallDirectBetween(Position robot1, Position robot2) {
        boolean     wallBetween = false;
        int         i;
        
        switch(robot1.direction) {
            case LEFT: {
                if ((robot2.y == robot1.y) && (robot2.x < robot1.x)) {
                    for (i = robot1.x - 1; i > robot2.x; i--) {
                        if (containsWallSafe(i, robot1.y)) wallBetween = true;
                    }
                }
                break;
            }
            case UP: {
                if ((robot2.x == robot1.x) && (robot2.y < robot1.y)) {
                    for (i = robot1.y - 1; i > robot2.y; i--) {
                        if (containsWallSafe(robot1.x, i)) wallBetween = true;
                    }
                }
                break;
            }
            case RIGHT: {
                if ((robot2.y == robot1.y) && (robot2.x > robot1.x)) {
                    for (i = robot1.x + 1; i < robot2.x; i++) {
                        if (containsWallSafe(i, robot1.y)) wallBetween = true;
                    }
                }
                break;
            }
            case DOWN: {
                if ((robot2.x == robot1.x) && (robot2.y > robot1.y)) {
                    for (i = robot1.y + 1; i < robot2.y; i++) {
                        if (containsWallSafe(robot1.x, i)) wallBetween = true;
                    }
                }
                break;
            }
        }

        return wallBetween;
    }
    
    private Position getPosition(Robot rob) {
        Position    pos = new Position(rob.getPositionX(), rob.getPositionY(), rob.getDirection());
        
        return pos;
    }
    
    private boolean isBiDirection(Direction dir1, Direction dir2) {
        boolean bidirect = false;
        
        switch (dir1) {
            case LEFT: if (dir2 == Direction.LEFT || dir2 == Direction.RIGHT) bidirect = true; break;
            case UP: if (dir2 == Direction.UP || dir2 == Direction.DOWN) bidirect = true; break;
            case RIGHT: if (dir2 == Direction.LEFT || dir2 == Direction.RIGHT) bidirect = true; break;
            case DOWN: if (dir2 == Direction.UP || dir2 == Direction.DOWN) bidirect = true; break;
        }
        
        return bidirect;
    }
    
    private Position performStep(Position oldPosition) {
        Position    newPos;
        
        newPos = (Position) oldPosition.clone();
        switch (oldPosition.direction) {
            case LEFT: {
                newPos.x--;
                break;
            }
            case UP: {
                newPos.y--;
                break;
            }
            case RIGHT: {
                newPos.x++;
                break;
            }
            case DOWN: {
                newPos.y++;
                break;
            }
        }
        
        return newPos;
    }
    
    private Position performTurn(Position oldPosition, boolean clockwise) {
        Position    newPos;
        
        newPos = oldPosition;
        switch (oldPosition.direction) {
            case LEFT: newPos.direction = clockwise ? Direction.UP : Direction.DOWN;  break;
            case UP: newPos.direction = clockwise ? Direction.RIGHT : Direction.LEFT; break;
            case RIGHT: newPos.direction = clockwise ? Direction.DOWN : Direction.UP; break;
            case DOWN: newPos.direction = clockwise ? Direction.LEFT : Direction.RIGHT; break;
        }
        
        return newPos;
    }
    
    private int getDistance(Position myPos, Robot enemy) {
        Position    enemyPos = getPosition(enemy);
        int         dist = 0;
        // MANHATTAN DISTANCE, WALLS ARE NOT ASSUMED
        // TODO: REPLACE WITH REAL DISTANCE BASED ON PATH LENGTH
        
        dist = Math.abs(myPos.x - enemyPos.x) + Math.abs(myPos.y - enemyPos.y);
        return dist;
    }
    
    private int getQuarter(Position myPos, Robot enemy) {
        // 2 | 1
        // -----
        // 3 | 4
        
        int         quarter = 1;
        Position    enemyPos = getPosition(enemy);
        
        if (myPos.x <= enemyPos.x) {
            if (myPos.y > enemyPos.y) {
                quarter = 1;
            } else {
                quarter = 4;
            }
        } else {
            if (myPos.y >= enemyPos.y) {
                quarter = 2;
            } else {
                quarter = 3;
            }
        }
        
        return quarter;
    }
    
}

