package cz.muni.fi.fibot.r2d2;

import java.util.Random;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import java.util.Queue;
import java.util.List;
import java.util.LinkedList;
import java.util.concurrent.LinkedBlockingQueue;
import java.io.IOException;

/**
 *
 * @author Jakub a Jirka
 */
public class R2D2 implements Strategy {
    
    // zakladni promenne hry
    private World world;
    private Robot robot;
    private int roundno = 0;
    private int remains;
    private int width, height;
    private int gamemap[][];
    
    // parametry strategie
    private static final int GOLD_DEVIATION = 2;
    private static final int UNSURE_OBJECTIVE = 2;
    private static final int EVAL_OBJECTIVE_MODIFIER = 2;
    private static final int EVAL_OBJECTIVE_BASE = 20;
    private int EVAL_OBJECTIVE = EVAL_OBJECTIVE_BASE;
    private static final int EVAL_ENERGY = 100;
    private static final int EVAL_POINTS = 100;
    private static final int MAX_DIST = 1000000;
    private static final int MIN_ALPHABETAEVAL = -16777216;
    private static final int MAX_ALPHABETAEVAL = 16777216;
    private static final int AVERAGE_TOTAL_PROFIT = 100;
    
    // konstanty pro objekty na mape
    private static final int idwall = 1;
    private static final int idbattery = 2;
    private static final int idrobot = 4;
    private static final int idtreasure = 8;
    
    // konstanty pro smer
    private static final int dirleft = 0;
    private static final int dirup = 1;
    private static final int dirright = 2;
    private static final int dirdown = 3;
    
    // konstanty pro akce
    private static final int moveleft = 0;
    private static final int moveright = 1;
    private static final int movestep = 2;
    private static final int moveshoot = 3;
    private static final int movetake = 4;
    private static final int moverest = 5;
    private static final int moverecharge = 6;
    
    // maximalni hloubka alfa-bety
    private static final int startevaluate2 = 0; // kolik kol do konce pouzit evaluate2 fci
    private static final int maxalphabetarange = 16;
    private static final int maxalphabetadepth = 4*maxalphabetarange; // pro maximalne 4 roboty
    
    private Random rndgen;
    
    // maximalni hloubka alphabety pri "umrtveni" nekterych robotu
    private int deadmaxalphabetarange;
    
    // kontrolni vypisy
    private static final boolean verbose = false;
    
    // kontrola timeoutu
    private static final int timeoutpollinginterval = 10000;
    private int timeoutpollingcnt;
    private boolean timeoutemergencyexit;
    
    // souradnice na mape
    private class Point {
        int i, j;
    }
    
    // prvek fronty pro hledani nejkratsich cest
    private class QElem {
        int i, j, dir, energy;
    }
    
    // prvek pole poslednich pozic vsech robotu
    private class RobotLastState {
        int i, j, dir, energy, points;
    }
    
    // prvek pole nejkratsich cest
    private class Vertex {
        int dist;
        // pouziva se pouze pri BackwardSearchu a napocitava se pred dist, resp.
        // nejprve se napocitaji cesty bez recharge, vzdalenosti se zkopiruji do
        // distnorecharge a potom se napocitaji cesty s recharge.
        int distnorecharge;
        boolean visited;
        //int pi, pj, pdir, penergy;
        //Strategy.BotAction paction;
    }
    
    /* Pole pro nejkratsi cesty */
    private int sp[][][];
    //private int spnorecharge[][][];
    private int spdir[][][][];
    
    /* Roboti mrtvi pro AlphaBetu (nemuzou me ovlivnit do urciteho poctu tahu)*/
    private boolean deadrobots[];
    
    /* Promenne, pro vypocet rozlozeni zlata */
    private int goldmap[][];
    private boolean visited[][];
    
    private class EvaluatedPoint {
        int i, j, dir, value;
        
        public EvaluatedPoint(int _i, int _j, int _dir, int _value) {
            i = _i;
            j = _j;
            dir = _dir;
            value = _value;
        }
    }
    
    /* Promenna pro strategicky cil */
    private EvaluatedPoint objective;
    
    // posledni stav sveta
    private RobotLastState RobotLastStates[];
    private int lastgamemapstate[][];
    
    
    private int alldiststowalls[][][];
    private int maxenergy;
    private int restenergy;
    private int treasurevaluesum;
    private int ntreasures;
    private int nbatteries;
    private Vertex bfs[][][][];
    private int visitedmaxenergy[][][];
    private int di[];
    private int dj[];
    private Point treasurelist[];
    private List<Point> ctl; // closest treasure list
    private List<Point> rtl; // reachable treasure list
    private Point batterylist[];
    private List<Point> cbl; // closest battery list
    private List<Point> rbl; // reachable treasure list
    private int renergy[];
    private int rpoints[];
    private int rdir[];
    private int rposi[], rposj[];
    private int newrposi[];
    private int newrposj[];
    private int newrenergy[];
    private int newrpoints[];
    private int newrdir[];
    private int lastrposi[][];
    private int lastrposj[][];
    private int lastrenergy[][];
    private int lastrpoints[][];
    private int lastrdir[][];
    private int lastnewrenergy[][];
    private int lastnewrposi[][];
    private int lastnewrposj[][];
    
    private int nrobots;
    private int r2d2id;
    private int possiblemoves[][];
    private int rootr2d2possiblemoves[];
    private int nrootr2d2possiblemoves;
    private int npossiblemoves[];
    private int canhit[];
    private int lastpossiblemoves[][][];
    private int lastnpossiblemoves[][];
    private int lastcanhit[][];
    private boolean alreadyhit[];
    private boolean lastalreadyhit[][];
    private int allmoves[];
    private int lastallmoves[][];
    
    private boolean lastr2d2rootmovesvalid;
    private int r2d2rootmovevalues[];
    
    private int finaldecisionint;
    private int killeralphacutoffs;
    private int killerbetacutoffs;
    private int alphacutoffs;
    private int betacutoffs;
    private int betanocutoffs;
    private int betanocutoffsspecial;
    private int generatemovescalled;
    private int alphabetacalled;
    private int evaluatecalled;
    private int alphanodes;
    private int betanodes;
    
    private int killermoves[][];
    private boolean killermovesvalid[];
    private long starttime;
    private boolean guardianAngel;
    
    private long maxdecisiontime = 0;
    private int minreachedalphabetarange = maxalphabetarange;
    private int maxreachedalphabetarange = 0;
    
    // slouzi k zjistovani, kteri roboti se daji povazovat za mrtve, na zacatku
    // nevime, jak se alphabeta bude ustihavat, tak to nastavime takto:
    // 4 roboti: 4, 3 roboti: 8, 2 roboti: 12, 1 robot: 12
    private int lastreachedalphabetarange[];
    
    List<Robot> robots;
    
    private boolean isInMap(int i, int j) {
        return (i >= 0 && i < width && j >= 0 && j < height);
    }
    
    private int imin(int a, int b) {
        return (a < b ? a : b);
    }
    private int imax(int a, int b) {
        return (a > b ? a : b);
    }
    
    private int iabs(int a) {
        return (a >= 0 ? a : -a);
    }
    
    private int isign(int a) {
        return (a > 0 ? 1 : (a < 0 ? - 1 : 0));
    }
    
    /* Creates a new instance of R2D2 */
    public R2D2() {
        // directions left, up, right, down
        di = new int[4];
        dj = new int[4];
        di[0] = -1;
        di[1] = 0;
        di[2] = 1;
        di[3] = 0;
        dj[0] = 0;
        dj[1] = -1;
        dj[2] = 0;
        dj[3] = 1;
    }
    
    public void setWorld(World world) {
        if (verbose) {
            System.out.println("set world called, roundno = " + roundno);
            System.out.println("world.getRobots.size() = " + world.getRobots().size());
        }
        this.world = world;
    }
    
    public void initialize() {
        
    }
    
    private void analyzeWorldandRobots() {
        if (verbose) {
            System.out.println("nrobots = " + nrobots);
            System.out.println("r2d2id = " + r2d2id);
        }
        
        if (roundno == 0) {
            maxdecisiontime = 0;
            maxreachedalphabetarange = 0;
            minreachedalphabetarange = maxalphabetarange;
            
            guardianAngel = false;
            width = world.getDimensionX();
            height = world.getDimensionY();
            maxenergy = world.getMaxRobotEnergy();
            restenergy = world.getRestRobotEnergy();
            remains = world.getRoundsCount()+1;
            gamemap = new int[width][height];
            lastgamemapstate = new int[width][height];
            bfs = new Vertex[width][height][4][maxenergy + 1];
            for (int i = 0; i < width; i++)
                for (int j = 0; j < height; j++)
                    for (int k = 0; k < 4; k++)
                        for (int l = 0; l <= maxenergy; l++)
                            bfs[i][j][k][l] = new Vertex();
            alldiststowalls = new int[width][height][4]; // left up right down
            visitedmaxenergy = new int[width][height][4]; // left up right down
            
            nrobots = world.getRobots().size();
            robots = world.getRobots();
            possiblemoves = new int[nrobots][7];
            rootr2d2possiblemoves = new int[7];
            npossiblemoves = new int[nrobots];
            canhit = new int[nrobots];
            lastpossiblemoves = new int[maxalphabetadepth + 1][nrobots][7];
            lastnpossiblemoves = new int[maxalphabetadepth + 1][nrobots];
            lastcanhit = new int[maxalphabetadepth + 1][nrobots];
            alreadyhit = new boolean[nrobots];
            lastalreadyhit = new boolean[maxalphabetadepth + 1][nrobots];
            allmoves = new int[nrobots];
            lastallmoves = new int[maxalphabetadepth + 1][nrobots];
            
            newrposi = new int[nrobots];
            newrposj = new int[nrobots];
            newrenergy = new int[nrobots];
            newrpoints = new int[nrobots];
            newrdir = new int[nrobots];
            lastrposi = new int[maxalphabetadepth + 1][nrobots];
            lastrposj = new int[maxalphabetadepth + 1][nrobots];
            lastnewrposi = new int[maxalphabetadepth + 1][nrobots];
            lastnewrposj = new int[maxalphabetadepth + 1][nrobots];
            lastrenergy = new int[maxalphabetadepth + 1][nrobots];
            lastnewrenergy = new int[maxalphabetadepth + 1][nrobots];
            lastrpoints = new int[maxalphabetadepth + 1][nrobots];
            lastrdir = new int[maxalphabetadepth + 1][nrobots];
            rposi = new int[nrobots];
            rposj = new int[nrobots];
            renergy = new int[nrobots];
            rpoints = new int[nrobots];
            rdir = new int[nrobots];
            
            killermoves = new int[maxalphabetadepth + 1][nrobots];
            killermovesvalid = new boolean[maxalphabetadepth + 1];
            
            RobotLastStates = new RobotLastState[nrobots];
            for (int i = 0; i < nrobots; i++)
                RobotLastStates[i] = new RobotLastState();
            
            r2d2rootmovevalues = new int[7];
            sp = new int[nrobots][width][height];
            //spnorecharge = new int[nrobots][width][height];
            spdir = new int[nrobots][width][height][4];
            
            deadrobots = new boolean[nrobots];
            lastreachedalphabetarange = new int[nrobots];
            for (int i = 0; i < nrobots - 1; i++)
                lastreachedalphabetarange[i] = maxalphabetarange;
            if (nrobots <= 4) {
                for (int i = 0; i <= nrobots - 1; i++) {
                    lastreachedalphabetarange[i] = 4*(4 - nrobots + 1 + i);
                    if (verbose) {
                        System.out.println("lastreachedalphabetarange[" + i + "] = " + lastreachedalphabetarange[i]);
                    }
                }
            }
            
            rndgen = new Random();
        }
        
        roundno++;
        remains--;
        
        
        treasurevaluesum = 0;
        ntreasures = 0;
        nbatteries = 0;
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                gamemap[i][j] = 0;
                if (world.containsWall(i, j))
                    gamemap[i][j] |= idwall;
                if (world.containsBattery(i, j)) {
                    gamemap[i][j] |= idbattery;
                    nbatteries++;
                }
                if (world.containsRobot(i, j))
                    gamemap[i][j] |= idrobot;
                if (world.containsTreasure(i, j)) {
                    gamemap[i][j] |= idtreasure*(2*world.getTreasureValue(i, j) + 1);
                    treasurevaluesum += world.getTreasureValue(i, j);
                    ntreasures++;
                }
                lastgamemapstate[i][j] = gamemap[i][j];
            }
        }
        treasurelist = new Point[ntreasures];
        batterylist = new Point[nbatteries];
        int k = 0, l = 0;
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
                if (verbose) {
                    System.out.print(gamemap[i][j] + " ");
                }
                if ((gamemap[i][j] & idtreasure) > 0) {
                    treasurelist[k] = new Point();
                    treasurelist[k].i = i;
                    treasurelist[k].j = j;
                    k++;
                }
                if ((gamemap[i][j] & idbattery) > 0) {
                    batterylist[l] = new Point();
                    batterylist[l].i = i;
                    batterylist[l].j = j;
                    l++;
                }
            }
            if (verbose) {
                System.out.println("");
            }
        }
        if (verbose) {
            System.out.println("ntreasures = " + ntreasures + ", k = " + k);
            System.out.println("nbatteries = " + nbatteries + ", l = " + l);
        }
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                if ((gamemap[i][j] & idwall) > 0) {
                    alldiststowalls[i][j][dirleft] = 0;
                    alldiststowalls[i][j][dirup] = 0;
                } else {
                    if (j == 0) {
                        alldiststowalls[i][j][dirup] = 1;
                    } else {
                        alldiststowalls[i][j][dirup] =
                                alldiststowalls[i][j - 1][dirup] + 1;
                    }
                    if (i == 0) {
                        alldiststowalls[i][j][dirleft] = 1;
                    } else {
                        alldiststowalls[i][j][dirleft] =
                                alldiststowalls[i - 1][j][dirleft] + 1;
                    }
                }
                if ((gamemap[width - i - 1][height - j - 1] & idwall) > 0) {
                    alldiststowalls[width - i - 1][height - j - 1][dirdown] = 0;
                    alldiststowalls[width - i - 1][height - j - 1][dirright] = 0;
                } else {
                    if (j == 0) {
                        alldiststowalls[width - i - 1][height - j - 1][dirdown] = 1;
                    } else {
                        alldiststowalls[width - i - 1][height - j - 1][dirdown] =
                                alldiststowalls[width - i - 1][height - j][dirdown] + 1;
                    }
                    if (i == 0) {
                        alldiststowalls[width - i - 1][height - j - 1][dirright] = 1;
                    } else {
                        alldiststowalls[width - i - 1][height - j - 1][dirright] =
                                alldiststowalls[width - i][height - j - 1][dirright] + 1;
                    }
                }
            }
        }
        
        if (verbose) {
            System.out.println("alldisttowalls dirLEFT:");
            for (int j = 0; j < height; j++) {
                for (int i = 0; i < width; i++) {
                    System.out.print(alldiststowalls[i][j][dirleft] + " ");
                }
                System.out.println("");
            }
            System.out.println("alldisttowalls dirUP:");
            for (int j = 0; j < height; j++) {
                for (int i = 0; i < width; i++) {
                    System.out.print(alldiststowalls[i][j][dirup] + " ");
                }
                System.out.println("");
            }
            System.out.println("alldisttowalls dirRIGHT:");
            for (int j = 0; j < height; j++) {
                for (int i = 0; i < width; i++) {
                    System.out.print(alldiststowalls[i][j][dirright] + " ");
                }
                System.out.println("");
            }
            System.out.println("alldisttowalls dirDOWN:");
            for (int j = 0; j < height; j++) {
                for (int i = 0; i < width; i++) {
                    System.out.print(alldiststowalls[i][j][dirdown] + " ");
                }
                System.out.println("");
            }
        }
        
        for (int i = 0; i < nrobots; i++) {
            Robot r = robots.get(i);
            RobotLastStates[i].i = rposi[i] = r.getPositionX();
            RobotLastStates[i].j = rposj[i] = r.getPositionY();
            RobotLastStates[i].energy = renergy[i] = r.getEnergy();
            RobotLastStates[i].points = rpoints[i] = r.getPoints();
            switch (r.getDirection()) {
                case LEFT: rdir[i] = dirleft; break;
                case UP: rdir[i] = dirup; break;
                case RIGHT: rdir[i] = dirright; break;
                case DOWN: rdir[i] = dirdown; break;
            }
            RobotLastStates[i].dir = rdir[i];
        }
    }
    
    public void setRobot(Robot robot) {
        if (verbose) {
            System.out.println("set robot called, roundno = " + roundno);
        }
        
        this.robot = robot;
        
        r2d2id = robot.getId();
    }
    
    private String MoveIntToString(int moveint) {
        switch (moveint) {
            case movestep:
                return "Krok";
            case moveleft:
                return "Levo";
            case moveright:
                return "Pravo";
            case moveshoot:
                return "Strela";
            case movetake:
                return "Vzeti";
            case moverest:
                return "Odpocinek";
            case moverecharge:
                return "Dobiti";
        }
        return "Zadny pohyb";
    }
    
    private int shootingProfit(int e1, int e2, int lremains) {
        int profit = 0;
        int olde1, olde2;
        
        if (e1 == 0 || e2 == 0)
            return 0;
        
        while (lremains > 0) {
            olde1 = e1;
            olde2 = e2;
            e1--;
            e2--;
            if (olde1 > 1) {
                e2 -= (olde2 - 1)/2;
                profit += (olde2 - 1)/2;
            }
            if (olde2 > 1) {
                e1 -= (olde1 - 1)/2;
            }
            if (olde1 == 1) {
                e1 = restenergy;
            }
            if (olde2 == 1) {
                e2 = restenergy;
            }
            lremains--;
        }
        
        return profit;
    }
    
    private int Evaluate(int rndsremain, int energyloss, int objreached, boolean tstbool) {
        // rndsremain - zbyvajici kola do konce hry (muze byt 0)
        // energyloss - energie ztracena zasahy (body ostatnich robotu, zatim nezohlednuje dvojite zasahy)
        
        int value;
        evaluatecalled++;
        if (objective != null) {
            if (objreached < 0) {
                Vertex v = bfs[newrposi[r2d2id]][newrposj[r2d2id]][newrdir[r2d2id]][newrenergy[r2d2id]];
                if (v.visited) {
                    value = EVAL_OBJECTIVE*(- v.dist*objective.value)
                    + (EVAL_ENERGY*(- energyloss))/2
                            + EVAL_POINTS*(newrpoints[r2d2id]);
                } else {
                    value = 0;
                }
            } else {
                value = //EVAL_OBJECTIVE*objective.value // objective uz bylo dosazeno, dist = 0
                        (EVAL_ENERGY*(- energyloss))/2
                        + EVAL_POINTS*(newrpoints[r2d2id]) +
                        + EVAL_POINTS*objreached;
            }
        } else {
            // uprednostnuje chozeni, aby se zbytecne nedobijelo
            value =
                    EVAL_ENERGY*((restenergy - 1)*imin(rndsremain - ((maxenergy - newrenergy[r2d2id] - 1)/(restenergy - 1) + 1), 0))
                    + (EVAL_ENERGY*(- energyloss))/2
                    + EVAL_POINTS*(newrpoints[r2d2id]);
            //+ EVAL_POINTS*(iabs(RobotLastStates[r2d2id].i - newrposi[r2d2id]) + iabs(RobotLastStates[r2d2id].j - newrposj[r2d2id]))/2;
        }
        
        if (verbose) {
            System.out.println("Value for (" + newrposi[r2d2id] + ", " + newrposj[r2d2id] + ", " +
                    newrdir[r2d2id] + ", " + newrenergy[r2d2id] + ") = " + value);
        }
        
        return value;
    }
    
    private int Evaluate2(int rndsremain, int energyloss, boolean tstbool) {
        // rndsremain - zbyvajici kola do konce hry (muze byt 0)
        // energyloss - energie ztracena zasahy (body ostatnich robotu, zatim nezohlednuje dvojite zasahy)
        
        int otherid = (r2d2id + 1) % nrobots;
        int value;
        evaluatecalled++;
        
        if (rndsremain == 0) {
            value = newrpoints[r2d2id] + newrenergy[r2d2id] - newrpoints[otherid] - newrenergy[otherid];
        } else {
            value = newrpoints[r2d2id] - newrpoints[otherid];
        }
        
        return value;
    }
    
    private int Evaluate3() {
        evaluatecalled++;
        
        int value = newrpoints[r2d2id] + newrenergy[r2d2id];
        
        return value;
    }
    
    private void GetPossibleMoves(int depth, boolean root, int rndsremain) {
        generatemovescalled++;
        
        // bude asi nutne zmenit poradi prochazeni moznosti, aby silnejsi tahy byly driv
        int newi, newj;
        int minenergy = 1; // energie musi byt vetsi nez minenergy, aby se vygenerovala jina akce nez rest/rchrg
        
        if (rndsremain == 1)
            minenergy  = 0;
        if ((!lastr2d2rootmovesvalid && verbose) && root) for (int rid = 0; rid < nrobots; rid++) {
            System.out.println("Robot #" + rid + ": " + rposi[rid] + ", " + rposj[rid] +
                    ", " + rdir[rid] + ", " + renergy[rid] + ", " + rpoints[rid]);
        }
        for (int rid = 0; rid < nrobots; rid++) {
            // v poslednim kole je nutne pripustit klesnuti energie na nula
            // a uvazit i dalsi akce krome dobijeni, to tady zatim neni
            npossiblemoves[rid] = 0;
            possiblemoves[rid][0] = -1;
            if (!deadrobots[rid]) {
                // akce v poradi podle priority, to je potreba jeste upravit
                if (renergy[rid] > minenergy) {
                    // vykalkulujeme roboty na dostrel
                    int closestrobotid = -1;
                    int mindist = imax(width, height);
                    for (int r2id = 0; r2id < nrobots; r2id++) if (rid != r2id) {
                        if (rposi[rid] == rposi[r2id] &&
                                isign(dj[rdir[rid]]) == isign(rposj[r2id] - rposj[rid])) {
                            // teoreticky mozny zasah v ose y
                            if (iabs(rposj[r2id] - rposj[rid]) < mindist) {
                                mindist = iabs(rposj[r2id] - rposj[rid]);
                                closestrobotid = r2id;
                            }
                        }
                        if (rposj[rid] == rposj[r2id] &&
                                isign(di[rdir[rid]]) == isign(rposi[r2id] - rposi[rid])) {
                            // teoreticky mozny zasah v ose x
                            if (iabs(rposi[r2id] - rposi[rid]) < mindist) {
                                mindist = iabs(rposi[r2id] - rposi[rid]);
                                closestrobotid = r2id;
                            }
                        }
                        // obe podminky byt splneny nemuzou
                    }
                    
                    if (closestrobotid >= 0 &&
                            alldiststowalls[rposi[rid]][rposj[rid]][rdir[rid]] > mindist) {
                        canhit[rid] = closestrobotid;
                        possiblemoves[rid][npossiblemoves[rid]] = moveshoot;
                        npossiblemoves[rid]++;
                    }
                }
                if (renergy[rid] > minenergy) {
                    if ((gamemap[rposi[rid]][rposj[rid]] & idtreasure) > 0) {
                        possiblemoves[rid][npossiblemoves[rid]] = movetake;
                        npossiblemoves[rid]++;
                    }
                }
                if (renergy[rid] > minenergy && (rid != r2d2id || !guardianAngel)) {
                    newi = rposi[rid] + di[rdir[rid]];
                    newj = rposj[rid] + dj[rdir[rid]];
                    if (isInMap(newi, newj) && (gamemap[newi][newj] & idwall) == 0) {
                        possiblemoves[rid][npossiblemoves[rid]] = movestep;
                        npossiblemoves[rid]++;
                    }
                }
                // aby se neuprednostnovalo otaceni na jednu stranu
                if (!root || rid != r2d2id || rndgen.nextInt(2) == 0) {
                    if (renergy[rid] > minenergy) {
                        possiblemoves[rid][npossiblemoves[rid]] = moveleft;
                        npossiblemoves[rid]++;
                    }
                    if (renergy[rid] > minenergy) {
                        possiblemoves[rid][npossiblemoves[rid]] = moveright;
                        npossiblemoves[rid]++;
                    }
                } else {
                    if (renergy[rid] > minenergy) {
                        possiblemoves[rid][npossiblemoves[rid]] = moveright;
                        npossiblemoves[rid]++;
                    }
                    if (renergy[rid] > minenergy) {
                        possiblemoves[rid][npossiblemoves[rid]] = moveleft;
                        npossiblemoves[rid]++;
                    }
                }
                if (renergy[rid] >= 1 && (rid != r2d2id || !guardianAngel || rndsremain == 1)) {
                    if ((gamemap[rposi[rid]][rposj[rid]] & idbattery) > 0) {
                        possiblemoves[rid][npossiblemoves[rid]] = moverecharge;
                        npossiblemoves[rid]++;
                    }
                }
                if (renergy[rid] >= 1 && (rid != r2d2id || !guardianAngel || renergy[rid] == 1)) {
                    possiblemoves[rid][npossiblemoves[rid]] = moverest;
                    npossiblemoves[rid]++;
                }
            }
            if (npossiblemoves[rid] == 0)
                npossiblemoves[rid] = 1; // je "mozny" pohyb -1 nastaveny na zacatku, tedy zadny pohyb
        }
        
        if ((!lastr2d2rootmovesvalid && verbose) && root) for (int rid = 0; rid < nrobots; rid++) {
            System.out.print("Mozne pohyby robota #" + rid + ": ");
            for (int moveid = 0; moveid < npossiblemoves[rid]; moveid++) {
                System.out.print(MoveIntToString(possiblemoves[rid][moveid]));
                if (moveid == npossiblemoves[rid] - 1)
                    System.out.println("");
                else
                    System.out.print(", ");
            }
        }
    }
    
    // predpoklada ze vsechny pohyby jsou platne (krome -1)!!!
    private int DoMove(int rid, int moveid) {
        int move;
        int alreadyhitsetrid = -1;
        
        move = possiblemoves[rid][moveid];
        allmoves[rid] = move;
        
        if (verbose) {
            System.out.println("DoMove called");
            System.out.println("Robot #" + rid + ": " + rposi[rid] + ", " + rposj[rid] +
                    ", " + rdir[rid] + ", " + renergy[rid] + ", " + rpoints[rid] +
                    ", move = " + move);
        }
        
        if (rid == r2d2id) {
            for (int i = 0; i < nrobots; i++) {
                newrenergy[i] = imax(renergy[i] - 1, 0);
                alreadyhit[i] = false;
            }
        }
        
        newrposi[rid] = rposi[rid];
        newrposj[rid] = rposj[rid];
        newrdir[rid] = rdir[rid];
        // newrenergy musi byt nastavena v predstihu, aby se tady neprepsala
        newrpoints[rid] = rpoints[rid];
        
        switch (move) {
            case moveleft:
                newrdir[rid] = (rdir[rid] + 3) % 4;
                break;
            case moveright:
                newrdir[rid] = (rdir[rid] + 1) % 4;
                break;
            case movestep:
                newrposi[rid] = rposi[rid] + di[rdir[rid]];
                newrposj[rid] = rposj[rid] + dj[rdir[rid]];
                break;
            case moveshoot:
                newrpoints[rid] = rpoints[rid] +
                        (imax(renergy[canhit[rid]] - 1, 0))/2;
                // pokud zasazeny robot provadi platnou recharge, pak na
                // konci kola bude mit plnou energii, takze se tady upravovat nebude
                // stejne tak je potreba osetrit pripad, kdy uz robotovi odecteno bylo
                if (!alreadyhit[canhit[rid]]) {
                    alreadyhitsetrid = canhit[rid];
                    if (verbose && deadrobots[alreadyhitsetrid]) {
                        System.err.println("!!!!!!!!!DEAD ROBOTS SPATNE NAPOCITANO");
                    }
                    newrenergy[alreadyhitsetrid] -= imax(renergy[alreadyhitsetrid] - 1, 0)/2;
                    alreadyhit[alreadyhitsetrid] = true;
                }
                break;
            case movetake:
                newrpoints[rid] = rpoints[rid] +
                        gamemap[rposi[rid]][rposj[rid]] / (2*idtreasure);
                gamemap[rposi[rid]][rposj[rid]] &= ~idtreasure;
                break;
            case moverest:
                // zde se musi nechat energie prerust i pres maxenergy a upravit az potom
                // pokud je totiz robot zasazen mohla by se energie spocitat spatne
                newrenergy[rid] += restenergy;
                break;
            case moverecharge:
                // pokud se robot strilejici na tohoto robota zpracovava pozdeji, odecte mu
                // neco z energie, proto tady nastavime takovou energii, aby i po odecteni
                // zustala vetsi nez maxenergy
                newrenergy[rid] = 100*maxenergy;
                break;
        }
        
        return alreadyhitsetrid;
    }
    
    private int ResolveCollisions(int depth) {
        int restorenew = 0;
        
        // je potreba si ulozit informace o moznych tazich
        if (depth != 1) {
            // uklada se pouze pokud se na dalsi urovni nevola evaluate
            for (int rid = 0; rid < nrobots; rid++) {
                lastalreadyhit[depth][rid] = alreadyhit[rid];
                lastallmoves[depth][rid] = allmoves[rid];
                lastcanhit[depth][rid] = canhit[rid];
                lastnpossiblemoves[depth][rid] = npossiblemoves[rid];
                for (int moveid = 0; moveid < npossiblemoves[rid]; moveid++)
                    lastpossiblemoves[depth][rid][moveid] = possiblemoves[rid][moveid];
            }
        }
        // problem - kolize 3 a vice robotu - vyreseno
        boolean collision = false;
        int collisioni = -1, collisionj = -1;
        do {
            collision = false;
            for (int rid = 0; rid < nrobots && !collision; rid++) {
                for (int r2id = rid + 1; r2id < nrobots && !collision; r2id++) {
                    if (newrposi[rid] == newrposi[r2id] &&
                            newrposj[rid] == newrposj[r2id]) {
                        collision = true;
                        collisioni = newrposi[rid];
                        collisionj = newrposj[rid];
                    }
                }
            }
            
            if (collision) {
                // kolize; robotu, ktery udelal step, se rusi prikaz
                for (int rid = 0; rid < nrobots; rid++) {
                    if (restorenew == 0) {
                        lastnewrposi[depth][rid] = newrposi[rid];
                        lastnewrposj[depth][rid] = newrposj[rid];
                    }
                    if (newrposi[rid] == collisioni &&
                            newrposj[rid] == collisionj) {
                        if (verbose && deadrobots[rid]) {
                            System.err.println("!!!!!!!!!DEAD ROBOTS SPATNE NAPOCITANO");
                        }
                        if (allmoves[rid] == movestep) {
                            newrposi[rid] = rposi[rid];
                            newrposj[rid] = rposj[rid];
                        }
                    }
                }
                restorenew = 1;
            }
        } while (collision);
        // upravime gamemap a rposi, rposj,...
        for (int rid = 0; rid < nrobots; rid++) {
            if (newrenergy[rid] > maxenergy)
                restorenew |= 2;
            // odstranime robota z jeho stare pozice
            gamemap[rposi[rid]][rposj[rid]] &= ~idrobot;
        }
        if ((restorenew & 2) > 0) {
            for (int rid = 0; rid < nrobots; rid++) {
                // odpocivani se mohla prilis zvysit energie
                // my si ale potrebujeme ulozit zvysenou energii, aby jsme ji potom
                // mohli obnovit
                lastnewrenergy[depth][rid] = newrenergy[rid];
                if (newrenergy[rid] > maxenergy)
                    newrenergy[rid] = maxenergy;
            }
        }
        // v posledni urovni se new a old parametry robotu neprohazuji, bylo by to zbytecne
        if (depth != 1) {
            for (int rid = 0; rid < nrobots; rid++) {
                gamemap[newrposi[rid]][newrposj[rid]] |= idrobot;
                lastrposi[depth][rid] = rposi[rid];
                lastrposj[depth][rid] = rposj[rid];
                lastrenergy[depth][rid] = renergy[rid];
                lastrpoints[depth][rid] = rpoints[rid];
                lastrdir[depth][rid] = rdir[rid];
                rposi[rid] = newrposi[rid];
                rposj[rid] = newrposj[rid];
                renergy[rid] = newrenergy[rid];
                rpoints[rid] = newrpoints[rid];
                rdir[rid] = newrdir[rid];
            }
        } else {
            for (int rid = 0; rid < nrobots; rid++)
                gamemap[newrposi[rid]][newrposj[rid]] |= idrobot;
        }
        
        return restorenew;
    }
    
    private void UnResolveCollisions(int depth, int restorenew) {
        if (depth != 1) {
            // obnovuje se pouze pokud na dalsi urovni neni evaluate
            for (int rid = 0; rid < nrobots; rid++) {
                allmoves[rid] = lastallmoves[depth][rid];
                alreadyhit[rid] = lastalreadyhit[depth][rid];
                // obnovime informace o moznych tazich
                canhit[rid] = lastcanhit[depth][rid];
                npossiblemoves[rid] = lastnpossiblemoves[depth][rid];
                for (int moveid = 0; moveid < lastnpossiblemoves[depth][rid]; moveid++)
                    possiblemoves[rid][moveid] = lastpossiblemoves[depth][rid][moveid];
            }
            
            for (int rid = 0; rid < nrobots; rid++) {
                gamemap[rposi[rid]][rposj[rid]] &= ~idrobot;
                if ((restorenew & 1) > 0) {
                    newrposi[rid] = lastnewrposi[depth][rid];  // muze se lisit od rposi, v pripade kolize
                    newrposj[rid] = lastnewrposj[depth][rid];  // muze se lisit od rposj, v pripade kolize
                } else {
                    newrposi[rid] = rposi[rid];
                    newrposj[rid] = rposj[rid];
                }
                if ((restorenew & 2) > 0) {
                    newrenergy[rid] = lastnewrenergy[depth][rid]; // muze se lisit od renergy[rid], pokud ta byla oriznuta na maxenergy
                } else {
                    newrenergy[rid] = renergy[rid];
                }
                newrpoints[rid] = rpoints[rid];
                newrdir[rid] = rdir[rid];
            }
            for (int rid = 0; rid < nrobots; rid++) {
                gamemap[lastrposi[depth][rid]][lastrposj[depth][rid]] |= idrobot;
                rposi[rid] = lastrposi[depth][rid];
                rposj[rid] = lastrposj[depth][rid];
                renergy[rid] = lastrenergy[depth][rid];
                rpoints[rid] = lastrpoints[depth][rid];
                rdir[rid] = lastrdir[depth][rid];
            }
        } else {
            // v posledni urovni se new a old parametry robotu neprohazuji, bylo by to zbytecne
            // je ale potreba obnovit newenergy a newpozici
            for (int rid = 0; rid < nrobots; rid++) {
                if ((restorenew & 1) > 0) {
                    newrposi[rid] = lastnewrposi[depth][rid];
                    newrposj[rid] = lastnewrposj[depth][rid];
                }
                if ((restorenew & 2) > 0) {
                    newrenergy[rid] = lastnewrenergy[depth][rid];
                }
                gamemap[newrposi[rid]][newrposj[rid]] &= ~idrobot;
            }
            for (int rid = 0; rid < nrobots; rid++)
                gamemap[rposi[rid]][rposj[rid]] |= idrobot;
        }
    }
    
    private void UnDoMove(int rid, int moveid, int alreadyhitsetrid) {
        int move = possiblemoves[rid][moveid];
        
        if (move == movetake)
            gamemap[rposi[rid]][rposj[rid]] |= idtreasure;
        
        if (alreadyhitsetrid >= 0) {
            alreadyhit[alreadyhitsetrid] = false;
            newrenergy[alreadyhitsetrid] += imax(renergy[alreadyhitsetrid] - 1, 0)/2;
        }
        
        if (verbose) {
            System.out.println("UnDoMove called");
            System.out.println("Robot #" + rid + ": " + rposi[rid] + ", " + rposj[rid] +
                    ", " + rdir[rid] + ", " + renergy[rid] + ", " + rpoints[rid] +
                    ", move = " + move);
        }
    }
    
    private boolean getTimeout() {
        long currenttime = System.currentTimeMillis();
        return (currenttime - starttime > 4800);
    }
    
    
    // vola se se zvolenou hloubkou a alpha = mineval; beta = maxeval;
    // v objreached je pocet kol, ktere uplynuly od doby, kdy byl dosazen
    // objective (-1 pokud objective dosazeno nebylo, -2 pokud objective vubec
    // neexistuje)
    private int AlphaBeta(int depth, int alpha, int beta, int otherrid, boolean root,
            boolean firstindepth, int rndsremain, int energyloss, int objreached,
            boolean tstbool) {
        boolean newtstbool = tstbool;
        int newobjreached = objreached;
        alphabetacalled++;
        
        if (timeoutemergencyexit)
            return beta;
        
        if (timeoutpollingcnt == 0) {
            timeoutpollingcnt = timeoutpollinginterval;
            if (robot.getTimeout()) {
                timeoutemergencyexit = true;
                return beta;
            }
        } else {
            timeoutpollingcnt--;
        }
        
        int val, killermove, killermoveid;
        int alreadyhitsetrid, oldnewrenergy, energylossplus;
        int restorenew;
        if (depth == 0) {
            if (remains > startevaluate2) {
//                if (rndsremain == 0)
//                    return Evaluate3();
                return Evaluate(rndsremain, energyloss, objreached, newtstbool);
            } else {
                return Evaluate2(rndsremain, energyloss, newtstbool);
            }
        } else {
            if (firstindepth)
                killermovesvalid[depth] = false;
            if (depth % nrobots == 0) { // ja jsem na tahu
                alphanodes++;
                GetPossibleMoves(depth, root, rndsremain);
                if (root && !lastr2d2rootmovesvalid) {
                    // v rootr2d2possiblemoves jsou mozne koreneve akce robota r2d2
                    // pouze pri spousteni alphabety pro prvni hloubku se jejich poradi
                    // rovna poradi napocitanem v GetPossibleMoves
                    nrootr2d2possiblemoves = npossiblemoves[r2d2id];
                    for (int moveid = 0; moveid < npossiblemoves[r2d2id]; moveid++) {
                        rootr2d2possiblemoves[moveid] = possiblemoves[r2d2id][moveid];
                    }
                }
                killermoveid = -1;
                if (killermovesvalid[depth]) {
                    killermove = killermoves[depth][r2d2id];
                    for (int moveid = 0; moveid < npossiblemoves[r2d2id]; moveid++) {
                        if (possiblemoves[r2d2id][moveid] == killermove)
                            killermoveid = moveid;
                    }
                    
                    if (killermoveid >= 0) {
                        alreadyhitsetrid = DoMove(r2d2id, killermoveid);
                        val = -AlphaBeta(depth - 1, -beta, -alpha, 0, false, firstindepth,
                                rndsremain, energyloss, newobjreached, newtstbool);
                        firstindepth = false;
                        UnDoMove(r2d2id, killermoveid, alreadyhitsetrid);
                        if (verbose && root) {
                            System.out.println("Killer " + MoveIntToString(possiblemoves[r2d2id][killermoveid]) + " value = " + val);
                        }
                        if (val >= beta) {
                            killermovesvalid[depth] = true;
                            killermoves[depth][r2d2id] = possiblemoves[r2d2id][killermoveid];
                            killeralphacutoffs++;
                            alphacutoffs++;
                            return beta;
                        }
                        if (val > alpha) {
                            alpha = val;
                            if (root && !timeoutemergencyexit)
                                finaldecisionint = possiblemoves[r2d2id][killermoveid];
                        }
                    }
                }
                if (root && lastr2d2rootmovesvalid) {
                    if (nrootr2d2possiblemoves != npossiblemoves[r2d2id]) {
                        if (verbose) {
                            System.err.println("!!!!!!!!!!!!!!!!!!! ERROR IN ROOTR2D2MOVES !!!!!!!!!!!!!!!!!!!");
                        }
                    }
                    
                    // setridime tahy podle zminula zjistenych hodnot
                    int tmp;
                    for (int i = 1; i < nrootr2d2possiblemoves; i++) {
                        for (int j = 0; j < nrootr2d2possiblemoves - i; j++) {
                            if (r2d2rootmovevalues[rootr2d2possiblemoves[j]] <
                                    r2d2rootmovevalues[rootr2d2possiblemoves[j + 1]]) {
                                tmp = rootr2d2possiblemoves[j];
                                rootr2d2possiblemoves[j] = rootr2d2possiblemoves[j + 1];
                                rootr2d2possiblemoves[j + 1] = tmp;
                            }
                        }
                    }
                    
                    for (int moveid = 0; moveid < npossiblemoves[r2d2id]; moveid++) {
                        possiblemoves[r2d2id][moveid] = rootr2d2possiblemoves[moveid];
                    }
                }
                for (int moveid = 0; moveid < npossiblemoves[r2d2id]; moveid++) if (moveid != killermoveid) {
                    alreadyhitsetrid = DoMove(r2d2id, moveid);
                    val = -AlphaBeta(depth - 1, -beta, -alpha, 0, false, firstindepth,
                            rndsremain, energyloss, newobjreached, newtstbool);
                    firstindepth = false;
                    UnDoMove(r2d2id, moveid, alreadyhitsetrid);
                    if (root) {
                        if (verbose) {
                            System.out.println(MoveIntToString(possiblemoves[r2d2id][moveid]) + " value = " + val);
                        }
                        r2d2rootmovevalues[possiblemoves[r2d2id][moveid]] = val;
                    }
                    if (val >= beta) {
                        killermovesvalid[depth] = true;
                        killermoves[depth][r2d2id] = possiblemoves[r2d2id][moveid];
                        alphacutoffs++;
                        return beta;
                    }
                    if (val > alpha) {
                        alpha = val;
                        if (root && !timeoutemergencyexit)
                            finaldecisionint = possiblemoves[r2d2id][moveid];
                    }
                }
            } else { // nekdo z ostatnich je na tahu jsou na tahu
                betanodes++;
                
                if (otherrid == r2d2id)
                    otherrid++;
                killermoveid = -1;
                if (killermovesvalid[depth]) {
                    killermove = killermoves[depth][otherrid];
                    for (int moveid = 0; moveid < npossiblemoves[otherrid]; moveid++) {
                        if (possiblemoves[otherrid][moveid] == killermove)
                            killermoveid = moveid;
                    }
                    
                    if (killermoveid >= 0) {
                        oldnewrenergy = newrenergy[otherrid];
                        energylossplus = 0;
                        alreadyhitsetrid = DoMove(otherrid, killermoveid);
                        if (alreadyhitsetrid == r2d2id)
                            energylossplus = imax(renergy[alreadyhitsetrid] - 1, 0)/2;
                        if (depth % nrobots == 1) {
                            restorenew = ResolveCollisions(depth);
                            if (objreached == -1) {
                                if (newrposi[r2d2id] == objective.i &&
                                        newrposj[r2d2id] == objective.j &&
                                        (objective.dir < 0 ||
                                        objective.dir == newrdir[r2d2id])) {
                                    newobjreached = 0;
                                }
                            } else {
                                if (objreached >= 0)
                                    newobjreached = objreached + 1;
                            }
                            val = -AlphaBeta(depth - 1, -beta, -alpha, otherrid + 1, false, firstindepth,
                                    rndsremain - 1, energyloss + energylossplus, newobjreached, newtstbool);
                            UnResolveCollisions(depth, restorenew);
                        } else {
                            val = AlphaBeta(depth - 1, alpha, beta, otherrid + 1, false, firstindepth,
                                    rndsremain, energyloss + energylossplus, newobjreached, newtstbool);
                        }
                        firstindepth = false;
                        UnDoMove(otherrid, killermoveid, alreadyhitsetrid);
                        newrenergy[otherrid] = oldnewrenergy;
                        if (val >= beta) {
                            killermovesvalid[depth] = true;
                            killermoves[depth][otherrid] = possiblemoves[otherrid][killermoveid];
                            killerbetacutoffs++;
                            betacutoffs++;
                            return beta;
                        } else {
                            betanocutoffs++;
                            if (beta > 0 && depth == 1)
                                betanocutoffsspecial++;
                        }
                        if (val > alpha) {
                            alpha = val;
                        }
                    }
                }
                for (int moveid = 0; moveid < npossiblemoves[otherrid]; moveid++) if (moveid != killermoveid) {
                    oldnewrenergy = newrenergy[otherrid];
                    energylossplus = 0;
                    alreadyhitsetrid = DoMove(otherrid, moveid);
                    if (alreadyhitsetrid == r2d2id)
                        energylossplus = imax(renergy[alreadyhitsetrid] - 1, 0)/2;
                    if (depth % nrobots == 1) {
                        restorenew = ResolveCollisions(depth);
                        if (objreached == -1) {
                            if (newrposi[r2d2id] == objective.i &&
                                    newrposj[r2d2id] == objective.j &&
                                    (objective.dir < 0 ||
                                    objective.dir == newrdir[r2d2id])) {
                                newobjreached = 0;
                            }
                        } else {
                            if (objreached >= 0)
                                newobjreached = objreached + 1;
                        }
                        val = -AlphaBeta(depth - 1, -beta, -alpha, otherrid + 1, false, firstindepth,
                                rndsremain - 1, energyloss + energylossplus, newobjreached, newtstbool);
                        UnResolveCollisions(depth, restorenew);
                    } else {
                        val = AlphaBeta(depth - 1, alpha, beta, otherrid + 1, false, firstindepth,
                                rndsremain, energyloss + energylossplus, newobjreached, newtstbool);
                    }
                    firstindepth = false;
                    UnDoMove(otherrid, moveid, alreadyhitsetrid);
                    newrenergy[otherrid] = oldnewrenergy;
                    if (val >= beta) {
                        killermovesvalid[depth] = true;
                        killermoves[depth][otherrid] = possiblemoves[otherrid][moveid];
                        betacutoffs++;
                        return beta;
                    } else {
                        if (beta > 0 && depth == 1)
                            betanocutoffsspecial++;
                        betanocutoffs++;
                    }
                    if (val > alpha) {
                        alpha = val;
                    }
                }
            }
        }
        
        return alpha;
    }
    
    private void ForwardSearch(int starti, int startj, int startdir, int startenergy,
            boolean norobot, boolean rechargeallowed) {
        int dir = startdir;
        int energy = startenergy;
        int newi, newj, newdir, newenergy;
        int origi = starti;
        int origj = startj;
        int testint = (norobot ? (idwall | idrobot) : idwall);
        int nedges = 0, nvertices = 0;
        int queuesize, maxqueuesize = 0, maxdist = 0;
        int movescount = 5;
        Strategy.BotAction action;
        Vertex v;
        QElem qe = new QElem();
        Queue<QElem> q = new LinkedList<QElem>();
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                for (int k = 0; k < 4; k++) {
                    visitedmaxenergy[i][j][k] = -1;
                    for (int l = 0; l <= maxenergy; l++)
                        bfs[i][j][k][l].visited = false;
                }
            }
        }
        
        if (!rechargeallowed)
            movescount = 4;
        
        v = bfs[origi][origj][dir][energy];
        visitedmaxenergy[origi][origj][dir] = energy;
        v.dist = 0;
        v.visited = true;
        //v.pi = -1;
        qe.i = origi;
        qe.j = origj;
        qe.dir = dir;
        qe.energy = energy;
        
        q.add(qe);
        while (!q.isEmpty()) {
            if (q.size() > maxqueuesize)
                maxqueuesize = q.size();
            qe = q.remove();
            nvertices++;
            //System.out.println("qe.dist = " + bfs[qe.i][qe.j][qe.dir][qe.energy].dist);
            
            for (int i = 0; i < movescount; i++) {
                newi = qe.i;
                newj = qe.j;
                newdir = qe.dir;
                newenergy = qe.energy - 1;
                if (newenergy >= 0) { // robot neni mrtvy
                    action = Strategy.BotAction.STEP; // hack aby nehazel warningy
                    switch (i) {
                        case 0: // STEP
                            newi = qe.i + di[newdir];
                            newj = qe.j + dj[newdir];
                            action = Strategy.BotAction.STEP;
                            break;
                        case 1: // RIGHT
                            newdir = (qe.dir + 1) % 4;
                            action = Strategy.BotAction.RIGHT;
                            break;
                        case 2: // LEFT
                            newdir = (qe.dir + 3) % 4;
                            action = Strategy.BotAction.LEFT;
                            break;
                        case 3: // REST
                            newenergy = imin(qe.energy + restenergy - 1, maxenergy);
                            action = Strategy.BotAction.REST;
                            break;
                        case 4: // RECHARGE
                            newenergy = maxenergy;
                            action = Strategy.BotAction.RECHARGE;
                            break;
                    }
                    
                    if (i < 3) {
                        nedges++;
                        // budeme predpokladat, ze zadny robot nechce umrit proto newenergy > 0
                        if (newenergy > 0 &&
                                isInMap(newi, newj) &&
                                !bfs[newi][newj][newdir][newenergy].visited &&
                                ((gamemap[newi][newj] & testint) == 0 ||
                                (newi == origi && newj == origj))) {
                            if (newenergy > visitedmaxenergy[newi][newj][newdir]) {
                                visitedmaxenergy[newi][newj][newdir] = newenergy;
                                QElem newqe = new QElem();
                                newqe.dir = newdir;
                                newqe.i = newi;
                                newqe.j = newj;
                                newqe.energy = newenergy;
                                
                                bfs[newi][newj][newdir][newenergy].visited = true;
                                bfs[newi][newj][newdir][newenergy].dist =
                                        bfs[qe.i][qe.j][qe.dir][qe.energy].dist + 1;
                                if (bfs[newi][newj][newdir][newenergy].dist > maxdist)
                                    maxdist = bfs[newi][newj][newdir][newenergy].dist;
                                // tyto udaje nejsou potreba
                                //bfs[newi][newj][newdir][newenergy].pi = qe.i;
                                //bfs[newi][newj][newdir][newenergy].pj = qe.j;
                                //bfs[newi][newj][newdir][newenergy].pdir = qe.dir;
                                //bfs[newi][newj][newdir][newenergy].penergy = qe.energy;
                                //bfs[newi][newj][newdir][newenergy].paction = action;
                                
                                q.add(newqe);
                            }
                        }
                    } else {
                        nedges++;
                        if (!bfs[newi][newj][newdir][newenergy].visited &&
                                (i == 3 || (gamemap[newi][newj] & idbattery) > 0)) {
                            if (newenergy > visitedmaxenergy[newi][newj][newdir]) {
                                visitedmaxenergy[newi][newj][newdir] = newenergy;
                                QElem newqe = new QElem();
                                newqe.dir = newdir;
                                newqe.i = newi;
                                newqe.j = newj;
                                newqe.energy = newenergy;
                                
                                bfs[newi][newj][newdir][newenergy].visited = true;
                                bfs[newi][newj][newdir][newenergy].dist =
                                        bfs[qe.i][qe.j][qe.dir][qe.energy].dist + 1;
                                if (bfs[newi][newj][newdir][newenergy].dist > maxdist)
                                    maxdist = bfs[newi][newj][newdir][newenergy].dist;
                                // tyto udaje nejsou potreba
                                //bfs[newi][newj][newdir][newenergy].pi = qe.i;
                                //bfs[newi][newj][newdir][newenergy].pj = qe.j;
                                //bfs[newi][newj][newdir][newenergy].pdir = qe.dir;
                                //bfs[newi][newj][newdir][newenergy].penergy = qe.energy;
                                //bfs[newi][newj][newdir][newenergy].paction = action;
                                
                                q.add(newqe);
                            }
                        }
                    }
                }
            }
        }
        if (verbose) {
            System.out.println("ForwardSearch: nvertices = " + nvertices + ", nedges = " + nedges +
                    ", maxqueuesize = " + maxqueuesize + ", maxdist = " + maxdist);
        }
    }
    
    private void BackwardSearch(int starti, int startj, int startdir, int startenergy,
            boolean norobot, boolean rechargeallowed) {
        int dir = startdir;
        int energy = startenergy;
        int newi, newj, newdir, newenergy;
        int origi = starti;
        int origj = startj;
        int energyhi, energylo;
        int testint = (norobot ? (idwall | idrobot) : idwall);
        int nedges = 0; int nvertices = 0;
        int queuesize, maxqueuesize = 0, maxdist = 0;
        int movescount = 5;
        
        if (startenergy < 0) {
            // pouzivame optimalizovany backward search, proto v pripade nezadane energie
            // pouzijeme nejmensi moznou, protoze kam se dostanu s ni, tam se dostanu i s
            // vetsi
            startenergy = 1;
            energy = startenergy;
        }
        
        Strategy.BotAction action;
        Vertex v = new Vertex();
        QElem qe;
        Queue<QElem> q = new LinkedList<QElem>();
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
                for (int k = 0; k < 4; k++) {
                    visitedmaxenergy[i][j][k] = maxenergy + 1; // nyni je to minenergy
                    for (int l = 0; l <= maxenergy; l++)
                        bfs[i][j][k][l].visited = false;
                }
            }
        }
        
        if (startdir < 0) {// all directions
            for (dir = 0; dir < 4; dir++) {
                if (startenergy < 0) { // all energies
                    // zatim tam chceme dojit s nenulovou energii, pozdeji
                    // uvazime i variantu energy == 0, obdobnou zmenou projde i AlphaBeta search
                    for (energy = 1; energy <= maxenergy; energy++) {
                        v = bfs[origi][origj][dir][energy];
                        v.dist = 0;
                        v.visited = true;
                        //v.pi = -1;
                        qe = new QElem();
                        qe.i = origi;
                        qe.j = origj;
                        qe.dir = dir;
                        qe.energy = energy;
                        
                        q.add(qe);
                    }
                    visitedmaxenergy[origi][origj][dir] = 1;
                } else {
                    v = bfs[origi][origj][dir][energy];
                    v.dist = 0;
                    v.visited = true;
                    //v.pi = -1;
                    
                    visitedmaxenergy[origi][origj][dir] = energy;
                    
                    qe = new QElem();
                    qe.i = origi;
                    qe.j = origj;
                    qe.dir = dir;
                    qe.energy = energy;
                    
                    q.add(qe);
                }
            }
        } else if (startenergy < 0) { // all non-zero energies
            for (energy = 1; energy <= maxenergy; energy++) {
                v = bfs[origi][origj][dir][energy];
                v.dist = 0;
                v.visited = true;
                //v.pi = -1;
                
                qe = new QElem();
                qe.i = origi;
                qe.j = origj;
                qe.dir = dir;
                qe.energy = energy;
                
                q.add(qe);
            }
            visitedmaxenergy[origi][origj][dir] = 1;
        } else {
            v = bfs[origi][origj][dir][energy];
            v.dist = 0;
            v.visited = true;
            //v.pi = -1;
            
            visitedmaxenergy[origi][origj][dir] = energy;
            
            qe = new QElem();
            qe.i = origi;
            qe.j = origj;
            qe.dir = dir;
            qe.energy = energy;
            
            q.add(qe);
        }
        
        if (!rechargeallowed)
            movescount = 4;
        
        while (!q.isEmpty()) {
            if (q.size() > maxqueuesize)
                maxqueuesize = q.size();
            qe = q.remove();
            nvertices++;
            //System.out.println("qe.i = " + qe.i + ", qe.j = " + qe.j + ", qe.dir = " + qe.dir + ", qe.energy = " + qe.energy);
            
            for (int i = 0; i < movescount; i++) {
                newi = qe.i;
                newj = qe.j;
                newdir = qe.dir;
                newenergy = qe.energy + 1; // reverse search!!!
                action = Strategy.BotAction.STEP; // hack aby nehazel warningy
                switch (i) {
                    case 0: // STEP
                        newi = qe.i + di[(newdir + 2)%4]; // bereme opacny smer
                        newj = qe.j + dj[(newdir + 2)%4]; // bereme opacny smer
                        action = Strategy.BotAction.STEP;
                        break;
                    case 1: // RIGHT
                        newdir = (qe.dir + 3) % 4; // otoceni doleva (reversed right)
                        action = Strategy.BotAction.RIGHT;
                        break;
                    case 2: // LEFT
                        newdir = (qe.dir + 1) % 4; // otoceni doprava (reversed left)
                        action = Strategy.BotAction.LEFT;
                        break;
                    case 3: // REST, zde je vice moznosti pri energy == maxenergy
                        action = Strategy.BotAction.REST;
                        break;
                    case 4: // RECHARGE, zde je vice moznosti pri energy == maxenergy
                        action = Strategy.BotAction.RECHARGE;
                        break;
                }
                
                if (i < 3) {
                    nedges++;
                    if (newenergy <= maxenergy &&
                            isInMap(newi, newj) &&
                            !bfs[newi][newj][newdir][newenergy].visited &&
                            (gamemap[newi][newj] & testint) == 0) {
                        if (newenergy < visitedmaxenergy[newi][newj][newdir]) {
                            visitedmaxenergy[newi][newj][newdir] = newenergy;
                            QElem newqe = new QElem();
                            newqe.dir = newdir;
                            newqe.i = newi;
                            newqe.j = newj;
                            newqe.energy = newenergy;
                            
                            bfs[newi][newj][newdir][newenergy].visited = true;
                            bfs[newi][newj][newdir][newenergy].dist =
                                    bfs[qe.i][qe.j][qe.dir][qe.energy].dist + 1;
                            if (bfs[newi][newj][newdir][newenergy].dist > maxdist)
                                maxdist = bfs[newi][newj][newdir][newenergy].dist;
                            // tyto udaje nejsou potreba
                            //bfs[newi][newj][newdir][newenergy].pi = qe.i;
                            //bfs[newi][newj][newdir][newenergy].pj = qe.j;
                            //bfs[newi][newj][newdir][newenergy].pdir = qe.dir;
                            //bfs[newi][newj][newdir][newenergy].penergy = qe.energy;
                            //bfs[newi][newj][newdir][newenergy].paction = action;
                            
                            q.add(newqe);
                        }
                    }
                } else {
                    if ((qe.energy > 1) && (i==3 || (i == 4 && (gamemap[newi][newj] & idbattery) > 0))) {
                        if (i == 3) {
                            if (qe.energy - restenergy > 0) {
                                energylo = qe.energy - restenergy + 1;
                                energyhi = energylo;
                            } else {
                                energylo = 1;
                                energyhi = 1;
                            }
                        } else { // i == 4
                            energylo = 1;
                            energyhi = 1;
                        }
                        for (newenergy = energylo; newenergy <= energyhi; newenergy++) {
                            nedges++;
                            if (!bfs[newi][newj][newdir][newenergy].visited) {
                                if (newenergy < visitedmaxenergy[newi][newj][newdir]) {
                                    visitedmaxenergy[newi][newj][newdir] = newenergy;
                                    QElem newqe = new QElem();
                                    newqe.dir = newdir;
                                    newqe.i = newi;
                                    newqe.j = newj;
                                    newqe.energy = newenergy;
                                    
                                    bfs[newi][newj][newdir][newenergy].visited = true;
                                    bfs[newi][newj][newdir][newenergy].dist =
                                            bfs[qe.i][qe.j][qe.dir][qe.energy].dist + 1;
                                    if (bfs[newi][newj][newdir][newenergy].dist > maxdist)
                                        maxdist = bfs[newi][newj][newdir][newenergy].dist;
                                    // tyto udaje nejsou potreba
                                    //bfs[newi][newj][newdir][newenergy].pi = qe.i;
                                    //bfs[newi][newj][newdir][newenergy].pj = qe.j;
                                    //bfs[newi][newj][newdir][newenergy].pdir = qe.dir;
                                    //bfs[newi][newj][newdir][newenergy].penergy = qe.energy;
                                    //bfs[newi][newj][newdir][newenergy].paction = action;
                                    
                                    q.add(newqe);
                                }
                            }
                        }
                    }
                }
            }
        }
        
        if (verbose) {
            System.out.println("BackwardSearch: nvertices = " + nvertices + ", nedges = " + nedges +
                    ", maxqueuesize = " + maxqueuesize + ", maxdist = " + maxdist);
        }
    }
    
    private int BotActionToInt(Strategy.BotAction action) {
        int ret = -1;
        
        if (action != null) {
            switch (action) {
                case STEP:
                    ret = movestep;
                    break;
                case LEFT:
                    ret = moveleft;
                    break;
                case RIGHT:
                    ret = moveright;
                    break;
                case RECHARGE:
                    ret = moverecharge;
                    break;
                case TAKE:
                    ret = movetake;
                    break;
                case SHOOT:
                    ret = moveshoot;
                    break;
                case REST:
                    ret = moverest;
                    break;
            }
        }
        
        return ret;
    }
    
    private int DirectionToInt(Robot.Direction direction) {
        int ret = -1;
        
        switch (direction) {
            case LEFT: ret = dirleft; break;
            case UP: ret = dirup; break;
            case RIGHT: ret = dirright; break;
            case DOWN: ret = dirdown; break;
        }
        
        return ret;
    }
    
    // pozor, bude hlasit chybu, pokud robot provede nejakou "sebevrazednou" akci
    // v situaci kdy ma energii 1 (s energii 1 se generuje jen odpocinek nebo dobiti)
    private boolean CheckMoveGenerator() {
        int restorenew;
        boolean errorfound = false;
        boolean errorinlastrobotstates = false;
        int errorinlrsrid = -1;
        boolean errorinlastgamemap = false;
        int errorinlgmi = -1;
        int errorinlgmj = -1;
        boolean errorinalreadyhit = false;
        int errorinahrid = -1;
        boolean errorinrestorednewenergy = false;
        int errorinrnerid = -1;
        boolean errorinmovesimulationrobot = false;
        int errorinmsrrid = -1;
        boolean errorinmovesimulationgamemap = false;
        int errorinmsgmi = -1;
        int errorinmsgmj = -1;
        boolean errorinmovegeneration = false;
        int errorinmgrid = -1;
        int robotactions[] = new int[nrobots];
        int robotmoveids[] = new int[nrobots];
        int lastnewenergystate[] = new int[nrobots];
        boolean lastguardianAngel = guardianAngel;
        // guardianAngel zabranuje generovani nejakych pohybu, pro kontrolu generatoru
        // pohybu je potreba jej vypnout
        guardianAngel = false;
        
        for (int rid = 0; rid < nrobots; rid++) {
            if (rposi[rid] != RobotLastStates[rid].i ||
                    rposj[rid] != RobotLastStates[rid].j ||
                    rdir[rid] != RobotLastStates[rid].dir ||
                    renergy[rid] != RobotLastStates[rid].energy ||
                    rpoints[rid] != RobotLastStates[rid].points) {
                errorinlastrobotstates = true;
                errorinlrsrid = rid;
            }
        }
        
        for (int i = 0; i < width; i++)
            for (int j = 0; j < height; j++) {
            if (gamemap[i][j] != lastgamemapstate[i][j]) {
                errorinlastgamemap = true;
                errorinlgmi = i;
                errorinlgmj = j;
            }
            }
        
        if (errorinlastrobotstates) {
            System.out.println("Error in last robot states:" + errorinlrsrid);
            errorfound = true;
        }
        
        if (errorinlastgamemap) {
            System.out.println("Error in last gamemap: " + errorinlgmi + ", " + errorinlgmj);
            errorfound = true;
        }
        
        if (!errorfound) {
            GetPossibleMoves(8, false, remains);
            
            for (int rid = 0; rid < nrobots; rid++) {
                robotactions[rid] = BotActionToInt(robots.get(rid).getLastAction());
                boolean actionfound = false;
                for (int moveid = 0; moveid < npossiblemoves[rid]; moveid++) {
                    if (robotactions[rid] == possiblemoves[rid][moveid]) {
                        actionfound = true;
                        robotmoveids[rid] = moveid;
                    }
                }
                if (!actionfound) { // nebyl tah nevygenerovan proto ze je ekvivalentni "nicnedelani"?, pokud ne je to chyba v generatoru
                    Robot r = robots.get(rid);
                    // diky slozitim pravidlum upravy energie robota, neni mozne presne urcit, zda
                    // robotova akce byla ekvivalentni "nicnedelani", jistotu, ze robot udelal
                    // nejakou akci mame pouze v pripade, kdy se zmeni nektery z jeho dalsich parametru.
                    // Kdyz nemame jistotu, ze neco udelal, predpokladame ze nic neudelal (abychom nehlasili
                    // zbytecne falesne poplachy), pripadna chyba se totiz z vysokou pravdepodobnosti
                    // detekuje dale
                    if (r.getPositionX() != RobotLastStates[rid].i ||
                            r.getPositionY() != RobotLastStates[rid].j ||
                            DirectionToInt(r.getDirection()) != RobotLastStates[rid].dir ||
                            /*r.getEnergy() != RobotLastStates[rid].energy ||*/
                            r.getPoints() != RobotLastStates[rid].points) {
                        errorinmovegeneration = true;
                        errorinmgrid = rid;
                    } else {
                        possiblemoves[rid][0] = -1;
                        robotmoveids[rid] = 0;
                    }
                }
            }
            
            if (errorinmovegeneration) {
                System.out.println("Error in move generation: " + errorinmgrid);
                errorfound = true;
            }
            
            if (!errorfound) {
                int alreadyhitset[] = new int[nrobots];
                alreadyhitset[r2d2id] = DoMove(r2d2id, robotmoveids[r2d2id]);
                
                for (int rid = 0; rid < nrobots; rid++) if (rid != r2d2id) {
                    lastnewenergystate[rid] = newrenergy[rid];
                    alreadyhitset[rid] = DoMove(rid, robotmoveids[rid]);
                }
                
                restorenew = ResolveCollisions(5);
                for (int rid = 0; rid < nrobots; rid++) {
                    Robot r = robots.get(rid);
                    if (rposi[rid] != r.getPositionX() ||
                            rposj[rid] != r.getPositionY() ||
                            rdir[rid] != DirectionToInt(r.getDirection()) ||
                            renergy[rid] != r.getEnergy() ||
                            rpoints[rid] != r.getPoints()) {
                        errorinmovesimulationrobot = true;
                        errorinmsrrid = rid;
                    }
                }
                for (int i = 0; i < width; i++) {
                    for (int j = 0; j < height; j++) {
                        int gamemapij = 0;
                        if (world.containsWall(i, j))
                            gamemapij |= idwall;
                        if (world.containsBattery(i, j)) {
                            gamemapij |= idbattery;
                        }
                        if (world.containsRobot(i, j))
                            gamemapij |= idrobot;
                        if (world.containsTreasure(i, j)) {
                            gamemapij |= idtreasure*(2*world.getTreasureValue(i, j) + 1);
                        }
                        if (((gamemap[i][j] & idtreasure) > 0 && gamemapij != gamemap[i][j]) ||
                                ((gamemap[i][j] & idtreasure) == 0 && gamemapij != gamemap[i][j] % (2*idtreasure))) {
                            errorinmovesimulationgamemap = true;
                            errorinmsgmi = i;
                            errorinmsgmj = j;
                        }
                        
                    }
                }
                UnResolveCollisions(5, restorenew);
                
                for (int rid = nrobots - 1; rid >= 0; rid--) if (rid != r2d2id) {
                    UnDoMove(rid, robotmoveids[rid], alreadyhitset[rid]);
                    newrenergy[rid] = lastnewenergystate[rid];
                }
                
                UnDoMove(r2d2id, robotmoveids[r2d2id], alreadyhitset[r2d2id]);
                
                for (int rid = 0; rid < nrobots; rid++) {
                    if (alreadyhit[rid]) {
                        errorinalreadyhit = true;
                        errorinahrid = rid;
                    }
                    if (rid != r2d2id && newrenergy[rid] != imax(0, renergy[rid] - 1)) {
                        errorinrestorednewenergy = true;
                        errorinrnerid = rid;
                    }
                    if (rposi[rid] != RobotLastStates[rid].i ||
                            rposj[rid] != RobotLastStates[rid].j ||
                            rdir[rid] != RobotLastStates[rid].dir ||
                            renergy[rid] != RobotLastStates[rid].energy ||
                            rpoints[rid] != RobotLastStates[rid].points) {
                        errorinlastrobotstates = true;
                        errorinlrsrid = rid;
                    }
                }
                
                for (int i = 0; i < width; i++)
                    for (int j = 0; j < height; j++) {
                    if (gamemap[i][j] != lastgamemapstate[i][j]) {
                        errorinlastgamemap = true;
                        errorinlgmi = i;
                        errorinlgmj = j;
                    }
                    }
                
                if (errorinmovesimulationrobot) {
                    System.out.println("Error in move simulation robot: " + errorinmsrrid);
                    errorfound = true;
                }
                
                if (errorinmovesimulationgamemap) {
                    System.out.println("Error in move simulation gamemap: " + errorinmsgmi + ", " + errorinmsgmj);
                    errorfound = true;
                }
                
                
                if (errorinalreadyhit) {
                    System.out.println("Error in alreadyhit: " + errorinahrid);
                    errorfound = true;
                }
                
                if (errorinrestorednewenergy) {
                    System.out.println("Error in restored new energy: " + errorinrnerid);
                    errorfound = true;
                }
                
                if (errorinlastrobotstates) {
                    System.out.println("Error in restored last robot states: " + errorinlrsrid);
                    errorfound = true;
                }
                
                if (errorinlastgamemap) {
                    System.out.println("Error in restored last gamemap: " + errorinlgmi + ", " + errorinlgmj);
                    errorfound = true;
                }
            }
        }
        
        guardianAngel = lastguardianAngel;
        
        return (!errorfound);
    }
    
    /* Pomocna metoda, pro vypocet rozlozeni zlata.
     * Posledni parametr neprimo umerne urcuje rozprostreni pokladu na sousedni pole */
    private void propagateValue(int i, int j, int value, int w) {
        LinkedList<EvaluatedPoint> fronta = new LinkedList<EvaluatedPoint>();
        fronta.addLast(new EvaluatedPoint(i, j, -1, value));
        while (!fronta.isEmpty()) {
            EvaluatedPoint p = fronta.removeFirst();
            if (visited[p.i][p.j]  || p.value == 0 || world.containsWall(p.i,p.j)) continue;
            visited[p.i][p.j] = true;
            goldmap[p.i][p.j] += p.value;
            for (int k = 0; k <= 3; k++) {
                if (isInMap(p.i+di[k], p.j+dj[k])) {
                    fronta.addLast(new EvaluatedPoint(p.i+di[k], p.j+dj[k], -1, p.value/w));
                }
            }
        }
    }
    
    // kombinace z prvku 0..n-1 o k prvcich
    int[] next_comb(int[] a, int n, int k) {
        int i = k - 1, j;
        
        if (i < 0) {
            return null;
        }
        
        if (a[i] == n - 1) {
            while (i > 0 && a[i - 1] + 1 == a[i])
                i--;
            
            if (i == 0) {
                return null;
            } else {
                a[i - 1]++;
                for (j = i; j < k; j++)
                    a[j] = a[i - 1] + j - i + 1;
                return a;
            }
        } else {
            a[i]++;
            return a;
        }
    }
    
    private void ComputeDeadRobots() {
        long startcdrtime = System.currentTimeMillis();
        // tohle bude chtit trochu optimalizovat (napr nezkoumat pole kam se r2d2 dostane
        // za vic nez maxalphabetarange kroku
        int mindist[] = new int[nrobots];
        int endangergraph[][] = new int[nrobots][nrobots];
        int upb, val, vali, valj;
        
        for (int rid = 0; rid < nrobots; rid++) {
            for (int rid2 = 0; rid2 < nrobots; rid2++) if (rid2 != rid) {
                endangergraph[rid][rid2] = maxalphabetarange + 1;
            }
        }
        
        for (int rid = 0; rid < nrobots; rid++) {
            for (int i = 0; i < width; i++) {
                for (int j = 0; j < height; j++) {
                    if ((gamemap[i][j] & idwall) == 0) {
                        for (int l = 0; l < 4; l++) if (spdir[rid][i][j][l] < maxalphabetarange + 1) {
                            for (int rid2 = 0; rid2 < nrobots; rid2++) if (rid2 != rid) {
                                if (imax(spdir[rid][i][j][l], sp[rid2][i][j]) < endangergraph[rid][rid2])
                                    endangergraph[rid][rid2] = imax(spdir[rid][i][j][l], sp[rid2][i][j]);
                            }
                            // pro (vali != i || valj !=j) pricteme k distancim jednicku, protoze
                            // aby se roboti nestojici na stejnem poli ovlivnili musi udelat
                            // nejakou akci
                            mindist[rid] = spdir[rid][i][j][l] + 1;
                            for (int rid2 = 0; rid2 < nrobots; rid2++) if (rid2 != rid) {
                                mindist[rid2] = MAX_DIST;
                            }
                            upb = alldiststowalls[i][j][l];
                            val = 1;
                            vali = i + di[l];
                            valj = j + dj[l];
                            while (val < upb) {
                                for (int rid2 = 0; rid2 < nrobots; rid2++) if (rid2 != rid) {
                                    if (sp[rid2][vali][valj] + 1 < mindist[rid2])
                                        mindist[rid2] = sp[rid2][vali][valj] + 1;
                                }
                                vali += di[l];
                                valj += dj[l];
                                val++;
                            }
                            for (int rid2 = 0; rid2 < nrobots; rid2++) if (rid2 != rid) {
                                if (imax(mindist[rid], mindist[rid2]) < endangergraph[rid][rid2])
                                    endangergraph[rid][rid2] = imax(mindist[rid], mindist[rid2]);
                            }
                        }
                    }
                }
            }
        }
        
        for (int rid = 0; rid < nrobots; rid++) {
            for (int rid2 = 0; rid2 < nrobots; rid2++) if (rid2 != rid) {
                endangergraph[rid][rid2] = imin(endangergraph[rid][rid2], endangergraph[rid2][rid]);
            }
        }
        
        boolean bestdeadrobots[] = new boolean[nrobots];
        int bestmaxexpectedalphabetarange = 0;
        int bestmaxpossiblealphabetarange = 0;
        int maxexpectedalphabetarange;
        int maxpossiblealphabetarange;
        for (int rid = 0; rid < nrobots; rid++) {
            bestdeadrobots[rid] = false;
        }
        for (int ndead = 0; ndead <= nrobots - 1; ndead++) {
            int a[] = new int[ndead];
            int robotrids[] = new int[nrobots - 1];
            int j = 0;
            for (int rid = 0; rid < nrobots; rid++) if (rid != r2d2id) {
                robotrids[j] = rid;
                j++;
            }
            for (int i = 0; i < ndead; i++) {
                a[i] = i;
            }
            
            // tohle je potreba lip nastavit
            boolean currentdeadrobots[] = new boolean[nrobots];
            do {
                maxpossiblealphabetarange = maxalphabetarange;
                // id mrtvych robotu: robotrids[a[0]] ... robotrids[a[ndead - 1]]
                // spocita maximalni korektni moznou hloubku alhabety v pripade
                // ze urciti roboti jsou mrtvi
                for (int rid = 0; rid < nrobots; rid++) {
                    currentdeadrobots[rid] = false;
                }
                if (verbose) {
                    System.out.print("Dead robot ids: ");
                }
                for (int i = 0; i < ndead; i++) {
                    currentdeadrobots[robotrids[a[i]]] = true;
                    if (verbose) {
                        System.out.print(robotrids[a[i]] + " ");
                    }
                }
                if (verbose) {
                    System.out.println("");
                }
                for (int i = 0; i < ndead; i++) {
                    int rid = robotrids[a[i]];
                    for (int rid2 = 0; rid2 < nrobots; rid2++) if (!currentdeadrobots[rid2]) {
                        if (maxpossiblealphabetarange > endangergraph[rid][rid2] - 1)
                            maxpossiblealphabetarange = endangergraph[rid][rid2] - 1;
                    }
                }
                maxexpectedalphabetarange = imin(maxpossiblealphabetarange, lastreachedalphabetarange[ndead]);
                if (verbose) {
                    System.out.println("maxpossiblealphabetarange = " + maxpossiblealphabetarange);
                    System.out.println("maxexpectedalphabetarange = " + maxexpectedalphabetarange);
                }
                if (maxexpectedalphabetarange > bestmaxexpectedalphabetarange ||
                        (maxexpectedalphabetarange == bestmaxexpectedalphabetarange &&
                        maxpossiblealphabetarange > bestmaxpossiblealphabetarange)) {
                    bestmaxexpectedalphabetarange = maxexpectedalphabetarange;
                    bestmaxpossiblealphabetarange = maxpossiblealphabetarange;
                    for (int rid = 0; rid < nrobots; rid++) {
                        bestdeadrobots[rid] = currentdeadrobots[rid];
                    }
                }
            } while((a = next_comb(a, nrobots - 1, ndead)) != null);
        }
        
        for (int rid = 0; rid < nrobots; rid++) {
            deadrobots[rid] = bestdeadrobots[rid];
        }
        deadmaxalphabetarange = bestmaxpossiblealphabetarange;
        
        long endcdrtime = System.currentTimeMillis();
        if (verbose) {
            System.out.println("cdrtime = " + (endcdrtime - startcdrtime));
        }
    }
    
    private int canShoot() {
        int ii = rposi[r2d2id];
        int jj = rposj[r2d2id];
        int cnt = 1;
        int ub = alldiststowalls[ii][jj][rdir[r2d2id]];
        int targetrid = -1, targeti, targetj;
        boolean istarget = false;
        while (cnt < ub) {
            ii += di[rdir[r2d2id]];
            jj += dj[rdir[r2d2id]];
            cnt++;
            if (!isInMap(ii, jj)) {
                if (verbose) {
                    System.err.println("!!!!!!!!!CHYBA V alldiststowalls");
                }
                break;
            }
            if ((gamemap[ii][jj] & idrobot) > 0) {
                istarget = true;
                break;
            }
        }
        boolean robotfound = false;
        for (int rid = 0; rid < nrobots; rid++) {
            if (rposi[rid] == ii && rposj[rid] == jj) {
                targetrid = rid;
                targeti = ii;
                targetj = jj;
                robotfound = true;
                break;
            }
        }
        if (!robotfound && istarget) {
            if (verbose) {
                System.err.println("!!!!!!!!!!!!!!!!!!!!!CHYBA V canShoot");
            }
        }
        
        return targetrid;
    }
    
    public Strategy.BotAction decide() {
        try {
            
            starttime = System.currentTimeMillis();
            long endtime;
            
            if (verbose && roundno > 0 && !CheckMoveGenerator()) {
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!THERE IS AN ERROR IN MOVE GENERATOR!!!!!!!!!!!!!!!!!!!!!!!!!!");
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
                System.err.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
            }
            
            analyzeWorldandRobots();
            
            /* Naplneni promennych popisujicich robota */
            int posi = robot.getPositionX();
            int posj = robot.getPositionY();
            int energy = robot.getEnergy();
            int dir = 0;
            int depth;
            int lastfinaldecisionint;
            Strategy.BotAction action;
            Vertex v = new Vertex();
            switch (robot.getDirection()) {
                case LEFT: dir = 0; break;
                case UP: dir = 1; break;
                case RIGHT: dir = 2; break;
                case DOWN: dir = 3; break;
            }
            
            // pokud zbyva do konce vice nez startevaluate2 kol, pokracuje se normalne, pokud min nebo stejne
            // pouziva se funkce evaluate2 a na cil se sere
            deadmaxalphabetarange = maxalphabetarange;
            if (remains > startevaluate2) {
                if (energy == 1) {
                    if (remains == 1) {
                        if ((gamemap[posi][posj] & idtreasure) > 0 &&
                                gamemap[posi][posj]/(2*idtreasure) > restenergy) {
                            if (verbose) {
                                System.out.println("Brani");
                                if (roundno == world.getRoundsCount()) {
                                    System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                                    System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                                    System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                                }
                            }
                            return Strategy.BotAction.TAKE;
                        } else if ((gamemap[posi][posj] & idbattery) > 0) {
                            if (verbose) {
                                System.out.println("Dobiti");
                                if (roundno == world.getRoundsCount()) {
                                    System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                                    System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                                    System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                                }
                            }
                            return Strategy.BotAction.RECHARGE;
                        } else {
                            if (verbose) {
                                System.out.println("Odpocinek");
                                if (roundno == world.getRoundsCount()) {
                                    System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                                    System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                                    System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                                }
                            }
                            return Strategy.BotAction.REST;
                        }
                    } else {
                        if (verbose) {
                            System.out.println("Odpocinek");
                            if (roundno == world.getRoundsCount()) {
                                System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                                System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                                System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                            }
                        }
                        return Strategy.BotAction.REST;
                    }
                }
                
                // vyhodnoceni vyhodnosti strileni
                int possibletargetrid = canShoot();
                if (verbose) {
                    System.out.println("possibletargetrid = " + possibletargetrid);
                }
                if (possibletargetrid < 0) {
                    EVAL_OBJECTIVE = EVAL_OBJECTIVE_BASE;
                } else {
                    int sprofit = shootingProfit(renergy[r2d2id], renergy[possibletargetrid], remains);
                    if (verbose) {
                        System.out.println("sprofit = " + sprofit);
                    }
                    if (sprofit > (remains*AVERAGE_TOTAL_PROFIT)/world.getRoundsCount()) {
                        EVAL_OBJECTIVE /= EVAL_OBJECTIVE_MODIFIER;
                    } else {
                        EVAL_OBJECTIVE = imin(EVAL_OBJECTIVE_BASE, EVAL_OBJECTIVE_MODIFIER*EVAL_OBJECTIVE);
                    }
                    if (verbose) {
                        System.out.println("EVAL_OBJECTIVE = " + EVAL_OBJECTIVE);
                    }
                }
                
                long startsptime = System.currentTimeMillis();
                
                /* Nalezeni nejkratsich cest pro vsechny roboty*/
                for (int rid = 0; rid < nrobots; rid++) {
                    /*
                    ForwardSearch(rposi[rid], rposj[rid], rdir[rid], renergy[rid], true, false);
                    for (int i = 0; i < width; i++) {
                        for (int j = 0; j < height; j++) {
                            spnorecharge[rid][i][j] = MAX_DIST;
                            for (int l = 0; l < 4; l++) {
                                for (int m = 0; m <= maxenergy; m++) {
                                    v = bfs[i][j][l][m];
                                    if (v.visited) {
                                        // tady se uvazuje, ze cizi robot mi muze obsadit pole i s nulovou energii,
                                        // ForwardSearch nulovou energii ale nepovoluje, takze to vlastne pouze
                                        // korektne osetri roboty kteri jsou mrtvi od zacatku (startovni
                                        // pole ForwardSearch vyplni i s nulovou energii)
                                        if (v.dist < spnorecharge[rid][i][j] && (m > 0 || rid != r2d2id)) {
                                            spnorecharge[rid][i][j] = v.dist;
                                        }
                                    }
                                }
                            }
                        }
                    }
                     */
                    ForwardSearch(rposi[rid], rposj[rid], rdir[rid], renergy[rid], true, true);
                    for (int i = 0; i < width; i++) {
                        for (int j = 0; j < height; j++) {
                            sp[rid][i][j] = MAX_DIST;
                            for (int l = 0; l < 4; l++) {
                                spdir[rid][i][j][l] = MAX_DIST;
                                for (int m = 0; m <= maxenergy; m++) {
                                    v = bfs[i][j][l][m];
                                    if (v.visited) {
                                        // tady se uvazuje, ze cizi robot mi muze obsadit pole i s nulovou energii,
                                        // ForwardSearch nulovou energii ale nepovoluje, takze to vlastne pouze
                                        // korektne osetri roboty kteri jsou mrtvi od zacatku (startovni
                                        // pole ForwardSearch vyplni i s nulovou energii)
                                        if (v.dist < sp[rid][i][j] && (m > 0 || rid != r2d2id)) {
                                            sp[rid][i][j] = v.dist;
                                        }
                                        if (v.dist < spdir[rid][i][j][l] && (m > 0 || rid != r2d2id)) {
                                            spdir[rid][i][j][l] = v.dist;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                
                long endsptime = System.currentTimeMillis();
                if (verbose) {
                    System.out.println("sptime = " + (endsptime - startsptime));
                }
                
                // VVVVVVVVVVVVVVVVVVVVVV vypocet mrtvych robotu VVVVVVVVVVVVVVVVVVVVVV
                ComputeDeadRobots();
                
                
                if (verbose) {
                    System.out.print("deadrobots: ");
                    for (int rid = 0; rid < nrobots; rid++)
                        System.out.print(deadrobots[rid] + " ");
                    System.out.println("");
                }
                // AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
                if (verbose) {
                    for (int rid = 0; rid < nrobots; rid++) {
                        System.out.println("Robot #" + rid + " shortest paths");
                        for (int j = 0; j < height; j++) {
                            for (int i = 0; i < width; i++) {
                                System.out.print(sp[rid][i][j] + " ");
                            }
                            System.out.println("");
                        }
                    }
                    
                    System.out.println("Treasure list:");
                    for (int i = 0; i < ntreasures; i++) {
                        System.out.print("(" + treasurelist[i].i + ", " + treasurelist[i].j + "), ");
                    }
                    System.out.println("");
                    
                    System.out.println("Battery list:");
                    for (int i = 0; i < nbatteries; i++) {
                        System.out.print("(" + batterylist[i].i + ", " + batterylist[i].j + "), ");
                    }
                    System.out.println("");
                }
                
                /* Zapomen na cil, pokud to neni poklad ani baterie */
                /* resi i situaci, kdy jsem prave sebral poklad, ktery byl objective */
                if (objective != null
                        && (gamemap[objective.i][objective.j] & idbattery) == 0
                        && (gamemap[objective.i][objective.j] & idtreasure) == 0) {
                    objective = null;
                }
                
                /* ### Kontrola cile ### */
                if (objective != null) {
                    if (objective.i == posi && objective.j == posj) {
                        if (world.containsTreasure(posi,posj)) {
                            // AlphaBeta rozhodne, zda se poklad ma vzit nebo ne
                            /*
                            objective = null;
                            if (verbose || true) {
                                System.out.println("Brani");
                                if (roundno == world.getRoundsCount()) {
                                    System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                                    System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                                    System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                                }
                            }
                            return Strategy.BotAction.TAKE;
                             */
                        }
                        if (world.containsBattery(posi,posj)) {
                            if (remains == 1) {
                                if (verbose) {
                                    System.out.println("Dobiti");
                                    if (roundno == world.getRoundsCount()) {
                                        System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                                        System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                                        System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                                    }
                                }
                                return Strategy.BotAction.RECHARGE;
                            } else {
                                /* Otacej se a strilej. Asi nechat bezet alfa-betu. */
                                guardianAngel = true;
                            }
                        }
                    }
                /* Tohle je dulezite! Pokud na cilovem poli nekdo stoji (ale nejsem to ja),
                 * zapomen na cil.
                 * Je potreba tuto podminku zohlednit pri vyberu cile, tak aby se za cil
                 * nevybralo pole na nemz nekdo stoji. */
                    if ((objective.i != posi || objective.j != posj) &&
                            world.containsRobot(objective.i, objective.j)) {
                        objective = null;
                    }
                    
                    /* Tohle je dulezite! Pokud k cili nestihnu dojit, zvolim si jiny */
                    if (objective != null && remains <= sp[r2d2id][objective.i][objective.j]) {
                        objective = null;
                    }
                }
                
                /* ### Volba cile ### */
                // cbl a ctl se pro ucely vypisu pocitaji pokazde
                /* Vypocet rozlozeni pokladu */
                goldmap = new int[width][height];
                for (int n = 0; n < ntreasures; n++) {
                    visited = new boolean[width][height];
                    propagateValue(treasurelist[n].i, treasurelist[n].j, world.getTreasureValue(treasurelist[n].i, treasurelist[n].j), GOLD_DEVIATION);
                }
                
                /* Vyber pokladu a baterii, ke kterym mas nejbliz a ke kterym dojdu */
                ctl = new LinkedList<Point>(); rtl = new LinkedList<Point>();
                for (int n = 0; n < ntreasures; n++) {
                    // Vypocet dosazitelnych pokladu na nichz nikdo nestoji, poklady se hned neberou, takze
                    // musim zde testovat, zda na pokladu nahodou nestojim
                    if (remains > sp[r2d2id][treasurelist[n].i][treasurelist[n].j]
                            && ((treasurelist[n].i == posi && treasurelist[n].j == posj) ||
                            (gamemap[treasurelist[n].i][treasurelist[n].j] & idrobot) == 0)) {
                        rtl.add(treasurelist[n]);
                        // Vypocet nejblizsich pokladu
                        int min = MAX_DIST; boolean stejne = false;
                        for (int rid = 0; rid < nrobots; rid++) {
                            if (min > sp[rid][treasurelist[n].i][treasurelist[n].j]) {
                                min = sp[rid][treasurelist[n].i][treasurelist[n].j];
                                stejne = false;
                            } else if (min == sp[rid][treasurelist[n].i][treasurelist[n].j]) {
                                stejne = true;
                            }
                        }
                        if (min == sp[r2d2id][treasurelist[n].i][treasurelist[n].j]
                                && (stejne == false || goldmap[treasurelist[n].i][treasurelist[n].j] >
                                1.5*gamemap[treasurelist[n].i][treasurelist[n].j]/(2*idtreasure))) {
                            ctl.add(treasurelist[n]);
                        }
                    }
                }
                cbl = new LinkedList<Point>(); rbl = new LinkedList<Point>();
                for (int n = 0; n < nbatteries; n++) {
                    // Vypocet dosazitelnych baterii na nichz nikdo nestoji (muze se stat ze uz stojim na baterii ja)
                    if (remains > sp[r2d2id][batterylist[n].i][batterylist[n].j]
                            && ((batterylist[n].i == posi && batterylist[n].j == posj) ||
                            (gamemap[batterylist[n].i][batterylist[n].j] & idrobot) == 0)) {
                        rbl.add(batterylist[n]);
                        // Vypocet nejblizich baterii
                        int min = MAX_DIST; boolean stejne = false;
                        for (int rid = 0; rid < nrobots; rid++) {
                            if (min > sp[rid][batterylist[n].i][batterylist[n].j]) {
                                min = sp[rid][batterylist[n].i][batterylist[n].j];
                                stejne = false;
                            } else if (min == sp[rid][batterylist[n].i][batterylist[n].j]) {
                                stejne = true;
                            }
                        }
                        if (min == sp[r2d2id][batterylist[n].i][batterylist[n].j] && stejne == false) {
                            cbl.add(batterylist[n]);
                        }
                    }
                }
                
                if (verbose) {
                    if (ctl.isEmpty())
                        System.out.println("No reachable treasure!");
                    else {
                        System.out.print("RTL WITH RECHARGE: ");
                        for (Object o:rtl) {
                            Point p = (Point) o;
                            System.out.print("(" + p.i + ", " + p.j + ") = " + sp[r2d2id][p.i][p.j] + ", ");
                        }
                        System.out.println("");
                        /*
                        System.out.print("RTL WITHOUT RECHARGE: ");
                        for (Object o:rtl) {
                            Point p = (Point) o;
                            System.out.print("(" + p.i + ", " + p.j + ") = " + spnorecharge[r2d2id][p.i][p.j] + ", ");
                        }
                        System.out.println("");
                         */
                    }
                    if (ctl.isEmpty())
                        System.out.println("No closest treasure!");
                    else {
                        System.out.print("CTL WITH RECHARGE: ");
                        for (Object o:ctl) {
                            Point p = (Point) o;
                            System.out.print("(" + p.i + ", " + p.j + ") = " + sp[r2d2id][p.i][p.j] + ", ");
                        }
                        System.out.println("");
                        /*
                        System.out.print("CTL WITHOUT RECHARGE: ");
                        for (Object o:ctl) {
                            Point p = (Point) o;
                            System.out.print("(" + p.i + ", " + p.j + ") = " + spnorecharge[r2d2id][p.i][p.j] + ", ");
                        }
                        System.out.println("");
                         */
                    }
                    if (cbl.isEmpty())
                        System.out.println("No closest battery!");
                    else  {
                        System.out.print("CBL WITH RECHARGE: ");
                        for (Object o:cbl) {
                            Point p = (Point) o;
                            System.out.print("(" + p.i + ", " + p.j + ") = " + sp[r2d2id][p.i][p.j] + ", ");
                        }
                        System.out.println("");
                        /*
                        System.out.print("CBL WITHOUT RECHARGE: ");
                        for (Object o:cbl) {
                            Point p = (Point) o;
                            System.out.print("(" + p.i + ", " + p.j + ") = " + spnorecharge[r2d2id][p.i][p.j] + ", ");
                        }
                        System.out.println("");
                         */
                    }
                }
                /* Volba cile */
                if (objective == null) {
                    int i, j, besti = -1, bestj = -1, mindist = MAX_DIST;
                    if (!rtl.isEmpty()) {
                        if (!ctl.isEmpty()) {
                        /* Pujdu pro poklad, ktery maximalizuje rozdil mezi jeho hodnotou a vzdalenosti.
                         * Pokud maji vsechny poklady tento rozdil zaporny, budu strilet a odpocivat. */
                            for (Object o:ctl) {
                                Point p = (Point) o;
                                i = p.i;
                                j = p.j;
                                /* Minimalizuje se vzdalenost - hodnota goldmap
                                 * zaroven musi platit:
                                 * - dojit pro poklad, vzit ho a pak odpocivat (co to jde) je lepsi nez nikam nechodit a odpocivat
                                 * - dojit pro poklad, vzit ho a pak odpocivat (co to jde) je lepsi nez jit na baterku a dobit se
                                 */
                                if ((sp[r2d2id][i][j] - goldmap[i][j]) < mindist
                                        && (remains - sp[r2d2id][i][j]-1)*(restenergy - 1) > imin(renergy[r2d2id], sp[r2d2id][i][j]+1) - gamemap[i][j]/(2*idtreasure) &&
                                        (cbl.isEmpty() || (remains - sp[r2d2id][i][j]-1)*(restenergy - 1) > maxenergy - gamemap[i][j]/(2*idtreasure)) &&
                                        ((remains - sp[r2d2id][i][j]-1)*(restenergy - 1) + gamemap[i][j]/(2*idtreasure) > imin(maxenergy,remains*(restenergy - 1)))) {
                                    mindist = sp[r2d2id][i][j] - goldmap[i][j];
                                    besti = i;
                                    bestj = j;
                                }
                            }
                            if (besti >= 0) {
                                /* Nastaveni cile. */
                                objective = new EvaluatedPoint(besti, bestj, -1, gamemap[besti][bestj]/(2*idtreasure));
                            }
                        } else {
                        /* Pokusim se dojit pro poklad, ktery maximalizuje rozdil mezi jeho hodnotou a vzdalenosti.
                         * Pokud maji vsechny poklady tento rozdil zaporny, budu strilet a odpocivat. */
                            for (Object o:rtl) {
                                Point p = (Point) o;
                                i = p.i;
                                j = p.j;
                                /* Minimalizuje se vzdalenost - hodnota goldmap
                                 * zaroven musi platit:
                                 * - dojit pro poklad, vzit ho a pak odpocivat (co to jde) je lepsi nez nikam nechodit a odpocivat
                                 * - dojit pro poklad, vzit ho a pak odpocivat (co to jde) je lepsi nez jit na baterku a dobit se
                                 */
                                if ((sp[r2d2id][i][j] - goldmap[i][j]) < mindist
                                        && (remains - sp[r2d2id][i][j]-1)*(restenergy - 1) > imin(renergy[r2d2id], sp[r2d2id][i][j]+1) - gamemap[i][j]/(2*idtreasure) &&
                                        (cbl.isEmpty() || (remains - sp[r2d2id][i][j]-1)*(restenergy - 1) > maxenergy - gamemap[i][j]/(2*idtreasure)) &&
                                        ((remains - sp[r2d2id][i][j]-1)*(restenergy - 1) + gamemap[i][j]/(2*idtreasure) > imin(maxenergy,remains*(restenergy - 1)))) {
                                    mindist = sp[r2d2id][i][j] - goldmap[i][j];
                                    besti = i;
                                    bestj = j;
                                }
                            }
                            if (besti >= 0) {
                                /* Nastaveni cile. Vaha bude mensi, protoze je cil nejisty. */
                                objective = new EvaluatedPoint(besti, bestj, -1, gamemap[besti][bestj]/(2*idtreasure)/UNSURE_OBJECTIVE);
                            }
                        }
                    }
                    if (objective == null) {
                        /* Volba mezi strilenim a baterii. */
                        /* Nejdriv si napocitam nejkratsi cestu k vystreleni */
                        int minshootingdistance = MAX_DIST;
                        int minshootingi = -1;
                        int minshootingj = -1;
                        int minshootingdir = -1;
                        int minshootingrid = -1;
                        int mincriterium = MAX_DIST; // kriteriem je vzdalenost a energie ciloveho robota
                        
                        for (int rid = 0; rid < nrobots; rid++) if (rid != r2d2id) {
                            for (int l = 0; l < 4; l++) {
                                int ii = rposi[rid];
                                int jj = rposj[rid];
                                int ub = alldiststowalls[ii][jj][l];
                                int cnt = 1;
                                while (cnt < ub) {
                                    ii += di[l];
                                    jj += dj[l];
                                    if (spdir[r2d2id][ii][jj][(l + 2)%4] - renergy[rid] < mincriterium) {
                                        mincriterium = spdir[r2d2id][ii][jj][(l + 2)%4] - renergy[rid];
                                        minshootingdistance = spdir[r2d2id][ii][jj][(l + 2)%4];
                                        minshootingi = ii;
                                        minshootingj = jj;
                                        minshootingdir = (l + 2)%4;
                                        minshootingrid = rid;
                                    }
                                    cnt++;
                                }
                            }
                        }
                        
                        if (verbose) {
                            System.out.println("minshootingdistance = " + minshootingdistance);
                            if (minshootingdistance != MAX_DIST) {
                                System.out.println("sprofit = " + shootingProfit(imax(1,renergy[r2d2id]-minshootingdistance), imax(1,renergy[minshootingrid]-minshootingdistance), remains - minshootingdistance));
                            }
                        }
                        
                        /* Nestrilej pokud nemas na koho nebo
                         * pokud jit k cili a strilet po nem neprinese tolik bodu jako stat a odpocivat
                         * a zaroven se po vyhledani cile nemuzes dobit na max. energii */
                        if (minshootingdistance == MAX_DIST ||
                                (shootingProfit(imax(1,renergy[r2d2id]-minshootingdistance), imax(1,renergy[minshootingrid]-minshootingdistance), remains - minshootingdistance) < maxenergy
                                && (remains - minshootingdistance)*(restenergy - 1) < maxenergy)) {
                            if (!rbl.isEmpty()) {
                                if (!cbl.isEmpty()) {
                                    /* Obsadim nejblizsi baterii a budu cekat do posledniho kola. */
                                    for (Object o:cbl) {
                                        Point p = (Point) o;
                                        i = p.i;
                                        j = p.j;
                                        if (sp[r2d2id][i][j] < mindist) {
                                            mindist = sp[r2d2id][i][j];
                                            besti = i;
                                            bestj = j;
                                        }
                                    }
                                    /* Nastaveni cile. */
                                    objective = new EvaluatedPoint(besti, bestj, -1, maxenergy);
                                } else {
                                    /* Pokusim se obsadit nejblizsi baterii a budu cekat do posledniho kola. */
                                    for (Object o:rbl) {
                                        Point p = (Point) o;
                                        i = p.i;
                                        j = p.j;
                                        if (sp[r2d2id][i][j] < mindist) {
                                            mindist = sp[r2d2id][i][j];
                                            besti = i;
                                            bestj = j;
                                        }
                                    }
                                    /* Nastaveni cile. */
                                    objective = new EvaluatedPoint(besti, bestj, -1, maxenergy);
                                }
                            }
                        } else {
                            /* Najdu si obet po ktere budu strilet */
                            objective = new EvaluatedPoint(minshootingi, minshootingj, minshootingdir, 1);
                        }
                    }
                }
                
                if (verbose) {
                    if (objective != null)
                        System.out.println("Objective = ("+objective.i+","+objective.j+") with value = "+objective.value);
                    else
                        System.out.println("Objective = No objective");
                }
                
                if (objective != null) {
                    //odebereme robota
                    gamemap[posi][posj] &= ~idrobot;
                    // musi se ohodnocovat i pole, kde jsou ostatni roboti
                    // s moznych tahu v alphabete se kolize vyresi samy (snad :))
                    // problem - pri vyse uvedene zmene by me mohl backward search zavest
                    // do ulicky zablokovane robotem, ale vyradit pole, na kterych jsou ostatni
                    // roboti jako uplne neperspektivni se mi uplne nechce.
                    BackwardSearch(objective.i, objective.j, objective.dir, -1, true, true);
                    
                    
                    //doplnime uplne okoli robota
                    for (int i = posi - maxalphabetarange; i <= posi + maxalphabetarange; i++) {
                        for (int j = posj - maxalphabetarange; j <= posj + maxalphabetarange; j++) {
                            if (isInMap(i, j) && (gamemap[i][j] & idwall) == 0) {
                                for (int rd = 0; rd < 4; rd++) {
                                    int rmindist = 100*width*height;
                                    for (int re = 1; re <= maxenergy; re++) {
                                        v = bfs[i][j][rd][re];
                                        if (v.visited && v.dist < rmindist) {
                                            rmindist = v.dist;
                                        } else {
                                            v.dist = rmindist;
                                            v.visited = true;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    
                    // zas ho tam pridame
                    gamemap[posi][posj] |= idrobot;
                    
                    v = bfs[posi][posj][dir][energy];
                    if (verbose) {
                        if (v.visited)
                            System.out.println("backward search successfull, mindist = " + v.dist);
                        else
                            System.out.println("backward search failed");
                    }
                } else {
                    /* Pokud nemam cil, necham alfu-betu at neco vymysli. */
                }
            }
            
            // vvvvvvvvvvvvvvv AlphaBeta algorithm vvvvvvvvvvvvvvvvvvvvv
            int startobjreached;
            if (objective != null) {
                if (rposi[r2d2id] == objective.i &&
                        rposj[r2d2id] == objective.j &&
                        (objective.dir < 0 ||
                        objective.dir == rdir[r2d2id])) {
                    startobjreached = 0;
                } else {
                    startobjreached = -1;
                }
            } else {
                startobjreached = -2;
            }
            lastr2d2rootmovesvalid = false;
            lastfinaldecisionint = -1;
            nrootr2d2possiblemoves = -1;
            timeoutemergencyexit = false;
            depth = imin(2*nrobots, nrobots*remains);
            while (!timeoutemergencyexit && depth/nrobots <= remains &&
                    depth/nrobots <= maxalphabetarange &&
                    depth/nrobots <= deadmaxalphabetarange) {
                if (verbose) System.out.println("Starting AlphaBeta with depth = " + depth);
                killeralphacutoffs = 0;
                killerbetacutoffs = 0;
                alphacutoffs = 0;
                betacutoffs = 0;
                betanocutoffs = 0;
                betanocutoffsspecial = 0;
                alphanodes = 0;
                betanodes = 0;
                generatemovescalled = 0;
                alphabetacalled = 0;
                evaluatecalled = 0;
                finaldecisionint = -1;
                timeoutpollingcnt = timeoutpollinginterval;
                for (int i = 0; i <= maxalphabetadepth; i++)
                    killermovesvalid[i] = false;
                AlphaBeta(depth, MIN_ALPHABETAEVAL, MAX_ALPHABETAEVAL, 0, true, true,
                        remains, 0, startobjreached, true);
                // I kdyz se alphabeta neustiha uplne, vezmeme jeji castecny vysledek, pokud uz nejaky ma.
                // Protoze tahy jsou usporadany z minule alphabety, muze se takto vybrat pouze stejny nebo
                // lepsi tah.
                if (finaldecisionint >= 0)
                    lastfinaldecisionint = finaldecisionint;
                if (!timeoutemergencyexit) {
                    lastr2d2rootmovesvalid = true;
                    if (verbose) {
                        System.out.println("finaldecisionint = " + finaldecisionint);
                        System.out.println("killeralphacutoffs = " + killeralphacutoffs);
                        System.out.println("killerbetacutoffs = " + killerbetacutoffs);
                        System.out.println("alphacutoffs = " + alphacutoffs);
                        System.out.println("betacutoffs = " + betacutoffs);
                        System.out.println("betanocutoffs = " + betanocutoffs);
                        System.out.println("betanocutoffsspecial = " + betanocutoffsspecial);
                        System.out.println("alphanodes = " + alphanodes);
                        System.out.println("betanodes = " + betanodes);
                        System.out.println("generatemovescalled = " + generatemovescalled);
                        System.out.println("alphabetacalled = " + alphabetacalled);
                        System.out.println("evaluatecalled = " + evaluatecalled);
                    }
                    depth += nrobots;
                }
            }
            finaldecisionint = lastfinaldecisionint;
            
            int nactuallydead = 0;
            for (int rid = 0; rid < nrobots; rid++) {
                if (deadrobots[rid] || renergy[rid] == 0)
                    nactuallydead++;
            }
            
            // pokud nebyl search oriznut pomoci deadmaxalphabetarange, tak
            // nastavime lastalphabetareachedrange
            if (deadmaxalphabetarange > (depth - nrobots)/nrobots)
                lastreachedalphabetarange[nactuallydead] = (depth - nrobots)/nrobots;
            
            if ((depth - nrobots)/nrobots < remains && (depth - nrobots)/nrobots < minreachedalphabetarange)
                minreachedalphabetarange = (depth - nrobots)/nrobots;
            
            if ((depth - nrobots)/nrobots > maxreachedalphabetarange)
                maxreachedalphabetarange = (depth - nrobots)/nrobots;
            // AAAAAAAAAAAAAA AlphaBeta algorithm AAAAAAAAAAAAAAAAAAAAAA
            
            
            action = Strategy.BotAction.REST;
            switch (finaldecisionint) {
                case movestep:
                    if (verbose) System.out.println("Krok");
                    action = Strategy.BotAction.STEP;
                    break;
                case moveleft:
                    if (verbose) System.out.println("Vlevo");
                    action = Strategy.BotAction.LEFT;
                    break;
                case moveright:
                    if (verbose) System.out.println("Vpravo");
                    action = Strategy.BotAction.RIGHT;
                    break;
                case moveshoot: {
                    if (verbose) System.out.println("Strela");
                    action = Strategy.BotAction.SHOOT;
                    break;
                }
                case movetake:
                    if (verbose) System.out.println("Brani");
                    action = Strategy.BotAction.TAKE;
                    break;
                case moverest:
                    if (verbose) System.out.println("Odpocivani");
                    action = Strategy.BotAction.REST;
                    break;
                case moverecharge:
                    if (verbose) System.out.println("Dobijeni");
                    action = Strategy.BotAction.RECHARGE;
                    break;
            }
            
            // System.gc();
            endtime = System.currentTimeMillis();
            if ((endtime - starttime) > maxdecisiontime)
                maxdecisiontime = (endtime - starttime);
            
            if (verbose) {
                System.out.println("Robot #" + r2d2id + ", roundno = " + roundno +
                        ", decision time = " + (endtime - starttime) + " ms");
                if (roundno == world.getRoundsCount()) {
                    System.out.println("maxdecisiontime = " + maxdecisiontime + " ms") ;
                    System.out.println("minreachedalphabetarange = " + minreachedalphabetarange);
                    System.out.println("maxreachedalphabetarange = " + maxreachedalphabetarange);
                }
            }
            return action;
        } catch (Exception e) {
            if (verbose) {
                System.err.println("EXCEPTION OCCURED");
            }
            return Strategy.BotAction.REST;
        }
    }
    
    
}
