package cz.muni.fi.fibot.resistanceisfutile;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import java.util.Collections;
import java.lang.Math;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedHashSet;

public class ResistanceIsFutile implements Strategy {

	private World world;
	
	private Robot robot;
	
	//coord je datovy typ pro uchovani souradnic	
	private class coord {
		int x,y;

		protected coord(int a, int b) {
			x = a;
			y = b;
		}
		
		protected coord(coord coordinates) {
			x = coordinates.x;
			y = coordinates.y;
		}
	}
	
	private class thing implements Comparable<thing> {
		coord pos; //souradnice objektu v poli
		float value; //hodnota objektu
		int initval; //puvodni hodnota objektu. Tyka se jen pokladu
		
		protected thing(int a, int b, int val) {
			pos = new coord(a,b);
			initval = val;
			value = val;
		}

		protected thing() {
			pos = new coord(0,0);
			value = -1;
			initval = -1;
		}
		
		protected thing(thing otherThing) {
			pos = new coord(otherThing.pos.x,otherThing.pos.y);
			value = otherThing.value;
			initval = otherThing.initval;
		}
		
		public int compareTo(thing otherThing) {
			return -(new Float(value)).compareTo(otherThing.value);
		}
	}

	private class valAndDir {
		double val;
		Robot.Direction dir;

		protected valAndDir() {
			val = 0;
		}
	}

	private class rangeAndDir {
		int val;
		LinkedHashSet<Robot.Direction> dirs;

		protected rangeAndDir() {
			val = 0;
			dirs = new LinkedHashSet<Robot.Direction>();
		}
	}

	private class RobotsDestiny {
		thing chosenObject;
		List<Strategy.BotAction> vision;

		protected RobotsDestiny() {
			chosenObject = new thing();
			vision = new ArrayList<Strategy.BotAction>();
		}
	}
	
	public void setWorld(World world) {
		this.world = world;
	}
	
	public void setRobot(Robot robot) { 
		this.robot = robot;
	} 
	
	private int dimX;
	private int dimY;
	
	Robot.Direction LEFT = Robot.Direction.LEFT; //pro kratsi kod
        Robot.Direction RIGHT = Robot.Direction.RIGHT;
	Robot.Direction UP = Robot.Direction.UP;
	Robot.Direction DOWN = Robot.Direction.DOWN;
	
	private List<thing> GoldList; //seznam pokladu
	private List<thing> BatList; //seznam baterii
	RobotsDestiny Dest;
	
	public ResistanceIsFutile() {
		GoldList = new ArrayList<thing> ();
		BatList = new ArrayList<thing> ();
		Dest = new RobotsDestiny ();
	}
	
	public void initialize() {
 		dimX = world.getDimensionX(); //definice konstant	
 		dimY = world.getDimensionY();	
		for(int i = 0; i != dimX; i++)
			for(int j = 0; j != dimY; j++) {
				if(world.containsTreasure(i,j))
					GoldList.add(new thing(i,j,world.getTreasureValue(i,j)));
				if(world.containsBattery(i,j)) BatList.add(new thing(i,j,0));
			}
		evalHuddleOfGoldP();
	}
	
	/* Vytvori instanci tridy "rangeAndDir" pro kazdou polozku pole. Hdnoty policek,
	 * na kterych se nachazi zed nebo robot inicializuje na -1*/
	private void initArray(rangeAndDir[][] array) {
		for(int i = 0; i != dimX; i++)
			for(int j = 0; j != dimY; j++) {
				array[i][j] = new rangeAndDir();
				if(world.containsWall(i,j)) array[i][j].val = -1;
			}
	}
	
	/* Najde pozici klice "key" ve 2D poli, ale jen pokud, policko
	 * na kterem klic lezi neni locknuty. Neni locknuty = odpovidajici 
	 * hodnota v "arrBool" je false*/
	private coord getKey(rangeAndDir[][] array, boolean[][] arrBool, int key) {
		coord pos = new coord(-1,-1);

		for(int x = 0; x != dimX; x++)
			for(int y = 0; y != dimY; y++)
				if(array[x][y].val == key && !arrBool[x][y]) {
					pos.x = x;
					pos.y = y;
				}
		return pos; //pri nenalezi odemceneho klice se vrati -1
	}
	
	//checkuje, zda-li jsou uz ohodnocena vsechna pristupna policka pole "array"
	private boolean Done(rangeAndDir[][] array) {
		for(int x = 0; x != dimX; x++)
			for(int y = 0; y != dimY; y++)
				if(array[x][y].val == 0) return false;
		return true;
	}
	
	//jen vypis pole "array" pro debug
	/*private void writeArray(rangeAndDir[][] array) {
		for(int y = 0; y != dimY; y++){
			for(int x = 0; x != dimX; x++) {
				System.out.print(array[x][y].val + " ");
				if(array[x][y].val < 10 && array[x][y].val >= 0) System.out.print(" ");
			}	
			System.out.print("\n");	
		}
		System.out.print("\n----\n");
	}
	*/

	//private void writeTreas(List<thing> GoldList) {
		
	/*	for(int i = 0; i < dimY; i++) {
			for(int j = 0; j < dimX; j++) {
				if(world.containsTreasure(j,i)) 
					System.out.printf("%4d",world.getTreasureValue(j,i));
				else System.out.printf("   O");
			}
			System.out.print("\n");
		}
	System.out.println("\n-----------");
	*/
/*	for(int i = 0; i < dimY; i++) {
			for(int j = 0; j < dimX; j++) {
				if(world.containsTreasure(j,i)) {
					for(int a = 0; a < GoldList.size(); a++) 
						if(j == GoldList.get(a).pos.x && GoldList.get(a).pos.y == i) {
							System.out.printf("%-5.2f",GoldList.get(a).value);
							break;
						}
				}
				else System.out.printf("  O  ");
			}
			System.out.print("\n");
		}
		System.out.println("\n***********");
	}
*/
	/* Vyplni 2D pole "array" cisly, vyjadrujicimi vzdalenost od pokladu.
	 * Vzdalenost je chapana jako pocet akci, ktere by musel robot vykonat,
	 * aby se z pocatecni pozice na toto misto dostal. Take vzeti pokladu je
	 * akce, a tak pole okolo startovni pozice jsou rovnou ohodnocena dvojkou.*/
	private void roundForRange(coord gold, rangeAndDir[][] array) {
		boolean[][] arrBool = new boolean[dimX][dimY]; //pro locknuti zpracovaneho policka
		coord pos = new coord(gold.x,gold.y);	
		int omega = 1;
		array[pos.x][pos.y].val = omega;
		
		if(pos.x + 1 < dimX && array[pos.x+1][pos.y].val >= 0) {
			array[pos.x+1][pos.y].val = omega + 1; 
			array[pos.x+1][pos.y].dirs.add(LEFT); 
		}
		
		if(pos.x > 0 && array[pos.x-1][pos.y].val >= 0) {
			array[pos.x-1][pos.y].val = omega + 1; 
			array[pos.x-1][pos.y].dirs.add(RIGHT); 
		}
		
		if(pos.y + 1 < dimY && array[pos.x][pos.y+1].val >= 0) {
			array[pos.x][pos.y+1].val = omega + 1; 
			array[pos.x][pos.y+1].dirs.add(UP); 
		}
		
		if(pos.y > 0 && array[pos.x][pos.y-1].val >= 0) {
			array[pos.x][pos.y-1].val = omega + 1; 
			array[pos.x][pos.y-1].dirs.add(DOWN); 
		}
		arrBool[pos.x][pos.y] = true;
		
		while (!Done(array)) {	
			omega++;
			pos = getKey(array,arrBool,omega);
	
			while(pos.x >= 0) {
	
				if (pos.x + 1 < dimX) {
					int i = array[pos.x][pos.y].dirs.contains(LEFT) ? omega + 1 : omega + 2;
					if ((array[pos.x + 1][pos.y].val > i) || (array[pos.x + 1][pos.y].val == 0)) {
						array[pos.x + 1][pos.y].val = i;
						array[pos.x + 1][pos.y].dirs.clear();
						array[pos.x + 1][pos.y].dirs.add(LEFT);
					} else if (array[pos.x + 1][pos.y].val == i)
						array[pos.x + 1][pos.y].dirs.add(LEFT);
				}
				if (pos.x > 0) {
					int i = array[pos.x][pos.y].dirs.contains(RIGHT) ? omega + 1 : omega + 2;
					if ((array[pos.x - 1][pos.y].val > i) || (array[pos.x - 1][pos.y].val == 0)) {
						array[pos.x - 1][pos.y].val = i;
						array[pos.x - 1][pos.y].dirs.clear();
						array[pos.x - 1][pos.y].dirs.add(RIGHT);
					} else if (array[pos.x - 1][pos.y].val == i)
						array[pos.x - 1][pos.y].dirs.add(RIGHT);
				}
				if (pos.y + 1 < dimY) {
					int i = array[pos.x][pos.y].dirs.contains(UP) ? omega + 1 : omega + 2;
					if ((array[pos.x][pos.y + 1].val > i) || (array[pos.x][pos.y + 1].val == 0)) {
						array[pos.x][pos.y + 1].val = i;
						array[pos.x][pos.y + 1].dirs.clear();
						array[pos.x][pos.y + 1].dirs.add(UP);
					} else if (array[pos.x][pos.y + 1].val == i)
						array[pos.x][pos.y + 1].dirs.add(UP);
				}
				if (pos.y > 0) {
					int i = array[pos.x][pos.y].dirs.contains(DOWN) ? omega + 1 : omega + 2;
					if ((array[pos.x][pos.y - 1].val > i) || (array[pos.x][pos.y - 1].val == 0)) {
						array[pos.x][pos.y - 1].val = i;
						array[pos.x][pos.y - 1].dirs.clear();
						array[pos.x][pos.y - 1].dirs.add(DOWN);
					} else if (array[pos.x][pos.y - 1].val == i)
						array[pos.x][pos.y - 1].dirs.add(DOWN);
				}
				
				arrBool[pos.x][pos.y] = true;
				pos = getKey(array,arrBool,omega);
			}
		}
		//writeArray(array);
	}

	//nic zvlastniho:), jen nuluje 2D pole
	private void makeZero(rangeAndDir[][] array) {
		for(int o = 0; o != dimX; o++)
			for(int p = 0; p != dimY; p++) {
				array[o][p].val = 0;
				if(world.containsWall(o,p)) array[o][p].val = -1;
			}
	}
	
	/* Pokud kterykoliv robot na mape vzal v min. kole poklad, je tento poklad
	 * odstranen z pole objektu, ktere na mape jsou umisteny.*/
	private void updateGoldList() {
		for(Robot mech: world.getRobots())
			if(mech.getLastAction() == BotAction.TAKE)
				for(int i = 0; i < GoldList.size(); i++) {
					if(mech.getPositionX() == GoldList.get(i).pos.x && 
						mech.getPositionY() == GoldList.get(i).pos.y) {
						evalHuddleOfGoldM(i);
						GoldList.remove(i);
					}
				}	
	}
	
	/* Bere v uvahu zvyseni hodnoty pokladu na zaklade blizkosti jinych pokladu
	 * (robot miri tam, kde je zlata nejvice). Nevztahuje se na baterie.*/
	private void evalHuddleOfGoldP() {
		rangeAndDir[][] array = new rangeAndDir[dimX][dimY];
		initArray(array);
		float hodnota;
		for(int i = 0; i != GoldList.size(); i++) {
			roundForRange(GoldList.get(i).pos,array);
			
			for(int j = 0; j != GoldList.size(); j++)
				if(i != j) {
	//GoldList.get(j).initval += ((float) GoldList.get(i).initval / array[GoldList.get(j).pos.x][GoldList.get(j).pos.y].val);
			int d = array[GoldList.get(j).pos.x][GoldList.get(j).pos.y].val;
			int v = GoldList.get(i).initval;
			GoldList.get(i).value += v / (1 + (Math.pow(Math.E, (((2 * (d - 0)) / Math.sqrt(world.getDimensionX() * world.getDimensionY())) + 1.0)) / Math.sqrt(2 * Math.PI)));
				/*	hodnota = 
					GoldList.get(i).initval + 
					(float)(1 - Math.pow(array[GoldList.get(j).pos.x][GoldList.get(j).pos.y].val, Math.PI/2));
					if(hodnota > 0) GoldList.get(j).value += hodnota;
				*/}
			makeZero(array);
		}
		//writeTreas(GoldList);
	}
	
	/* Jestlize nektery robot vykona akci TAKE a nejaky poklad zmizi z mapy,
	 * musi se odecist odpovidajici hodnota od pokladu, ktere jsou okolo nej.*/
	private void evalHuddleOfGoldM(int j) {
		rangeAndDir[][] array = new rangeAndDir[dimX][dimY];
		initArray(array);
		roundForRange(GoldList.get(j).pos,array); 
		float hodnota;	
		for(int i = 0; i != GoldList.size(); i++) {
			if(i != j) {
			GoldList.get(j).value -= GoldList.get(j).initval / (1 + (Math.pow(Math.E, (((2 * (array[GoldList.get(i).pos.x][GoldList.get(i).pos.y].val - 0)) / Math.sqrt(world.getDimensionX() * world.getDimensionY())) + 1.0)) / Math.sqrt(2 * Math.PI)));
/*				hodnota =
				GoldList.get(j).initval + 
				(float)(1 - Math.pow(array[GoldList.get(i).pos.x][GoldList.get(i).pos.y].val,Math.PI/2));
				if(hodnota > 0) GoldList.get(i).value -= hodnota;
*/	//GoldList.get(i).initval -= (float) GoldList.get(j).initval / array[GoldList.get(i).pos.x][GoldList.get(i).pos.y].val;
			}
		}
	}

	/* Do vzdalenosti robota od objektu vezme v uvahu i jeho natoceni vzhledem
	 * k oklnim proudum, tedy vzhledem ke smeru sipek na okolnich polickach. 
	 * Podminky nize osetruji pripad, kdy je robot u nektere ze sten. Napriklad robot
	 * je u spodni steny a je natocen nahoru(UP), pak ale nemuzeme pristupovat
	 * k poli, ktere lezi za nim, protoze tam ani zadne pole neni. Vsechny tyto
	 * moznosti tyto podminky osetruji. Upravi se jen hodnoty policek, ktere
	 * existuji. Navic se jeste hleda policko v okoli robota s nejmensi 
	 * hodnotou. Na konci se vrati prave koordinaty tohoto policka. Vzdalenost
	 * robota od pokladu se uklada do int[] realRage(jeden ze zpusobu, jak
	 * mit moznost menit original predaneho paramatru je, hodit ho do pole.*/
	private coord evalDirection(Robot Mech, rangeAndDir[][] array, int a, int b, int[] realRange) {
		int range = 2601;
		coord pos = new coord(Mech.getPositionX(),Mech.getPositionY());
		coord start = new coord(0,0);
		Robot.Direction dir = Mech.getDirection();
		realRange[0] = 2601; //tak tohle zacina byt me oblibene cislo:))
		//writeArray(array);
		//policka za robotem
		if(!(pos.x + 1 == dimX && dir == LEFT) && !(pos.x == 0 && dir == RIGHT) &&
			!(pos.y + 1 == dimY && dir == UP) && !(pos.y == 0 && dir == DOWN) &&
			!world.containsWall(pos.x+a,pos.y-b)) {
			
			if(array[pos.x+a][pos.y-b].val == 1) //tohle jde resit lip, to vis dobre..jsem uz unaveny...
				switch(dir) {
					case UP: array[pos.x+a][pos.y-b].dirs.add(DOWN); break;
					case DOWN: array[pos.x+a][pos.y-b].dirs.add(UP); break;
					case LEFT: array[pos.x+a][pos.y-b].dirs.add(RIGHT); break;
					case RIGHT: array[pos.x+a][pos.y-b].dirs.add(LEFT);
					}

			if(array[pos.x+a][pos.y-b].dirs.contains(rightOf(rightOf(dir))))
				range = array[pos.x+a][pos.y-b].val + 3;
			else if(array[pos.x+a][pos.y-b].dirs.contains(rightOf(dir)) || array[pos.x+a][pos.y-b].dirs.contains(leftOf(dir)))
				range = array[pos.x+a][pos.y-b].val + 4;
			else if(array[pos.x+a][pos.y-b].dirs.contains(dir)) 
				range = array[pos.x+a][pos.y-b].val + 5;
			
			start.x = pos.x + a;
			start.y = pos.y - b;
		}
		
		//policka nalevo od robota
		if(!(pos.x + 1 == dimX && dir == DOWN) && !(pos.x == 0 && dir == UP) &&
			!(pos.y + 1 == dimY && dir == LEFT) && !(pos.y == 0 && dir == RIGHT) && 
			!world.containsWall(pos.x+b,pos.y+a)) {
		
			if(array[pos.x+b][pos.y+a].val == 1)
				switch(dir) {
					case UP: array[pos.x+b][pos.y+a].dirs.add(LEFT); break;
					case DOWN: array[pos.x+b][pos.y+a].dirs.add(RIGHT); break;
					case LEFT: array[pos.x+b][pos.y+a].dirs.add(DOWN); break;
					case RIGHT: array[pos.x+b][pos.y+a].dirs.add(UP);
				}
			
			if(array[pos.x+b][pos.y+a].dirs.contains(leftOf(dir)))
				range = array[pos.x+b][pos.y+a].val + 2;
			else if(array[pos.x+b][pos.y+a].dirs.contains(dir) || array[pos.x+b][pos.y+a].dirs.contains(rightOf(rightOf(dir))))
				range = array[pos.x+b][pos.y+a].val + 3;
			else if(array[pos.x+b][pos.y+a].dirs.contains(rightOf(dir)))
				range = array[pos.x+b][pos.y+a].val + 4;

			if(range < realRange[0]) {
				realRange[0] = range;
				start.x = pos.x + b;
				start.y = pos.y + a;
			}
		}
		
		//policka napravo od robota
		if(!(pos.x + 1 == dimX && dir == UP) && !(pos.x == 0 && dir == DOWN) &&
			!(pos.y + 1 == dimY && dir == RIGHT) && !(pos.y == 0 && dir == LEFT) &&
			!world.containsWall(pos.x-b,pos.y-a)) {
				
			if(array[pos.x-b][pos.y-a].val == 1)
				switch(dir) {
					case UP: array[pos.x-b][pos.y-a].dirs.add(RIGHT); break;
					case DOWN: array[pos.x-b][pos.y-a].dirs.add(LEFT); break;
					case LEFT: array[pos.x-b][pos.y-a].dirs.add(UP); break;
					case RIGHT: array[pos.x-b][pos.y-a].dirs.add(DOWN);
				}
			
			if(array[pos.x-b][pos.y-a].dirs.contains(rightOf(dir)))
				range = array[pos.x-b][pos.y-a].val + 2;
			else if(array[pos.x-b][pos.y-a].dirs.contains(dir) || array[pos.x-b][pos.y-a].dirs.contains(rightOf(rightOf(dir))))
				range = array[pos.x-b][pos.y-a].val + 3;
			else if(array[pos.x-b][pos.y-a].dirs.contains(leftOf(dir)))
				range = array[pos.x-b][pos.y-a].val + 4;
	
			if(range < realRange[0]) {
				realRange[0] = range;
				start.x = pos.x - b;
				start.y = pos.y - a;
			}
		}

		//policka pred robotem
		if(!(pos.x + 1 == dimX && dir == RIGHT) && !(pos.x == 0 && dir == LEFT) &&
			!(pos.y + 1 == dimY && dir == DOWN) && !(pos.y == 0 && dir == UP) &&
			!world.containsWall(pos.x-a,pos.y+b)) {
			
			if(array[pos.x-a][pos.y+b].val == 1)
				switch(dir) {
					case UP: array[pos.x-a][pos.y+b].dirs.add(UP); break;
					case DOWN: array[pos.x-a][pos.y+b].dirs.add(DOWN); break;
					case LEFT: array[pos.x-a][pos.y+b].dirs.add(LEFT); break;
					case RIGHT: array[pos.x-a][pos.y+b].dirs.add(RIGHT);
				}
			
			if(array[pos.x-a][pos.y+b].dirs.contains(dir))
				range = array[pos.x-a][pos.y+b].val + 1;
			else if(array[pos.x-a][pos.y+b].dirs.contains(rightOf(dir)) || array[pos.x-a][pos.y+b].dirs.contains(leftOf(dir)))
				range = array[pos.x-a][pos.y+b].val + 2;
			else if(array[pos.x-a][pos.y+b].dirs.contains(rightOf(rightOf(dir))))
				range = array[pos.x-a][pos.y+b].val + 3;

			if(range < realRange[0]) {
				realRange[0] = range;
				start.x = pos.x - a;
				start.y = pos.y + b;
			}
		}
		//writeArray(array);	
		return start;
	}
	
	/* Slouzi jako predstupen k evalDirection. To proto, ze velkou potiz
	 * dela pritomnost robota u nejake ze sten. Clovek by mohl lehce pristoupit
	 * mimo pole "array". Tedy 3. a 4. argument fce evalDireciton(...) rika
	 * jake hodnoty se maji pricitat k souradnicim "y" a "x" pri urcitem natoceni*/	
	private coord findOptimalStart(Robot Mech, rangeAndDir[][] array, int[] realRange) {
		Robot.Direction dir = Mech.getDirection();
		coord start = new coord(Mech.getPositionX(),Mech.getPositionY());

		if(array[start.x][start.y].val != 1) {
			switch(dir) {
				case LEFT:   start = evalDirection(Mech,array,1,0,realRange); break;
				case UP:     start = evalDirection(Mech,array,0,-1,realRange); break;
				case RIGHT:  start = evalDirection(Mech,array,-1,0,realRange); break;
				case DOWN:   start = evalDirection(Mech,array,0,1,realRange);
			}			 
		}
		else realRange[0] = 1;
			
		return start;	
	}	

	public Robot lookUP() {
		for(int i=robot.getPositionY()-1; i>=0 && !world.containsWall(robot.getPositionX(),i); i--)
			if(world.containsRobot(robot.getPositionX(),i)) {
				for(Robot mech:world.getRobots()) 
					if(mech.getPositionX()==this.robot.getPositionX() && mech.getPositionY()==i) 
						return mech;
		}
		return null;
	}
	
	public Robot lookDOWN() {
			for(int i=robot.getPositionY()+1; i<dimY && !world.containsWall(robot.getPositionX(),i); i++)
				if(world.containsRobot(robot.getPositionX(),i)) {
					for(Robot mech:world.getRobots()) 
						if(mech.getPositionX()==this.robot.getPositionX() && mech.getPositionY()==i) 
							return mech;
		}
		return null;
	}
	
	public Robot lookLEFT() {
			for(int i=robot.getPositionX()-1; i>=0 && !world.containsWall(i,robot.getPositionY()); i--)
				if(world.containsRobot(i,robot.getPositionY())) {
					for(Robot mech:world.getRobots()) 
						if(mech.getPositionX()==i && mech.getPositionY()==this.robot.getPositionY()) 
							return mech;
		}
		return null;
	}
	
	public Robot lookRIGHT() {
			for(int i=robot.getPositionX()+1; i<dimX && !world.containsWall(i,robot.getPositionY()); i++)
				if(world.containsRobot(i,robot.getPositionY())) {
					for(Robot mech:world.getRobots())
						if(mech.getPositionX()==i && mech.getPositionY()==this.robot.getPositionY()) 
							return mech;
		}
		return null;
	}

	private boolean inFrontRobot(Robot r) {
		if (r == null)
			return false;
		int x = r.getPositionX();
		int y = r.getPositionY();
		switch (r.getDirection()) {
			case LEFT:
				if ((x == 0) || (world.containsRobot(x - 1, y)))
					return true;
				break;
			case RIGHT:
				if ((x == world.getDimensionX() - 1) || (world.containsRobot(x + 1, y)))
					return true;
				break;
			case UP:
				if ((y == 0) || (world.containsRobot(x, y - 1)))
					return true;
				break;
			case DOWN:
				if ((y == world.getDimensionY() - 1) || (world.containsRobot(x, y + 1)))
					return true;
				break;

		}
		return false;
	}
	
	private boolean inFrontWall(int x, int y, Robot.Direction d) {
		switch (d) {
			case LEFT:
				if ((x == 0) || (world.containsWall(x - 1, y)))
					return true;
				break;
			case RIGHT:
				if ((x == world.getDimensionX() - 1) || (world.containsWall(x + 1, y)))
					return true;
				break;
			case UP:
				if ((y == 0) || (world.containsWall(x, y - 1)))
					return true;
				break;
			case DOWN:
				if ((y == world.getDimensionY() - 1) || (world.containsWall(x, y + 1)))
					return true;
				break;

		}
		return false;
	}
	private boolean inFrontWall(Robot r) {
		if (r == null)
			return false;
		int x = r.getPositionX();
		int y = r.getPositionY();
		switch (r.getDirection()) {
			case LEFT:
				if ((x == 0) || (world.containsWall(x - 1, y)))
					return true;
				break;
			case RIGHT:
				if ((x == world.getDimensionX() - 1) || (world.containsWall(x + 1, y)))
					return true;
				break;
			case UP:
				if ((y == 0) || (world.containsWall(x, y - 1)))
					return true;
				break;
			case DOWN:
				if ((y == world.getDimensionY() - 1) || (world.containsWall(x, y + 1)))
					return true;
				break;

		}
		return false;
	}

	private boolean inFront(Robot r) {
		if (r == null)
			return false;
		int x = r.getPositionX();
		int y = r.getPositionY();
		switch (r.getDirection()) {
			case LEFT:
				if ((x == 0) || (world.containsWall(x - 1, y)) || (world.containsRobot(x - 1, y)))
					return true;
				break;
			case RIGHT:
				if ((x == world.getDimensionX() - 1) || (world.containsWall(x + 1, y)) || (world.containsRobot(x + 1, y)))
					return true;
				break;
			case UP:
				if ((y == 0) || (world.containsWall(x, y - 1)) || (world.containsRobot(x, y - 1)))
					return true;
				break;
			case DOWN:
				if ((y == world.getDimensionY() - 1) || (world.containsWall(x, y + 1)) || (world.containsRobot(x, y + 1)))
					return true;
				break;

		}
		return false;
	}

	public valAndDir vratMaxLook() {
		valAndDir max = new valAndDir();
		Robot[] sortMap = { lookLEFT(),lookUP(),lookRIGHT(),lookDOWN() } ;
		Robot.Direction d = robot.getDirection();

		max.val = 0;

		Robot r = sortMap[d.ordinal()];
		if (r != null) {
			max.dir = d;
			max.val = ((r.getEnergy() / 2) - 1);
		}

		double v;
		
		r = sortMap[leftOf(d).ordinal()];
		if (r != null) {
			if ((r.getDirection() == d) || (Math.abs(r.getDirection().ordinal() - d.ordinal()) == 2) || inFront(r))
				v = (r.getEnergy() / 2) - 3;
			else
				v = r.getEnergy() / 4;
			if (v > max.val) {
				max.dir = leftOf(d);
				max.val = v;
			}
		}

		r = sortMap[rightOf(d).ordinal()];
		if (r != null) {
			if ((r.getDirection() == d) || (Math.abs(r.getDirection().ordinal() - d.ordinal()) == 2) || inFront(r))
				v = (r.getEnergy() / 2) - 3;
			else
				v = r.getEnergy() / 4;
			if (v > max.val) {
				max.dir = rightOf(d);
				max.val =  v;
			}
		}

		r = sortMap[rightOf(rightOf(d)).ordinal()];
		if (r != null) {
			v = r.getEnergy() / 4;
			if (v > max.val) {
				max.dir = rightOf(d);
				max.val =  v;
			}
		}

		return max;
	}

	/* Zhodnocuje pozici robota vuci pokladum a take pozici nejblizsiho nepratelskeho 
	 * robota vuci pokladu. Tedy cim bliz jsem pokladu tim ma pro me poklad relativni 
	 * vetsi cenu. Cim dal jsem, tim mensi. Naopak je tomu u nepratelskeho robota
	 * (uvazuje se jen ten nejmene vzdaleny pokladu). Vraci nejhodnotnejsi poklad.*/
	private thing evalTreasures() {
		rangeAndDir[][] array = new rangeAndDir[dimX][dimY];
		initArray(array);
		int[] realRange = new int[1];
		int minEnemyRange = 2601; 
		List<thing> GList = new ArrayList<thing> ();	
		for (int i = 0; i != GoldList.size(); i++) GList.add(new thing(GoldList.get(i))); //kopirovani GoldListu:)
		float difference, apocalypsa; 	
		
		for(int i = 0; i != GList.size(); i++) {
			roundForRange(GList.get(i).pos,array);
			findOptimalStart(robot,array,realRange);
			//System.out.printf("rr %d @ (%d, %d)\n", realRange[0], GList.get(i).pos.x, GList.get(i).pos.y);
			GList.get(i).value /= 1 + (Math.pow(Math.E, (((2 * (realRange[0] - 0)) / Math.sqrt(world.getDimensionX() * world.getDimensionY())) + 1.0)) / Math.sqrt(2 * Math.PI));
			
	//		for(Robot Bot: world.getRobots()) //nalezani pokladu nejblizsiho robota
	//			if(Bot.getId() != robot.getId()) {
	//				findOptimalStart(Bot,array,realRange); 
	//				if(realRange[0] < minEnemyRange) minEnemyRange = realRange[0];
	//			}
	//		if(minEnemyRange != 2601) {	
	//			GList.get(i).value *= 1 + (Math.pow(Math.E, (((2 * (minEnemyRange - 0)) / Math.sqrt(world.getDimensionX() * world.getDimensionY())) + 1.0)) / Math.sqrt(2 * Math.PI));
				// apocalypsa = GList.get(i).value + (1 - (float) Math.pow(minEnemyRange,Math.PI/2));	
				// if(apocalypsa > 0) GList.get(i).value -= apocalypsa;
	//		}
			makeZero(array);
		}	
	//	writeTreas(GoldList);	
	//	writeTreas(GList);
		Collections.sort(GList);
		return GList.get(0);
	}
	
	/* Ohodnocuje baterie a vybira nejhodnotnejsi z nich, kterou vraci.*/
	private thing evalBatteries() {
		rangeAndDir[][] array = new rangeAndDir[dimX][dimY];
		initArray(array);
		int[] realRange = new int[1];
		
		for(int i = 0; i != BatList.size(); i++) {
			roundForRange(BatList.get(i).pos,array);
			findOptimalStart(robot,array,realRange);
			
			BatList.get(i).value = 
				(world.getMaxRobotEnergy() - robot.getEnergy() - 
				(realRange[0] * world.getRestRobotEnergy())) * 
				((dimX + dimY) / (4 * (BatList.size() + world.getRobots().size()))); 
			
			if(BatList.get(i).value < 0) BatList.get(i).value = 0;
			makeZero(array);
		}
		
		Collections.sort(BatList);
		return BatList.get(0);
	}
	
	/* Pro kazdeho robota odhadne jeho predpokladany cil(baterie/poklad).
	 * Tato fce je predstupen pro analyzu cesty k predpokladanemu cili*/
	private boolean beginToSee() {
		boolean anyObject = false;
		
		if(GoldList.size() > 0) {	
			Dest.chosenObject = evalTreasures();
			anyObject = true;
		}
		
		if(BatList.size() > 0) {
			thing battery = evalBatteries();
			if(battery.value > Dest.chosenObject.value) Dest.chosenObject = battery;
			anyObject = true;
		}
		
		return anyObject;
	}
	
	private Robot.Direction goToStream(coord dest) {
		coord pos = new coord(robot.getPositionX(),robot.getPositionY());
		Robot.Direction dir = robot.getDirection();
		boolean goOn = true;
		
		while(goOn) {
			
			if(pos.x == dest.x && pos.y == dest.y) break;
		
			if((dir == UP && pos.x - 1 == dest.x) || (dir == DOWN && pos.x + 1 == dest.x) ||
				(dir == LEFT && pos.y + 1 == dest.y) || (dir == RIGHT && pos.y - 1 == dest.y)) {
				
				Dest.vision.add(BotAction.LEFT);
				switch(dir) {
					case UP: dir = LEFT; break;
					case DOWN: dir = RIGHT; break;
					case LEFT: dir = DOWN; break;
					case RIGHT: dir = UP;
				}
			}
			
			if((dir == UP && pos.x + 1 == dest.x) || (dir == DOWN && pos.x - 1 == dest.x) ||
				(dir == LEFT && pos.y - 1 == dest.y) || (dir == RIGHT && pos.y + 1 == dest.y)) {
				
				Dest.vision.add(BotAction.RIGHT);
				switch(dir) {
					case UP: dir = RIGHT; break;
					case DOWN: dir = LEFT; break;
					case LEFT: dir = UP; break;
					case RIGHT: dir = DOWN;
				}
			}

			if((dir == UP && pos.y + 1 == dest.y) || (dir == DOWN && pos.y - 1 == dest.y) ||
				(dir == LEFT && pos.x + 1 == dest.x) || (dir == RIGHT && pos.x - 1 == dest.x)) { 
				
				Dest.vision.add(BotAction.LEFT); 
				switch(dir) {
					case UP: dir = LEFT; break;
					case DOWN: dir = RIGHT; break;
					case LEFT: dir = DOWN; break;
					case RIGHT: dir = UP;
				}
			}
		
			if((dir == UP && pos.y - 1 == dest.y) || (dir == DOWN && pos.y + 1 == dest.y) ||
				(dir == LEFT && pos.x - 1 == dest.x) || (dir == RIGHT && pos.x + 1 == dest.x)) {
				
				Dest.vision.add(BotAction.STEP);
				goOn = false;
			}
		}
		return dir; 
	}
	
	private void followStream(rangeAndDir[][] array, coord position, Robot.Direction dir) {
		coord pos = new coord(position);
		
		while(true) {
			if(pos.x == Dest.chosenObject.pos.x && pos.y == Dest.chosenObject.pos.y) break;

			if(array[pos.x][pos.y].dirs.contains(dir)) {
				Dest.vision.add(BotAction.STEP);
				switch(dir) {
					case UP: pos.y--; break;
					case DOWN: pos.y++; break;
					case LEFT: pos.x--; break;
					case RIGHT: pos.x++;
				}
			} else if(array[pos.x][pos.y].dirs.contains(leftOf(dir))) {
				Dest.vision.add(BotAction.LEFT);
				dir = leftOf(dir);
			} else if(array[pos.x][pos.y].dirs.contains(rightOf(dir))) {
				Dest.vision.add(BotAction.RIGHT);
				dir = rightOf(dir);
			}
		}
	}	
	
	private void forecastActions() {
		rangeAndDir[][] array = new rangeAndDir[dimX][dimY];
		initArray(array);
		int[] realRange = new int[1];
		Robot.Direction dir;
		
		roundForRange(Dest.chosenObject.pos,array);
		//writeArray(array); 
		coord dest = findOptimalStart(robot,array,realRange);
		dir = goToStream(dest);
		followStream(array,dest,dir);
	}
	
	private Robot.Direction leftOf(Robot.Direction d) {
		switch (d) {
			case LEFT: return DOWN;
			case UP: return LEFT;
			case DOWN: return RIGHT;
		}
		return UP; /* case RIGHT */
	}

	private Robot.Direction rightOf(Robot.Direction d) {
		switch (d) {
			case RIGHT: return DOWN;
			case DOWN: return LEFT;
			case UP: return RIGHT;
		}
		return UP; /* case LEFT */
	}

	private boolean shootable() {
		Robot r;
		if (((r = lookUP()) != null) && (r.getDirection() == DOWN))
			return true;
		if (((r = lookDOWN()) != null) && (r.getDirection() == UP))
			return true;
		if (((r = lookRIGHT()) != null) && (r.getDirection() == LEFT))
			return true;
		if (((r = lookLEFT()) != null) && (r.getDirection() == RIGHT))
			return true;
		return false;
	}

	public BotAction decide() {

		updateGoldList();
		if(robot.getEnergy() < 2) return ((world.containsBattery(robot.getPositionX(), robot.getPositionY()) && (!shootable())) ? BotAction.RECHARGE : BotAction.REST);
		
		if(Dest.chosenObject.value >= 0 && robot.getPositionX() == Dest.chosenObject.pos.x &&
			robot.getPositionY() == Dest.chosenObject.pos.y) {
			
			Dest.chosenObject.value = -1; //huch?
			if(world.containsTreasure(robot.getPositionX(),robot.getPositionY())) return BotAction.TAKE;
			if(world.containsBattery(robot.getPositionX(),robot.getPositionY()) && (!shootable())) return BotAction.RECHARGE;
		}

		if(Dest.vision.isEmpty() || !(
		world.containsTreasure(Dest.chosenObject.pos.x, Dest.chosenObject.pos.y)
		|| world.containsBattery(Dest.chosenObject.pos.x, Dest.chosenObject.pos.y)
		)) {
			Dest.vision.clear();
			if(beginToSee()) forecastActions();
		}
				
		valAndDir rd = vratMaxLook();
		float v = Dest.chosenObject.value;
		int d = Dest.vision.size();
		Robot.Direction dir = robot.getDirection();

		if (rd.val > (v / (1 + (Math.pow(Math.E, (((2 * (d - 0)) / Math.sqrt(world.getDimensionX() * world.getDimensionY())) + 1.0)) / Math.sqrt(2 * Math.PI))))) {
				if (dir == rd.dir)
					return BotAction.SHOOT;
				if (rightOf(dir) == rd.dir) {
					Dest.vision.clear();
					return BotAction.RIGHT;
				}
				if (leftOf(dir) == rd.dir) {
					Dest.vision.clear();
					return BotAction.LEFT;
				}
				return BotAction.LEFT;
		}

		BotAction action = BotAction.REST;
		if(!Dest.vision.isEmpty()) {
			action = Dest.vision.get(0);
			if ((action == BotAction.STEP) && inFrontRobot(robot))
				return BotAction.SHOOT;

			Dest.vision.remove(0);
		}
		return action; //konecne rozhodnuti:))
	}
}
