package cz.muni.fi.fibot.robik;

import com.sun.org.apache.xerces.internal.impl.xpath.regex.Match;
import java.util.PriorityQueue;
import java.util.Random;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Stack;

public class Robik implements Strategy {

    protected World world;
    protected Robot robot;
    private int state;
    private Random random;
    protected Set<Position> globalBatteries;
    protected Set<Treasure> globalTreasures;
    protected int activeRobots;
    static int MAX_EXPANSIONS = 10;

    public Robik() {
        state = 2;
        activeRobots = 0;
        this.random = new Random();
    }

    public void setWorld(World world) {
        //System.out.print("Nastavuji svet Robikovi...");
        this.world = world;
    }

    public void setRobot(Robot robot) {
        this.robot = robot;
    }
    
    public void initialize() {
        //System.out.print("Volam initialize...");
        activeRobots = world.getRobots().size();
        globalBatteries = new HashSet<Position>();
        globalTreasures = new HashSet<Treasure>();
        for(int i = 0; i < world.getDimensionX(); i++)
            for (int j = 0; j < world.getDimensionY(); j++) {
                if (world.containsBattery(i,j))
                    globalBatteries.add(new Position(i,j));
                if (world.containsTreasure(i,j)) {
                    Treasure newTreasure = new Treasure(i,j,world);
                    newTreasure.countDistances(world);
                    globalTreasures.add(newTreasure);
                }
            }
        //System.out.println("Hotovo");
        
        // Pokusny vypis vzdalenosti
        /*try {
            PrintWriter pw = new PrintWriter(new File("C:/pokus.txt"));
            for(Treasure t: globalTreasures){
                pw.println("Poklad na souradnicich (" + t.positionX + "," + t.positionY + ")");
                for(int j=0; j < world.getDimensionY(); j++){
                    pw.print("\t");
                    for(int i=0; i < world.getDimensionX(); i++)
                        pw.print(((t.distances[i][j][Direction.UP.ordinal()] == Integer.MAX_VALUE)? "#": t.distances[i][j][Direction.UP.ordinal()]) + "\t\t\t");
                    pw.println();
                    for(int i=0; i < world.getDimensionX(); i++)
                        pw.print(((t.distances[i][j][Direction.LEFT.ordinal()] == Integer.MAX_VALUE)? "#": t.distances[i][j][Direction.LEFT.ordinal()]) + "\t\t" + ((t.distances[i][j][Direction.RIGHT.ordinal()] == Integer.MAX_VALUE)? "#" : t.distances[i][j][Direction.RIGHT.ordinal()]) + "\t");
                    pw.println();
                    pw.print("\t");
                    for(int i=0; i < world.getDimensionX(); i++)
                        pw.print(((t.distances[i][j][Direction.DOWN.ordinal()] == Integer.MAX_VALUE)? "#": t.distances[i][j][Direction.DOWN.ordinal()]) + "\t\t\t");
                    pw.println();
                }
                    
            }
            pw.close();
        } catch (IOException e) {
            System.out.println("Soubor pokus.txt se nepodarilo vytvorit!");
            return;
        }*/
    }
    
    public Strategy.BotAction decide() {

        //System.out.println("Vytvarim strom.");
        StateTree tree = new StateTree(this);        
        
        if (this.world.getRestRobotEnergy() > 0 && this.robot.getEnergy() < 2
            && this.robot.getEnergy() < this.world.getMaxRobotEnergy()) {
            return BotAction.REST;
	}
	if(world.containsTreasure(robot.getPositionX(),robot.getPositionY())) {
            //System.out.println("Sbiram poklad v hodnote" + world.getTreasureValue(robot.getPositionX(),robot.getPositionY()));
            return BotAction.TAKE;
        }
        BotAction action = tree.bfs();
        //System.out.println("Strom vratil akci " + action);
        return action;
    }

}

class Position {
    protected int positionX;
    protected int positionY;

    public Position(int x, int y) {
        positionX = x;
        positionY = y;
    }
    
    public boolean equals(Object obj) {
        if(obj instanceof Position) {
            Position p = (Position)obj;
            return this.positionX == p.positionX && this.positionY == p.positionY;
        } else return false;
    }
    
    public int hashCode() {
        return this.positionX * this.positionY;
    }
}

class Treasure extends Position{
    protected int[][][] distances;
    protected int value;
    
    private class TempDist {
        protected Position position;
        protected int distance;
        protected Direction dir;
        
        public TempDist(int x, int y, int dist, Direction dir) {
            position = new Position(x,y);
            distance = dist;
            this.dir = dir;
        }
    }
    
    public Treasure(int x, int y, World w) {
        super(x,y);
        distances = new int[w.getDimensionX()][w.getDimensionY()][Direction.values().length];
        this.value = w.getTreasureValue(x,y);
    }
    
    public void countDistances(World w) {
        //System.out.print("Zacinam pocitat vzdalenosti pokladu...");
        for(int i = 0; i < w.getDimensionX(); i++)
            for(int j = 0; j < w.getDimensionY(); j++)
                for(int k = 0; k < Direction.values().length; k++)
                    distances[i][j][k] = Integer.MAX_VALUE;
        List<TempDist> DistCount = new ArrayList<TempDist>();
        DistCount.add(new TempDist(positionX, positionY, 0, Direction.UP));
        distances[positionX][positionY][Direction.UP.ordinal()] = 0;
        DistCount.add(new TempDist(positionX, positionY, 0, Direction.RIGHT));
        distances[positionX][positionY][Direction.RIGHT.ordinal()] = 0;
        DistCount.add(new TempDist(positionX, positionY, 0, Direction.DOWN));
        distances[positionX][positionY][Direction.DOWN.ordinal()] = 0;
        DistCount.add(new TempDist(positionX, positionY, 0, Direction.LEFT));
        distances[positionX][positionY][Direction.LEFT.ordinal()] = 0;
        while(!DistCount.isEmpty()) {
            TempDist td = DistCount.remove(0);
            switch(td.dir){
                case UP: if((td.position.positionY + 1 < w.getDimensionY()) && !w.containsWall(td.position.positionX, td.position.positionY + 1) &&
                            distances[td.position.positionX][td.position.positionY + 1][Direction.UP.ordinal()] > td.distance + 1){
                                distances[td.position.positionX][td.position.positionY + 1][Direction.UP.ordinal()] = td.distance + 1;
                                DistCount.add(new TempDist(td.position.positionX, td.position.positionY + 1,td.distance + 1, Direction.UP));
                         }
                         if(distances[td.position.positionX][td.position.positionY][Direction.LEFT.ordinal()] > td.distance + 1) {
                            distances[td.position.positionX][td.position.positionY][Direction.LEFT.ordinal()] = td.distance + 1;
                            DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.LEFT));
                         }
                         if(distances[td.position.positionX][td.position.positionY][Direction.RIGHT.ordinal()] > td.distance + 1) {
                            distances[td.position.positionX][td.position.positionY][Direction.RIGHT.ordinal()] = td.distance + 1;
                            DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.RIGHT));
                         }
                         break;
                case RIGHT: if((td.position.positionX - 1 >= 0) && !w.containsWall(td.position.positionX - 1, td.position.positionY) &&
                                distances[td.position.positionX - 1][td.position.positionY][Direction.RIGHT.ordinal()] > td.distance + 1){
                                distances[td.position.positionX - 1][td.position.positionY][Direction.RIGHT.ordinal()] = td.distance + 1;
                                DistCount.add(new TempDist(td.position.positionX - 1, td.position.positionY,td.distance + 1, Direction.RIGHT));
                            }
                            if(distances[td.position.positionX][td.position.positionY][Direction.UP.ordinal()] > td.distance + 1) {
                                distances[td.position.positionX][td.position.positionY][Direction.UP.ordinal()] = td.distance + 1;
                                DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.UP));
                            }
                            if(distances[td.position.positionX][td.position.positionY][Direction.DOWN.ordinal()] > td.distance + 1) {
                                distances[td.position.positionX][td.position.positionY][Direction.DOWN.ordinal()] = td.distance + 1;
                                DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.DOWN));
                            }
                            break;
                case DOWN: if((td.position.positionY - 1 >= 0) && !w.containsWall(td.position.positionX, td.position.positionY - 1) &&
                                distances[td.position.positionX][td.position.positionY - 1][Direction.DOWN.ordinal()] > td.distance + 1){
                                distances[td.position.positionX][td.position.positionY - 1][Direction.DOWN.ordinal()] = td.distance + 1;
                                DistCount.add(new TempDist(td.position.positionX, td.position.positionY - 1,td.distance + 1, Direction.DOWN));
                           }
                           if(distances[td.position.positionX][td.position.positionY][Direction.LEFT.ordinal()] > td.distance + 1) {
                               distances[td.position.positionX][td.position.positionY][Direction.LEFT.ordinal()] = td.distance + 1;
                               DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.LEFT));
                           }
                           if(distances[td.position.positionX][td.position.positionY][Direction.RIGHT.ordinal()] > td.distance + 1) {
                               distances[td.position.positionX][td.position.positionY][Direction.RIGHT.ordinal()] = td.distance + 1;
                               DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.RIGHT));
                           }
                           break;
                case LEFT: if((td.position.positionX + 1 < w.getDimensionX()) && !w.containsWall(td.position.positionX + 1, td.position.positionY) &&
                                distances[td.position.positionX + 1][td.position.positionY][Direction.LEFT.ordinal()] > td.distance + 1){
                                distances[td.position.positionX + 1][td.position.positionY][Direction.LEFT.ordinal()] = td.distance + 1;
                                DistCount.add(new TempDist(td.position.positionX + 1, td.position.positionY,td.distance + 1, Direction.LEFT));
                           }
                           if(distances[td.position.positionX][td.position.positionY][Direction.UP.ordinal()] > td.distance + 1) {
                               distances[td.position.positionX][td.position.positionY][Direction.UP.ordinal()] = td.distance + 1;
                               DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.UP));
                           }
                           if(distances[td.position.positionX][td.position.positionY][Direction.DOWN.ordinal()] > td.distance + 1) {
                               distances[td.position.positionX][td.position.positionY][Direction.DOWN.ordinal()] = td.distance + 1;
                               DistCount.add(new TempDist(td.position.positionX, td.position.positionY, td.distance + 1, Direction.DOWN));
                           }
                           break;
            }
        }
        //System.out.println("Hotovo.");
    }
        
    public boolean equals(Object obj) {
        if(obj instanceof Treasure) {
            Treasure t = (Treasure)obj;
            return this.positionX == t.positionX && this.positionY == t.positionY;
        }
        else return false;
    }

    public int hashCode() {
        return this.positionX * this.positionY;
    }
}

class OurRobot implements Cloneable{
    protected Robot robot;
    protected Direction direction;
    protected int positionX;
    protected int positionY;
    protected int energy;
    protected int points;
    
    public OurRobot() {
        
    }
    
    protected Object clone() {
        OurRobot newRobot = new OurRobot();
        newRobot.robot = this.robot;
        newRobot.direction = this.direction;
        newRobot.positionX = this.positionX;
        newRobot.positionY = this.positionY;
        newRobot.energy = this.energy;
        newRobot.points = this.points;
        return newRobot;
    }
}

class StateNode implements Cloneable, Comparable{
    protected List<OurRobot> robots;
    protected Set<Treasure> treasures;
    int level;
    double value;
    BotAction firstMove;
    
    public StateNode(int level, BotAction firstMove){
        robots = new ArrayList<OurRobot>();
        treasures = new HashSet<Treasure>();
        this.level = level;
        this.value = Double.MAX_VALUE;
        this.firstMove = firstMove;
    }
    
    public boolean containsTreasure(int x, int y) {
        for(Treasure t : treasures) {
            if(t.positionX == x && t.positionY == y) return true;
        }
        return false;
    }
    
    public Treasure getTreasure(int x, int y) {
        for(Treasure t : treasures) {
            if(t.positionX == x && t.positionY == y) return t;
        }
        return null;        
    }
    
    public OurRobot getOurRobot(int x, int y) {
        for(OurRobot r : robots) {
            if(r.positionX == x && r.positionY == y) return r;
        }
        return null;
    }
    
    public OurRobot getOurRobot(Robot robot) {
        for(OurRobot r : robots) {
            if(r.robot == robot) return r;
        }
        return null;
    }
    
    protected Object clone() {
        StateNode newStateNode = new StateNode(this.level, this.firstMove);
        newStateNode.value = this.value;
        newStateNode.robots = new ArrayList<OurRobot>();
        for(OurRobot r : this.robots)
            newStateNode.robots.add((OurRobot)r.clone());
        newStateNode.treasures = new HashSet<Treasure>();
        for(Treasure t : this.treasures)
            newStateNode.treasures.add(t);
        return newStateNode;
    }

    public int compareTo(Object obj) {
            StateNode sn = (StateNode)obj;
            if(this.value < sn.value)   // vetsi hodnota je v heuristice brana jako lepsi, proto se
                return 1;              // radi v PrirityQueue od nejvetsi po nejmensi
            else if(this.value > sn.value)
                return -1;
            else
                return 0;
    }
    
    public String toString(){
        String returnValue=new String("");
        returnValue+="Uroven: "+this.level+"; ohodnoceni: "+this.value+"; first move: "+this.firstMove;
        /*if(this.firstMove==BotAction.SHOOT){returnValue+="SHOOT";}
        else if(this.firstMove==BotAction.STEP){returnValue+="STEP";}
        else if(this.firstMove==BotAction.LEFT){returnValue+="LEFT";}
        else if(this.firstMove==BotAction.RIGHT){returnValue+="RIGHT";}
        else if(this.firstMove==BotAction.RECHARGE){returnValue+="RECHARGE";}
        else if(this.firstMove==BotAction.TAKE){returnValue+="TAKE";}*/
        return returnValue;
    }
}

class StateTree {
    protected StateNode root;
    protected Robik robik; // ukazatel na naseho Robota
    protected PriorityQueue<StateNode> bfsQueue;
    protected Set<Listik> stateSet = new HashSet<Listik>();
    
    public StateTree(Robik robik) {
        this.robik = robik;
        bfsQueue = new PriorityQueue<StateNode>();
        root = new StateNode(0,null);
        for(Robot rb: robik.world.getRobots()) {
            OurRobot or = new OurRobot();
            or.robot = rb;
            or.direction = rb.getDirection();
            or.energy = rb.getEnergy();
            or.points = rb.getPoints();
            or.positionX = rb.getPositionX();
            or.positionY = rb.getPositionY();
            root.robots.add(or);
        }
        for(Treasure tr: robik.globalTreasures) {
            if(robik.world.containsTreasure(tr.positionX,tr.positionY))
                root.treasures.add(tr);
        }
    }
    
    public BotAction bfs() {
        int expansions = 0;
        StateNode newStateNode = null;
        
        /*Listik hash=new Listik(root.robots.size()*3);
        for(int i=0; i<root.robots.size();i++){
            hash.listik[3*i+0]=root.robots.get(i).positionX;
            hash.listik[3*i+1]=root.robots.get(i).positionY;
            hash.listik[3*i+2]=root.robots.get(i).direction.ordinal();
        }
        stateSet.add(hash);*/
        bfsQueue.add(root);
        while((newStateNode = bfsQueue.poll()) != null) {
            //System.out.println("Ctu prvni prvek z fronty. Ohodnoceni:" + newStateNode.value + ", uroven: " + newStateNode.level);
            if(expansions++ < Robik.MAX_EXPANSIONS){
                expandState(newStateNode, new ArrayList<BotAction>());
               // System.out.println("Pocet prvku ve fronte: " + bfsQueue.size());
            }
            else {
                return newStateNode.firstMove;
            }
        }
        return null;
    }
    
    private double evaluate(StateNode stateNode){
        double ourHeuristic = 0;
        double suma = 0;
        
        for(OurRobot or : stateNode.robots){
            if(or.robot == robik.robot)
                ourHeuristic = evaluateRobot(or, stateNode);
            else
                suma += evaluateRobot(or, stateNode);
        }
        
        return /*(suma / stateNode.robots.size()) */ ourHeuristic;
        //return stateNode.getOurRobot(robik.robot).points;
    }
    
    private double evaluateRobot(OurRobot robot, StateNode stateNode){
        double TVN = 20;     // nasobic hodnoty pokladu
        double TDP = 3;   // mocnina vzdalenosti pokladu
        double ORDM = 8;    // maximalni zapocitana vzdalenost ostatnich robotu od pokladu
        double SCM = 20;     // nasobitel skore robota
        double EPM = 10;   //nasobitel za energii robota
        double value = 0;
        
        /*pricitani hodnot pokladu*/
        for(Treasure treasure: stateNode.treasures){
            value += (TVN*treasure.value)/
                    Math.pow(1+treasure.distances[stateNode.getOurRobot(robot.robot).positionX]
                                             [stateNode.getOurRobot(robot.robot).positionY]
                                             [stateNode.getOurRobot(robot.robot).direction.ordinal()],TDP);
            /*odecitani vzdalenosti ostatnich robotu od pokladu*/
            for(OurRobot r : stateNode.robots)
                if(r != robot){
                    double otherRobotsDistance = treasure.distances[r.positionX][r.positionY][r.direction.ordinal()];
                    if(otherRobotsDistance < ORDM)
                        value /= ORDM / (otherRobotsDistance + 1);
                }
        }
        /*pricteni skore*/
        value += robot.points * SCM;
        
        /*pricteni bodu za energii*/
        //value += Math.abs((robik.world.getMaxRobotEnergy() / 2) - robot.energy) * EPM;
        double max2 = Math.abs(robik.world.getMaxRobotEnergy() / 2);
        value += (((max2 - Math.abs(max2 - robot.energy))) / max2) * EPM;
        
        //System.out.println("Zvracim hodnotu: " + value);
        return value;        
    }
    
    class Listik{
        public int[] listik;
        
        public Listik(int size){
            this.listik=new int[size];
        }
        
        public boolean equals(Object obj) {
            Listik l = (Listik)obj;
            for(int i = 0; i < listik.length; i++)
                if(listik[i] != l.listik[i])
                    return false;
            return true;
        }
        
        public int hashCode() {
            return listik[0] * listik[1];
        }
    }
    
    
    private void expandState(StateNode stateNode, ArrayList<BotAction> botActions){
        
        if(botActions.size() == stateNode.robots.size()) {
            StateNode newStateNode = (StateNode)stateNode.clone();
            executeTurn(newStateNode, botActions);
            if(newStateNode.firstMove == null)
                newStateNode.firstMove = botActions.get(stateNode.robots.indexOf(stateNode.getOurRobot(robik.robot)));

            Listik hash = new Listik(newStateNode.robots.size()*3);
            boolean allRest = true;
            for(int i = 0; i < newStateNode.robots.size(); i++){
                hash.listik[3*i+0] = newStateNode.robots.get(i).positionX;
                hash.listik[3*i+1] = newStateNode.robots.get(i).positionY;
                hash.listik[3*i+2] = newStateNode.robots.get(i).direction.ordinal();
                if(botActions.get(i) != BotAction.REST) allRest = false;
            }
            if(!stateSet.contains(hash) || !allRest){
                stateSet.add(hash);         

                newStateNode.value = evaluate(newStateNode);
                newStateNode.level = stateNode.level + 1;
                bfsQueue.add(newStateNode);
                //System.out.println("expanduji");
            }else{
                //System.out.println("odreazavam");
            }
            //System.out.println(newStateNode);
            /*String temp = "";
            for(BotAction ba : botActions)
                if(ba != null)
                    temp += ba.toString() + " ";
                else
                    temp += "null" + " ";
            System.out.println("Vkladam novy uzel s akcemi: " + temp);*/
        }
        else {
            OurRobot robot = stateNode.robots.get(botActions.size());
            if(robot.energy <= 0) {
                botActions.add(null);
                expandState(stateNode,botActions);
                botActions.remove(botActions.size() - 1);
            }
            else {
                for(BotAction action: BotAction.values()) {
                    boolean actionOK = false;
                    switch(action) {
                        case TAKE:
                            if(stateNode.containsTreasure(robot.positionX,robot.positionY)) {
                                actionOK = true;
                            } break;
                        case RECHARGE:
                            if(robik.world.containsBattery(robot.positionX,robot.positionY)) {
                                actionOK = true;
                            } break;
                        case STEP:
                            if(robot.direction == Direction.UP && robot.positionY > 0
                                    && !robik.world.containsWall(robot.positionX,robot.positionY - 1))
                                actionOK = true;
                            if(robot.direction == Direction.RIGHT && robot.positionX < robik.world.getDimensionX() - 1
                                    && !robik.world.containsWall(robot.positionX + 1,robot.positionY))
                                actionOK = true;
                            if(robot.direction == Direction.DOWN && robot.positionY < robik.world.getDimensionY() - 1
                                    && !robik.world.containsWall(robot.positionX,robot.positionY + 1))
                                actionOK = true;
                            if(robot.direction == Direction.LEFT && robot.positionX > 0
                                    && !robik.world.containsWall(robot.positionX - 1,robot.positionY))
                                actionOK = true;
                            break;
                        case SHOOT:
                            if(testShoot(robot,stateNode) != null)
                                actionOK = true;
                            break;
                        default: actionOK = true; break;
                        
                    }
                    if(actionOK) {
                        botActions.add(action);
                        expandState(stateNode,botActions);
                        botActions.remove(botActions.size() - 1);
                    }
                }
            }
        }
    }
    
    private void executeTurn(StateNode stateNode, ArrayList<BotAction> botActions) {
        int[] damage = new int[stateNode.robots.size()];
        Position[] positions = new Position[stateNode.robots.size()];
        
        for(OurRobot r : stateNode.robots) r.energy--;
        
        for(OurRobot r : stateNode.robots) {
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.SHOOT) {
                OurRobot damagedRobot = testShoot(r,stateNode);
                if(damagedRobot != null) {
                    damage[stateNode.robots.indexOf(damagedRobot)] = (damagedRobot.energy) / 2;
                    r.points += (damagedRobot.energy) / 2;
                }
            }
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.STEP) {
                switch(r.direction) {
                    case UP: positions[stateNode.robots.indexOf(r)] = new Position(r.positionX, r.positionY - 1); break;
                    case RIGHT: positions[stateNode.robots.indexOf(r)] = new Position(r.positionX + 1, r.positionY); break;
                    case DOWN: positions[stateNode.robots.indexOf(r)] = new Position(r.positionX, r.positionY + 1); break;
                    case LEFT: positions[stateNode.robots.indexOf(r)] = new Position(r.positionX - 1, r.positionY); break;
                }
            }
            else
                positions[stateNode.robots.indexOf(r)] = new Position(r.positionX, r.positionY);
        }
        for(int i = 0 ; i < damage.length; i++) {
            if(damage[i] > 0)
                stateNode.robots.get(i).energy -= damage[i];
        }
        for(int i = 0; i < positions.length; i++) {
            boolean isEqual = false;
            for(int j = i+1; j < positions.length; j++)
                if(positions[i] != null && positions[i].equals(positions[j])) {
                    positions[j] = null;
                    isEqual = true;
                }
            if(isEqual == true)
                positions[i] = null;
        }
        for(int i = 0; i < positions.length; i++) {
            if(positions[i] != null) {
                stateNode.robots.get(i).positionX = positions[i].positionX;
                stateNode.robots.get(i).positionY = positions[i].positionY;
            }
        }
        for(OurRobot r : stateNode.robots) {
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.LEFT) {
                switch(r.direction) {
                    case UP: r.direction = Direction.LEFT; break;
                    case RIGHT: r.direction = Direction.UP; break;
                    case DOWN: r.direction = Direction.RIGHT; break;
                    case LEFT: r.direction = Direction.DOWN; break;
                }
            }
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.RIGHT) {
                switch(r.direction) {
                    case UP: r.direction = Direction.RIGHT; break;
                    case RIGHT: r.direction = Direction.DOWN; break;
                    case DOWN: r.direction = Direction.LEFT; break;
                    case LEFT: r.direction = Direction.UP; break;
                }
            }
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.TAKE) {
               Treasure tr = stateNode.getTreasure(r.positionX, r.positionY);
               if(tr != null) {
                   r.points += robik.world.getTreasureValue(tr.positionX, tr.positionY);
                   stateNode.treasures.remove(tr);
               }
            }
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.REST) {
                r.energy += robik.world.getRestRobotEnergy();
                if(r.energy > robik.world.getMaxRobotEnergy())
                    r.energy = robik.world.getMaxRobotEnergy();
            }
            if(botActions.get(stateNode.robots.indexOf(r)) == BotAction.RECHARGE) {
                if(robik.world.containsBattery(r.positionX, r.positionY))
                    r.energy = robik.world.getMaxRobotEnergy();
            }
        }
        
        
    }
    
    private boolean positionOnMap(int x, int y) {
        return x >= 0 && x < robik.world.getDimensionX() &&
               y >= 0 && y < robik.world.getDimensionY();
    }
    
    private OurRobot testShoot(OurRobot robot, StateNode stateNode) {
        int dx = 0, dy = 0;
        int x = robot.positionX;
        int y = robot.positionY;
        switch(robot.direction){
            case UP: dx = 0; dy = -1; break;
            case RIGHT: dx = 1; dy = 0; break;
            case DOWN: dx = 0; dy = 1; break;
            case LEFT: dx = -1; dy = 0; break;
        }
        x += dx; y += dy;
        while(positionOnMap(x, y)) {
            if(robik.world.containsWall(x,y)) return null;
            OurRobot shootedRobot = stateNode.getOurRobot(x,y);
            if(shootedRobot != null && shootedRobot.energy > 0) return shootedRobot;
            x += dx; y += dy;
        }
        return null;
    }
}
