package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.World;

/**
 *  Třída reprezuntující pozici na hrací ploše
 */
public class Position {
    
    private int x; // x-ová souřadnice
    private int y; // y-ová souřadnice
    private Robot.Direction d; // směr

    /**
     * Konstruktor pro pozici zkískanou ze zadaného robota
     *
     * @param robot
     *      robot
     */  
    public Position(Robot robot) {
        this.setX(robot.getPositionX());
        this.setY(robot.getPositionY());
        this.setD(robot.getDirection());
    }
    
    
    /**
     * Konstruktor pro pozici se směrem
     *
     * @param x
     *      x-ová souřadnice
     * @param y   
     *      y-ová souřadnice
     * @param d   
     *      směr     
     */  
    public Position(int x, int y, Robot.Direction d) {
        this.setX(x);
        this.setY(y);
        this.setD(d);
    }
    
    
    /**
     * Konstruktor pro pozici bez směru
     *
     * @param x
     *      x-ová souřadnice
     * @param y
     *      y-ová souřadnice
     */  
    public Position(int x, int y) {
        this.setX(x);
        this.setY(y);
        this.setD(null);
    }
    
    /**
     * Konstruktor pro kopii pozice
     *
     * @param p
     *      pozice
     */  
    public Position(Position p) {
        this.setX(p.getX());
        this.setY(p.getY());
        this.setD(p.getD());
    }
    
    
    /**
     * Vrať x-ovou souřadnici
     *
     * @return
     *      x-ová souřadnice
     */  
    public int getX() {
        return x;
    }

    
    /**
     * Vrať y-ovou souřadnici
     *
     * @return    
     *      y-ová souřadnice
     */  
    public int getY() {
        return y;
    }
    

    /**
     * Vrať směr
     *
     * @return 
     *      směr
     */   
    public Robot.Direction getD() {
        return d;
    }
    
    
    /** 
     * Nastav x-ovou souřadnici
     *
     * @param x
     *      x-ová souřadnice
     */   
    public void setX(int x) {
        this.x = x;
    }

    
    /**
     * Nastav y-ovou souřadnici
     *
     * @param y
     *      y-ová souřadnice
     */ 
    public void setY(int y) {
        this.y = y;
    }
    

    /**
     * Nastav směr
     *
     * @param d
     *      směr
     */ 
    public void setD(Robot.Direction d) {
        this.d = d;
    }
    
    
    /**
     * Převodní funkce (int -> Robot.Direction)
     * <pre>
     *      0 -> UP
     *      1 -> RIGHT
     *      2 -> DOWN
     *      3 -> LEFT
     * </pre>
     *
     * @param d   
     *      číselná reprezentace směru
     * @return
     *      směr
     * @throws NullPointerException 
     *      v případě, že je parametr d mimo množinu {0,1,2,3}
     */       
    public static Robot.Direction intToDir(int d) {
        switch (d) {
            case 0: return Robot.Direction.UP;
            case 1: return Robot.Direction.RIGHT;
            case 2: return Robot.Direction.DOWN;
            case 3: return Robot.Direction.LEFT;
        }    
        throw new IllegalArgumentException("Číselná reprezentace pozice mimo množinu {0,1,2,3}.");
    }
    
    
    /**
     * Převodní funkce (Robot.Direction -> int)
     * <pre>
     *      UP    -> 0
     *      RIGHT -> 1
     *      DOWN  -> 2
     *      LEFT  -> 3
     * </pre>
     *
     * @param d
     *      směr
     * @return 
     *      číselná repreznetace směru
     */      
    public static int dirToInt(Robot.Direction d) {
        switch (d) {
            case UP: return 0;
            case RIGHT: return 1;
            case DOWN: return 2;
            default: return 3;
        }           
    }    
    
    public boolean isCorrect(World world) {
        int dimX = world.getDimensionX();
        int dimY = world.getDimensionY();
        if (dimX <= x || dimY <= y || x < 0 || y < 0 || world.containsWall(x, y)) return false; //offside nebo zeď
        if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy()==0) return false; //mrtvola
        return true;        
    }
    
    public String toString() {
        String output;
        
        output = "("+x+", "+y;
        if (d == null) {
            output += ")"; 
        } else {
            output += ", "; 
            switch (d) {
                case UP: output += "UP"; break;
                case RIGHT: output += "RIGHT"; break;
                case DOWN: output += "DOWN"; break;
                default: output += "LEFT"; break;
            }
            output += ")"; 
        }
        
        return output;
    }
    
    public int toInt() {
        return dirToInt(d) | y<<2 | x<<8;
    }
    
}
