package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Strategy;

import java.util.Map;
import java.util.HashMap;

public class Rest {
    
    World world;  // svět
    Robot robot;  // můj robot
    Battery battery;
    
    TarjanWithEnergy tarjanWithEnergyRobot; // tarjan mého robota
    Map<Robot,TarjanWithEnergy> tarjanWithEnergyEnemies; //tarjani nepřátel
    
    Position bestBattery=null;
        
    int roundsToEnd;   //počet kol do konce
    
    /**
     * Konstruktor, inic. svět, robota a tarjany
     *
     * @param world
     *      svět
     * @param robot
     *      náš robot
     * @param tarjanRobot
     *      tarjan mého robota
     * @param tarjanEnemies
     *      tarjani nepřátel
     */
    public Rest(World world, Robot robot, TarjanWithEnergy tarjanWithEnergyRobot, Map<Robot,TarjanWithEnergy> tarjanWithEnergyEnemies, Battery battery) {
        this.world = world;
        this.robot = robot;
        this.battery = battery; 
        this.tarjanWithEnergyRobot = tarjanWithEnergyRobot;
        this.tarjanWithEnergyEnemies = tarjanWithEnergyEnemies;         
    }
    
    /* Vraci atraktivitu nejvýhodnejšího odpočívání
     * 
     * @return Double
     *      koeficient atraktivnosti
     */
    public double mostAttractValue() {
        roundsToEnd = world.getRoundsCount() - world.getRound();
        //Log.print(Log.Level.MEDIUM, "R:: roundsToEnd: " + roundsToEnd + " energie: " +  robot.getEnergy());
        
        double restAttractivity;
            
        restAttractivity = Math.min(world.getMaxRobotEnergy(), roundsToEnd * world.getRestRobotEnergy()); //kolik jeste stihnu nazrat
        restAttractivity = Math.min(restAttractivity, world.getMaxRobotEnergy() - robot.getEnergy() + 1); //cistej zisk
        restAttractivity = restAttractivity / roundsToEnd; //zisk na kolo
        //Log.print(Log.Level.MEDIUM, "R::     restAttractivity: " + restAttractivity);
        if (restAttractivity > 0) {
            //znevýhodnit pokud mě někdo múže trefit
            Sniper sniper = new Sniper(world, robot, tarjanWithEnergyRobot, tarjanWithEnergyEnemies);            
            if (sniper.distToShootAs() < roundsToEnd) {                 
                restAttractivity = restAttractivity / (roundsToEnd - sniper.distToShootAs());
                //Log.print(Log.Level.MEDIUM, "R::     restAattractivity: " + restAttractivity + " korekce");
            }
            if (sniper.distToShootAs() == 0 && roundsToEnd>1) {
                restAttractivity=0;
            }             
        }        
        
        double rechargeAttractivity=0;
        int minDistance = Tarjan.NO_PATH;
        for (Position b : battery.batteries) {
            if (minDistance > tarjanWithEnergyRobot.distance(b)) {
                minDistance = tarjanWithEnergyRobot.distance(b);
                bestBattery = b;
            }
        }        
        if (minDistance < roundsToEnd) {            
            rechargeAttractivity = world.getMaxRobotEnergy()-robot.getEnergy();
            rechargeAttractivity = rechargeAttractivity/roundsToEnd;            
            if (roundsToEnd-minDistance > 5) {
                rechargeAttractivity=rechargeAttractivity/(roundsToEnd-minDistance);            
            }
        }
        //Log.print(Log.Level.MEDIUM, "R::     rechargeAttractivity: " + rechargeAttractivity + " baterry="+bestBattery+" distance="+minDistance);
        
        
        if (rechargeAttractivity > 0) {
            return rechargeAttractivity;            
        } else {            
            bestBattery=null;
            return restAttractivity;            
        }
    }

    /**
     * Vrací akci kterou mám provést abych dosáhl na mostAttractValue()
     * 
     * @return Strategy.BotAction
     *      akce která se má provést
     */
    public Strategy.BotAction mostAttractAction() {
        if (bestBattery==null) {
            return Strategy.BotAction.REST;
        } else {
            if (roundsToEnd==1) 
                return Strategy.BotAction.RECHARGE;
            if (tarjanWithEnergyRobot.action(bestBattery) != null)
                return tarjanWithEnergyRobot.action(bestBattery);
            if (robot.getEnergy()<2)
                return Strategy.BotAction.REST;        
        }  
        return Strategy.BotAction.LEFT;
    }   
    
    
}
