package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Strategy;

import java.util.Map;
import java.util.HashMap;

/**
 * Třída pro "substrategie" týkající se střílení.
 */
public class Sniper {
        
    World world;  // svět
    Robot robot;  // můj robot
    TarjanWithEnergy tarjanWithEnergyRobot; // tarjan mého robota
    Map<Robot,TarjanWithEnergy> tarjanWithEnergyEnemies; //tarjani nepřátel
    
    private Position lastBestPosition = new Position(0,0); //nevýhodnějsí pozice, kerou našlo poslední volání mostAttractValue()
    
    
    /**
     * Konstruktor, inic. svět, robota a tarjany
     *
     * @param world
     *      svět
     * @param robot
     *      náš robot
     * @param tarjanRobot
     *      tarjan mého robota
     * @param tarjanEnemies
     *      tarjani nepřátel
     */
    public Sniper(World world, Robot robot, TarjanWithEnergy tarjanWithEnergyRobot, Map<Robot,TarjanWithEnergy> tarjanWithEnergyEnemies) {
        this.world = world;
        this.robot = robot;
        this.tarjanWithEnergyRobot = tarjanWithEnergyRobot;
        this.tarjanWithEnergyEnemies = tarjanWithEnergyEnemies;      
    }
     
    
    /**
     * Kolik bodu ziskam strelo do nepritele pokud vystrelim za x kol? 
     *
     * @param enemy
     *      nepřítel
     * @param rounds
     *      počet kol
     * @return double
     *      odhadnutá energie
     */
    private boolean IsRobotBetween(Robot robot2, Robot robot1) {
        if (robot1.getPositionX() == robot2.getPositionX()) {
            for (int i = Math.min(robot1.getPositionY(), robot2.getPositionY())+1; i < Math.max(robot1.getPositionY(), robot2.getPositionY()); i++) {
                if (world.containsRobot(robot1.getPositionX(), i)) return true;
            }
            return false;
        } 
        if (robot1.getPositionY() == robot2.getPositionY()) {
            for (int i = Math.min(robot1.getPositionX(), robot2.getPositionX())+1; i < Math.max(robot1.getPositionX(), robot2.getPositionX()); i++) {
                if (world.containsRobot(i, robot1.getPositionY())) return true;
            }
            return false;
        }        
        return false;        
    }  
    
    /**
     * Kolik bodu ziskam strelo do nepritele pokud vystrelim za x kol? 
     *
     * @param enemy
     *      nepřítel
     * @param rounds
     *      počet kol
     * @return double
     *      odhadnutá energie
     */
    private double enemyEnergy(Robot enemy, int rounds) {
        int enemyEnergy;
        
        if (rounds==0) return Math.floor(((double) (enemy.getEnergy() - 1)/2)); 
        
        if (enemy.getEnergy()-rounds < 2) enemyEnergy = world.getRestRobotEnergy();
        else  enemyEnergy = enemy.getEnergy() - 1;
        
        //Log.print(Log.Level.MEDIUM, "enemyEnergy: rounds=" + rounds + " energie="+ enemy.getEnergy());
        //Log.print(Log.Level.MEDIUM, "enemyEnergy: " + Math.max(0,Math.floor(((double) enemyEnergy-rounds)/2)));
                    
        return  Math.max(0,Math.floor(((double) enemyEnergy-rounds)/2));               
    }  
    
    /**
     * Jak dlouho vydrží nepřítel na pozici kde je
     *
     * @param enemy
     *      nepřítel
     * @return int
     *      počet kol
     */
    public int stayInXY(Robot enemy) { 
        return Math.min(stayInX(enemy), stayInY(enemy));
    }  
    
    /**
     * Jak dlouho vydrží nepřítel na ose Y
     *
     * @param enemy
     *      nepřítel
     * @return int
     *      počet kol
     */
    private int stayInY(Robot enemy) { 
        int minimum = Tarjan.NO_PATH;
        int actUp = Tarjan.NO_PATH, actDown = Tarjan.NO_PATH;
        
        for (int i = 0; i < world.getDimensionX(); i++) {
            
            if (enemy.getPositionY() > 0) 
                actUp = tarjanWithEnergyEnemies.get(enemy).distance(new Position(i, enemy.getPositionY() - 1));
            if (enemy.getPositionY() + 1 < world.getDimensionY())
                actDown = tarjanWithEnergyEnemies.get(enemy).distance(new Position(i, enemy.getPositionY() + 1));         
            
            minimum=Math.min(minimum, Math.min(actUp, actDown)); 
        }
        return minimum - 1;
    }  
    
    
    /**
     * Jak dlouho vydrží nepřítel na ose X
     *
     * @param enemy
     *      nepřítel
     * @return int
     *      počet kol
     */
    private int stayInX(Robot enemy) {
        int minimum = Tarjan.NO_PATH;
        int actLeft = Tarjan.NO_PATH, actRight = Tarjan.NO_PATH;
        
        for (int i = 0; i < world.getDimensionY(); i++) {
            
            if (enemy.getPositionX() > 0) 
                actLeft = tarjanWithEnergyEnemies.get(enemy).distance(new Position(enemy.getPositionX() - 1, i));
            if (enemy.getPositionX() + 1 < world.getDimensionX())
                actRight = tarjanWithEnergyEnemies.get(enemy).distance(new Position(enemy.getPositionX() + 1, i));         
            
            minimum=Math.min(minimum, Math.min(actLeft, actRight));            
        }
        return minimum - 1;
    }
    
    
    /**
     * Jak rychle se dostanu na nepřítelovu osu X s otočením na něj
     *
     * @param enemy
     *      nepřítel
     * @return int
     *      počet kol
     */
    private int distToX(Robot enemy) {
        int minDistance = Tarjan.NO_PATH; 
       
        int x, y;
              
        x = enemy.getPositionX();                
        y = enemy.getPositionY() - 1;               
        while (true) {
            if (y < 0 || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance,tarjanWithEnergyRobot.distance(new Position(x, y, Robot.Direction.DOWN)));            
            y--;
        } 
        
        y = enemy.getPositionY() + 1;
        while (true) {
            if (y >= world.getDimensionY() || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance,tarjanWithEnergyRobot.distance(new Position(x, y, Robot.Direction.UP)));            
            y++;
        }                      
        
        return minDistance;
    }
    
    
    /**
     * Jak rychle se dostanu na nepřítelovu osu Y s otočením na něj
     *
     * @param enemy
     *      nepřítel
     * @return int
     *      počet kol
     */
    private int distToY(Robot enemy) {
      int minDistance = Tarjan.NO_PATH; 
       
        int x, y;
              
        x = enemy.getPositionX() - 1;                
        y = enemy.getPositionY();               
        while (true) {
            if (x < 0 || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance,tarjanWithEnergyRobot.distance(new Position(x, y, Robot.Direction.RIGHT)));            
            x--;
        } 
        
        x = enemy.getPositionX() + 1;
        while (true) {
            if (x >= world.getDimensionX() || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance,tarjanWithEnergyRobot.distance(new Position(x, y, Robot.Direction.LEFT)));            
            x++;
        }                      
        
        return minDistance;           
    }
    
    
    /**
     * Vrací atraktivitu nejatraktivnější střelecké pozice
     * 
     * @return Double
     *      koeficient atraktivnosti
     */
    public double mostAttractValue() {
        
        double maxAttract = 0;
        double actualXAttract = 0, actualYAttract = 0; 
         
        int x, y;        
        Position actXPosition = new Position(0, 0);
        Position actYPosition = new Position(0, 0);
                
        for (Robot enemy : world.getRobots()) { 
            if (enemy == robot) continue; //nebudu prece strilet na sebe
            //Log.print(Log.Level.MEDIUM,"");
            //Log.print(Log.Level.MEDIUM, "S:: distToX(stayInX): " + distToX(enemy) +"("+stayInX(enemy)+")"+ " DistToY(stayInY): " + distToY(enemy) +"("+stayInY(enemy)+")");
            //Log.print(Log.Level.MEDIUM,"S::  xxxxxxxxxxx");
            
            //vyhodnost jit na X nepritele
            actualXAttract = (double) enemyEnergy(enemy, distToX(enemy)) / (distToX(enemy) + 1); 
            if (distToX(enemy)==0 && IsRobotBetween(robot,enemy)) actualXAttract=0;
            //Log.print(Log.Level.MEDIUM, "S::    actualXAttract: " + actualXAttract);
            if (stayInX(enemy) < distToX(enemy)) { //muze utyct znevyhodni tuto varianto 
                actualXAttract = Math.max(((double )world.getRestRobotEnergy()-2)/3, actualXAttract);
                actualXAttract = actualXAttract/(1+distToX(enemy)-stayInX(enemy));
                //Log.print(Log.Level.MEDIUM, "S::     actualXAttract: " + actualXAttract + " korekce");
            }             
            
            //Log.print(Log.Level.MEDIUM,"S::   yyyyyyyyyyyy");
            
            //vyhodnost jit na Y nepritele            
            actualYAttract = (double) enemyEnergy(enemy,distToY(enemy)) / (distToY(enemy) + 1);
            if (distToY(enemy)==0 && IsRobotBetween(robot,enemy)) actualYAttract=0;
            //Log.print(Log.Level.MEDIUM, "S::     actualYAttract: " + actualYAttract);
            if (stayInY(enemy) < distToY(enemy)) { //muze utyct znevyhodni tuto varianto 
                actualYAttract = Math.max((world.getRestRobotEnergy()-2)/3, actualYAttract);
                actualYAttract = actualYAttract/(1+distToY(enemy)-stayInY(enemy));
                //Log.print(Log.Level.MEDIUM, "S::     actualYAttract: " + actualYAttract + " korekce");
            }  
            
            
            if (actualXAttract > Math.max(maxAttract, actualYAttract) && distToX(enemy) != Tarjan.NO_PATH) {
                //zatim je nejvyhodnejsi jit na X, ted musim zjistit, jestli zhora nebo zdola a na jakou pozici
                
                //kontrola zhora
                x = enemy.getPositionX();
                y = enemy.getPositionY() - 1;
                actXPosition.setX(x);
                actXPosition.setD(Robot.Direction.DOWN);                                            
                while (true) {
                    actXPosition.setY(y);
                
                    if (y < 0 || world.containsWall(x, y)) break;
                    if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
                    
                    if (tarjanWithEnergyRobot.distance(actXPosition) == distToX(enemy)) {
                        lastBestPosition.setX(x);
                        lastBestPosition.setY(y);
                        lastBestPosition.setD(Robot.Direction.DOWN);
                        break;
                    }
                    y--;
                } 
                
                //kontrola zdola
                y = enemy.getPositionY() + 1;
                actXPosition.setD(Robot.Direction.UP);                                            
                while (true) {
                    actXPosition.setY(y);
                
                    if (y >= world.getDimensionY() || world.containsWall(x, y)) break;
                    if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
                    
                    if (tarjanWithEnergyRobot.distance(actXPosition) == distToX(enemy)) {
                        lastBestPosition.setX(x);
                        lastBestPosition.setY(y);
                        lastBestPosition.setD(Robot.Direction.UP);
                        break;
                    }
                    y++;
                }  
                
                maxAttract = actualXAttract;
            } 
            
            if (actualYAttract > Math.max(maxAttract, actualXAttract) && distToY(enemy) != Tarjan.NO_PATH) {
                //zatim je nejvyhodnejsi jit na Y, ted musim zjistit, jestli zleva nebo zprava a na jakou pozici
                
                //kontrol zleva
                x = enemy.getPositionX() - 1;                    
                y = enemy.getPositionY();                    
                actYPosition.setY(y);                    
                actYPosition.setD(Robot.Direction.RIGHT);                     
                while (true) {
                    actYPosition.setX(x);
                
                    if (x < 0 || world.containsWall(x, y)) break;
                    if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
                    
                    if (tarjanWithEnergyRobot.distance(actYPosition) == distToY(enemy)) {
                        lastBestPosition.setX(x);
                        lastBestPosition.setY(y);
                        lastBestPosition.setD(Robot.Direction.RIGHT);
                        break;
                    }
                    x--;
                } 
                
                //kontrola zprava
                x = enemy.getPositionX() + 1;                    
                actYPosition.setD(Robot.Direction.LEFT);                                            
                while (true) {
                    actYPosition.setX(x);
                
                    if (x >= world.getDimensionX() || world.containsWall(x, y)) break;
                    if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
                    
                    if (tarjanWithEnergyRobot.distance(actYPosition) == distToY(enemy)) {
                        lastBestPosition.setX(x);
                        lastBestPosition.setY(y);
                        lastBestPosition.setD(Robot.Direction.LEFT);
                        break;
                    }
                    x++;
                }            
                
                maxAttract = actualYAttract;
            }
                        
        }
                
        return maxAttract;       
    }
    
    
    /**
     * Vrací akci kterou mám provést pokud chci využít nevýhodnějsí pozice, kerou našlo poslední volání mostAttractValue()
     * 
     * @return Strategy.BotAction
     *      akce která se má provést
     */
    public Strategy.BotAction mostAttractAction() {
        if (tarjanWithEnergyRobot.action(lastBestPosition) != null) { // pokud nejsem na místě
            return tarjanWithEnergyRobot.action(lastBestPosition);
        }
        else { 
            return Strategy.BotAction.SHOOT;
        }   
    }   
    
   
    /**
     * Jak rychle se dostane nepřítel na misto odkud by nás mohl trefit
     * 
     * @param enemy
     *      nepřítel
     * @return int
     *      počet kol      
     */
    private int distToShootAs(Robot enemy) {
        int minDistance = Tarjan.NO_PATH; 
       
        int x, y;
              
        x = robot.getPositionX();                
        y = robot.getPositionY() - 1;               
        while (true) {
            if (y < 0 || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance, tarjanWithEnergyEnemies.get(enemy).distance(new Position(x, y, Robot.Direction.DOWN)));            
            y--;
        } 
        
        y = robot.getPositionY() + 1;
        while (true) {
            if (y >= world.getDimensionY() || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance, tarjanWithEnergyEnemies.get(enemy).distance(new Position(x, y, Robot.Direction.UP)));            
            y++;
        }                      
              
        x = robot.getPositionX() - 1;                
        y = robot.getPositionY();               
        while (true) {
            if (x < 0 || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance, tarjanWithEnergyEnemies.get(enemy).distance(new Position(x, y, Robot.Direction.RIGHT)));            
            x--;
        } 
        
        x = robot.getPositionX() + 1;
        while (true) {
            if (x >= world.getDimensionX() || world.containsWall(x, y)) break;
            if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy() == 0) break;
            minDistance = Math.min(minDistance, tarjanWithEnergyEnemies.get(enemy).distance(new Position(x, y, Robot.Direction.LEFT)));            
            x++;
        }                      
        
        if (minDistance==0 && IsRobotBetween(robot, enemy)) minDistance=1;
        return minDistance;          
    }
    
    /**
     * Jak nejrychleji se dostane libovolný nepřítel na místo odkud by nas mohl trefit
     * 
     * @return int
     *      počet kol   
     */
    public int distToShootAs() {
        int minDistance = Tarjan.NO_PATH;
            
        for (Robot enemy : world.getRobots()) { 
            if (enemy == robot) continue; //nepocitaj sebe na 
            minDistance = Math.min(distToShootAs(enemy), minDistance);
        }
        return minDistance;
    }
   
}
