package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
      
public class TarjanWithEnergy {
    
    Tarjan tarjan;
    World world;
    Robot robot;    
    
    Battery battery;
    
    public Position targetBattery;
         
    public TarjanWithEnergy(World world, Robot robot, Tarjan tarjan, Battery battery) {
       this.world = world;
       this.robot = robot;
       this.tarjan = tarjan;   
       this.battery = battery;
    }
    
    public int distance(Position position) {       
                
        int distance = tarjan.distance(position);
        
        if (distance - robot.getEnergy() > 0 && distance != Tarjan.NO_PATH) {           
            distance += 1 + (distance - (robot.getEnergy()+1))/(world.getRestRobotEnergy()-1);
        }
        
        if (distance > world.getRoundsCount() - world.getRound())
            distance = Tarjan.NO_PATH;
        
        return distance;
    }    
    
    public int distanceWithBattery(Position position) {         
        int distance, distanceRest, distanceRestBattery; 
        int minDistBattery = Tarjan.NO_PATH;
        Position minBattery=null, actBattery;
                
        distanceRest = distance = tarjan.distance(position);
        
        if (distance - robot.getEnergy() > 0 && distance != Tarjan.NO_PATH) {           
            distanceRest += 1 + (distance - (robot.getEnergy()+1))/(world.getRestRobotEnergy()-1);
        }
        
        for (Position b : battery.batteries) {
           if (distance > tarjan.distance(b)) {
                 
                actBattery=new Position(b);   
                
                if (tarjan.distance(new Position(b.getX(), b.getY(), Robot.Direction.UP)) == tarjan.distance(new Position(b.getX(), b.getY()))) 
                    actBattery.setD(Robot.Direction.UP);                   
                else if (tarjan.distance(new Position(b.getX(), b.getY(), Robot.Direction.RIGHT))  == tarjan.distance(new Position(b.getX(), b.getY()))) 
                    actBattery.setD(Robot.Direction.RIGHT); 
                else if (tarjan.distance(new Position(b.getX(), b.getY(), Robot.Direction.DOWN))  == tarjan.distance(new Position(b.getX(), b.getY()))) 
                    actBattery.setD(Robot.Direction.DOWN); 
                else 
                    actBattery.setD(Robot.Direction.LEFT); 
                    
                int distToBatery = tarjan.distance(b);
                if (distToBatery - robot.getEnergy() > 0) {           
                   distToBatery += 1 + (distToBatery - (robot.getEnergy()+1))/(world.getRestRobotEnergy()-1);
                }   
                
                int distFromBatery = battery.distance(actBattery,position);
                if (distFromBatery - world.getMaxRobotEnergy() > 0) {           
                   distFromBatery += 1 + (distFromBatery - ( world.getMaxRobotEnergy()+1))/(world.getRestRobotEnergy()-1);
                }
                
                if (minDistBattery > distToBatery+1+distFromBatery) {
                    minDistBattery=distToBatery+1+distFromBatery;
                    minBattery=actBattery;                  
                }         
            }               
           
        }     
        
        if (distanceRest > world.getRoundsCount() - world.getRound())
            distanceRest = Tarjan.NO_PATH;
        if (minDistBattery > world.getRoundsCount() - world.getRound())
            minDistBattery = Tarjan.NO_PATH;        
       
        //Log.print(Log.Level.HIGH, "   Tarjan::  position=" + position + " distance=" + distance + " distanceRest=" + distanceRest+ " minDistBattery=" + minDistBattery);
            
        if (distanceRest >= minDistBattery) {
            //Log.print(Log.Level.HIGH, "   Tarjan:: battery=" + minBattery);
        
            targetBattery = minBattery;
            return minDistBattery;
        } else { 
            targetBattery = null;
            return distanceRest;
        }
    }    
    
    public Strategy.BotAction action(Position position) {
        if (robot.getEnergy()<2) return Strategy.BotAction.REST;               
        else return tarjan.action(position);    
    }
        
}
