package cz.muni.fi.fibot.solnysloup;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Strategy.BotAction;
import cz.muni.fi.fibot.World;
import java.util.PriorityQueue;
import java.util.Random;

public class SolnySloup implements Strategy {
    private class StrategyPolicko implements Comparable {
        private int x;
        private int y;
        private int vzdalenost;
        private StrategyPolicko predchudce;
        private Direction smer;
        
        public StrategyPolicko(int x, int y) {
            this(x, y, Integer.MAX_VALUE, null);
        }
        
        public StrategyPolicko(int x, int y, int vzdalenost, Direction smer) {
            this.x = x;
            this.y = y;
            this.vzdalenost = vzdalenost;
            this.smer = smer;
            this.predchudce = null;
        }
        
        public int getX() {
            return x;
        }
        
        public int getY() {
            return y;
        }
        
        public int getVzdalenost() {
            return vzdalenost;
        }
        
        public void setVzdalenost(int vzdalenost) {
            this.vzdalenost = vzdalenost;
        }
        
        public StrategyPolicko getPredchudce() {
            return predchudce;
        }
        
        public void setPredchudce(StrategyPolicko predchudce) {
            this.predchudce = predchudce;
        }
        
        public Direction getSmer() {
            return smer;
        }
        
        public void setSmer(Direction smer) {
            this.smer = smer;
        }
        
        public int compare(Object o){
            return compareTo(o);
        }
        
        public int compareTo(Object o){
            if (!(o instanceof StrategyPolicko)) { throw new ClassCastException("Class is not Policko"); }
            return 0-(new Integer(((StrategyPolicko) o).getVzdalenost())).compareTo(new Integer(this.getVzdalenost()));
        }
        
    }
    
    private World world;
    private Robot robot;
    
    private StrategyPolicko[][][] mapa;
    
    private PriorityQueue [] polePokladu;
    private PriorityQueue [] poleRobotu;
    
    private Robot [] roboti;
    private Strategy.BotAction [] akce;
    private int [] shootVzdalenost;
    
    private int pocetRobotu;
    private int mujBot;
    private Robot cil;
    
    public SolnySloup() {
        // inicializace
    }
    
    /* Gives robot implementation some information about world.
      This method can be called anytime between steps  */
    public void setWorld(World world) {
        this.world = world;
        //updateMapa();
    }
    
    /* Gives robot implementation some information about itself.
      This method can be called anytime between steps */
    public void setRobot(Robot robot) {
        this.robot = robot;
    }
    
    public void initialize() {
        if (world == null) { return ; }
        if (robot == null) { return ; }
        pocetRobotu = world.getRobots().size();
        akce = new Strategy.BotAction [pocetRobotu];
        mapa = new StrategyPolicko [pocetRobotu][world.getDimensionX()][world.getDimensionY()];
        polePokladu = new PriorityQueue [pocetRobotu];
        poleRobotu = new PriorityQueue [pocetRobotu];
        shootVzdalenost = new int [pocetRobotu];
        roboti = new Robot [pocetRobotu];
        for (int i = 0; i < pocetRobotu; i++){
            roboti[i] = (Robot) world.getRobots().toArray()[i];
            if ((roboti[i]).equals(robot)) {mujBot = i;}
            shootVzdalenost[i] = Integer.MAX_VALUE;
        }
        cil = null;
    }
    
    /* According to information about robot itself and about World, robot implementation should
       decide what to do in some specified time (about 5 s) */
    public Strategy.BotAction decide() {
        
        // umiram nebo konec hry => odpocivej
        if ((robot.getEnergy() <= 1)||(world.getRound() >= world.getRoundsCount()-1)) { return Strategy.BotAction.REST; }
        // stojim na pokladu => seber ho
        if (world.containsTreasure(robot.getPositionX(), robot.getPositionY())) { return Strategy.BotAction.TAKE; }
        
// inicializace
        BotAction vyslednaAkce = BotAction.REST;
        for (int i = 0; i < pocetRobotu; i++){
            polePokladu[i] = new PriorityQueue();
            poleRobotu[i] = new PriorityQueue();
            spoctiVzdalenosti(roboti[i], i);
            najdiRoboty(poleRobotu[i], i);
            akce[i] = universalStrategy(roboti[i], i);
        }
        
// pocatecni strategie - ziskej body strelbou
        int max = 0;
        for (int j = 1; j < pocetRobotu; j++) { // zjisti max energii ze vesch botu
            if ((j != mujBot)&&(roboti[j].getEnergy() > roboti[max].getEnergy())) {
                max = j;
            }
        }
        // pokud ma robot hodne energie a existuje k nemu cesta, pak ho jdi zastrelit
        if ((roboti[max].getEnergy() > world.getRestRobotEnergy())
        &&(true)) {
            vyslednaAkce =  shootStrategy();
        } else {
// jdu hlavne pro poklady nebo strilet?
            if (Math.pow(2,Math.sqrt(world.getRestRobotEnergy())+1) > prumernaCenaPokladu()) {
                vyslednaAkce = shootStrategy();
            } else {
                if (polePokladu[mujBot].size() == 0) {
                    vyslednaAkce = shootStrategy();
                } else {
                    vyslednaAkce = treasureStrategy();
                }
            }
        }
        return vyslednaAkce;
    }
    
    private int prumernaCenaPokladu() {
        int result = 0;
        for (int i = 0; i < polePokladu[mujBot].size(); i++){
            result = result + world.getTreasureValue(((StrategyPolicko) polePokladu[mujBot].toArray()[i]).getX(),((StrategyPolicko) polePokladu[mujBot].toArray()[i]).getY());
        }
        return (result/polePokladu[mujBot].size());
    }
    
    private Strategy.BotAction universalStrategy(Robot bot, int index) {
        
        return Strategy.BotAction.REST;
    }
    
    private Strategy.BotAction shootStrategy() {
        if ((cil == null) || (cil.getEnergy() < world.getRestRobotEnergy()/2)) {
            int [] boti = new int [pocetRobotu];
            for (int i=0; i < pocetRobotu; i++) {
                boti[i] = i;
            }
            for (int i=0; i < pocetRobotu-1; i++) {
                for (int j=i+1; j < pocetRobotu; j++) {
                    int a = roboti[boti[i]].getEnergy()/shootVzdalenost[boti[i]];
                    int b = roboti[boti[j]].getEnergy()/shootVzdalenost[boti[j]];
                    if (a == b) {
                        if (new Random().nextBoolean()) {
                            int pom=boti[i];
                            boti[i] = boti[j];
                            boti[j] = pom;
                        }
                    }
                    if (a < b) {
                        int pom=boti[i];
                        boti[i] = boti[j];
                        boti[j] = pom;
                    }
                }
            }
            cil = roboti[boti[0]];
        }
        return jdiNa(mapa[mujBot][robot.getPositionX()][robot.getPositionY()],mapa[mujBot][cil.getPositionX()][cil.getPositionY()]);
    }
    
    private Strategy.BotAction treasureStrategy() {
        return jdiNa(mapa[mujBot][robot.getPositionX()][robot.getPositionY()],(StrategyPolicko) polePokladu[mujBot].peek());
    }
    
    private void najdiRoboty(PriorityQueue<StrategyPolicko> frontaRobotu, int index) {
        Robot bot = roboti[index];
        for (int i = 0; i < pocetRobotu; i++) {
            Robot r = roboti[i];
            if ((r != null)&&(r.getEnergy() != 0)&&(!r.equals(bot))&&(mapa[index][r.getPositionX()][r.getPositionY()].getVzdalenost() < Integer.MAX_VALUE)) { // sebe, mrtvoly a nedosazitelne neuvazujeme
                StrategyPolicko pom = mapa[index][r.getPositionX()][r.getPositionY()];
                StrategyPolicko roh = null;
                boolean prekazka = false;
                while ((roh == null)&&(!mapa[index][bot.getPositionX()][bot.getPositionY()].equals(pom.getPredchudce()))){
                    if (pom.getVzdalenost()-pom.getPredchudce().getVzdalenost() > 1) { // cesta se zataci
                        roh = pom; // zapamatuj si nejblizsi zatacku od protivnika
                    }
                    pom = pom.getPredchudce();
                    if (world.containsRobot(pom.getX(),pom.getY())) {
                        prekazka = true; // v ceste stoji jinej zivej robot, pricti vzdalenost
                    }
                }
                
                int vzdalenost = 0;
                if (roh == null) { // robot je v rovine semnou (nahore/dole/vlevo/vpravo)
                    vzdalenost = pom.getVzdalenost() - 1; // staci se otocit, chodit tam nemusi => odecist 1
                } else { // robot neni v rovine semnou (je nekde "za zatackou")
                    vzdalenost = roh.getVzdalenost() - 1; // staci dojit k rohu a otocit se, chodit krok uz nemusi => odecist 1
                }
                if (prekazka) {
                    vzdalenost += 3; // v ceste stoji jinej zivej robot, pricti konstantu
                }
                frontaRobotu.add(new StrategyPolicko(r.getPositionX(), r.getPositionY(), vzdalenost, r.getDirection()));
                if (index == mujBot) { shootVzdalenost[i] = vzdalenost; }
            }
        }
    }
    
    private Strategy.BotAction jdiNa(StrategyPolicko odkud, StrategyPolicko kam) {
        Strategy.BotAction akce = BotAction.REST;
        
        if ((Math.abs(odkud.getX()-kam.getX()) > 1) || (Math.abs(odkud.getY()-kam.getY()) > 1)) { // nejsou to sousedni policka, upravit, kam chceme jit (vedle)
            while (!odkud.equals(kam.getPredchudce())) {
                kam = kam.getPredchudce();
            }
        }
        
        if (odkud.getX() == kam.getX()) {
            if (odkud.getY() < kam.getY()) {
                if (odkud.getSmer() == Direction.DOWN) {
                    akce = BotAction.STEP;
                } else {
                    if (odkud.getSmer() == Direction.RIGHT) {
                        akce = BotAction.RIGHT;
                    } else {
                        if (odkud.getSmer() == Direction.LEFT) {
                            akce = BotAction.LEFT;
                        } else {
                            // otoc se o 180 stupnu - VYHODNE !!!!!! DODELAT
                            akce = BotAction.LEFT;
                        }
                    }
                }
            } else {
                if (odkud.getSmer() == Direction.UP) {
                    akce = BotAction.STEP;
                } else {
                    if (odkud.getSmer() == Direction.LEFT) {
                        akce = BotAction.RIGHT;
                    } else {
                        if (odkud.getSmer() == Direction.RIGHT) {
                            akce = BotAction.LEFT;
                        } else {
                            // otoc se o 180 stupnu - VYHODNE !!!!!! DODELAT
                            akce = BotAction.LEFT;
                        }
                    }
                }
            }
        } else {
            if (odkud.getX() < kam.getX()) {
                if (odkud.getSmer() == Direction.RIGHT) {
                    akce = BotAction.STEP;
                } else {
                    if (odkud.getSmer() == Direction.UP) {
                        akce = BotAction.RIGHT;
                    } else {
                        if (odkud.getSmer() == Direction.DOWN) {
                            akce = BotAction.LEFT;
                        } else {
                            // otoc se o 180 stupnu - VYHODNE !!!!!! DODELAT
                            akce = BotAction.LEFT;
                        }
                    }
                }
            } else {
                if (odkud.getSmer() == Direction.LEFT) {
                    akce = BotAction.STEP;
                } else {
                    if (odkud.getSmer() == Direction.DOWN) {
                        akce = BotAction.RIGHT;
                    } else {
                        if (odkud.getSmer() == Direction.UP) {
                            akce = BotAction.LEFT;
                        } else {
                            // otoc se o 180 stupnu - VYHODNE !!!!!! DODELAT
                            akce = BotAction.LEFT;
                        }
                    }
                }
            }
        }
        return akce;
    }
    
    private void spoctiPole(int xxx, int yyy, int vzdal, Direction smer, StrategyPolicko nej, PriorityQueue<StrategyPolicko> fronta, StrategyPolicko[][] mapaSveta) {
        if ((xxx < world.getDimensionX())&&(xxx >= 0)&&(yyy < world.getDimensionY())&&(yyy >= 0)
        &&(!((world.containsRobot(xxx,yyy))&&(world.getRobotOn(xxx,yyy).getEnergy()<=0)))
        &&(!world.containsWall(xxx,yyy))) { // test omezeni hraciho planu, mrtvol a zdi
            if (mapaSveta[xxx][yyy].getVzdalenost() < Integer.MAX_VALUE) { // !!! je uz ve fronte ke zpracovani => nepridavat
                if (mapaSveta[xxx][yyy].getVzdalenost() > vzdal) { // existuje kratsi cesta
                    mapaSveta[xxx][yyy].setVzdalenost(vzdal);
                    mapaSveta[xxx][yyy].setSmer(smer);
                    mapaSveta[xxx][yyy].setPredchudce(nej);
                } else {} // tudy je to delsi...
            } else { // neni jeste ve fronte ke zpracovani => pridat ; ma hodnotu maxint
                mapaSveta[xxx][yyy].setVzdalenost(vzdal);
                mapaSveta[xxx][yyy].setSmer(smer);
                mapaSveta[xxx][yyy].setPredchudce(nej);
                fronta.add(mapaSveta[xxx][yyy]);
            }
        }
    }
    
    private void spoctiVzdalenosti(Robot bot, int index) {
        
        for (int i = 0; i < world.getDimensionX(); i++) {
            for (int j = 0; j < world.getDimensionY(); j++) {
                mapa[index][i][j] = new StrategyPolicko(i, j, Integer.MAX_VALUE, null);
            }
        }
        mapa[index][bot.getPositionX()][bot.getPositionY()].setSmer(bot.getDirection());
        mapa[index][bot.getPositionX()][bot.getPositionY()].setVzdalenost(0);
        
        PriorityQueue poleKPocitani = new PriorityQueue<StrategyPolicko>();
        poleKPocitani.add(mapa[index][bot.getPositionX()][bot.getPositionY()]);
        
        while (!poleKPocitani.isEmpty()) {
            StrategyPolicko nej = (StrategyPolicko) poleKPocitani.poll();
            int xova = nej.getX();
            int yova = nej.getY();
            int vzdal = nej.getVzdalenost();
            
            if (nej.getSmer() == Direction.UP) {
                /* ocisluj sousedni 4 policka */
                spoctiPole(xova+0, yova-1, vzdal+1, Direction.UP, nej, poleKPocitani, mapa[index]); // policko PO smeru natoceni
                spoctiPole(xova+0, yova+1, vzdal+3, Direction.DOWN, nej, poleKPocitani, mapa[index]); // policko PROTI smeru natoceni
                spoctiPole(xova-1, yova+0, vzdal+2, Direction.LEFT, nej, poleKPocitani, mapa[index]); // policko VLEVO od smeru natoceni
                spoctiPole(xova+1, yova+0, vzdal+2, Direction.RIGHT, nej, poleKPocitani, mapa[index]); // policko VPRAVO od smeru natoceni
            } else {
                if (nej.getSmer() == Direction.DOWN) {
                    /* ocisluj sousedni 4 policka */
                    spoctiPole(xova+0, yova+1, vzdal+1, Direction.DOWN, nej, poleKPocitani, mapa[index]); // policko PO smeru natoceni
                    spoctiPole(xova+0, yova-1, vzdal+3, Direction.UP, nej, poleKPocitani, mapa[index]); // policko PROTI smeru natoceni
                    spoctiPole(xova+1, yova+0, vzdal+2, Direction.RIGHT, nej, poleKPocitani, mapa[index]); // policko VLEVO od smeru natoceni
                    spoctiPole(xova-1, yova+0, vzdal+2, Direction.LEFT, nej, poleKPocitani, mapa[index]); // policko VPRAVO od smeru natoceni
                } else {
                    if (nej.getSmer() == Direction.LEFT) {
                        /* ocisluj sousedni 4 policka */
                        spoctiPole(xova-1, yova+0, vzdal+1, Direction.LEFT, nej, poleKPocitani, mapa[index]); // policko PO smeru natoceni
                        spoctiPole(xova+1, yova+0, vzdal+3, Direction.RIGHT, nej, poleKPocitani, mapa[index]); // policko PROTI smeru natoceni
                        spoctiPole(xova+0, yova+1, vzdal+2, Direction.DOWN, nej, poleKPocitani, mapa[index]); // policko VLEVO od smeru natoceni
                        spoctiPole(xova+0, yova-1, vzdal+2, Direction.UP, nej, poleKPocitani, mapa[index]); // policko VPRAVO od smeru natoceni
                    } else {
                        // Direction.RIGHT)
                        /* ocisluj sousedni 4 policka */
                        spoctiPole(xova+1, yova+0, vzdal+1, Direction.RIGHT, nej, poleKPocitani, mapa[index]); // policko PO smeru natoceni
                        spoctiPole(xova-1, yova+0, vzdal+3, Direction.LEFT, nej, poleKPocitani, mapa[index]); // policko PROTI smeru natoceni
                        spoctiPole(xova+0, yova-1, vzdal+2, Direction.UP, nej, poleKPocitani, mapa[index]); // policko VLEVO od smeru natoceni
                        spoctiPole(xova+0, yova+1, vzdal+2, Direction.DOWN, nej, poleKPocitani, mapa[index]); // policko VPRAVO od smeru natoceni
                    }
                }
            }
            
            if (world.containsTreasure(xova,yova)) {
                polePokladu[index].add(mapa[index][xova][yova]);
            }
        }
    }
}
