package cz.muni.fi.fibot.vtipnynazev;

import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.Robot;

public class MyRobot {
	/**
	 * Enum of all possible directions for Robot
	 * 
	 * @author Jiri Syrovy
	 * 
	 */
	public int direction;
	public int id;
	public int energy;
	public int positionX;
	public int lastX;
	public int positionY;
	public int lastY;
	public int points;
	public int newEnergy;
	public boolean reverted;
	public MyWorld world;
	public Strategy.BotAction pendingAction;
	public Strategy.BotAction lastAction;

	/**
	 * Constructor must be called with parameters !
	 */
	public MyRobot() {
	}

	/**
	 * Robot constructor - usually called by WorldLoader
	 * 
	 * @param direction -
	 *            initial robot direction
	 * @param posX -
	 *            initial robot position (x coordinate)
	 * @param posY -
	 *            initial robot position (y coordinate)
	 * @param energy -
	 *            initial robot energy
	 * @param world -
	 *            world where the robot lives ;)
	 */
	public MyRobot(int direction, int posX, int posY, MyWorld world) {
		this.direction = direction;
		this.positionX = posX;
		this.lastX = posX;
		this.lastY = posY;
		this.reverted = false;
		this.positionY = posY;
		this.points = 0;
		this.energy = world.getMaxRobotEnergy();
		this.newEnergy = this.energy;
		this.world = world;
	}

	public MyRobot(Robot old) {
		id = old.getId();
		if (old.getDirection() == Robot.Direction.DOWN)
			direction = Coord.DOWN;
		else if  (old.getDirection() == Robot.Direction.UP)
			direction = Coord.UP;
		else if (old.getDirection() == Robot.Direction.LEFT)
			direction = Coord.LEFT;
		else if (old.getDirection() == Robot.Direction.RIGHT)
			direction = Coord.RIGHT;
		positionX = old.getPositionX();
		positionY = old.getPositionY();
		energy = old.getEnergy();
		newEnergy = old.getEnergy();
		lastAction = old.getLastAction();
		lastX = positionX;
		lastY = positionY;
		reverted = false;
		points = old.getPoints();
	}
	
	public MyRobot clone() {
		MyRobot robot = new MyRobot();
		robot.id = this.id;
		robot.direction = this.direction;
		robot.positionX = this.positionX;
		robot.positionY = this.positionY;
		robot.world = this.world;
		robot.energy = this.energy;
		robot.newEnergy = this.energy;
		robot.lastX = this.lastX;
		robot.lastY = this.lastY;
		robot.reverted = this.reverted;
		robot.points = this.points;
		robot.lastAction = this.lastAction;
		robot.pendingAction = this.pendingAction;
		return robot;
		
	}

	/**
	 * Getter for strategy
	 * 
	 * @return Strategy - strategy for this robot
	 */
	public void commit() {
		if (this.reverted) {
			this.positionX = this.lastX;
			this.positionY = this.lastY;
		}
		this.energy = this.newEnergy;
	}

	/**
	 * @return direction of the robot
	 */
	public int getDirection() {
		return direction;
	}

	/**
	 * @return x coordinate
	 */
	public int getPositionX() {
		return this.positionX;
	}

	/**
	 * @return y coordinate
	 */
	public int getPositionY() {
		return this.positionY;
	}

	/**
	 * @return robot's energy
	 */
	public int getEnergy() {
		return this.energy;
	}

	/**
	 * 
	 */
	public void revert() {
		this.reverted = true;
	}

	public boolean getReverted() {
		return this.reverted;
	}

	public void doAction() {
		if (this.energy <= 0) {
			this.newEnergy = this.energy;
			return;
		} else {
			this.energy--;
			this.newEnergy = this.energy;
			this.reverted = false;
			this.lastX = this.positionX;
			this.lastY = this.positionY;
		}
		this.lastAction = this.pendingAction;
	}

	public void turnRight() {
		this.direction = (direction + 1) % 4;
	}

	public void turnLeft() {
		this.direction = (direction + 3) % 4;
	}

	public void doStep() {
		int x = this.positionX + Coord.dir[direction].m_x;
		int y = this.positionY + Coord.dir[direction].m_y;
		if (!this.world.containsWall(x, y)) {
			this.positionX = x;
			this.positionY = y;
		}
	}

	public MyRobot getShot() {
		MyRobot result = null;
		int x = this.positionX, y = this.positionY;
		switch (this.direction) {
		case Coord.UP:
			while (y > 0 && !world.containsRobot(x, y - 1) && !world.containsWall(x, y - 1)) {
				y--;
			}
			y -= 1;
			break;
		case Coord.DOWN:
			while (y < world.getDimensionY() - 1 && !world.containsRobot(x, y + 1) && !world.containsWall(x, y + 1))
				y++;
			y += 1;
			break;
		case Coord.LEFT:
			while (x > 0 && !world.containsRobot(x - 1, y) && !world.containsWall(x - 1, y))
				x--;
			x -= 1;
			break;
		case Coord.RIGHT:
			while (x < world.getDimensionX() - 1 && !world.containsRobot(x + 1, y) && !world.containsWall(x + 1, y))
				x++;
			x += 1;
			break;
		}
		for (MyRobot robot : world.getRobots()) {
			if (robot.getPositionX() == x && robot.getPositionY() == y) {
				result = robot;
				break;
			}
		}
		return result;
	}

	public void shoot() {
		MyRobot strickenRobot = getShot();
		if (strickenRobot != null) {
			this.points += strickenRobot.energy / 2;
			strickenRobot.newEnergy = strickenRobot.energy - strickenRobot.energy / 2;
		}
	}

	public void take() {
		this.points += this.world.removeTreasure(this.positionX, this.positionY);
	}

	public void rest() {
		this.newEnergy += this.world.getRestRobotEnergy();
		if (this.newEnergy > this.world.getMaxRobotEnergy())
			this.newEnergy = this.world.getMaxRobotEnergy();
	}

	public void recharge() {
		if (this.world.containsBattery(this.positionX, this.positionY)) {
			this.newEnergy = this.world.getMaxRobotEnergy();
		}
	}

	/**
	 * @return how many points does the robot have
	 */
	public int getPoints() {
		return this.points;
	}

	public void setPoints(int points) {
		this.points = points;
	}

	public void addPoints(int points) {
		this.points += points;
	}

	/**
	 * @return Returns the pendingAction.
	 */
	public Strategy.BotAction getPendingAction() {
		return pendingAction;
	}

	/**
	 * returns bots last action as string
	 * @return String
	 */
	public String getLastActionString() {
		if (this.lastAction != null) {
			switch (this.lastAction) {
			case LEFT:
				return "left";
			case RIGHT:
				return "right";
			case STEP:
				return "step";
			case SHOOT:
				return "shoot";
			case REST:
				return "rest";
			case TAKE:
				return "take";
			case RECHARGE:
				return "rchrg";
			default:
				return "-";
			}
		} else
			return "-";
	}

	/**
	 * returns last bots action
	 * @return Strategy.BotAction
	 */
	public Strategy.BotAction getLastAction() {
		return this.lastAction;
	}

	/**
	 * @param pendingAction
	 *            The pendingAction to set.
	 */
	public void setPendingAction(Strategy.BotAction pendingAction) {
		this.pendingAction = pendingAction;
	}

	/**
	 * @param world
	 *            The world to set.
	 */
	public void setWorld(MyWorld world) {
		this.world = world;
	}

	/**
	 * @return Returns the id.
	 */
	public int getId() {
		return id;
	}

	/**
	 * @param id
	 *            The id to set.
	 */
	public void setId(int id) {
		this.id = id;
	}

	/**
	 * @return Returns the lastX.
	 */
	public int getLastX() {
		return lastX;
	}

	/**
	 * @return Returns the lastY.
	 */
	public int getLastY() {
		return lastY;
	}
}
