package cz.muni.fi.fibot.vtipnynazev;

import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.vtipnynazev.purestrat.IPureStrat;
import java.util.*;
import java.util.concurrent.*;

public class MyWorld {
	public static final int NUMSPOTS = 20;
	public int m_numRobots;
	public List<IPureStrat> m_strats;
	public MyRobot m_me;
	public int m_meid;
	public int m_subround;
	public List<Coord> m_treasures;
	public List<Coord> m_batteries;
	public SortedMap<Double, Coord> m_spots;
	public Coord[] m_fixedList;
	/**
	 * Class that describes world, where robots live
	 * 
	 * @author Administrator
	 *
	 */
	public List<MyRobot> robots;
	public int dimensionX;
	public int dimensionY;
	public int gameMap[][];
	public static final int wall = 1;
	public static final int battery = 2;
	public static final int robotConst = 4;
	public static final int treasure = 256;
	public int restRobotEnergy;
	public int maxRobotEnergy;
	public int round;
	public int roundsCount;

	public void setMe() {
		m_me = robots.get(m_meid);
	}

	public Strategy.BotAction doStep() {
		Strategy.BotAction result = m_strats.get(m_meid).getNextAction(this);
		m_me.pendingAction = result;
		if (m_meid == m_numRobots - 1)
			commit();
		m_meid = (m_meid + 1) % m_numRobots; //next robot
		setMe();
		++m_subround;
		return result;
	}

	public void commit() {
		Queue<MyRobot> finishedQueue = new LinkedBlockingQueue<MyRobot>();
		for (MyRobot robot : robots)
			finishedQueue.add(robot);
		boolean clean = false;
		// action initialization 
		for (MyRobot robot : robots) {
			try {
				// decrease energy, ...
				robot.doAction();
			} catch (Exception e) {
				// Do nothing
				finishedQueue.remove(robot);
			}
		}
		// do some shooting
		for (MyRobot robot : robots) {
			if (robot.getPendingAction() == Strategy.BotAction.SHOOT) {
				robot.shoot();
				finishedQueue.remove(robot);
			}
		}
		// evaluate shooting
		for (MyRobot robot : robots) {
			robot.commit();
		}
		// do some steps
		for (MyRobot robot : finishedQueue) {
			if (robot.getPendingAction() == Strategy.BotAction.STEP) {
				robot.doStep();
			}
		}
		// evaluate steps
		while (!clean) {
			clean = true;
			for (MyRobot robot : finishedQueue) {
				for (MyRobot robotComp : robots) {
					if (robot != robotComp
							&& !robot.getReverted()
							&& ((robot.getPositionX() == robotComp.getPositionX() && robot.getPositionY() == robotComp.getPositionY())
									|| (robot.getPositionX() == robotComp.getLastX() && robot.getLastX() == robotComp.getPositionX() && robot.getPositionY() == robotComp.getPositionY()) || (robot
									.getPositionY() == robotComp.getLastY()
									&& robotComp.getPositionY() == robot.getLastY() && robot.getPositionX() == robotComp.getPositionX()))) {
						robot.revert();
						robotComp.revert();
						clean = false;
					}
				}
			}
			for (MyRobot robot : robots) {
				robot.commit();
			}
		}
		// do (and evalute) another actions
		for (MyRobot robot : finishedQueue) {
			if (robot.getPendingAction() == Strategy.BotAction.LEFT) {
				robot.turnLeft();
			} else if (robot.getPendingAction() == Strategy.BotAction.RIGHT) {
				robot.turnRight();
			} else if (robot.getPendingAction() == Strategy.BotAction.TAKE) {
				robot.take();
			} else if (robot.getPendingAction() == Strategy.BotAction.REST) {
				robot.rest();
			} else if (robot.getPendingAction() == Strategy.BotAction.RECHARGE) {
				robot.recharge();
			}
			finishedQueue.remove(robot);
		}
		// save everything
		for (MyRobot robot : robots) {
			robot.commit();
		}
		// another round has finished
		this.round++;
	}

	/**
	 * @return Returns the round.
	 */
	public int getRound() {
		return round;
	}

	/**
	 * @param round
	 *            The round to set.
	 */
	public void setRound(int round) {
		this.round = round;
	}

	public MyWorld() {}

	public double getSpotScore(Coord spot) {
		final double COEF1 = 4.0;
		final double COEF2 = -8.0;
		final double COEF3 = 1.0;
		final double COEF4 = 1.0;
		int n = 0;
		int t = 0;
		int dist = 0;
		int x = spot.m_x;
		int y = spot.m_y;
		for (int i = -1; i <= 1; i += 2) {
			while (x + i >= 0 && x + i < dimensionX && !this.containsWall(x + i, y)) {
				x += i;
				++dist;
				if (this.containsBattery(x, y))
					++n;
				if (this.containsTreasure(x, y))
					++t;
			}
			x = spot.m_x;
		}
		for (int i = -1; i <= 1; i += 2) {
			while (y + i >= 0 && y + i < dimensionY && !this.containsWall(x, y + i)) {
				y += i;
				++dist;
				if (this.containsBattery(x, y))
					++n;
				if (this.containsTreasure(x, y))
					++t;
			}
			y = spot.m_y;
		}
		int mid = Math.abs(2 * spot.m_x - dimensionX + 1) + Math.abs(2 * spot.m_y - dimensionY + 1);
		double result = 
			COEF1 * dist / (dimensionX + dimensionY) + 
			COEF2 * mid / (2 * (dimensionX + dimensionY)) +
			COEF3 * n +
			COEF4 * t;
		return result;
	}

	public void setSpotsList() {
		Random jitterer = new Random();
		m_spots = new TreeMap<Double, Coord>();
		for (int x = 0; x < this.dimensionX; x++)
			for (int y = 0; y < this.dimensionY; y++) {
				if (!this.containsWall(x, y)) {
					Coord now = new Coord(x, y);
					double score = getSpotScore(now) + 0.00001 * jitterer.nextDouble();
					m_spots.put(-score, now);
				}
			}
		Iterator<Map.Entry<Double, Coord> > it = m_spots.entrySet().iterator();
		double val = 0.0;
		for (int i = 0; i < NUMSPOTS && it.hasNext(); ++i)
			val = it.next().getKey();
		m_spots = m_spots.headMap(val);
		
		ArrayList<Coord> fixedList = new ArrayList<Coord>();
		fixedList.addAll(this.m_treasures);
		it = m_spots.entrySet().iterator();
		for (; it.hasNext(); ) {
			fixedList.add(it.next().getValue());
		}
		m_fixedList = new Coord[fixedList.size()];
		for (int i = 0; i < m_fixedList.length; ++i)
			m_fixedList[i] = fixedList.get(i);

		/*
		System.out.println("---");
		for (it = m_spots.entrySet().iterator(); it.hasNext(); ) {
			Map.Entry<Double, Coord> n = it.next();
			System.out.println(n.getValue() + "\t" + (-n.getKey()));
		}
		*/
	}

	public MyWorld(cz.muni.fi.fibot.World world) {
		//set up a copy of World
		dimensionX = world.getDimensionX();
		dimensionY = world.getDimensionY();
		gameMap = new int[this.dimensionX][this.dimensionY];
		m_treasures = new LinkedList<Coord>();
		m_batteries = new LinkedList<Coord>();
		for (int x = 0; x < this.dimensionX; x++)
			for (int y = 0; y < this.dimensionY; y++) {
				gameMap[x][y] = world.getFieldValue(x, y);
				if (this.containsTreasure(x, y))
					m_treasures.add(new Coord(x, y));
				if (this.containsBattery(x, y))
					m_batteries.add(new Coord(x, y));
			}
		maxRobotEnergy = world.getMaxRobotEnergy();
		restRobotEnergy = world.getRestRobotEnergy();
		roundsCount = world.getRoundsCount();
		round = world.getRound();
		robots = new ArrayList<MyRobot>();
		for (cz.muni.fi.fibot.Robot robot : world.getRobots()) {
			MyRobot clonedRobot = new MyRobot(robot);
			clonedRobot.setWorld(this);
			robots.add(clonedRobot);
		}
		m_numRobots = robots.size();
		m_subround = round * m_numRobots;
		m_meid = 0;
		setMe();
		m_strats = new ArrayList<IPureStrat>();
		for (int i = 0; i < robots.size(); ++i)
			m_strats.add(null);
	}

	public MyWorld clone() {
		MyWorld world = new MyWorld();
		world.dimensionX = this.dimensionX;
		world.dimensionY = this.dimensionY;
		world.gameMap = new int[this.dimensionX][this.dimensionY];
		for (int x = 0; x < this.dimensionX; x++)
			for (int y = 0; y < this.dimensionY; y++) {
				world.gameMap[x][y] = this.gameMap[x][y];
			}
		world.maxRobotEnergy = this.maxRobotEnergy;
		world.restRobotEnergy = this.restRobotEnergy;
		world.roundsCount = this.roundsCount;
		world.round = this.round;
		List<MyRobot> robots = new ArrayList<MyRobot>();
		world.robots = robots;
		for (MyRobot robot : this.robots) {
			MyRobot clonedRobot = robot.clone();
			clonedRobot.setWorld(world);
			robots.add(clonedRobot);
		}
		world.m_subround = this.m_subround;
		world.m_numRobots = this.m_numRobots;
		world.m_meid = this.m_meid;
		world.setMe();
		world.m_strats = new ArrayList<IPureStrat>();
		for (IPureStrat strat : this.m_strats) {
			IPureStrat clonedStrat = strat;//strat == null ? null : strat.clone();//FIXME ok?
			world.m_strats.add(clonedStrat);
		}
		world.m_treasures = new LinkedList<Coord>();
		world.m_treasures.addAll(this.m_treasures);
		world.m_batteries = this.m_batteries;
		world.m_spots = this.m_spots;
		world.m_fixedList = this.m_fixedList;
		return world;
	}

	public void setDimensionX(int x) {
		this.dimensionX = x;
	}

	public void setDimensionY(int y) {
		this.dimensionY = y;
	}

	public boolean containsObject(int x, int y, int object) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		if ((this.gameMap[x][y] & object) == object)
			return true;
		else
			return false;
	}

	/**
	 * Returns value of treasure on (x,y)
	 * 
	 * @param x
	 *            coordinate
	 * @param y
	 *            coordinate
	 * @return treasure value
	 */
	public int getTreasureValue(int x, int y) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		return (this.gameMap[x][y] / 256);
	}

	/**
	 * Returns if is treasure on (x,y)
	 * 
	 * @param x
	 *            coordinate
	 * @param y
	 *            coordinate
	 * @return true/false
	 */
	public boolean containsTreasure(int x, int y) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		return 0 < (this.gameMap[x][y] / 256);
	}

	/**
	 * Returns if is wall on (x,y)
	 * 
	 * @param x
	 *            coordinate
	 * @param y
	 *            coordinate
	 * @return true/false
	 */
	public boolean containsWall(int x, int y) {
		return this.containsObject(x, y, wall);
	}

	/**
	 * Returns if is battery on (x,y)
	 * 
	 * @param x
	 *            coordinate
	 * @param y
	 *            coordinate
	 * @return true/false
	 */
	public boolean containsBattery(int x, int y) {
		return this.containsObject(x, y, battery);
	}

	/**
	 * Returns if is robot on (x,y)
	 * 
	 * @param x
	 *            coordinate
	 * @param y
	 *            coordinate
	 * @return true/false
	 */
	public boolean containsRobot(int x, int y) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		for (MyRobot robot : this.robots) {
			if (robot.getPositionX() == x && robot.getPositionY() == y)
				return true;
		}
		return false;
	}

	public MyRobot getRobotOn(int x, int y) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		for (MyRobot robot : this.robots) {
			if (robot.getPositionX() == x && robot.getPositionY() == y)
				return robot;
		}
		return null;
	}

	public int removeTreasure(int x, int y) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		int points = this.gameMap[x][y] / 256;
		this.gameMap[x][y] -= points * 256;
		return points;
	}

	/**
	 * Size of the world
	 * 
	 * @return x size
	 */
	public int getDimensionX() {
		return this.dimensionX;
	}

	/**
	 * Size of the world
	 * 
	 * @return y size
	 */
	public int getDimensionY() {
		return this.dimensionY;
	}

	/**
	 * How much energy does robot get after rest
	 * 
	 * @return energy
	 */
	public int getRestRobotEnergy() {
		return this.restRobotEnergy;
	}

	/**
	 * How much energy can robot have
	 * 
	 * @return energy
	 */
	public int getMaxRobotEnergy() {
		return this.maxRobotEnergy;
	}

	/**
	 * Robots list
	 * 
	 * @return unmodifiable robots list
	 */
	public List<MyRobot> getRobots() {
		return this.robots;
	}

	public void setGameMap(int[][] gameMap) {
		this.gameMap = gameMap;
	}

	public void setMaxRobotEnergy(int maxRobotEnergy) {
		this.maxRobotEnergy = maxRobotEnergy;
	}

	public void setRestRobotEnergy(int restRobotEnergy) {
		this.restRobotEnergy = restRobotEnergy;
	}

	public void setRobots(List<MyRobot> robots) {
		this.robots = robots;
	}

	/**
	 * @return Returns the roundsCount.
	 */
	public int getRoundsCount() {
		return roundsCount;
	}

	/**
	 * @param roundsCount
	 *            The roundsCount to set.
	 */
	public void setRoundsCount(int roundsCount) {
		this.roundsCount = roundsCount;
	}

	public int getFieldValue(int x, int y) {
		if (!(x < this.dimensionX && y < this.dimensionY && x >= 0 && y >= 0))
			throw new IllegalArgumentException("Coordinate(s) out of range");
		if (this.containsRobot(x, y))
			return this.gameMap[x][y] + MyWorld.robotConst;
		else
			return this.gameMap[x][y];
	}
}
