package cz.muni.fi.fibot.vtipnynazev;

import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.vtipnynazev.purestrat.*;
import java.util.*;

/**
 * main class for FIBOT competition
 * 
 * @author Radim Rehurek
 * 
 */

public class VtipnyNazev implements Strategy {
	public static final int MAXROUNDS = 300;
	public static final int MAXROBOTS = 10;
	public static final int LOOKAHEAD = 3;
	public static final int PICKBEST = 5;
	
	public MyWorld m_world;
	public Robot m_robot;
	public Random m_rand;
	public int m_myPos;
	public int m_startPly;
	public int m_finalPly;
	public Strategy.BotAction m_bestAction;
	public STurnShoot[][] m_turnShoot;
	public boolean[] m_reachable;
	public int m_lookAhead;
	public int m_nodes;
	public boolean m_fastExit;
	public long m_endTime;

	public final double jitter() {
		return 0.00001 * m_rand.nextDouble();
	}
	
	public int getMyPos() {
		for (int i = 0; i < m_world.robots.size(); ++i)
			if (m_world.robots.get(i).getId() == m_robot.getId())
				return i;
		return -1;
	}

	public SortedMap<Double, Coord> getSortedTreasures() {
		Coord now = new Coord(m_world.m_me.positionX, m_world.m_me.positionY);
		SortedMap<Double, Coord> result = new TreeMap<Double, Coord>();
		
		for (Coord pos : m_world.m_treasures) {
			double value = m_world.getTreasureValue(pos.m_x, pos.m_y) + jitter();
			int id = PathFinder.getId(pos);
			int dist = PathFinder.getDistance(now, m_world.m_me.direction, id);
			value /= dist;
			result.put(-value, pos);			
		}
		return result;
	}
	
	public SortedMap<Double, Coord> getSortedPoints() {
		Coord now = new Coord(m_world.m_me.positionX, m_world.m_me.positionY);
		SortedMap<Double, Coord> result = new TreeMap<Double, Coord>();
		
		Iterator<Map.Entry<Double, Coord> > it = m_world.m_spots.entrySet().iterator();
		for (; it.hasNext(); ) {
			Map.Entry<Double, Coord> n = it.next();
			double value = -n.getKey();
			Coord pos = n.getValue();
			int id = PathFinder.getId(pos);
			int dist;
			if (id == -1)
				dist = Integer.MAX_VALUE;
			else
				dist = PathFinder.getDistance(now, m_world.m_me.direction, id);
			value /= 2 * dist;
			result.put(-value, pos);
		}
		return result;
	}
	
	public SortedMap<Double, IPureStrat> generateStrats() {
		SortedMap<Double, IPureStrat> result = new TreeMap<Double, IPureStrat>();
		
		if (!m_reachable[m_world.m_meid] || m_world.m_me.energy < 1) {//we don't care what off-world robots do
			result.put(0.0, new SWait(0.0));
			return result;
		}
		
		if (m_world.m_me.energy <= SRecharge.ENERGYLIMIT) {//low on energy - heal
			result.put(0.0, new SRecharge());
			return result;
		}
		
		SortedMap<Double, Coord> pTreasures = getSortedTreasures();
		SortedMap<Double, Coord> pPoints = getSortedPoints();

		Iterator<Map.Entry<Double, Coord> > iter;
		for (iter = pPoints.entrySet().iterator(); iter.hasNext();) {
			Map.Entry<Double, Coord> e = iter.next();
			if (m_world.m_me.positionX == e.getValue().m_x && m_world.m_me.positionY == e.getValue().m_y) {
				result.put(e.getKey(), new SWait(-e.getKey()));
				iter.remove();
				break;
			}
		}
		
		pPoints.putAll(pTreasures);
		int left = PICKBEST;
		for (iter = pPoints.entrySet().iterator(); iter.hasNext() && left > 0; --left) {
			Map.Entry<Double, Coord> e = iter.next();
			if (m_world.containsTreasure(e.getValue().m_x, e.getValue().m_y)) {
				STreasure st = new STreasure(m_world, e.getValue(), -e.getKey());
				result.put(-st.getExpectedProfit(m_world), st);
			} else {
				SSnipe ss = new SSnipe(m_world, e.getValue(), -e.getKey());
				result.put(-ss.getExpectedProfit(m_world), ss);
			}
		}

		//TODO test jestli se da utect z prestrelky kdyz vypada nevyhodne
		//TODO eval - if mutual shooting eval profit till end vs retreating; when no more treasures, add spots dynamically = other bot positions
		if (result.size() <= 0)
			result.put(0.0, new SWait(0.0));
		
		return result;
	}

	public Strategy.BotAction decide() {
		m_endTime = System.currentTimeMillis() + 5000;
		m_bestAction = Strategy.BotAction.NOTHING;
		m_nodes = 0;
		m_startPly = m_world.m_subround;
		m_finalPly = Math.min(m_world.getRoundsCount() * m_world.m_numRobots - 1, m_world.m_subround + m_world.m_numRobots * m_lookAhead);
		m_fastExit = false;
		ply();
		if (m_fastExit)
			m_lookAhead = Math.max(1, m_lookAhead - 1);
		
		return m_bestAction;
	}
	

	public double[] evaluate() {
		double[] result = new double[m_world.m_numRobots];
		int sum = 0;
		for (int i = 0; i < result.length; ++i)
			sum += m_world.robots.get(i).points;
		for (int i = 0; i < result.length; ++i) {
			MyRobot robot = m_world.robots.get(i);
			double fSum = 1.0 * (sum - robot.points) / m_world.m_numRobots;	
			IPureStrat strat = m_world.m_strats.get(i); 
			double hValue = strat == null ? 0.0 : strat.getExpectedProfit(m_world);
			double hWeight = 1.0;
			if (robot.energy == 0)
				hWeight = 0.0;
			double nearEnd = 0.25 * Math.max(0, 4 - (m_world.roundsCount - m_world.round)); 
			result[i] = robot.points - 0.1 * fSum + nearEnd * robot.energy + hValue * hWeight * (1.0 - nearEnd);
		}
		return result;
	}
	
	public void setWorld(World world) {
		Coord.initDir(world.getDimensionX(), world.getDimensionY());
		m_world = new MyWorld(world);
	}

	public void setRobot(Robot robot) {
		m_robot = robot;
		m_reachable = new boolean[m_world.m_numRobots];
		Coord myPos = new Coord(m_robot.getPositionX(), m_robot.getPositionY());
		
		//mark unreachable robots
		for (int i = 0; i < m_world.m_numRobots; ++i) {
			MyRobot other = m_world.robots.get(i);
			m_reachable[i] = PathFinder.findPath(new Coord(other.positionX, other.positionY), Coord.LEFT, myPos, m_world) != null;
		}
		
		//remove unreachable treasures
		for (Iterator<Coord> iter = m_world.m_treasures.iterator(); iter.hasNext(); ) {
			Coord treasure = iter.next();
			if (PathFinder.findPath(myPos, Coord.LEFT, treasure, m_world) == null)
				iter.remove();
		}
	}

	public void initialize() {
		m_turnShoot = new STurnShoot[MAXROUNDS * MAXROBOTS][];
		m_rand = new Random();
		m_lookAhead = LOOKAHEAD;
		m_myPos = getMyPos();
		m_world.setSpotsList();
		PathFinder.precomputePaths(m_world.m_fixedList, m_world);
	}
	
	public List<MyRobot> getSameLevelRobots() {
		List<MyRobot> result = new LinkedList<MyRobot>();
		int oldDir = m_world.m_me.direction;
		for (m_world.m_me.direction = 0; m_world.m_me.direction < 4; ++m_world.m_me.direction) {
			MyRobot hit = m_world.m_me.getShot();
			if (hit != null)
				result.add(hit);
		}
		m_world.m_me.direction = oldDir;
		return result;
	}

	public double[] ply() {
		++m_nodes;
		if (m_nodes % 10000 == 0) {
			if (System.currentTimeMillis() > m_endTime) {
				m_fastExit = true;
				return null;
			}
		}
		if (m_world.m_subround >= m_finalPly/* || m_world.robots.get(this.m_myPos).energy < 1*/)//XXX stop when i'm dead?
			return evaluate();

		m_turnShoot[m_world.m_subround] = new STurnShoot[4]; 
		SortedMap<Double, MyWorld> branches = new TreeMap<Double, MyWorld>();
		List<MyRobot> sameLevel = getSameLevelRobots();
		for (MyRobot robot : sameLevel) {
			Coord dir = new Coord(
					Integer.signum(robot.getPositionX() - m_world.m_me.getPositionX()),
					Integer.signum(robot.getPositionY() - m_world.m_me.getPositionY())
					);
			int dirThen = Coord.getDir(dir);
			STurnShoot shootIt = new STurnShoot(m_world.m_me.direction, dirThen, robot.energy - robot.energy / 2);
			int hist = m_world.m_subround - Math.abs(shootIt.m_diff) * m_world.m_numRobots;
			if (hist >= m_startPly) {
				m_turnShoot[hist][shootIt.m_dirThen] = shootIt;
			}
		}
		
		boolean interrupted = 
			m_world.m_strats.get(m_world.m_meid) == null || 
			m_world.m_strats.get(m_world.m_meid).isInterrupted(m_world);
		boolean panic = m_world.m_me.energy <= SRecharge.ENERGYLIMIT;
		interrupted |= panic;
		
		if (interrupted) {//current strategy no longer valid; generate new list
			SortedMap<Double, IPureStrat> gen = generateStrats();
			for (Map.Entry<Double, IPureStrat> e : gen.entrySet()) {
				MyWorld newBranch = m_world.clone();
				newBranch.m_strats.set(m_world.m_meid, e.getValue());
				branches.put(e.getKey(), newBranch);
			}
		} else {
			branches.put(-m_world.m_strats.get(m_world.m_meid).getExpectedProfit(m_world), m_world.clone());
		}

		double[] best = null;

		for (MyWorld branch : branches.values()) {//branch out
			if (!branch.m_strats.get(branch.m_meid).isInterrupted(branch)) {
				Strategy.BotAction action = branch.doStep();
				MyWorld tmp = m_world;
				m_world = branch;
				double[] val = ply();
				m_world = tmp;
				if (m_fastExit)
					return null;
				if (best == null || val[m_world.m_meid] > best[m_world.m_meid]) {
					best = val;
					if (m_world.m_subround == m_startPly)
						m_bestAction = action;
				}
			}
		}
		
		for (STurnShoot ts : m_turnShoot[m_world.m_subround]) {//backpropagate shooting
			if (ts == null)
				continue;
			MyWorld branch = m_world.clone();
			branch.m_strats.set(m_world.m_meid, ts);
			if (!branch.m_strats.get(branch.m_meid).isInterrupted(branch)) {
				Strategy.BotAction action = branch.doStep();
				MyWorld tmp = m_world;
				m_world = branch;
				double[] val = ply();
				if (m_fastExit)
					return null;
				m_world = tmp;
				if (best == null || val[m_world.m_meid] > best[m_world.m_meid]) {
					best = val;
					if (m_world.m_subround == m_startPly)
						m_bestAction = action;
				}
			}
		}
		m_turnShoot[m_world.m_subround] = null; 
		return best;
	}
//TODO neprasi se chozeni kdyz se srazim s nekym? odecte se action a simulace mi krok neuzna? 
}
